#ifndef _ALG_BRESENHAM_H 
#define _ALG_BRESENHAM_H 
///////////////////////////////////////////////////////////////////////////////
#include "alexandresku.h"
namespace Vtl
{
namespace ImAlg
{
/*! 
 *  \brief    , , , 
 *   .  '' -  ,  
 *      ,     .
 *
 *******************************************************************/
namespace Bres
{
	/*!
	 * \brief 
	 *
	 ***************************************************************/
	template<typename Pixel> class Draw
	{
		const Pixel _color;
	public:
		Draw(Pixel color):_color(color) 
		{
		}
		void operator()(Pixel* ptr)
		{
			*ptr=_color;
		}
	};
	/*!
	 * \brief 
	 *
	 ***************************************************************/
	template<typename Pixel> class DrawDotted
	{
		uint32_t _mask;
		const Pixel _color1;
		const Pixel _color2;
	public:
		DrawDotted(Pixel color1,Pixel color2=0,unsigned mask=0xFFFF0000)
			:_mask(mask),_color1(color1),_color2(color2)  
		{
			_mask=0xFFFF0000;
		}
		void operator()(Pixel* ptr)
		{
			*ptr=(_mask&0x80000000)?_color1:_color2;
			uint32_t ma=_mask;
			__asm {
				rol ma,1;
			}
			_mask=ma;
		}
	};
	/*!
	 * \brief 
	 *
	 ***************************************************************/
	template<typename Pixel> class DrawGradient
	{
		const Pixel _color1;
		const Pixel _color2;
		const int _step;
		int pos;
	public:
		DrawGradient(Pixel color1,Pixel color2=0,int step)
			:_step(step),_color1(color1),_color2(color2)  
		{
			pos=step;
		}
		void operator()(Pixel* ptr)
		{
			*ptr=Pixel(color1*pos+color2*(1-pos));
			--pos;
			if(pos<=0) pos=step;
		}
	};
	/*!
	 * \brief    ,  ,  
	 *         ,  
	 *   .       X   
	 *          
	 *     .    
	 *    D,      .
	 *   Vtl::Alg::Bres::Draw, Vtl::Alg::Bres::DrawDotted .
	 *
	 ***************************************************************/
	template<typename Pixel,class D> class SetPixelGray
	{
		Pixel * const _data;
		const int _nls;
		D _draw;
		Pixel * ptr;
		int _x;
		int _y;
	public:
		SetPixelGray(Pixel* data,int nls,D draw) : _data(data),_nls(nls),_draw(draw) {}
		/// 
		void Set0(int x,int y)
		{
			_x=x;
			_y=y;
			ptr=_data+_nls*y+x;
			_draw(ptr);
		} 
		/// !!! x_inc,y_inc    -1, 0, 1 .    
		/// ,   ++    y_inc*nls+x_inc \n
		/// - -    ,    nls   y_inc;
		/// -        x_inc, y_inc  , 
		///        .
		///
		template<int x_inc,int y_inc> void move()
		{
			ptr+=y_inc*_nls+x_inc;
			_draw(ptr);
		}
	};
	/*!
	 * \brief   '' .   
	 *   \ref Lit_Wilton , Listing 6-2 .      
	         (Swap),   
		       ,  
		         
		    () .    
			  .   ,    
		  ,       ( 
		),     :
			\arg  ,      
			  ,  (x1,y1).
		    \arg  ,     move<,>() 
			          
			    x,   y,   {-1,0,+1}.
		    \arg            
			    (  - )  .

		    :
		\code
		template<...> class MyFunctor
		{
		public:
			MyFunctor(...)  {}
			///  ,    ,
			///       ,   .
			void Set0(int x,int y)
			{
			}
			// !!! x_inc,y_inc    -1, 0, 1 .  
			// ,   \b    
			//     x_inc,y_inc 
			template<int x_inc,int y_inc> void move()
			{
			}
		};
		\endcode

		    . class SetPixelGray.

		 :
		     ,     \f$(x_1,y_1) \f$   
		 \f$(x_2,y_2) \f$ . 
		   
		\f$ frac{| y_2 - y_1 |}{| x_2 - x_1 |} \le 1 \f$ , ..    , 
		  .
		\image html BresLine.gif "    "

		
		\f$ x_1 < x_2 \f$ 


	 *
	 
	 ***************************************************************/
	template<typename F> class Line
	{
		F& functor;
		int d;
		int Aincr;
		int Bincr;
	public:
		Line(F& f) : functor(f) 
		{
		}
		void Draw(int dx,int dy)
		{
			int	yincr;
			int	xincr;
			if (dx > 0) // determine increment for x 
			{
				xincr = 1;
			}
			else
			{
				xincr = -1;
				dx = -dx;
			}
			if (dy > 0) // determine increment for y 
			{
				yincr = 1;
			}
			else
			{
				yincr = -1;
				dy = -dy;
			}
			int vert=dy-dx;
			int dxy;
			if(vert<=0)
			{
				dxy=dx;
				Aincr = 2 * vert;
			}
			else
			{
				dxy=dy;
				Aincr = - 2 * vert;
			}
			Bincr = Aincr + 2*dxy;
			d =Aincr + dxy;
			if (xincr >= 0)				
			{
				if(vert <= 0)
				{
					if(yincr >0) bres<-1,+1,+1>(dxy);
					else         bres<-1,+1,-1>(dxy);
				}
				else
				{
					if(yincr >0) bres<+1,+1,+1>(dxy);
					else         bres<+1,-1,+1>(dxy);
				}

			}
			else //if(dx < 0)
			{
				if(vert <= 0)
				{
					if(yincr >0) bres<-1,-1,+1>(dxy);
					else         bres<-1,-1,-1>(dxy);

				}
				else
				{
					if(yincr >0) bres<+1,+1,-1>(dxy);
					else         bres<+1,-1,-1>(dxy);
				}
			}
/*
			else //if(dx == 0)
			{
				if(dy==0)
				{

				}
			}
*/
		}
	private:
		template<int is_vert,int plus_dir,int Increment> void move(int stride)
		{
			if(is_vert <= 0)
				functor.move<plus_dir,Increment>();//x_move
			else
				functor.move<Increment,plus_dir>();//y_move
		}
		template<int is_vert,int plus_dir,int Increment> void bres(int stride)
		{
			for( int t=0; t< stride; ++t)		
			{
				if (d >= 0)
				{
					d += Aincr;
					move<is_vert,plus_dir,Increment>(stride);
				}
				else					// set pixel B 
				{
					d += Bincr;
					move<is_vert,plus_dir,0>(stride);
				}
			}
		}
/*
		template<int plus_dir,int Increment> x_bres(int stride)
		{
			//d=plus_dir*d;
			for( int t=0; t< stride; ++t)		
			{
				if (d >= 0)
				{
					d += Aincr;
					functor.move<plus_dir,Increment>();
				}
				else					// set pixel B 
				{
					d += Bincr;
					functor.move<plus_dir,0>();
				}
			}
		}
		template<int Increment,int plus_dir> y_bres(int stride)
		{
			for( int t=0; t< stride; ++t)		
			{
				if (d >= 0)
				{
					d += Aincr;
					functor.move<Increment,plus_dir>();
				}
				else					// set pixel B 
				{
					d += Bincr;
					functor.move<0,plus_dir>();
				}
			}
		}
*/
	};





	/*!
	 * \brief   '' 
	 *
	 ***************************************************************/
	template<class T> class Circle
	{
		long	Asquared;
		long	TwoAsquared;
		long	Bsquared;
		long	TwoBsquared;
		// Listing 7-2 
		long	d;
		long	dx,dy;
	public:
		///xc,yc - center of ellipse 
		///a0,b0 - semiaxes 
		Circle(int xc,int yc,int a0 )
		{
			int	x = 0;
			int	y = a0;

			long	a = a0;				// use 32-bit precision 

			Asquared = a * a;		/* initialize values outside */
			TwoAsquared = 2 * Asquared;	/*  of loops */

			d = Asquared*(1 - a) + Asquared/4;
			dx = 0;
			dy = TwoAsquared * a;
		}
		void Draw()
		{

			while (dx<dy)
			{
				Set4Pixels( x, y, xc, yc);

				if (d > 0)
				{
					++x;
					--y;
					dy -= TwoAsquared;
					d -= dy;
				}
				else
				{
					++x;
				}
				dx += TwoAsquared;
				d += Asquared + dx;
			}


			d +=  - (dx+dy)/2;

			while (y>=0)
			{
				Set4Pixels( x, y, xc, yc);

				if (d < 0L)
				{
					++x;
					--y;
					dx += TwoAsquared;
					d += dx;
				}
				else
				{
					--y;
				}
				dy -= TwoAsquared;
				d += Asquared - dy;
			}
		}

		void Set4Pixels(int x,int y,int xc,int yc)	    /* set pixels by symmetry in 4 quadrants */
		{
			SetPixel( xc+x, yc+y);
			SetPixel( xc-x, yc+y);
			SetPixel( xc+x, yc-y);
			SetPixel( xc-x, yc-y);
		}
	};
	/*!
	 * \brief  ''     
	 *     .
	 * // Listing 7-2 
	 ***************************************************************/
	template<class F> class EllipseH
	{
		long	Asquared;
		long	TwoAsquared;
		long	Bsquared;
		long	TwoBsquared;
		// Listing 7-2 
		long	d;
		long	dx,dy;
	public:
		///xc,yc - center of ellipse 
		///a0,b0 - semiaxes 
		EllipseH(int xc,int yc,int a0,int b0 )
		{
			int	x = 0;
			int	y = b0;

			long	a = a0;				// use 32-bit precision 
			long	b = b0;

			Asquared = a * a;		/* initialize values outside */
			TwoAsquared = 2 * Asquared;	/*  of loops */
			Bsquared = b * b;
			TwoBsquared = 2 * Bsquared;

			d = Bsquared - Asquared*b + Asquared/4L;
			dx = 0;
			dy = TwoAsquared * b;
		}
		void Draw()
		{

			while (dx<dy)
			{
				Set4Pixels( x, y, xc, yc);

				if (d > 0L)
				{
					++x;
					--y;
					dy -= TwoAsquared;
					d -= dy;
				}
				else
				{
					++x;
				}
				dx += TwoBsquared;
				d += Bsquared + dx;
			}


			d += (3L*(Asquared-Bsquared)/2L - (dx+dy)) / 2L;

			while (y>=0)
			{
				Set4Pixels( x, y, xc, yc);

				if (d < 0L)
				{
					++x;
					--y;
					dx += TwoBsquared;
					d += dx;
				}
				else
				{
					--y;
				}
				dy -= TwoAsquared;
				d += Asquared - dy;
			}
		}

		void Set4Pixels(int x,int y,int xc,int yc)	    /* set pixels by symmetry in 4 quadrants */
		{
			SetPixel( xc+x, yc+y);
			SetPixel( xc-x, yc+y);
			SetPixel( xc+x, yc-y);
			SetPixel( xc-x, yc-y);
		}
	};
	/*!
	 * \brief  ''    
	 *    . 
	 *
	 ***************************************************************/
	template<class T> class Ellipse
	{
	};
	/*!
	 * \brief   . 
	 *
	 ***************************************************************/
	template<class T> class Fill
	{
	};
}//namespace Bres
}//namespace ImAlg
}//namespace Vtl
///////////////////////////////////////////////////////////////////////////////
#endif //_ALG_BRESENHAM_H 
