#ifndef _IMG_REGISTRATION_H
#define _IMG_REGISTRATION_H
/////////////////////////////////////////////////////////////////////////////////////////
#include <sstream>

#include "plf_kernels_1d.h"
#include "plf_io_filter.h"
#include "plf_filters_complex.h"
#include "plf_functors.h"
#include "plf_sys.h"

#include "storage.h"
#include "integral_transforms.h"
#include "imalg_tools.h"
#include "imalg_resampling.h"

#include "debug_test.h" 
/////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl {
	namespace ImAlg{
namespace Registration{
	template<typename Real_t,class Model> class Hartley;
}}}
namespace Tests
{
	void test_save_x(float *block,std::string fname_root,size_t w,size_t h,bool gray);
}

namespace Vtl
{
namespace ImAlg
{
/*!
 * \brief   
 */
namespace Registration
{
		/*!
		 *	\brief      Vtl::Plf.   
		 *               ,  
		 *               .   max_size_
		 *         ,      .  max_size_   
		 *           ,    ,  max_size_ ~ 8.
		 *
		 */
		template<typename Real_t>
		class Container
		{
		public:
			struct Point
			{
				int x;
				int y;
				Real_t value;
				Point(int x_, int y_,Real_t value_): x(x_), y(y_),value(value_) {}
				bool operator<(const Point& p)
				{
					return value<p.value;
				}
			};
			typedef std::list<Point> List;
			typedef typename List::const_iterator const_Iter;
		private:
			List lst;
			size_t max_size;
			size_t half_width;
			size_t half_height;
		public:
			Container(size_t img_size[],size_t max_size_=8)
				: max_size(max_size_),half_width(img_size[0]/2),half_height(img_size[1]/2)
			{
				lst.push_back(Point(0,0,0)); 
			}
			const_Iter begin() {return lst.begin();}
			const_Iter end() {return lst.end();}
			void Add(int x, int y,Real_t value)
			{
				Point p(x-half_width,y-half_height,value);
				if(p < *(lst.rbegin()))
				{
					if(lst.size()<max_size)
						lst.push_back(p); 
				}
				else
				{
					 
					for( List::iterator it=lst.begin(); it!=lst.end(); ++it)
					{
						if(*it < p)
						{
							lst.insert(it,p);
							if(lst.size()>=max_size)
							{
								lst.pop_back(); 
							}
							return;
						}
					}
				}
			}
			void print()
			{
				int counter=0;
				for( List::const_iterator it=lst.begin() ; it!=lst.end(); ++it)
				{
					std::cout << std::setw(5) << counter++ << "  " 
						      << std::setw(5) << (*it).x   << " , "
							  << std::setw(5) << (*it).y   << " , " 
							  << std::setprecision(3) << (*it).value <<std::endl;
				}
			}
			void printRatio()
			{
				int counter=0;
				if(lst.size()<=1)
				{
						std::cout << "--- Peaks ratio: second/main=  "  << "only " <<lst.size()<<" peqaks detected" <<std::endl;
				}
				else
				{
						List::const_iterator it=lst.begin() ;
						Real_t v0=(*it).value ;
						++it;
						Real_t v1=(*it).value ;
						std::cout << "--- Peaks ratio: second/main=  "  << v1/v0 <<std::endl;
				}
			}
		};
	namespace Impl
	{
		template<typename Real_t>
		inline void comp_Phase(Real_t* image_phase,size_t n[],bool windowing)
		{
			if(windowing)
				Impl::window(image_phase,n[0],n[1],n[0],Real_t(n[0]/2),Real_t(n[1]/2),Real_t(0));
			transform_and_zero_freq(image_phase,n);
			//  -  (highpass)
			//highpass_filtering(image_phase,n);
			Vtl::NumAlg::IntTransf::Hartley::F::Hartley2Phase<Real_t> pp(image_phase);
			Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n,pp);

		}

		template<typename Real_t>
		inline void flip_vert(Real_t* im,size_t n[])
		{
			for(size_t y=0; y< n[1]/2; ++y )
			{
				Real_t *p1=im+y*n[0];
				Real_t *p2=im+(n[1]-1-y)*n[0];
				for(size_t x=0; x<n[0]; ++x)
				{
					Real_t t=p1[x];
					p1[x]=p2[x];
					p2[x]=t;
				}
			}
		}
		template<typename Real_t>
		inline void transform_and_zero_freq(Real_t* im,size_t n[])
		{
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(im,n);
			Vtl::NumAlg::IntTransf::Hartley::fht2D_square(im,n[0],n[1]);
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(im,n);
//			Vtl::NumAlg::IntTransf::Hartley::AddQuadrants2D(im,n[0],n[1]);
			//    :
			size_t zero_freq_index=Vtl::NumAlg::IntTransf::Utils::offset_to_zero_freq<2,true>(n);
			//std::cout << " zero_freq_index = " << zero_freq_index <<std::endl; 
			im[zero_freq_index]=0;
		}

		template<typename Real_t>
		static void window(Real_t* image,int width,int height,int nls,Real_t center_x,Real_t center_y,Real_t fill_by)
		{
			//typedef Vtl::ImAlg::WndFun::Gauss<Real_t,0> Functor;
			//Functor wnd_x(Real_t(is.width/8),4);
			//Functor wnd_y(Real_t(is.height/8),4);

			typedef Vtl::ImAlg::WndFun::Hann<Real_t,0> Functor;
			//typedef Vtl::ImAlg::WndFun::Nuttall<Real_t,0> Functor;
			//typedef Vtl::ImAlg::WndFun::Rect<Real_t,0> Functor;
			Functor wnd_x(width);
			Functor wnd_y(height);

			Vtl::Plf::Kernel::ImageWin<Real_t> iw;
			iw.Init(wnd_x,wnd_y,width,height,center_x,center_y,fill_by);
			iw.Run(image,nls);
		}

		template<typename Real_t>
		inline void comp_AmpToPolar(Real_t* amp,Real_t* polar,size_t n[],Vtl::ImAlg::Resamp::LogGrid<float>& lsg)
		{
			Vtl::ImAlg::Resamp::InterpBilinear<Real_t> interp(amp,n[0],n[1],n[0],Real_t(0));
			float x0=float(n[0]/2);
			float y0=0;//float(n[1]/2);
			Vtl::ImAlg::Resamp::PhiGrid<Real_t> phig(0.0f,Real_t(M_PI),n[1]);
			Vtl::ImAlg::Resamp::CartToPolar(interp,x0,y0,polar,n[0],n[1],n[0],lsg,phig);
			//Vtl::ImAlg::Resamp::CartToPolar(interp,x0,y0,polar,n[0],n[1],n[0]);
			for(size_t i=0; i<n[1]; ++i)
			{
				polar[i*n[0]]=0;
				polar[i*n[0]+1]=0;
				polar[i*n[0]+2]=0;
			}
		}
		template<typename Real_t>
		inline void comp_Rotate(Real_t* input,Real_t* result,size_t n[],float matrix_2x2[])
		{
			Vtl::ImAlg::Resamp::InterpBilinear<Real_t> interp(input,n[0],n[1],n[0],Real_t(0));
			float x0=float(n[0]/2);
			float y0=float(n[1]/2);
			Vtl::ImAlg::Resamp::Affine(interp,x0,y0,x0,y0,matrix_2x2,result,n[0],n[1],n[0]);
			//Vtl::ImAlg::Resamp::RotateScale(interp,x0,y0,phi,scale,result,n[0],n[1],n[0]);
		}
		template<typename Real_t>
		inline void find_shift(Container<Real_t>& cont,Real_t* phase_1,Real_t* phase_2,Real_t* delta,size_t n[])
		{
			comp_phase_corr(phase_1,phase_2,delta,n);
			delta_correlation(delta,n);
			find_offset(cont,delta,n);
		}
		template<typename Real_t>
		inline void comp_phase_corr(Real_t* phase_1,Real_t* phase_2,Real_t* corr,size_t n[])
		{
			Vtl::NumAlg::IntTransf::Hartley::F::PhaseCorr<Real_t> pp(phase_1,phase_2,corr); 
			Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n,pp);
		}
		template<typename Real_t>
		inline void highpass_filtering_half(Real_t* im1,size_t n[])
		{
			
			typedef Windowing::Data<Real_t,1> Dat;
			//typedef Windowing::Data1Log<Real_t> Dat;
			typedef Windowing::Kernels::Reddy<Real_t> Fun;
			//typedef Windowing::Kernels::Reddy_2<Real_t> Fun;

			Fun filter(n[0],n[1]);
			Dat d(im1);
			int x0=(n[0]>>1)+1;
			int y0=0;
			Windowing::Win2D<Dat,Fun>(x0,y0,d,n[0],n[1]/2,n[0],filter);
			
		}
		template<typename Real_t>
		inline void highpass_filtering(Real_t* im1,size_t n[])
		{
/*			typedef Windowing::Data<Real_t,1> Dat;
			typedef Windowing::Kernels::Reddy<Real_t> Fun;
			Fun filter(n[0],n[1]);
			Dat d(im1);
			int x0=(n[0]>>1)+1;
			int y0=(n[1]>>1)+1;
			Windowing::Win2D<Dat,Fun>(x0,y0,d,n[0],n[1],n[0],filter);
*/
		}
		template<typename Real_t>
		inline void delta_correlation(Real_t* im,size_t n[])
		{
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(im,n);
			Vtl::NumAlg::IntTransf::Hartley::ifht2D(im,n[0],n[1]);
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(im,n);
			//Vtl::NumAlg::IntTransf::Utils::Hartley2Modul<Real_t> mod(im1);
			//Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n,mod);
		}
		template<typename Real_t>
		inline Real_t calc_total_sum(Real_t* im,size_t sz)
		{
			double sum=0;
			for(size_t y=0; y< sz ; ++y)
			{
				double tmp=0;
				for(size_t x=0; x< sz ; ++x)
				{
					//if(*im>0)
						tmp+=*im;
					++im;
				}
				sum+=tmp; 
				//std::cout << "sum=" << sum <<std::endl;
			}
			return Real_t(sum);
		}
		template<class Container_t,typename Real_t>
		inline void find_offset(Container_t& cont,Real_t* im,size_t n[])
		{
			Real_t threshold=calc_total_sum(im,n[0]);
			std::cout << "!! threshold !!=" << threshold <<std::endl;
			//threshold=28972625.732421875f;
			//std::cout << "!! threshold !!=" << threshold <<std::endl;
			//threshold=size*size*0.01f;
			threshold*=0.001f;
			//Real_t threshold=size*size*0.01f;

			//Input: im1,size
			Vtl::Plf::SrcArray<Real_t,Real_t> src("Src -phase correlation",im,n[0],n[1],n[0]);
			
			typedef Vtl::Plf::F11::Threshold<Real_t,std::less<Real_t> > PostThresh;
			PostThresh thresh(threshold,std::less<Real_t>(),0); 

			//typedef Vtl::Plf::F::HarrisNMS<Real_t> Func;
			typedef Vtl::Plf::Fpie::HarrisNMS_Smooth<Real_t> Func;
			Func functor;
			Vtl::Plf::EnvProc1to1<Real_t,Func,PostThresh > nms("nms",functor,thresh);
			Vtl::Plf::DstContainer<Real_t,Container_t> dst("Dst",cont);

			
			src.out.ConnectTo(&nms.in);
			nms.out.ConnectTo(&dst.in);

			std::string out_name="find_delta_correlation";
			Vtl::Plf::FiltersSystem sys(out_name+".dot");
	//------ Assing task------------------------
			sys.Assign(&src,&dst);
	//------ Initialize destinations -----------
			Vtl::Plf::ImgSize is=dst.in.getImageSize();
			dst.setDstImage(is);
			sys.Run(indicator,out_name.c_str());
			//cont.print();
			cont.printRatio();

		}
	}
	template<typename Real_t,unsigned MemBlocks>
	class Base
	{
	protected:
		typedef Storage<Real_t,MemBlocks> Storage_t;
		Storage_t storage_;
	protected:
		virtual ~Base() {}
		Base() {}
		void Init(typename Storage_t::Offsets_t& s)
		{
			storage_.Init(s);
		}
		template<unsigned n_img>
		Real_t* _img()
		{
			return storage_.block<n_img>();
		}
		template<unsigned n_img> 
		static std::string Name(std::string name)
		{
			std::ostringstream os;
			os << name <<"_" <<n_img;
			return os.str();
		}
		/*!
		 *           ,
		 *           ,  
		 *    ,      Init()  
		 *     2.    .
		 ******************************************************************************************/
		template<typename Iter> 
		void _Insert(Vtl::Plf::OutPortNoPostProc<Real_t>* input,Iter src_begin,Iter src_end,Real_t* img,int w,int h,int nls)
		{
			::memset(img,0,w*h*sizeof(img[0]));
			Vtl::Plf::DstArray<Real_t,Real_t> dst("dst");

			input->ConnectTo(&dst.in);
//			avrg.out().ConnectTo(&dst.in);

			Vtl::Plf::FiltersSystem sys("ins_1");
			sys.Assign(src_begin,src_end,&dst);
			Vtl::Plf::ImgSize is=dst.in.getImageSize();

			int y0=(h-is.height)/2;
			int x0=(w-is.width)/2;

			assert(x0>=0 || y0>=0);

			if(x0<0 || y0<0)
			{
				throw std::string("Bad init size in ReddyChatterji_Hartley");
			}
			Real_t* out_ptr=img+w*y0+x0;
			dst.setDstImage(out_ptr,is.width,is.height,nls);
			sys.Run(indicator,"Insert 1");
			


			Impl::window(out_ptr,is.width,is.height,nls,Real_t(is.width/2),Real_t(is.height/2),Real_t(0));

			typedef Vtl::ImAlg::WndFun::Nuttall<float,0> H0;
			H0 h0(7);
			Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> ker(h0);
			Vtl::ImAlg::Conv::Conv2D::CyclicInplace(img,w,h,nls,ker);
		}
	public:
	};

	/*!
	 * \brief    Reddy-Chatterji \ref Lit_ReddyChatterji , 
	 *             .   ,  ,  
	 *           ,      .    
	 *                 .  ,    ( 1.8).
	 */
	template<typename Real_t>
	class Hartley<Real_t,Vtl::ImAlg::Model::Offset<Real_t> >
		: public Base<Real_t,3>
	{
		size_t n[2]; //width,height
	public:
		typedef Vtl::ImAlg::Model::Offset<Real_t> Model;
	public:
		virtual ~Hartley()
		{
		}
		Hartley()
			: Base()
		{
		}
		void Init(size_t width,size_t height)
		{
			size_t sq=width*height;
			Storage_t::Offsets_t s={sq,sq,sq};
			// blocks
			// 0 - Image1 | Hartley of Image1 | Hartley of Image1 divided by Amplitude
			// 1 - Image2 | Hartley of Image2 | Hartley of Image2 divided by Amplitude
			// 2 - temporary for : 
			//       | phase correlation
			//		 | its hartley for delta function
			Base::Init(s);
			n[0]=width;
			n[1]=height;
		}
	private:
		template<unsigned n_img>
		Real_t* img_H()
		{
			return _img<n_img-1>();
		}
		Real_t* img_Tmp()
		{
			return _img<2>();
		}
	public:
		template<unsigned n_img,typename Iter> 
		void Insert(Vtl::Plf::OutPortNoPostProc<Real_t>* input,Iter src_begin,Iter src_end)
		{
			Base::_Insert(input,src_begin,src_end,img_H<n_img>(),n[0],n[1],n[0]);
			AfterInsert<n_img>();
		}



		void Run(Model& model)
		{
			//compute Phase correlation
			Container<Real_t> cont(n,8);
			Vtl::Dbg::TimeElapsed t;
			t.start();
			Impl::find_shift(cont,img_H<1>(),img_H<2>(),img_Tmp(),n);
			t.finish();
			Real_t dx=  Real_t((*(cont.begin())).x);
			Real_t dy=  Real_t((*(cont.begin())).y);
			model.set(dx,dy);

			std::cout << "dx=" << std::setw(6) << dx <<"   dy="<< std::setw(6) << dy <<std::endl;
			std::cout << "Compute translation :>" << t.GetString() <<std::endl; 
			Tests::test_save_x(storage_.block<2>(),"21_phase_corr_ofs",n[0],n[1],false);
		}
	private:
		template<unsigned n_img> 
		void AfterInsert()
		{
			bool windowing=false;
			bool highpass=false;

			Tests::test_save_x(img_H<n_img>(),Name<n_img>("00_orig"),n[0],n[1],true);
			Impl::transform_and_zero_freq(img_H<n_img>(),n);
			Vtl::NumAlg::IntTransf::Hartley::F::Hartley2Phase<Real_t> pp(img_H<n_img>());
			Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n,pp);
		}
};

	/*!
	 * \brief    Reddy-Chatterji \ref Lit_ReddyChatterji ,  \ref Lit_ImgRegBushBroom
	 *             .   ,  ,  
	 *           ,      .    
	 *                 .  ,    ( 1.8).
	 */
	template<typename Real_t>
	class Hartley<Real_t,Vtl::ImAlg::Model::SkewYOffset<Real_t> >
		: public Base<Real_t,6>
	{
		size_t n[2]; //width,height; width==height==size
	public:
		typedef Vtl::ImAlg::Model::SkewYOffset<Real_t> Model;
	public:
		virtual ~Hartley()
		{
		}
		Hartley()
			: Base()
		{
		}
		void Init(size_t size)
		{
//			this->size=size;
			size_t sq=size*size;
			Storage_t::Offsets_t s={sq,sq,sq,sq/2,sq/2,sq};
			// blocks
			// 0 - Image1 | Hartley of Image1 | Hartley of Image1 divided by Amplitude
			// 1 - Image2 | Hartley of Image2 | Hartley of Image2 divided by Amplitude
			// 2 - temporary for : 
			//         Amplitude (Fourier/Hartley - the same) 
			//       | its hartley for delta function
			//       | rotated and rescaled Hartley of Image2 divided by Amplitude
			//       | phase correlation
			//		 | its hartley for delta function
			// 3 - Image 1 Amplitude 
			// 4 - Image 2 Amplitude 
			// 5- Sino
			Base::Init(s);
			n[0]=size;
			n[1]=size;
		}
	private:
		template<unsigned n_img>
		Real_t* img_H()
		{
			return _img<n_img-1>();
		}
		Real_t* img_Tmp()
		{
			return _img<2>();
		}
		template<unsigned n_img>
		Real_t* img_Amp()
		{
			return _img<n_img+2>();
		}
		Real_t* img_Sino()
		{
			return _img<5>();
		}
	public:
		template<unsigned n_img,typename Iter> 
		void Insert(Vtl::Plf::OutPortNoPostProc<Real_t>* input,Iter src_begin,Iter src_end)
		{
			Base::_Insert(input,src_begin,src_end,img_H<n_img>(),n[0],n[1],n[0]);
			AfterInsert<n_img>();
		}
		void Run(Model& model,int w,int h) //,Real_t* mat,Real_t *dx_,Real_t *dy_)
		{
			model.setIdentity();

			Real_t cof=model.getCof(w,h);

			Real_t mat[4];
			mat[0]=1.0f;
			mat[1]=0.0f;
			mat[2]=0.0f;
			mat[3]=1.0f;
			_SkewOffset_2(mat);
//			_SkewOffset_2(model);

			//compute Phase correlation
			Container<Real_t> cont(n,8);
			Vtl::Dbg::TimeElapsed t;
			t.start();
			Impl::find_shift(cont,img_H<1>(),img_H<2>(),img_Tmp(),n);
			t.finish();
			Real_t dx=  Real_t((*(cont.begin())).x);
			Real_t dy=  Real_t((*(cont.begin())).y);
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<Real_t,2,2> MatrixA;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<Real_t,2> VectorB;
			using namespace Vtl::NumAlg::Lin::ublas;
			MatrixA A;
			VectorB b;
			A(0,0)=mat[0];
			A(1,0)=mat[2];
			A(0,1)=mat[1];
			A(1,1)=mat[3];
			b(0)=dx/cof;
			b(1)=dy/cof;
			b=prod(A,b);
			
			//*dx_=dx;
			//*dy_=dy;
			//model.offset_by((dx/cof),dy/cof);
			model.set(mat[2],b(0),b(1));
			//std::cout << "dx=" << std::setw(6) << dx <<"   dy="<< std::setw(6) << dy <<std::endl;
			std::cout << "Compute translation :>" << t.GetString() <<std::endl; 
			Tests::test_save_x(storage_.block<2>(),"21_phase_corr_ofs",n[0],n[1],false);
		}
	private:
		template<unsigned n_img> 
		void AfterInsert()
		{
			Tests::test_save_x(img_H<n_img>(),Name<n_img>("00_orig"),n[0],n[1],true);

			Vtl::Dbg::TimeElapsed t;
			t.start();
			Impl::transform_and_zero_freq(img_H<n_img>(),n);
			//highpass_filtering(img_H<n_img>(),n);
			Vtl::NumAlg::IntTransf::Hartley::F::Hartley2PhaseAmpDown<2,Real_t> pp(img_H<n_img>(),img_Amp<n_img>(),n);
			Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n,pp);
			Vtl::Plf::F11::Log<float> log(0.001);
			Vtl::ImAlg::Tools::copy_1_1(img_Amp<n_img>(),n[0],n[1]/2,n[0],log);

			t.finish();
			std::cout << "SKEW: compute Amplitude & Phase (image "<<n_img<<")* :>" << t.GetString() <<std::endl; 
			Tests::test_save_x(img_H<n_img>(),Name<n_img>("01_phase"),n[0],n[1],false);
			Tests::test_save_x(img_Amp<n_img>(),Name<n_img>("01_amp"),n[0],n[1],false);

			size_t nnn[]={n[0]};
			for(size_t y=0; y< n[1]/2; ++y)
			{
				Real_t *p1=img_Amp<n_img>()+n[0]*y;
				MinusMean(p1,n[0],p1);
				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<1,false>(p1,nnn);
				Vtl::NumAlg::IntTransf::Hartley::fht1D(p1,n[0]);
				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<1,false>(p1,nnn);
				p1[n[0]/2]=0;
				Vtl::NumAlg::IntTransf::Hartley::F::Hartley2Phase<Real_t> pp1(p1);
				Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<1>(nnn,pp1);
			}
			Tests::test_save_x(img_Amp<n_img>(),Name<n_img>("01_amp_linebyline_phase"),n[0],n[1]/2,false);
		}
		void MinusMean(const Real_t *data,size_t len,Real_t *out)
		{
			float mean=0;
			for(size_t x=0; x<len ;++x)
			{
				mean+=data[x];
			}
			mean/=len;
			for(size_t x=0; x<len ;++x)
			{
				//float c=1;
				float c=exp(-Vtl::sqr((x-len/2.0f)/(len/4.0f))/2.0f);
				out[x]=(data[x]-mean)*c;
			}

		}
		void _SkewOffset_2(float* mat)
		{
			const size_t nn=n[0];
			size_t nnn[]={nn};
			std::vector<Real_t> tmp(2*nn);
			Real_t *v1=&(tmp[0]);
			Real_t *v2=v1+nn;
			for(size_t y=0; y< n[1]/2; ++y)
			{
				Real_t *p1=img_Amp<1>()+n[0]*y;
				Real_t *p2=img_Amp<2>()+n[0]*y;
				Real_t *r=img_Tmp()+n[0]*y+n[0]*n[1]/2;
///----------------------------------------

				Vtl::NumAlg::IntTransf::Hartley::F::PhaseCorr<Real_t> pp(p1,p2,r); 
				Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<1>(nnn,pp);

				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<1,false>(r,nnn);
				Vtl::NumAlg::IntTransf::Hartley::ifht1D(r,nn);
				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<1,false>(r,nnn);

			}
			Tests::test_save_x(img_Tmp()+n[0]*n[1]/2,"20_xxxxxxx",nn,n[1]/2,false);
			{
				Vtl::NumAlg::IntTransf::Utils::F::ReflectHalfToOther<2,Real_t,true> pp(img_Tmp()); 
				Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n,pp);
			}

			Tests::test_save_x(img_Tmp(),"20_xx",n[0],n[1],false);

			{
				Real_t angle_from=Real_t(-M_PI/16);
				Real_t angle_to=Real_t(M_PI/16);
				Hough(angle_from,angle_to,img_Tmp(),img_Sino(),n[0],n[1]);
	
				Container<Real_t> cont(n,8);
				Impl::find_offset(cont,img_Sino(),n);
				float phi_t= float((*(cont.begin())).y +n[1]/2);
				//phi_t -=1;
				float phi=(angle_to-angle_from)/n[1] * phi_t + angle_from;
				float rho= float((*(cont.begin())).x);

				mat[0]=1;
				mat[1]=0;
				mat[2]=tan(phi);
				mat[3]=1;

				Real_t mat_rev[]={1,-mat[2],0,1};
				Impl::comp_Rotate(img_H<2>(),img_Tmp(),n,mat_rev);


				Tests::test_save_x(img_Tmp(),"20_phase_2_skew",n[0],n[1],false);

				::memcpy(img_H<2>(),img_Tmp(),sizeof(Real_t)*n[0]*n[1]);
				//std::cout << "Inverse FFT (compute rotation/scale):>" << t.GetString() <<std::endl; 


			}
		}
		static void Hough(Real_t angle_from,Real_t angle_to,Real_t* img,Real_t* sino,size_t width,size_t height)
		{//Hough

			size_t n[]={width,height};
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(img,n);
			Vtl::NumAlg::IntTransf::Hartley::fht2D(img,n[0],n[1]);
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(img,n);

			Vtl::ImAlg::Resamp::LinGrid<Real_t> lng(-Real_t(n[1]/2*1.4),Real_t(n[1]/2*1.4),n[1]);
			//Vtl::ImAlg::Resamp::LinGrid<Real_t> lng(-Real_t(n[1]/2),Real_t(n[1]/2),n[1]);
			Vtl::ImAlg::Resamp::PhiGrid<Real_t> phig(angle_from,angle_to,n[0]);
			Vtl::ImAlg::Resamp::InterpBilinear<Real_t> interp(img,n[0],n[1],n[0],Real_t(0));
			Vtl::ImAlg::Resamp::CartToPolar(interp,Real_t(n[0]/2),Real_t(n[1]/2),sino,n[0],n[1],n[0],lng,phig);

			Tests::test_save_x(sino,"20_xx_polar",n[0],n[1],false);

			typedef Vtl::ImAlg::WndFun::DoG<Real_t,0> Filt;
			Filt f(4,n[0],1);
			Vtl::Plf::Kernel::FKern<Real_t,Vtl::Plf::Kernel::KT_Win,false> ker(f,n[0],Real_t(n[0]/2),Real_t(0));
			for(size_t y=0; y<n[1] ;++y )
			{
				Real_t *py=sino+n[0]*y;
				Vtl::Plf::Kernel::WindowData(py,ker.begin(),ker.end(),1,0);
				for(size_t x=0; x<n[0]; ++x)
				{
					//Real_t c=exp(-Vtl::sqr((x-n[0]/2.0f)/(n[0]/2.0f))/2.0f);
					Real_t c=1;
					c-=exp(-Vtl::sqr((x-n[0]/2.0f)/(4.0f))/2.0f);
					py[x]*=c;
				}
			}

			Tests::test_save_x(sino,"20_xx_polar_f",n[0],n[1],false);
			Real_t *py=sino;
			size_t nn[]={n[0]};
			for(size_t y=0; y<n[1] ;++y, py+=n[0])
			{
				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<1,false>(py,nn);
				Vtl::NumAlg::IntTransf::Hartley::fht1D(py,n[0]);
				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<1,false>(py,nn);
			}
			Tests::test_save_x(sino,"20_xx_sino",n[0],n[1],false);
		}

	};

	/*!
	 * \brief    Reddy-Chatterji \ref Lit_ReddyChatterji ,  \ref Lit_ImgRegBushBroom
	 *             .   ,  ,  
	 *           ,      .    
	 *          .  ,    ( 1.8).
	 */
	template<typename Real_t>
	class Hartley<Real_t,Vtl::ImAlg::Model::Similarity<Real_t> >
		: public Base<Real_t,5>
	{
	public:
		typedef Vtl::ImAlg::Model::Similarity<Real_t> Model;
	private:
		size_t size;
		Vtl::ImAlg::Resamp::LogGrid<float> lsg;
		size_t n[2];
		bool is_edged;
	public:
		virtual ~Hartley()
		{
		}
		Hartley()
			: Base()
		{
		}
		void Init(size_t size, bool is_edged)
		{
			this->size=size;
			this->is_edged=is_edged;
			size_t sq=size*size;
			Storage_t::Offsets_t s={sq,sq,sq,sq,sq};
			// blocks
			// 0 - Image1 | Hartley of Image1 | Hartley of Image1 divided by Amplitude
			// 1 - Image2 | Hartley of Image2 | Hartley of Image2 divided by Amplitude
			// 2 - temporary for : 
			//         Amplitude (Fourier/Hartley - the same) 
			//       | log-polar amplitudes phase correlation 
			//       | its hartley for delta function
			//       | rotated and rescaled Hartley of Image2 divided by Amplitude
			//       | phase correlation
			//		 | its hartley for delta function
			// 3 - Image 1 Amplitude in log-polar
			// 4 - Image 2 Amplitude in log-polar
			//storage_.Init(s);
			Base::Init(s);
			n[0]=Size();
			n[1]=Size();
			//size_t n[]={Size(),Size()};
			lsg=Vtl::ImAlg::Resamp::LogGrid<float>(2.0f,float(std::min(n[0],n[1])/2),n[0]);
		}
		size_t Size() const {return size;}
	private:
		template<unsigned n_img>
		Real_t* img_H()
		{
			//return storage_.block<n_img-1>();
			return _img<n_img-1>();
		}
		Real_t* img_Tmp()
		{
			//return storage_.block<2>();
			return _img<2>();
		}
		template<unsigned n_img>
		Real_t* img_Amp()
		{
			//return storage_.block<n_img+2>();
			return _img<n_img+2>();
		}
				
	public:
		template<unsigned n_img,typename Iter> 
		void Insert(Vtl::Plf::OutPortNoPostProc<Real_t>* input,Iter src_begin,Iter src_end)
		{
			Base::_Insert(input,src_begin,src_end,img_H<n_img>(),Size(),Size(),Size());
			AfterInsert<n_img>();
		}



		void Run(Model& model,int w,int h)
		{
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<Real_t,2,2> MatrixA;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<Real_t,2> VectorB;
			using namespace Vtl::NumAlg::Lin::ublas;

			Real_t cof=model.getCof(w,h);


			Vtl::Dbg::TimeElapsed t;
			t.start();
			Container<Real_t> cont_amp1(n,8);
			find_rotate_scale(cont_amp1);
			t.finish();

			Model::value_type phi= - float(M_PI*(*(cont_amp1.begin())).y)/n[1];
			Model::value_type scale = lsg.values_ratio( (*(cont_amp1.begin())).x );

			//-----------------------------------------------------------------------------------
			std::cout << "Inverse FFT (compute rotation/scale):>" << t.GetString() <<std::endl; 
			Container<Real_t> cont1(n,8);
			Container<Real_t> cont2(n,8);
			Container<Real_t>& cont=cont1;
			//phi=phi
			model.set(phi,scale,Real_t(0),Real_t(0));
			angle_shift_specif(cont1,model);
			Real_t a0=(*(cont1.begin())).value;
			std::cout << "Delta function amplitude a0= "<< a0 << std::endl;
			//phi=phi+180
			model.set(phi+Real_t(M_PI),scale,Real_t(0),Real_t(0));
			angle_shift_specif(cont2,model);
			Real_t a1=(*(cont2.begin())).value;
			std::cout << "Delta function amplitude a1= "<< a1 << std::endl;
			//    phi  phi+M_PI
			if (abs(a1)>abs(a0))
			{
				std::cout << "phi=180 deg+phi " << std::endl;
				cont=cont2;
				phi+=Real_t(M_PI);
			}
			else
			{
				std::cout << "phi=phi " << std::endl;
				cont=cont1;
			}
			//     
			Real_t dx=  Real_t((*(cont.begin())).x);
			Real_t dy=  Real_t((*(cont.begin())).y);
			model.set(-phi,scale,-dx/cof,-dy/cof);
			model.Inverse();
		}
	private:
		template<unsigned n_img> 
		void AfterInsert()
		{
			bool windowing=false;
			bool highpass=false;
			if (!is_edged)
			{
				for(size_t y=0; y< n[1]/2; ++y)
				{
					Real_t *p=img_H<n_img>()+n[0]*y;
					MinusMean(p,n[0],p);
				}
			}
			_RotScaleOffset_1<n_img>(highpass,windowing);
		}
		void MinusMean(const Real_t *data,size_t len,Real_t *out)
		{
			float mean=0;
			for(size_t x=0; x<len ;++x)
			{
				mean+=data[x];
			}
			mean/=len;
			for(size_t x=0; x<len ;++x)
			{
				//float c=1;
				float c=exp(-Vtl::sqr((x-len/2.0f)/(len/4.0f))/2.0f);
				out[x]=(data[x]-mean)*c;
			}
		}
		//--Reddy-Chatterji
		template<unsigned n_img> 
		void _RotScaleOffset_1(bool highpass,bool windowing)
		{
			Vtl::Dbg::TimeElapsed t;
			t.start();
			//window(img_H<n_img>(),n);
			Impl::transform_and_zero_freq(img_H<n_img>(),n);
			//highpass_filtering(img_H<n_img>(),n);
//				Vtl::NumAlg::IntTransf::Hartley::F::Hartley2PhaseAmp<Real_t,amp_in_up_half_plane> pp(hartley_phase,amp);
			Vtl::NumAlg::IntTransf::Hartley::F::Hartley2PhaseAmpDown<2,Real_t> pp(img_H<n_img>(),img_Tmp(),n);
			//Vtl::NumAlg::IntTransf::Utils::ProcPair::Hartley2PhaseAmpDown<2,Real_t> pp(img_H<n_img>(),img_Tmp(),n);
			Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n,pp);
			Vtl::Plf::F11::Log<float> log(0.001);
			Vtl::ImAlg::Tools::copy_1_1(img_Tmp(),n[0],n[1]/2,n[0],log);


			Impl::comp_AmpToPolar(img_Tmp(),img_Amp<n_img>(),n,lsg);

			t.finish();
			std::cout << "compute Amplitude & Phase (image "<<n_img<<")* :>" << t.GetString() <<std::endl; 
			Tests::test_save_x(img_H<n_img>(),Name<n_img>("01_phase"),n[0],n[1],false);
			Tests::test_save_x(img_Tmp(),Name<n_img>("01_amp"),n[0],n[1],false);
			Tests::test_save_x(img_Amp<n_img>(),Name<n_img>("01_amp_logpolar"),n[0],n[1],false);

			t.start();
			if(windowing) 
				Impl::window(img_Amp<n_img>(),n[0],n[1],n[0],Real_t(n[0]/2),Real_t(n[1]/2),Real_t(0));

			Impl::transform_and_zero_freq(img_Amp<n_img>(),n);
			//  -  (highpass)
			if(highpass) 
				Impl::highpass_filtering(img_Amp<n_img>(),n);

			Vtl::NumAlg::IntTransf::Hartley::F::Hartley2Phase<Real_t> ppp(img_Amp<n_img>());
			Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n,ppp);

			t.finish();
			std::cout << "compute Phase of LogPolar Amplitude (image "<< n_img <<")* :>" << t.GetString() <<std::endl; 
			Tests::test_save_x(img_Amp<n_img>(),Name<n_img>("05_phase_hartley_phase"),n[0],n[1],false);
		}
		void angle_shift_specif(Container<Real_t>& cont,const Model& model)
		{
			model.Interp(img_H<2>(),n[0],n[1],n[0],img_Tmp(),n[0],Real_t(0));
			//Tests::test_save_x(storage_.block<2>(),"20_phase_2_rotate",n[0],n[1],false);
			::memcpy(img_Amp<1>(),img_Tmp(),sizeof(Real_t)*n[0]*n[1]);
			//img_Amp<1>()       
//			Vtl::Dbg::TimeElapsed t;
//			t.start();
			// 
			Impl::find_shift(cont,img_H<1>(),img_Amp<1>(),img_Tmp(),n);
//			t.finish();
//			std::cout << "Compute translation :>" << t.GetString() <<std::endl; 
			//Tests::test_save_x(storage_.block<2>(),"21_phase_corr_ofs",n[0],n[1],false);
		}

		void find_rotate_scale(Container<Real_t>& cont_amp1)
		{
			Vtl::Dbg::TimeElapsed t;

				t.start();
				Impl::find_shift(cont_amp1,img_Amp<1>(),img_Amp<2>(),img_Tmp(),n);
				t.finish();
				std::cout << "Compute delta functions for rotation/scale:>" << t.GetString() <<std::endl; 
				Tests::test_save_x(img_Tmp(),"11_phase_delta",n[0],n[1],false);

				std::cout << "===== Peaks for roattion/scale:>" << t.GetString() <<std::endl; 

				for(Container<Real_t>::const_Iter it=cont_amp1.begin(); it!=cont_amp1.end(); ++it)
				{
					Container<Real_t>::Point p=*it;
					float phi=float(p.y);
					phi=(phi/n[1])*180.0f;
					float scale=lsg.values_ratio(p.x);

					std::cout << std::setprecision(3) 
						<< "scale=" << std::setw(6) << scale  
						<<"   phi="<< std::setw(6) << phi 
						<< " deg   a=" << p.value <<std::endl;
				}
		}				
	};



}//namespace Registration
}//namespace ImAlg
}//namespace Vtl
/////////////////////////////////////////////////////////////////////////////////////////
#endif //_IMG_REGISTRATION_H
