#ifndef _IMG_REGISTRATION_OFLOW_H
#define _IMG_REGISTRATION_OFLOW_H
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include "plf_kernels_1d.h"
#include "plf_functors.h"
#include "imalg_tools.h"
#include "storage.h"
#include "img_trans_models.h"
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

namespace Vtl
{
namespace ImAlg
{
namespace Registration
{

	/*!
	 * \brief           
	 *        (Kanade-Lukas \ref Lit_KLT_TK , Shi-Tomasi \ref Lit_KLT_ST, Projective transform \ref Lit_Zokai )
	 *
	 *
	 *******************************************************************************************/
	template<typename T,unsigned ColorPlanes_,bool UseMask_>
	class OFlowAllocStrategy
	{
	public:
		typedef T value_type;
		enum {UseMask=UseMask_};
		enum {ColorPlanes=ColorPlanes_};
		enum {ImgPlane=5 }; //0-I1, 1-I2, 2-I1 transformed 3-gx, 4-gy , 5-mask
		enum {ImgLevel=ImgPlane*ColorPlanes+(UseMask?1:0)};
		typedef Storage<T,ImgLevel> Storage_t;
		struct Size
		{
			size_t w;
			size_t h;
			size_t ofs;
		};
		std::vector<Size> sizes;
		size_t w;
		size_t h;
		size_t ofs;
		size_t level;
	private:
		Storage_t storage_;
	public:
		OFlowAllocStrategy() {}
		virtual ~OFlowAllocStrategy() {}
		void Init(size_t width,size_t height)
		{
			Init(width,height,width,height);
		}
		void Init(size_t width,size_t height,size_t width_min,size_t height_min)
		{
			sizes.resize(0);
			size_t sq=0;
			for( ; width>=width_min && height>=height_min ; width/=2, height/=2 )
			{
				Size s;
				s.w=width;
				s.h=height;
				s.ofs=sq;
				sq+=s.w*s.h;
				sizes.push_back(s);
			}
			Storage_t::Offsets_t s;
			for(int i=0; i< Storage_t::N ; ++i)
			{
				s[i]=sq;
			}
			storage_.Init(s);
			set_Level(0);
		}
		size_t Levels() const { return sizes.size(); }
		void set_Level(size_t level) 
		{ 
			this->level=level;
			w=sizes[level].w;
			h=sizes[level].h;
			ofs=sizes[level].ofs;
		}
		size_t get_Level() const { return level;}
		size_t W() const { return w;}
		size_t H() const { return h;}
		template<unsigned n_plane>
		T* I1()
		{
			return ofs+storage_.block<ImgPlane*n_plane>();
		}
		template<unsigned n_plane>
		T* I2()
		{
			return ofs+storage_.block<ImgPlane*n_plane+1>();
		}
		template<unsigned n_plane>
		T* It()
		{
			return ofs+storage_.block<ImgPlane*n_plane+2>();
		}
		template<unsigned n_plane>
		T* gx()
		{
			return ofs+storage_.block<ImgPlane*n_plane+3>();
		}
		template<unsigned n_plane>
		T* gy()
		{
			return ofs+storage_.block<ImgPlane*n_plane+4>();
		}
		T* mask()
		{
			return ofs+storage_.block<ImgLevel-1>();
		}
	};


	//! \brief  ,   !
	namespace Impl
	{
		//! \brief  ,   !
		namespace OFlow
		{
			template<unsigned n,typename T,class M>
			T G(const M& model,T x,T y,T gx,T gy) 
			{
				return model.Dx<n>(x,y)*gx+model.Dy<n>(x,y)*gy;
			}

			template<unsigned n>
			struct BuildG
			{
				template<class M,class VectorG,typename T> 
				static void vec(VectorG& GG,const M& model,T x,T y,T gx,T gy)
				{
					GG[n-1]=G<n-1>(model,x,y,gx,gy);
					BuildG<n-1>::vec(GG,model,x,y,gx,gy);
				}
			};
			template<>
			struct BuildG<0>
			{
				template<class M,class VectorG,typename T> 
				static void vec(VectorG& GG,const M& model,T x,T y,T gx,T gy)
				{
				}
			};

			template<bool UseMask,class MatrixH,class VectorB>
			class BuildHB
			{
				MatrixH& H;
				VectorB& B;
			public:
				typedef typename VectorB::value_type value_type;
			public:
				BuildHB(MatrixH& H_,VectorB& B_) : H(H_),B(B_) 
				{
					H = Vtl::NumAlg::Lin::ublas::zero_matrix<MatrixH::value_type>( H.size1(), H.size2() );
					B = Vtl::NumAlg::Lin::ublas::zero_vector<VectorB::value_type>( B.size() );
				}
				template<typename T>
				void operator()(VectorB& G,T deltaI,T mask)
				{
					B+=Vtl::Alg::mult_if<UseMask>(G*deltaI,mask);
					H+=Vtl::Alg::mult_if<UseMask>(outer_prod(G,G),mask);
				}
			};
			template<bool UseMask,class VectorB>
			class BuildB
			{
				VectorB& B;
			public:
				typedef typename VectorB::value_type value_type;
			public:
				BuildB(VectorB& B_) : B(B_) 
				{
					B = Vtl::NumAlg::Lin::ublas::zero_vector<VectorB::value_type>( B.size() );
				}
				template<typename T>
				void operator()(VectorB& G,T deltaI,T mask)
				{
					B+=Vtl::Alg::mult_if<UseMask>(G*deltaI,mask);
				}
			};


			template<typename T>
			class OFlowHelper
			{
			public:
				template<typename S>
				static void ins(S* in,int w, int h,int nls,T* data)
				{
					Vtl::Plf::F11::ConvertMult<T> to_T(T(1./255));
					Vtl::ImAlg::Tools::copy_1_1(in, w, h, nls, data, w, to_T);
				}
#if 0
				/*!
				*     .     
				*  (   )
			 *
			 ****************************************************************************/
				static void ins(T* in,int w, int h,int nls,T* data,int w_new,int h_new)
				{
					T mat[4]={float(w_new)/float(w),0,0,float(h_new)/float(h)};//Scale matrix
					Vtl::ImAlg::Resamp::InterpBilinear<Img::Pixel> interp((*img)[0],w,h,nls);
					float x0=float(w/2);
					float y0=float(h/2);
					float x1=float(w_new/2);
					float y1=float(h_new/2);
					Vtl::ImAlg::Resamp::Affine(interp,x0,y0,x1,y1,mat,data,w_new,h_new,w_new);
				}
#endif
				/*!
				*  data   2.
				*         
				*  1  2   . \n
				*         ,  
				*            
				*           .
			 *
			 ****************************************************************************/
				static void dif(T* data1,T* data2,T* gx,T* gy,int w, int h,float sigma)
				{
					Vtl::Plf::F11::DoNothing<T> no;
					Vtl::ImAlg::Conv::FastGaussConvolve<float,0> g0(sigma);
					Vtl::ImAlg::Conv::FastGaussConvolve<float,1> g1(sigma);

					Vtl::ImAlg::Tools::copy_1_1(data2, w, h, w, gx, w, no);
					Vtl::ImAlg::Tools::copy_1_1(data2, w, h, w, gy, w, no);
					Vtl::ImAlg::Conv::Conv2D::conv2DInplace(gx,w,h,w,g1,g0);
					Vtl::ImAlg::Conv::Conv2D::conv2DInplace(gy,w,h,w,g0,g1);
					Vtl::ImAlg::Conv::Conv2D::conv2DInplace(data2,w,h,w,g0,g0);
					Vtl::ImAlg::Conv::Conv2D::conv2DInplace(data1,w,h,w,g0,g0);

					Tests::SaveImage("a_data1.tif",data1,w,h,w,T(0),T(255));
					Tests::SaveImage("a_data2.tif",data2,w,h,w,T(0),T(255));
					Tests::SaveImage("a_gx.tif",gx,w,h,w,T(128/(4.*255)),T(4*255));
					Tests::SaveImage("a_gy.tif",gy,w,h,w,T(128/(4.*255)),T(4*255));
				}
				template<class Model,class BuildX>
				static void build(const Model& model,BuildX& HB,T* I1,T* I2,T* gx,T* gy,size_t w,size_t h,T* mask,float sigma,T fill)
				{
					using namespace Vtl::NumAlg::Lin::ublas;
					c_vector<typename BuildX::value_type,Model::N> GG;

					T cof=T(2./std::max(w,h));
					for(size_t iy=0; iy<h ; ++iy)
						for(size_t ix=0; ix<w ; ++ix)
						{

							T x=T(int(ix)-int(w/2))*cof;
							T y=T(int(iy)-int(h/2))*cof;
							size_t iofs=iy*w+ix;
							if(I1[iofs] != fill)
							{
								BuildG<Model::N>::vec(GG,model,x,y,gx[iofs],gy[iofs]);
								GG/=cof;					
								HB(GG,I1[iofs]-I2[iofs],mask[iofs]);
							}
						}
				}

			};
		} //namespace OFlow
	}//namespace Impl


	/*!
	 * \brief     , .  
	 *         Vtl::ImAlg::Registration::Model   AllocStrategy.
	 *
	 *******************************************************************************************/
	template<class Model,class AllocStrategy >
	class OFlow
	{
		Model& model;
		AllocStrategy& mem;
	public:
		typedef typename AllocStrategy::value_type T;
		enum {UseMask=AllocStrategy::UseMask};
		enum {ColorPlanes=AllocStrategy::ColorPlanes};
	public:
		OFlow(Model& model_, AllocStrategy& mem_): model(model_), mem(mem_)
		{
		}
		virtual ~OFlow()
		{
		}
		template<class D>
		void Insert1(D* data,int w,int h,int nls)
		{
			mem.set_Level(0);
			if(w!=mem.W() || h!=mem.H())
				throw "Error OFlow::Insert1()";

			Impl::OFlow::OFlowHelper<T>::ins(data,w,h,nls,mem.I1<0>());
		}
		template<class D>
		void Insert2(D* data,int w,int h,int nls)
		{
			mem.set_Level(0);
			if(w!=mem.W() || h!=mem.H())
				throw "Error OFlow::Insert1()";

			Impl::OFlow::OFlowHelper<T>::ins(data,w,h,nls,mem.I2<0>());
		}

		void AfterInsert()
		{
			//T fill=T(396./255.);
			T fill=T(256./255.);
			using namespace Impl::OFlow;
			using namespace Vtl::NumAlg::Lin;
			using namespace Vtl::NumAlg::Lin::ublas;
			//T sigma=2.0f;
			T sigma=1.2f;
			const int w=mem.W();
			const int h=mem.H();
			const int w2=w/2;
			const int h2=h/2;
			const int cof=std::max(w,h)/2;

			
			typedef c_matrix<double,Model::N,Model::N> MatrixH;
			typedef c_vector<double,Model::N> VectorB;
			Model model_direct(model);
			Model model_old(model);
			Model model_inverse(model);
			model_inverse.Inverse();
			MatrixH H;
			VectorB B;

			//std::cout << "dif" << std::endl;
			std::cout << "image " << w << "x" << h << std::endl;
			for(int iii=0 ; iii<2 ; ++iii,sigma*=1.5f)
			{
				OFlowHelper<T>::dif(mem.I1<0>(),mem.I2<0>(),mem.gx<0>(),mem.gy<0>(),mem.W(),mem.H(),sigma);

				BuildHB<UseMask,MatrixH,VectorB> to_build(H,B);
				OFlowHelper<T>::build(model_direct,to_build,mem.I1<0>(),mem.I2<0>(),mem.gx<0>(),mem.gy<0>(),mem.W(),mem.H(),mem.mask(),sigma,fill);

				MatrixH::value_type multiplier=matrix_max_abs_element(H);
				H/=multiplier;
				B/=multiplier;
				MatrixH Hinv;
				svd::inverse_save(Hinv,H,1.e-16);

				for(int iter=0 ; iter<15 ; ++iter)
				{
					std::cout << "iter " << iter << std::endl;
					model_old=model_direct;

					VectorB d_theta=prod(Hinv,B);


					model_direct.set(d_theta);
					model_direct.Inverse();
					model_inverse.UpdateByNewAfterThis(model_direct);
					model_direct=model_inverse;
					model_direct.Inverse();

					T err=cof*model_direct.max_err_at_rect(model_old,-1.0f,-1.0f,1.0f,1.0f);
					std::cout << "err=" << err << std::endl;


					model_inverse.Interp(mem.I1<0>(),w, h,w,mem.It<0>(),w,fill);

					{
						std::ostringstream os;
						os.fill('0');
						os<< "a_res_" << std::setw(2) << iter <<".tif";
						Tests::SaveImage(os.str(),mem.It<0>(),w,h,w,T(0),T(255));
					}

					BuildB<UseMask,VectorB> to_build(B);
					OFlowHelper<T>::build(model_direct,to_build,mem.It<0>(),mem.I2<0>(),mem.gx<0>(),mem.gy<0>(),mem.W(),mem.H(),mem.mask(),sigma,fill);
					B/=multiplier;
					if(err<0.005)
						break;
				}
			}
			//T err=model_direct.max_err_at_rect(model,-1.0f,-1.0f,1.0f,1.0f);
			//std::cout << "Total err=" << err*cof << std::endl;

			model=model_direct;
		}
	};
}//namespace Registration
}//namespace ImAlg
}//namespace Vtl
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_IMG_REGISTRATION_OFLOW_H
