#ifndef _WINDOWING_FUN_H 
#define _WINDOWING_FUN_H
/////////////////////////////////////////////////////////////////////////////
#include <vector>
#include <algorithm>
#define _USE_MATH_DEFINES
#include <cmath>
#include <assert.h>
#include "loki_static_check.h" 
#include "loki_TypeManip.h"
#include "alexandresku.h"
#include "spec_functions.h"
namespace Vtl
{
namespace ImAlg
{
	/*!
	 * \brief ,       .       ,
	 *   ,  ( , ..  )     , . ,
	 *      .
	 *
	 *     (     ) 
	 *          ,      1. 
	 *  " "   ,          
	 *    .  -        
	 *    ,    .
	 * 
	 *     (enum),    Base: \n
	 * enum { finit_support_left };        .\n
	 * enum { finit_support_right=FSR };        .\n
	 * enum { Deriv=D }; 0,   ,   ,    
	 * ( 2- )    .     
	 *        .\n
	 * enum { conv_norm_alg=NAC };        . 
	 *     D,        Vtl::Plf::Kernel::FKern.
	 *
	 *
	 *   : \n
	 * -    Vtl::Plf::Kernel::F::Traits<T,FS,NAC,D,SpecCentr>
	 * -    Name,  : static const char * Name() { return "Hamming"; }
	 * -   T operator(T x) const;        x.
	 * -      (  ). 
	 * -   
	 * \code
	 * T center() const
	 * T left() const
	 * T right() const
	 * \endcode
	 *     x,      ,   . 
	 *        .       (.  ),
	 *        ,      ,  .
	 *     -  ,          .
	 *   ,         ,   
	 *       left()  right()   ,    
	 *  .       ,       ,   
	 *     ,   .    .  ,     
	 *      .    ,         , 
	 *  ,          ,     .   
	 *  ,         ,     ,
	 *       (  ,     ( .\n
	 *  "" ,         (right()+left())/2.     ,   
	 *        .    -    
	 *     ,       .
	 *
	 *
	 *
	 * 
	 * 
	 * <b>:</b>
	 *     //  
	 *      ,     
	 *  .    .  Vtl::Plf::Kernel::F::Gauss.    
	 *          
	 *   -  ,    
	 *   .   ,    
	 *        Maple, 
	 *    9.5.        
	 *   Maple   5   .
	 * 
	 * \code
	 *    
	 * > restart;
	 * 
	 * ,   
	 * > assume(x,real,sigma,real,t,real,sigma>0,t>0); 
	 * 
	 *  ,   
	 * > g(x):=exp(-x^2/(2*sigma^2));
	 * 
	 *  5   x   g(x).
	 * > res := diff(g(x),x\$5);
	 * 
	 * > res := simplify(res); 
	 * 
	 * ,      t=(x/sigma)^2? 
	 *      .
	 * > res := algsubs( x^2=t*sigma^2,res);
	 * 
	 * 
	 * > res := simplify(res);
	 * 
	 * ,    x , ..  
	 *   ,   ,    
	 *    t     t  x. 
	 *    :
	 * > res := simplify(res*sigma^7);
	 * \endcode
	 * 
	 *           Copy/Paste.
	 *      ,     
	 *  :       
	 *    Maple "x^3"   ++ "x*x*x".  
	 *   Maple,      
	 *  :
	 * \code
	 * res := -exp(-1/2*t)*x*(-10*t+t^2+15)
	 * \endcode
	 * 
	 * <b>  :</b>
	 *      ( )  \ref Lit_Alexandresku
	 *    Loki.
	 *
	 *            Vtl::Plf::Kernels     
	 *  ,    ,     ,   
	 *          ,
	 *       .  ,     
	 *      (        
	 *         ,       
	 *        .      
	 *    Gauss     Hann, Hamming, Blackman, Nuttall ...
	 * 
	 ***************************************************************************************/
	namespace WndFun
	{
		/*!  Traits_common -    1D   .
		 *  - FS = 1,    , 0 -  ,   .
		 *  - FSL = true,       , false -  .
		 *  - FSR = true,       , false -  .
		 *  - NAC  ,      ()   Vtl::Plf::Kernel::Base   ,  NAC=D
		 *  - D=0 ,     ,    -   .
		 * 
		 * 
		 ***************************************************************************************/
		template<typename T,bool FSL,bool FSR,unsigned NAC,unsigned D> struct Traits_common
		{
			typedef T value_type;
			enum { finit_support=FSL };
			enum { finit_support_left=FSL };
			enum { finit_support_right=FSR };
			enum { conv_norm_alg=NAC };
			enum { Deriv=D };
		public:
			template<class Functor>
			static std::string FullName()
			{
				std::ostringstream os;
				if(Functor::Deriv>0)
					os << (d/dx)^<<Functor::Deriv << ' ';
				os << Functor::Name();
				return os.str();
			}
			template<class Functor>
			static std::string Properties()
			{
				std::ostringstream os;
				os << Functor::finit_support? "Finite" : "Infinite" << " support, "
				   << "normalization algorithm for convolution="<<Functor::conv_norm_alg;
				Properties(os,Loki::Int2Type<Functor::finit_support>() );
				return os.str();
			}
		private:
			template<class Functor>
			static void Properties(std::ostringstream &os,Loki::Int2Type<0> ){}
			template<class Functor>
			static void Properties(std::ostringstream &os,Loki::Int2Type<1> )
			{
				std::ostringstream os;
				os << Functor::spec_center? "\"Center\" is NOT in window center" : "\"Center\" is in window center, standard";

				return os.str();
			}
		protected:
			~Traits_common() {}
		};
		/*!  Traits -    1D   .
		 *  - FS = 1,    , 0 -  ,   .
		 *  - NAC  ,      ()   Vtl::Plf::Kernel::Base   ,  NAC=D
		 *  - D=0 ,     ,    -   .
		 *  - SpecCentr=0,  ""     , =1 -    
		 * 
		 * 
		 ***************************************************************************************/
//		template<typename T,unsigned FS,unsigned NAC,unsigned D,unsigned SpecCentr> struct Traits;
		template<typename T,unsigned FSL,unsigned FSR,unsigned NAC,unsigned D> struct Traits;
		template<typename T,unsigned NAC,unsigned D> struct Traits<T,false,false,NAC,D>
		: public Traits_common<T,false,false,NAC,D> 

		{
/*
//		private:
			const T half_size;
			const T n_half_width;
			const T size;
			const T center;
			const T left;
			const T right;
//
		public:
			Traits(T half_size_,T n_half_width_,T center_at) 
				: half_size(half_size_),n_half_width(n_half_width_),center(center_at)
			    , size(half_size_*n_half_width_+1),left(center-size/2),right(size-1) {}
*/
		public:
			Traits(T half_size_,T n_half_width_,T center_at) {}
		protected:
			~Traits() {}
		private:
			Traits();
		};
		template<typename T,unsigned NAC,unsigned D> struct Traits<T,true,true,NAC,D>
		: public Traits_common<T,true,true,NAC,D> 
		{
/*		private:
			const T _size;
			const T _center;
			const T _left;
			const T _right;
		public:
*/
			Traits(T size_,T center_) /*: _size(size_) ,_center(center_), _left(center_-size_/2),_right(size_-1)*/ {}
		protected:
			~Traits() {}
		private:
			Traits();
		};

		/*!
		 * \brief  -    .  <b> </b>   
		 *     <b>      </b>! 
		 *     ,    
		 *  ,      class Direct
		 ********************************************************************************/
		template<typename T,typename TT,unsigned D=0,unsigned NAC=0> class Direct_Ref
			: public Traits<T,true,true,NAC,D>
		{
		public:
			typedef int arg_type; 
		private:
			const arg_type _size_;
			const TT* vals;
		public:
			static const char * Name() { return "HannDirect_Ref"; }
			Direct_Ref(int size_,int center_,const TT* values) : Traits<T,true,true,NAC,D>(T(size_),T(center_)),_size_(size_), vals(values) {}
			template<class TT> T operator()(TT x) const
			{
				assert( int(x)>=0 && int(x)< _size_ );
				return T(vals[ int(x) ]);
			}
			arg_type size() const {return _size_;}
			arg_type center() const {return arg_type((_size_-1)/2.0f);}
			arg_type left() const {return 0;}
			arg_type right() const {return _size_-1;}
		};
		/*!
		 * \brief    .
		 *
		 ********************************************************************************/
		template<typename T> struct Laws5
		{
			static const int  * Level()
			{
				static int d[]={ 1, 4,6, 4, 1};
				return d;
			}
			static const int  *Edge()  
			{
				static int d[]={-1,-2,0, 2, 1};
				return d;
			}
			static const int  *Spot()
			{
				static int d[]={-1, 0,2, 0,-1};
				return d;
			}
			static const int  *Ripple()
			{
				static int d[]={ 1,-4,6,-4, 1};
				return d;
			}
			struct L5 //Level
				: public Direct_Ref<T,int,0,0>
			{
				static const char * Name() { return "Laws.L5"; }
				L5() : Direct_Ref(5,2,Level()) {}
			};
			struct E5 //Level
				: public Direct_Ref<T,int,1,1>
			{
				static const char * Name() { return "Laws.E5"; }
				E5() : Direct_Ref(5,2,Edge()) {}
			};
			struct S5 //Level
				: public Direct_Ref<T,int,2,2>
			{
				static const char * Name() { return "Laws.S5"; }
				S5() : Direct_Ref(5,2,Spot()) {}
			};
			struct R5 //Level
				: public Direct_Ref<T,int,2,2>
			{
				static const char * Name() { return "Laws.R5"; }
				R5() : Direct_Ref(5,2,Ripple()) {}
			};
		};

		/*!
		 * \brief  -    .
		 *
		 ********************************************************************************/
		template<typename T,typename TT,unsigned D=0,unsigned NAC=0> class Direct
			: public Traits<T,true,true,NAC,D>
		{
			std::vector<TT> vals;
		public:
			static const char * Name() { return "Direct"; }
			Direct(int size_,int center_,const TT* values) : Traits(T(size_),T(center_)) 
			{
				vals.assign(values,values+size_); 
			}
			template<class TT> T operator()(TT x) const
			{
				assert( int(x)>=0 && int(x)< size );
				return T(vals[ int(x) ]);
			}
		};
		/*!
		 * \brief   . 
		 *   
		 *
		 * \f[ w(x)= \left\lbrace   \begin{array}{cc}
		 *		1 , \qquad if \quad 0 \le x < size-1  \\ 
		 *		0 , \qquad else 
		 *	\end{array} \right. \f] 
		 *
		 *
		 *   .
		 *
		 *****************************************************************************/
		template<typename T,unsigned D> class Rect
			: public Traits<T,true,true,D,D>
		{
		public:
			static const char * Name() { return "Rect"; }
			template<class TT> 
			Rect(TT size_) : Traits(T(size_),T(size_/2.0f)) {}
//			template<class TT> 
//			Rect(TT size_, const Rect& ) : Traits(T(size_),T(size_/2.0f)) {}
			template<class TT> T operator()(TT x) const
			{
				return result( T(x), Loki::Int2Type<D>() );
			}
		private:
			T result(T x,Loki::Int2Type<0>) const
			{
				return T(1);
			}
		};
		/*!
		 * \brief   (Bartlett - ). 
		 *   
		 *
		 * \f[ w(x)= \left\lbrace   \begin{array}{cc}
		 *		{ 2 x \over size-1 } , \qquad if \quad 0 \le x < {size-1 \over 2} \\ 
		 *		2- {2 x \over size-1 } , \qquad if \quad {size-1 \over 2} \le x \le size-1  > 0
		 *	\end{array} \right. \f] 
		 *
		 *    : \n
		 * \f[ w(x)= \left\lbrace   \begin{array}{cc}
		 *		{ 2 \over size-1 } , \qquad if \quad 0 \le x < {size-1 \over 2} \\ 
		 *		-{2 \over size-1 } , \qquad if \quad {size-1 \over 2} \le x \le size-1  > 0
		 *	\end{array} \right. \f] 
		 *
		 *
		 *      .
		 *
		 *****************************************************************************/
		template<typename T,unsigned D> class Bartlett
			: public Traits<T,true,true,D,D>
		{
		public:
			typedef T arg_type; 
		private:
			const T c;
			const arg_type _size_;
		public:
			static const char * Name() { return "Bartlett"; }
			template<class TT> 
			Bartlett(TT size_) : Traits(T(size_),T(size_/2.0f)),_size_(arg_type(size_)), c(T(2./(size_-1))) {}
//			template<class TT> 
//			Bartlett(TT size_, const Bartlett& ) : Traits(T(size_),T(size_/2.0f)), c(T(2./(size_-1))) {}
			template<class TT> T operator()(TT x) const
			{
				return result( T(x), Loki::Int2Type<D>() );
			}
			arg_type size() const {return _size_;}
			arg_type center() const {return arg_type((_size_-1)/2.0f);}
			arg_type left() const {return 0;}
			arg_type right() const {return _size_-1;}
		private:
			T result(T x,Loki::Int2Type<0>) const
			{
				T v=x*c;
				return T((v<1 ? v : 2-v));
			}
			T result(T x,Loki::Int2Type<1>) const
			{
				T v=x*c;
				return T( v < 1 ? c : (v>1 ? -c :0 ) );
			}
		};
		namespace Impl
		{
			namespace Gauss
			{
					template<unsigned n,unsigned m> struct c_mod
					{
						enum {v=c_mod<n-1,m-1>::v+(m+1)*c_mod<n-1,m+1>::v };
					};
					template<unsigned n> struct c_mod<n,-1>
					{
						enum {v=0 };
					};
					template<unsigned n> struct c_mod<n,n>
					{
						enum {v=1 };
					};
				template<unsigned n,unsigned m> int c()
				{
					return Vtl::Alg::minus_if_odd<((n-m)/2)&1>(int(c_mod<n,m>::v));
				}

				template<typename T,unsigned n,unsigned m> struct P
				{
					static T v(T x2) { return P<T,n,m+2>::v(x2)*x2+c<n,m>(); }
				};
				template<typename T,unsigned n> struct P<T,n,n>
				{
					static T v(T x2) { return T(c<n,n>()); }
				};
				template<unsigned n,typename T> T Q(T x2,T x)
				{
					return Vtl::Alg::mult_if_odd<n>(P<T,n,(n&1)>::v(x2),-x);
				}
			}
		}
		/*!
		 *
		 *
		 * \brief   (Gauss).   
		 * <TABLE WIDTH="100%" ALIGN="CENTER"  border=0><TR VALIGN="MIDDLE"><TD ALIGN="CENTER" NOWRAP>
		 * \f$ w(x)= exp( - {x^2 \over 2 \sigma^2 } ) \f$    
		 * </TD><TD WIDTH=10 ALIGN="RIGHT">   (1)   </TD></TR></TABLE>
		 * 
		 *   <typename T,unsigned D,bool norm_deriv=false> :
		 *  - T -       (   )
		 *  - D -  , 0 -  , 1 -  ...
		 *  - norm_deriv -  true,      (1),  false, 
		 *      (1)       ( \f$ 1/\sigma^D \f$).
		 *       -  ,    norm_deriv=false   
		 *       ,        .
		 *
		 * <b>       </b> 
		 *
		 *             
		 *  (      ),    
		 *           .
		 *
		 *   
		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=0><TR VALIGN="MIDDLE"><TD ALIGN="CENTER" NOWRAP> 
		 * \f$ G(\xi) = exp( - \xi^2/2) \f$     
		 * </TD><TD WIDTH=10 ALIGN="RIGHT">   (2)   </TD></TR></TABLE>
		 * 
		 * :
		 * 
		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=0><TR VALIGN="MIDDLE"><TD ALIGN="CENTER" NOWRAP> 
		 * \f$ w(x)= G(x / \sigma) \f$    
		 * </TD><TD WIDTH=10 ALIGN="RIGHT">   (3)   </TD></TR></TABLE>
		 *
		 *   :
		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=0><TR VALIGN="MIDDLE"><TD ALIGN="CENTER" NOWRAP> 
		 * \f$ {d w(x) \over d x } = { d G( \xi ) \over d \xi} { d \xi \over d x} = {1 \over \sigma} {d G \over d \xi}({x \over \sigma }) \f$    
		 * </TD><TD WIDTH=10 ALIGN="RIGHT">   (4a)   </TD></TR></TABLE>
		 *
		 * \f[ {d^2 w(x) \over d x^2 } =  {1 \over \sigma^2} {d^2 G \over d \xi^2}({x \over \sigma }) \f]    
		 * ...
		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=0><TR VALIGN="MIDDLE"><TD ALIGN="CENTER" NOWRAP> 
		 * \f$ {d^n w(x) \over d x^n } =  {1 \over \sigma^n} {d^n G \over d \xi^n}({x \over \sigma }) \f$    
		 * </TD><TD WIDTH=10 ALIGN="RIGHT">   (4)   </TD></TR></TABLE>
		 *          ( ) 
		 *   G.
		 *
		 *      \f$ G(\xi) \f$ .
		 *
		 * \f[ G^{(1)}(\xi)={d G(\xi) \over d \xi } = - { 2 \xi  \over 2 } exp( - \xi^2 / 2) = - \xi G(\xi) \f]    
		 * \f[ G^{(2)}(\xi)={d^2 G(\xi) \over d \xi^2 } = - G(\xi) - \xi G^{(1)}(\xi) = ( \xi^2 - 1)G(\xi) \f]    
		 *   
		 *
		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=0><TR VALIGN="MIDDLE"><TD ALIGN="CENTER" NOWRAP> 
		 * \f$ G^{(n)}(\xi)={d^n G(\xi) \over d \xi^n } = Q_n(\xi) G(\xi)  \f$    
		 * </TD><TD WIDTH=10 ALIGN="RIGHT">   (5)   </TD></TR></TABLE>
		 *
		 *  \f$ Q_n(\xi) \f$ -  n   \f$ \xi \f$.
		 *   (5)  (4),(4a) , 
		 * \f[ G^{(n+1)}(\xi)={d^{(n+1)} G(\xi) \over d \xi^n } = Q_n^{(1)}(\xi) G(\xi)+Q_n(\xi) G(\xi)^{(1)} = (Q_n^{(1)}(\xi) - \xi Q_n(\xi) )G(\xi)  \f]    
		 *
		 * :
		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=0><TR VALIGN="MIDDLE"><TD ALIGN="CENTER" NOWRAP> 
		 * \f$ Q_{n+1}(\xi) = - \xi Q_n(\xi) + Q_n^{(1)}(\xi)  \qquad , \qquad Q_0(\xi)=1 \f$           
		 * </TD><TD WIDTH=10 ALIGN="RIGHT">   (6)   </TD></TR></TABLE>
		 *        \f$ Q_{n}   \quad \forall n \f$.
		 *      ,        .
		 *
		 * 1)  \f$ Q_{n}(\xi) \f$        \f$ \xi \f$  
		 *    n    -   n.     [n/2]+1 . ,
		 *   \f[ Q_0(\xi) = 1 \f]
		 *   \f[ Q_1(\xi) = - \xi Q_0(\xi)+ Q_0'(\xi) = -\xi + 0 = - \xi \f]
		 *   \f[ Q_2(\xi) = - \xi Q_1(\xi)+ Q_1'(\xi) = -\xi (-\xi) - 1 =  \xi^2 - 1 \f]
		 *   (6)  ,   \f$ Q_n(\xi) \f$    
		 *    ,     (6)      -m , 0<=m<=n 
		 *   m+1   ,    (6) -   m   m-1 >=0, c  
		 *   m.
		 *
		 * 2)      :    \f$ Q_{n}(\xi) \f$ 
		 *        (     , . . 1). 
		 *      ,        enum   
		 *           \f$ Q_{n}(\xi) \f$ 
		 *      ,    runtime \ref Lit_Alexandresku
		 *
		 * 3)  \f$ Q_{n}(\xi) \f$    \f$ \xi^n \f$  \f$ c_{n,n} = (-1)^n \f$,      (6).
		 *
		 * 4)        \f$ c_{n,n} = (-1)^n \f$  
		 *     \f$ c_{n,n-2j} = (-1)^{n-j}|c_{n,n-2j}| \quad , \qquad j=0..[{n \over 2}]+1 \f$
		 *
		 * 5)    \f$ Q_{n}(\xi) = c_{n,n} \xi^n + c_{n,n-2} \xi^{n-2} + c_{n,n-4} \xi^{n-4} + \dots \f$
		 *       (6)      :
		 *    \f[ c_{n+1,m+1} = - c_{n,m} + (m+2) c_{n-1,m+2}\f]
		 *    ,  
		 *    \f[ c_{n,m} = - c_{n-1,m-1} + (m+1) c_{n-1,m+1}\f]
		 *       :
		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=0><TR VALIGN="MIDDLE"><TD ALIGN="CENTER" NOWRAP> 
		 *    \f$ |c_{n,m}| = |c_{n-1,m-1}| + (m+1) |c_{n-1,m+1}| \quad , \quad |c_{n,n}|=1 \quad , \quad |c_{n,-1}|=0 \f$
		 * </TD><TD WIDTH=10 ALIGN="RIGHT">   (7)   </TD></TR></TABLE>
		 *              
		 *    Vtl::Plf::Kernel::F::Impl::Gauss::c_mod <n,m>;
		 *      e -  
		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=0><TR VALIGN="MIDDLE"><TD ALIGN="CENTER" NOWRAP> 
		 *    \f$ c_{n,m} = (-1)^n (-1)^{[(n-m)/2]} |c_{n,m}| = (-1)^{[(n-m)/2]+n} |c_{n,m}|  \f$
		 * </TD><TD WIDTH=10 ALIGN="RIGHT">   (8)   </TD></TR></TABLE>
		 *     
		 *    Vtl::Plf::Kernel::F::Impl::Gauss int c<n,m>();
		 *
		 * 6)   1, 3, 4  \f$ Q_{n}(\xi) \f$   
		 * \f[ Q_{n}(\xi) = P_{n}(\xi^2) \left\lbrace   \begin{array}{cc} 
		 *		-\xi , \qquad if \quad n\%2=1  \\ 
		 *		1 , \qquad if \quad n\%2=0 
		 *	  \end{array}   \right. \f] 
		 *
		 *    \f$ P_{n}(\eta) \f$ 
		 *   -    \f$ \eta^{[{n over 2 }]-j} \qquad j=0 \dots [{n over 2}] \f$ 
		 *   -   \f$ \eta^{[{n over 2 }]} \f$  1 
		 *   -   
		 *     \f$ P_n(\eta)=|c_{n,n}| \eta^{[{n over 2 }]} - |c_{n,n}| \eta^{[{n over 2 }]-1} + |c_{n,n}| \eta^{[{n over 2 }]-2} - \dots \f$ 
		 *	        Vtl::Plf::Kernel::F::Impl::Gauss::P <T,n,m>.
		 *   -     \f$ |c_{n,n}| \f$    Vtl::Plf::Kernel::F::Impl::Gauss:: <n,m>.
		 *   
		 *  ,   \f$ Q_n(\xi) \f$        
		 *  template<unsigned n,typename T> T Vtl::Plf::Kernel::F::Impl::Gauss::Q(T x2,T x),    
		 *   x*x,   -  x,           (5).
		 *   ,      <b>  </b>,
		 *      \f$ |c_{n,n}| \f$     
		 *  \f$ P_n(\eta)  \f$ , \f$ Q_n(\xi) \f$      (5)
		 *              
		 *       ,      D  
		 *  Vtl::Plf::Kernel::Gauss.  runtime     ,    .
		 *
		 *
		 *****************************************************************************/
		template<typename T,unsigned D,bool norm_deriv=false> class Gauss
			: public Traits<T,false,false,D,D>
		{
		public:
			typedef T arg_type;
		private:
			const T mult;
			const T sigma;
			const T half_width;
		public:
			static const char * Name() { return "Gauss"; }
			template<class TT,class TTT> 
			Gauss(TT sigma_,TTT N_sigmas_) 
				: Traits<T,false,false,D,D>(T(sigma_),T(N_sigmas_),T(0)) , sigma(T(sigma_)), half_width(T(sigma_*N_sigmas_)) , mult(T(1.0f/Vtl::Alg::x_to_N<D>::value(sigma_))) {}
//			template<class TT> 
//			Gauss(TT new_wnd_size,const Gauss& obj) 
//				: Traits(T(new_wnd_size)/2/obj.n_half_width,obj.n_half_width) , mult(T(1./Vtl::Alg::x_to_N<D>::value(half_size))), wnd(new_wnd_size,obj.wnd) {}

			template<class TT> T operator()(TT x) const
			{
				T t=Vtl::sqr(T(x)/sigma);
				return T(Vtl::Alg::mult_if<norm_deriv>(Impl::Gauss::Q<D,double>(t,double(x))*exp(-t/2.),mult));
			}
			arg_type size() const {return right()-left();}
			arg_type center() const {return arg_type(0);}
			arg_type left() const {return -arg_type(half_width);}
			arg_type right() const {return arg_type(half_width);}
		};
		/*!
		 *
		 * \brief      ().   \f$ Dog(x)= e^{x^2 \over 2 \sigma_2^2} - e^{x^2 \over 2 \sigma_1^2} \f$ .
		 *
		 *   ,   DoG   .
		 *  3 : \n
		 * \param sigma_small -  ,   . (\f$ \sigma_1^2 \f$). 
		 * \param sigma_large -  ,    (\f$ \sigma_2^2 \f$). 
		 * 
		 * \param N_sigmas_large -     (   .
		 *
		 *  ,     size=2*sigma_large*N_sigmas_large+1 , 
		           
		 * size=half_wnd_size+1+half_wnd_size  half_wnd_size=sigma_large*N_sigmas_large.
		 *
		 *****************************************************************************/
		template<typename T,unsigned D> class DoG
			: public Traits<T,false,false,D,D>
		{
		public:
			typedef T arg_type;
		private:
			double large2small;
			const T sigma;
			const T half_width;
		public:
			static const char * Name() { return "DoG"; }
			template<typename T1,typename T2,typename T3> 
			DoG(T1 sigma_small,T2 sigma_large,T3 N_sigmas_large) 
				: Traits(T(sigma_large),T(N_sigmas_large),T(0)), sigma(T(sigma_large)),half_width(T(sigma_large*N_sigmas_large)), large2small(double(sigma_large)/double(sigma_small)) {}
			template<class TT> T operator()(TT x) const
			{
				return result( T(x), Loki::Int2Type<D>() );
			}
			arg_type size() const {return right()-left();}
			arg_type center() const {return arg_type(0);}
			arg_type left() const {return -arg_type(half_width);}
			arg_type right() const {return arg_type(half_width);}
		private:
			T result(T x,Loki::Int2Type<0>) const
			{
				T t2=T(x)/sigma;
				T t1=T(Vtl::sqr(t2*large2small)); 
				t2=Vtl::sqr(t2); 
				return T(exp(-t2/2.)-exp(-t1/2.));
			}
		};
		//---------------------------------------------
		template<typename T,unsigned D> class Sinc;
		/*!
		 *
		 * \brief     sinc(x) .   \f$ sinc(x)= {sin(x) \over x } \f$ .
		 *
		 *   : \n
		 * \param half_central_size -    , 
		 *   , \f$ cx=c*half\_central\_size=\pi \f$  \f$ ={ \pi \over half\_central\_size }\f$.
		 *  :\n
		 *  half_central_size=1,    sin(cx)  2,      =1,2,3,... \n
		 *  half_central_size=2,    sin(cx)  4      =2,4,6,8,... \n
		 *  half_central_size=3,    sin(cx)  6      =3,6,9,12,... \n
		 *  ..
		 * 
		 * \param N_lobes -    . 0 -   , 1 -  
		 *       ..  N_lobes        , 
		 *       , ,    
		 *  ,     .
		 *
		 *  ,     size=2*half_central_size*(N_lobes+1)+1 , 
		           
		 * size=half_wnd_size+1+half_wnd_size  half_wnd_size=half_central_size*(N_lobes+1).
		 *
		 *****************************************************************************/
		template<typename T> class Sinc<T,0>
			: public Traits<T,false,false,0,0>
		{
		public:
			typedef T arg_type;
		private:
			const T half_cent;
			const T half_width;
			const T c;
			static T sinc(T x) { return sin(x)/x; }
		public:
			static const char * Name() { return "Sinc"; }
			template<typename TT,typename TTT> 
			Sinc(TT half_central_size,TTT N_lobes) : Traits(T(half_central_size),T(N_lobes+1),T(0)), half_cent(T(half_central_size)),half_width(T(half_central_size*(N_lobes+1))), c(T(M_PI/half_central_size)) {}
			template<class TT> T operator()(TT x) const
			{
				T y=std::abs(T(c*x));
				return ((y<T(1.0e-20)) ? T(1) : sinc(y)) ; 
			}
			arg_type size() const {return right()-left();}
			arg_type center() const {return arg_type(0);}
			arg_type left() const {return -arg_type(half_width);}
			arg_type right() const {return arg_type(half_width);}
		};

		/*!
		 * \brief   (Kaiser). 
		 *   \f$ w(x)= { I_0(\alpha \sqrt{1-({2x \over size-1} -1 )^2} ) \over I_0(\alpha)}\f$ ,
		 *  \f$ I_0( . )\f$ -   
		 * 
		 *   
		 * \f$ W_K(\omega)= { size \cdot sinh(\zeta) \over I_0(\alpha) \zeta }\f$,
		 *  \f$ \zeta=\sqrt{\alpha^2-({ size \cdot \omega \over 2 })^2} \f$
		 * 
		 *        \f$ \alpha \f$
		 *   :
		 * \image html wndfun_KaiserWindow.png 
		 * \image html wndfun_KaiserWindowSpectra.png 
		 *
		 *****************************************************************************/
		template<typename T,unsigned D> class Kaiser
			: public Traits<T,true,true,D,D>
		{
		public:
			typedef T arg_type; 
		private:
			const T c;
			const T alpha;
			const arg_type _size_;
		public:
			static const char * Name() { return "Kaiser"; }
			template<class TT> 
			Kaiser(TT size_,T alpha_) : Traits(T(size_),T(size_/2.0f)),_size_(arg_type(size_)), c(T(2./(size_-1))), alpha(alpha_) {}
//			template<class TT> 
//			Kaiser(TT size_, const Kaiser& k) : Traits(T(size_),T(size_/2.0f)), c(T(2./(size_-1))), alpha(k.alpha) {}
			template<class TT> T operator()(TT x) const
			{
				return result( T(x), Loki::Int2Type<D>() );
			}
			arg_type size() const {return _size_;}
			arg_type center() const {return arg_type((_size_-1)/2.0f);}
			arg_type left() const {return 0;}
			arg_type right() const {return _size_-1;}
		private:
			T result(T x,Loki::Int2Type<0>) const
			{
				T v=x*c-T(1);
				return T( Vtl::NumAlg::Spf::bess::I<0>(alpha*sqrt(1-v*v)) / Vtl::NumAlg::Spf::bess::I<0>(alpha) );
			}
		};
		//-----------------------------------------------------------------
		namespace Impl
		{
			namespace common_Trig
			{
				template<typename T,unsigned D,unsigned K,unsigned I>  class cos_Member_base
				{
					const T b;
				public:
					template<typename TTT>
					cos_Member_base(T c,TTT* a) : b(T(*a)*Vtl::Alg::x_to_N<D>::value(c*I)){}
//					cos_Member_base(const cos_Member_base& cm) : b(cm.b) {}
					T operator()(T ckx) const
					{
		//				std::cout << "base:    K=" << K << "  I=" << I << "  b=" << b << "  ckx="<< ckx << std::endl;
						return b * Vtl::Alg::sincos<D+1>(ckx);
					}
				};
				template<typename T,unsigned D,unsigned K,unsigned I>  class cos_Member
					: public cos_Member_base<T,D,K,I>  
				{
					cos_Member<T,D,K,I+1> cm;  
				public:
					template<typename TTT>
					cos_Member(T c,TTT* a) : cos_Member_base(c,a), cm(c,a+1) {}
//					cos_Member(const cos_Member<T,D,K,I>& cm_) : cos_Member_base<T,D,K,I>(cm_), cm(cm_.cm) {}
					T operator()(T ckx,T cx) const
					{
		//				std::cout << "cm:  K=" << K << "  I=" << I << "  ckx="<< ckx << "   cx=" << cx << std::endl;
						return cos_Member_base::operator()(ckx)+cm(ckx+cx,cx);
					}
				};
				template<typename T,unsigned D,unsigned K>  class cos_Member<T,D,K,K>
					: public cos_Member_base<T,D,K,K>  
				{
				public:
					template<typename TTT>
					cos_Member(T c,TTT* a) : cos_Member_base<T,D,K,K>(c,a) {}
//					cos_Member(const cos_Member<T,D,K,K>& cm_) : cos_Member_base<T,D,K,K>(cm_) {}
					T operator()(T ckx,T cx) const
					{
		//				std::cout << "cmKK:  K=" << K << "  ckx="<< ckx << "   cx=" << cx << std::endl;
						return cos_Member_base::operator()(ckx);
					}
				};
			}
		}
		/*!
		 * \brief    .

			\code
			template<typename T,unsigned D,unsigned K>  class common_Trig;
			\endcode
			T -   (value_type) \n
			D - , 0 -, 1 -  , 2 -   .. \n
			K -    (. ), ..     K+1  ,    . \n

			    ,   .
			   :
			\f[  w(x) = a_0 + a_1 cos({2 \pi x \over N-1}) + a_2 cos({4 \pi x \over N-1}) + a_3 cos({6 \pi x \over N-1}) + a_4 cos({8 \pi x \over N-1})  \f]
			\f[  0 \le x \le N-1  \f]
			 
			\f[ c={2 \pi \over N-1} \f]
			       :
			\f[  w(x) = a_0 + a_1 cos( c x ) + a_2 cos( 2 c x ) + a_3 cos( 3 c x ) + a_4 cos( 4 c x )  \f]

			     :

		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=1>
		   <TR VALIGN="MIDDLE">
		   <TD>                                 </TD><TD>    a0     </TD><TD>      a1   </TD><TD>     a2    </TD><TD>     a3    </TD><TD>    a4 </TD><TD>   B  </TD><TD> dB </TD>
		   </TR>
		   <TR VALIGN="MIDDLE">
		   <TD>Hamming                              </TD><TD> 0.53836   </TD><TD> -0.46164   </TD><TD>    -      </TD><TD> -         </TD><TD>    -  </TD><TD> 1.37 </TD><TD> 43 </TD>
		   </TR>
		   <TR VALIGN="MIDDLE">
		   <TD>Hann                                 </TD><TD> 0.5       </TD><TD> -0.5	    </TD><TD>    -      </TD><TD> -         </TD><TD>    -  </TD><TD> 1.50 </TD><TD> 33 </TD> 
		   </TR>
		   <TR VALIGN="MIDDLE">
		   <TD>Blackman                             </TD><TD> 0.42      </TD><TD> -0.5       </TD><TD> 0.08	    </TD><TD> -	        </TD><TD>    -  </TD><TD> 1.73 </TD><TD> 60 </TD> 
		   </TR>
		   <TR VALIGN="MIDDLE">
		   <TD>BlackmanNuttall                     </TD><TD> 0.3635819 </TD><TD> -0.4891775 </TD><TD> 0.1365995 </TD><TD> -0.0106411 </TD><TD>    -  </TD><TD> 1.98 </TD><TD> 98 </TD> 
		   </TR>
		   <TR VALIGN="MIDDLE">
		   <TD>BlackmanHarris                      </TD><TD> 0.35875   </TD><TD> -0.48829   </TD><TD> 0.14128   </TD><TD> -0.01168   </TD><TD>    -  </TD><TD> 2.01 </TD><TD> 94 </TD> 
		   </TR>
		   <TR VALIGN="MIDDLE">
		   <TD>Nuttall, continuous first derivative </TD><TD> 0.355768  </TD><TD> -0.487396  </TD><TD> 0.144232  </TD><TD> -0.012604  </TD><TD>    -  </TD><TD> 2.02 </TD><TD> 95 </TD> 
		   </TR>
		   <TR VALIGN="MIDDLE">
		   <TD>Flat top window	                    </TD><TD> 1.0	    </TD><TD> -1.93      </TD><TD> 1.29      </TD><TD> -0.388     </TD><TD> 0.032 </TD><TD> 3.77 </TD><TD> 70 </TD>
		   </TR>
		   </TABLE>

			  ,           , 
			  (  )   1 ( ,     4.64). 
			    ,       1.

		 * <TABLE WIDTH="100%" ALIGN="CENTER" border=1>
		   <TR VALIGN="MIDDLE">
		   <TD>Flat top window	                    </TD><TD> 0.215517  </TD><TD> -0.415948  </TD><TD> 0.278017     </TD><TD> -0.278017  </TD><TD> 0.006897 </TD><TD> 3.77 </TD><TD> 70 </TD>
		   </TR>
		   </TABLE>

		      x     . 
			\f[ w^{(D)}(x)={d^D w(x) \over d x^D} \f]
			\f[ w^{(0)}(x)=w(x) \f]

			  :
		 
		    \f[ sincos(k,x) =  \left\lbrace   \begin{array}{cc}
		 *		sin( x ) , \qquad if \quad k \quad is \quad even \\ 
		 *		cos( x ) , \qquad if \quad k \quad is \quad odd
		 *	\end{array} \right. \f], 
		 * 
		 *   \f[ sisign(k,x) = (-1)^{ [ (k+1)/2 ] }  \f] 
		 *  \f$ { [ ... ] } \f$         , 
		 *  
		 *   \f[ dsin(k,x) = { d^k sin \over d x^k}(x) = (-1)^{ [ (k+1)/2 ] } \left\lbrace   \begin{array}{cc}
		 *		sin( x ) , \qquad if \quad k \quad is \quad even \\ 
		 *		cos( x ) , \qquad if \quad k \quad is \quad odd
		 *	\end{array} \right. = sisign(k,x) sincos(k,x) \f] 
		 *         Vtl::Alg,
		 * \see Vtl::Alg::sincos
		 * \see Vtl::Alg::minus_if_odd -   sisign,    sisign(k,x)=minus_if_odd<(k+1)/2>(x); 
		 * \see Vtl::Alg::dsin
		 *
		 *        :
		 *	\f[ w^{(D)}(x) = a_0 \delta_{D0} + \sum_{j=1}^4 { a_j (cj)^D dsin(D+1, c j x ) } \f] 
		 *	\f[ w^{(D)}(x) = a_0 \delta_{D0} + (-1)^{D+1 \over 2} \sum_{j=1}^4 { a_j (cj)^D sincos(D+1, c j x ) } \f] 
		 *	\f[ w^{(D)}(x) = a_0 \delta_{D0} + sisign( D+1 ,\sum_{j=1}^4 { a_j (cj)^D sincos(D+1, c j x ) } ) \f] 
		 *  \f$  \delta_{ij}  \f$ - .
		 *   ++     :
		 *	\f[ w^{(D)}(x) = a_0 \delta_{D0} + minus\_if\_odd<(D+1)/2> \lbrace \sum_{j=1}^4 { a_j * x_to_N<D>::value(cj) * sincos<D+1>( cjx ) } \rbrace \f] 
		 *
		 *      -       common_Trig
		 *     result()   ,     .
		 *   (    minus_if_odd)   
		 * Vtl::Plf::Kernel::F::Impl::cos_Member,       
		 *   Vtl::Plf::Kernel::F::Impl::cos_Member_base,      - 
		 *  Vtl::Plf::Kernel::F::Impl::cos_Member,        
		 *     ,      
		 *    ,     .
		 *
		 *       a_j * x_to_N<D>::value(cj)    ,  
		 *     ,         enum,
		 *       ,        .
		 *          .
		 *****************************************************************************/
		template<typename T,unsigned D,unsigned K>  class common_Trig
			: public Traits<T,true,true,D,D>
		{
		public:
			typedef T arg_type; 
		private:
			const T c;
			const T a0;
			const Impl::common_Trig::cos_Member<T,D,K,1> cm;  
			const arg_type _size_;
		public:
			template<typename TT,typename TTT>
			common_Trig(TT size_,TTT* a) : Traits(T(size_),T(size_/2.0f)),_size_(arg_type(size_)), c(T(2*M_PI/(size_-1))) , a0(T(a[0])), cm(c,a+1) {}
//			template<typename TT>
//			common_Trig(TT size_,const common_Trig<T,D,K>& cm_) : c(T(2*M_PI/(size_-1))) , a0(cm_.a0), cm(cm_.cm) {}
			template<typename TT>
			T operator()(TT x) const
			{
				return result(T(x),Loki::Int2Type<D==0>());
			}
			arg_type size() const {return _size_;}
			arg_type center() const {return arg_type((_size_-1)/2.0f);}
			arg_type left() const {return 0;}
			arg_type right() const {return _size_-1;}
		private:
			T result(T x,Loki::Int2Type<true>) const
			{
				T cx=c*x;
				return a0+cm(cx,cx); 
			}
			T result(T x,Loki::Int2Type<false>) const
			{
				T cx=c*x;
				return Vtl::Alg::minus_if_odd<(D+1)/2>(cm(cx,cx)); 
			}
		};
		/*!
		 * \brief   (Hann).   
		 * \f$ w(x)= { 1 \over 2 } [ 1 - cos( { 2 \pi x \over size - 1 ) }] \f$
		 *  0 <= x <= size-1 
		 * \see common_Trig
		 *
		 *****************************************************************************/
		template<typename T,unsigned D>  class Hann
			: public common_Trig<T,D,1>
		{
		public:
			static const char * Name() { return "Hann"; }
			template<typename TT>
			Hann(TT size_) : common_Trig<T,D,1>(size_,cof()) {} 
//			template<typename TT>
//			Hann(TT size_,const Hann& ) :common_Trig(size_,cof()) {}
		private:
			static double *cof()
			{
				static double a[2]={0.5,-0.5};
				return a;
			}
		};

		/*!
		 * \brief  e (Hamming).   
		 * \f$ w(x)=   0.54 - 0.46 cos( { 2 \pi x \over size - 1 ) } \f$ 
		 *  0 <= x <= size-1
		 * \see common_Trig
		 *
		 *****************************************************************************/
		template<typename T,unsigned D>  class Hamming
			: public common_Trig<T,D,1>
		{
		public:
			static const char * Name() { return "Hamming"; }
			template<typename TT>
			Hamming(TT size_) : common_Trig(size_,cof()) {} 
//			template<typename TT>
//			Hamming(TT size_,const Hamming& ) :common_Trig(size_,cof()) {}
		private:
			static double *cof()
			{
				static double a[2]={0.53836,-0.46164};
				return a;
			}
		};
		/*!
		 * \brief   (Blackman). \see common_Trig
		 *
		 *****************************************************************************/
		template<typename T,unsigned D>  class Blackman
			: public common_Trig<T,D,2>
		{
		public:
			static const char * Name() { return "Blackman"; }
			template<typename TT>
			Blackman(TT size_) : common_Trig(size_,cof()) {} 
//			template<typename TT>
//			Blackman(TT size_,const Blackman& ) :common_Trig(size_,cof()) {}
		private:
			static double *cof()
			{
				static double a[3]={0.42,-0.5,0.08};
				return a;
			}
		};
		/*!
		 * \brief  - (Blackman-Nuttall). \see common_Trig
		 *
		 *****************************************************************************/
		template<typename T,unsigned D>  class BlackmanNuttall
			: public common_Trig<T,D,3>
		{
		public:
			static const char * Name() { return "BlackmanNuttall"; }
			template<typename TT>
			BlackmanNuttall(TT size_) : common_Trig(size_,cof()) {} 
//			template<typename TT>
//			BlackmanNuttall(TT size_,const BlackmanNuttall& ) :common_Trig(size_,cof()) {}
		private:
			static double *cof()
			{
				static double a[4]={0.3635819,-0.4891775,0.1365995,-0.0106411};
				return a;
			}
		};
		/*!
		 * \brief  - (Blackman-Harris). \see common_Trig
		 *
		 *****************************************************************************/
		template<typename T,unsigned D>  class BlackmanHarris
			: public common_Trig<T,D,3>
		{
		public:
			static const char * Name() { return "BlackmanHarris"; }
			template<typename TT>
			BlackmanHarris(TT size_) : common_Trig(size_,cof()) {} 
//			template<typename TT>
//			BlackmanHarris(TT size_,const BlackmanHarris& ) :common_Trig(size_,cof()) {}
		private:
			static double *cof()
			{
				static double a[4]={0.35875,-0.48829,0.14128,-0.01168};
				return a;
			}
		};
		/*!
		 * \brief   (Nuttall),   . \see common_Trig
		 *
		 *****************************************************************************/
		template<typename T,unsigned D>  class Nuttall
			: public common_Trig<T,D,3>
		{
		public:
			static const char * Name() { return "Nuttall"; }
			template<typename TT>
			Nuttall(TT size_) : common_Trig(size_,cof()) {} 
//			template<typename TT>
//			Nuttall(TT size_,const Nuttall& ) :common_Trig(size_,cof()) {}
		private:
			static double *cof()
			{
				static double a[4]={0.355768,-0.487396,0.144232,-0.012604};
				return a;
			}
		};
		/*!
		 * \brief  - (FlatTop), . \see common_Trig
		 *
		 *****************************************************************************/
		template<typename T,unsigned D>  class FlatTop
			: public common_Trig<T,D,4>
		{
		public:
			static const char * Name() { return "FlatTop"; }
			template<typename TT>
			FlatTop(TT size_) : common_Trig(size_,cof()) {} 
//			template<typename TT>
//			FlatTop(TT size_,const FlatTop& ) :common_Trig(size_,cof()) {}
		private:
			static double *cof()
			{
				static double a[5]={0.215517,-0.415948,0.278017,-0.083621,0.006897};
				return a;
			}
		};
#if 0
		/*!
		 *	\brief   (),    \ref Lit_ReddyChatterji .
		 *   \f$ w(x)= { I_0(\alpha \sqrt{1-({2x \over size-1} -1 )^2} ) \over I_0(\alpha)}\f$ ,
		 *
		 *
		 *****************************************************************************/
		template <typename T> class Reddy
		{
			const T a;
			const T b;
		public:
			typedef T value_type;
			static const char * Name() { return "Reddy"; }
			template<class TT> 
			Reddy(TT width,TT height) : a(T(M_PI/width)),b(T(M_PI/height)) {}
			template<class TT> 
			Reddy(TT width,TT height, const Reddy& ) : a(T(M_PI/width)),b(T(M_PI/height)) {}
			T operator()(int x,int y) const
			{
				T v=cos(a*x)*cos(b*y);
				return (1-v)*(2-v);
			}
		};
		/*!
		 *	\brief   (),  .
		 *
		 *
		 *****************************************************************************/
		template <typename T> class Reddy_2
		{
			const T a;
			const T b;
		public:
			typedef T value_type;
			static const char * Name() { return "Reddy_2"; }
			template<class TT> 
			Reddy_2(TT width,TT height) : a(T(M_PI/width)),b(T(M_PI/height)) {}
			template<class TT> 
			Reddy_2(TT width,TT height, const Reddy_2& ) : a(T(M_PI/width)),b(T(M_PI/height)) {}
			T operator()(int x,int y) const
			{
				T v=cos(a*x)*cos(b*y);
				return (1-v)*(1-v)*(2-v);
			}
		};
#endif


	/*
	See MathLab Help for more windows:
		    Bartlett-Hann                 windowbarthannwin
		+++ Bartlett                      windowbartlett
		+++ Blackman                      windowblackman
		+   Blackman-Harris               windowblackmanharris
		    Bohman                        windowbohmanwin
		    Chebyshev                     windowchebwinFlat 
		+   Top                           windowflattopwin
		+++ Hann                          windowhann
		+   Kaiser                        windowkaiser
		+   Nuttall's Blackman-Harris     windownuttallwin
		    Parzen (de la Valle-Poussin)  windowparzenwin
		    Rectangular                   windowrectwin
		    Tapered cosine                windowtukeywin
		    Triangular                    windowtriang
	*/
	}//namespace WndFun
/** @} */ // end of ConvolutionKernelsDoxGroup  
}//namespace ImAlg
}//namespace Vtl
/////////////////////////////////////////////////////////////////////////////
#endif //_WINDOWING_FUN_H
