
#include "EssentialMatrix.h"

namespace Basics3D
{
	void EssentialMatrix(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_E,
		                   const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_F,
											 const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_K,
											 const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_K1)
	{
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> K1T;
		K1T = boost::numeric::ublas::trans(in_K1);

		out_E = boost::numeric::ublas::prod(K1T, in_F);
		out_E = boost::numeric::ublas::prod(out_E, in_K);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void EssentialMatrix(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
		                   const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_R,
											 const Vtl::NumAlg::Lin::ublas::c_vector<double, 3> &in_t)
	{
		C_TransformMatrices::SkewSymmetric(out, in_t);
		out = boost::numeric::ublas::prod(out, in_R);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void SingularValues(double out_sv[2],
		                  const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in)
	{
		// SVD decomposition of the essential matrix
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> U;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> V;
		Vtl::NumAlg::Lin::ublas::c_vector<double, 3> W;
		Vtl::NumAlg::Lin::ublas::c_vector<double, 3> RV1;

		U = in;

		Vtl::NumAlg::Lin::svd::decomp(U, W, V, RV1);

		int k = 0;
		double min = W[0];

		for (int i = 1; i < 3; i++)
		{
			if (W[i] < min)
			{
				min = W[i];
				k = i;
			}
		}

		int i = 0;
		int j = 0;
		for (; i < 2 && j < 3; i++, j++)
		{
			if (j == k)
			{
				j++;
			}
			out_sv[i] = W[j];
		}

		//std::cout << std::endl << std::endl << "Singular values of essential matrix"
		//	<< std::endl;
		//for (int i = 0; i < 3; i++)
		//{
		//	std::cout << W(i) << "  ";
		//}
		//std::cout << " | ";
		//for (int i = 0; i < 2; i++)
		//{
		//	std::cout << out_sv[i] << "  ";
		//}

		//double eps = 0.03;
		//if (fabs(W(0) - W(1)) < eps || fabs(W(0) - W(2)) < eps || fabs(W(2) - W(1)) < eps)
		//{
		//	std::cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!";
		//}
	}
}