
#include "FundamentalMatrix.h"

namespace Basics3D
{
	///////////////////////////////////////////////////////////////////////////////////////
	// Returns the minimal number of points pairs requared for computation of fund. matrix
	int C_FundamentalMatrix::SampleSize()
	{
		return 7;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	// Returns the distance threshold
	// in_sigma is inaccuracy of feature points detector
	double C_FundamentalMatrix::DistThresh(double in_sigma)
	{
		return sqrt(3.84) * in_sigma;
	}
	/////////////////////////////////////////////////////////////////////////////////////// 
	// Creates matrix A of the system Af = 0, 
	// where f is a vector containing elements of the fundamental matrix,
	//       strings of A: (x'x, x'y, x', y'x, y'y, y', x, y, 1)
	//       (x, y), (x', y') - corresponding points on two images
	void C_FundamentalMatrix::CreateSystem(Vtl::NumAlg::Lin::ublas::matrix<double> &out_M,
																				 const T_Points2DSet<C_Point2D> &in_Points)
	{
		int pairsNumb = in_Points.Size();

		if (pairsNumb < 7)
		{
			std::cout << "Can't create matrix from less than 7 points\n";
			return;
		}

		out_M.resize(pairsNumb, 9, false);

		for (int i = 0; i < pairsNumb; i++)
		{
			C_Point2D P1 = in_Points[2 * i];
			C_Point2D P2 = in_Points[2 * i + 1];

			out_M(i, 0) = P1.X() * P2.X();
			out_M(i, 1) = P1.Y() * P2.X();
			out_M(i, 2) = P2.X();
			out_M(i, 3) = P1.X() * P2.Y();
			out_M(i, 4) = P1.Y() * P2.Y();
			out_M(i, 5) = P2.Y();
			out_M(i, 6) = P1.X();
			out_M(i, 7) = P1.Y();
			out_M(i, 8) = 1;
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////
  // Creates system to find FM from homography matrices
	void C_FundamentalMatrix::CreateSystem
		(Vtl::NumAlg::Lin::ublas::matrix<double> &out_M,
		 const std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &in_Matrices)
	{
	  int matrNumb = in_Matrices.size();

		if (matrNumb < 2)
		{
			std::cout << "Can't create matrix from less than 2 input mtrices\n";
			return;
		}

		out_M.resize(matrNumb * 6, 9, false);

		for (int i = 0; i < matrNumb; i++)
		{
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> H = in_Matrices[i];
			
			out_M(6 * i, 0) = H(0, 0); 
			out_M(6 * i, 1) = 0;
			out_M(6 * i, 2) = 0;
			out_M(6 * i, 3) = H(1, 0);
			out_M(6 * i, 4) = 0;
			out_M(6 * i, 5) = 0;
			out_M(6 * i, 6) = H(2, 0);
			out_M(6 * i, 7) = 0;
			out_M(6 * i, 8) = 0;

			out_M(6 * i + 1, 0) = H(0, 1); 
			out_M(6 * i + 1, 1) = H(0, 0);
			out_M(6 * i + 1, 2) = 0;
			out_M(6 * i + 1, 3) = H(1, 1);
			out_M(6 * i + 1, 4) = H(1, 0);
			out_M(6 * i + 1, 5) = 0;
			out_M(6 * i + 1, 6) = H(2, 1);
			out_M(6 * i + 1, 7) = H(2, 0);
			out_M(6 * i + 1, 8) = 0;

			out_M(6 * i + 2, 0) = H(0, 2); 
			out_M(6 * i + 2, 1) = 0;
			out_M(6 * i + 2, 2) = H(0, 0);
			out_M(6 * i + 2, 3) = H(1, 2);
			out_M(6 * i + 2, 4) = 0;
			out_M(6 * i + 2, 5) = H(1, 0);
			out_M(6 * i + 2, 6) = H(2, 2);
			out_M(6 * i + 2, 7) = 0;
			out_M(6 * i + 2, 8) = H(2, 0);

			out_M(6 * i + 3, 0) = 0; 
			out_M(6 * i + 3, 1) = H(0, 1);
			out_M(6 * i + 3, 2) = 0;
			out_M(6 * i + 3, 3) = 0;
			out_M(6 * i + 3, 4) = H(1, 1);
			out_M(6 * i + 3, 5) = 0;
			out_M(6 * i + 3, 6) = 0;
			out_M(6 * i + 3, 7) = H(2, 1);
			out_M(6 * i + 3, 8) = 0;

			out_M(6 * i + 4, 0) = 0; 
			out_M(6 * i + 4, 1) = H(0, 2);
			out_M(6 * i + 4, 2) = H(0, 1);
			out_M(6 * i + 4, 3) = 0;
			out_M(6 * i + 4, 4) = H(1, 2);
			out_M(6 * i + 4, 5) = H(1, 1);
			out_M(6 * i + 4, 6) = 0;
			out_M(6 * i + 4, 7) = H(2, 2);
			out_M(6 * i + 4, 8) = H(2, 1);

			out_M(6 * i + 5, 0) = 0; 
			out_M(6 * i + 5, 1) = 0;
			out_M(6 * i + 5, 2) = H(0, 2);
			out_M(6 * i + 5, 3) = 0;
			out_M(6 * i + 5, 4) = 0;
			out_M(6 * i + 5, 5) = H(1, 2);
			out_M(6 * i + 5, 6) = 0;
			out_M(6 * i + 5, 7) = 0;
			out_M(6 * i + 5, 8) = H(2, 2);
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_FundamentalMatrix::SevenPointsAlg
			                (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out,
										   const Vtl::NumAlg::Lin::ublas::matrix<double> &in_System)
	{
		int m = in_System.size1();
		int n = in_System.size2();

		Vtl::NumAlg::Lin::ublas::matrix<double> U(m, n);
		Vtl::NumAlg::Lin::ublas::matrix<double> V(n, n);
		Vtl::NumAlg::Lin::ublas::vector<double> W(n);
		Vtl::NumAlg::Lin::ublas::vector<double> RV1(n);

		for (int i = 0; i < m; i++)
		{
			for (int j = 0; j < n; j++)
			{
				U(i, j) = in_System(i, j);
			}
		}

		Vtl::NumAlg::Lin::svd::decomp(U, W, V, RV1);

		double min1, min2;
		min1 = min2 = W(0);

		int k1, k2;
		k1 = k2 = 0;
		
		for (int j = 1; j < n; j++)
		{
			if (W(j) < min1)
			{
				min2 = min1;
				min1 = W(j);
				k2 = k1;
				k1 = j;
			}
		}

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3 ,3> A;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3 ,3> B;
		
		for (int i = 0; i < 3; i++)
		{
			for (int j = 0; j < 3; j++)
			{
			  A(i, j) = V(i * 3 + j, k1);
			  B(i, j) = V(i * 3 + j, k2);
			}
		}

		double a00, a01, a02, a10, a11, a12, a20, a21, a22,
			     b00, b01, b02, b10, b11, b12, b20, b21, b22;

		a00 = A(0, 0);   a01 = A(0, 1);   a02 = A(0, 2);
		a10 = A(1, 0);   a11 = A(1, 1);   a12 = A(1, 2);
    a20 = A(2, 0);   a21 = A(2, 1);   a22 = A(2, 2);

		b00 = B(0, 0);   b01 = B(0, 1);   b02 = B(0, 2);
    b10 = B(1, 0);   b11 = B(1, 1);   b12 = B(1, 2);
    b20 = B(2, 0);   b21 = B(2, 1);   b22 = B(2, 2);

		// det(alfa*H1 + (1-alfa)*H2) = 0;
		// a*alfa^3 + b*alfa^2 + c*alfa + d = 0;
		double a = - a20 * a11 * a02 + a20 * a01 * a12 - a10 * a01 * a22 + a10 * a21 * a02 
							 - a00 * a12 * a21 + a00 * a11 * a22 + b00 * b12 * b21 - b00 * b11 * b22 
							 - b20 * b01 * b12 + b20 * b11 * b02 - b10 * b21 * b02 + b10 * b21 * a02 
							 + b10 * a01 * a22 - b10 * b01 * a22 - a20 * a01 * b12 - a20 * b01 * a12
							 + a20 * b01 * b12 + a20 * a11 * b02 + a20 * b11 * a02 - a20 * b11 * b02 
							 - a10 * a21 * b02 + a10 * b01 * a22 - a10 * b21 * a02 + b10 * b01 * b22 
							 - b10 * a01 * b22 - a00 * a11 * b22 + a00 * b11 * b22 + b00 * a11 * b22 
							 + a10 * a01 * b22 - a10 * b01 * b22 - a00 * b12 * b21 + a00 * b12 * a21 
							 + a00 * a12 * b21 - a00 * b11 * a22 + b00 * a12 * a21 + b00 * b11 * a22 
							 - b00 * a11 * a22 - b00 * b12 * a21 - b00 * a12 * b21 + a10 * b21 * b02 
							 - b20 * a11 * b02 - b20 * b11 * a02 + b20 * a11 * a02 - b20 * a01 * a12 
							 + b20 * a01 * b12 + b20 * b01 * a12 - b10 * a21 * a02 + b10 * a21 * b02;

		double b =  3 * (- b00 * b12 * b21 + b00 * b11 * b22 + b20 * b01 * b12 
										 - b20 * b11 * b02 + b10 * b21 * b02 - b10 * b01 * b22)
						 +	2 * (- b10 * b21 * a02 + b10 * b01 * a22 - a20 * b01 * b12
										 + a20 * b11 * b02 + b10 * a01 * b22 - a00 * b11 * b22 
										 - b00 * a11 * b22 - a10 * a01 * b22 + a10 * b01 * b22
										 + a00 * b12 * b21 - b00 * b11 * a22 + b00 * b12 * a21
										 + b00 * a12 * b21 - a10 * b21 * b02 + b20 * a11 * b02
										 + b20 * b11 * a02 - b20 * a01 * b12 - b20 * b01 * a12
										 - b10 * a21 * b02)
						 - b10 * a01 * a22 + a20 * a01 * b12 + a20 * b01 * a12 - a20 * a11 * b02
						 - a20 * b11 * a02 + a10 * a21 * b02 - a10 * b01 * a22 + a10 * b21 * a02
						 + a00 * a11 * b22 - a00 * b12 * a21 - a00 * a12 * b21 + a00 * b11 * a22
						 - b00 * a12 * a21 + b00 * a11 * a22 - b20 * a11 * a02 + b20 * a01 * a12
						 + b10 * a21 * a02;

		double c = b20 * b01 * a12 + a20 * b01 * b12 + b10 * b21 * a02 + b00 * b11 * a22
						 - b00 * b12 * a21 + b20 * a01 * b12 + a00 * b11 * b22 - b10 * a01 * b22
						 - a00 * b12 * b21 - a20 * b11 * b02 - b10 * b01 * a22 + a10 * b21 * b02
						 + b10 * a21 * b02 - b20 * a11 * b02 - b20 * b11 * a02 - b00 * a12 * b21
						 - a10 * b01 * b22 + b00 * a11 * b22
						 + 3 * ( b00 * b12 * b21 + b10 * b01 * b22 - b00 * b11 * b22 + b20 * b11 * b02
										-b20 * b01 * b12 - b10 * b21 * b02);

		double d = - b00 * b12 * b21 + b00 * b11 * b22 - b10 * b01 * b22 
							 + b10 * b21 * b02 + b20 * b01 * b12 - b20 * b11 * b02;

		double x[3];

		// a*alfa^3 + b*alfa^2 + c*alfa + d = 0;
		if (abs(a) > 0.00000001)
		{
			double q, r, r2, q3;
			double x[3];

			//double M_PI = 3.141592653589793;  define    <cmath> !!!   (<cmath>)    #define _USE_MATH_DEFINES.   num_linalg.h.

			double M_2PI = 2 * M_PI;

			q = (b * b - 3.0 * c) / 9.0;
			r = (b * (2.0 * b * b - 9.0 * c) + 27.0 * d) / 54.0;
			r2 = r * r;
			q3 = q * q * q;

			if(r2 < q3) 
			{
				double t = acos(r / sqrt(q3));
				b /= 3.0;
				q = - 2.0 * sqrt(q);

				x[0] = q * cos(t / 3.0) - b;
				x[1] = q * cos((t + M_2PI) / 3.0) - b;
				x[2] = q * cos((t - M_2PI) / 3.0) - b;

				out.push_back(x[0] * A + (1.0 - x[0]) * B);
				out.push_back(x[1] * A + (1.0 - x[1]) * B);
				out.push_back(x[2] * A + (1.0 - x[2]) * B);
			}
			else 
			{
				double aa,bb;

				if(r <= 0.0)
				{
					r =- r;
				}

				aa = - pow(r + sqrt(r2 - q3), 1.0 /3.0); 

				if(aa != 0.0)
				{
					bb = q / aa;
				}
				else
				{
					bb = 0.0;
				}

				b /= 3.0;
				q = aa + bb;
				r = aa - bb;

				x[0] = q - b;
				x[1] = (-0.5) * q-b;
				x[2] = (sqrt(3.0) * 0.5) * fabs(r);

				out.push_back(x[0] * A + (1.0 - x[0]) * B);

				if(x[2] == 0.0)
				{
					out.push_back(x[1] * A + (1.0 - x[1]) * B);
				}
			}
		}
		else
		{
			// b*alfa^2 + c*alfa + d = 0
			if (abs(b) > 0.00000001)
			{
				a = b;
				b = c;
				c = d;

				double D = b*b - 4*a*c;
				if (D > 0.00000001)
				{
					D = sqrt(D);
					x[0] = (-b + D) / (2 * a);
					x[1] = (-b - D) / (2 * a);

					out.push_back(x[0] * A + (1.0 - x[0]) * B);
					out.push_back(x[1] * A +  (1.0 - x[1])* B);
				}
				else
				{
					if (D >= 0)
					{
						x[0] = -b / (2 * a);
						out.push_back(x[0] * A + (1.0 - x[0]) * B);
					}
				}
			}
			else
			{
				//c*alfa + d = 0
				if (abs(c) > 0.00000001)
				{
					x[0] = -d / c;
					out.push_back(x[0] * A + (1.0 - x[0]) * B);
				}
			}
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_FundamentalMatrix::EightPointsAlg
		                          (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
										           const Vtl::NumAlg::Lin::ublas::matrix<double> &in_System)
	{
		Vtl::NumAlg::Lin::ublas::c_vector<double, 9> vect;
		SolveSystem(vect, in_System);

		for (int i = 0; i < 3; i++)
		{
		  for (int j = 0; j < 3; j++)
			{
			  out(i, j) = vect(i * 3 + j);
			}
		}

		///*temp*/Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> forPrint = out;
		///*temp*/std::cout << std::endl << std::endl << "FM det != 0" << std::endl;
		///*temp*/T_MatrixOperations::Print(forPrint);
		///*temp*/std::cout << std::endl << std::endl;
		///*temp*/T_MatrixOperations::Scale(forPrint, forPrint);
		///*temp*/T_MatrixOperations::Print(forPrint);
		///*temp*/std::cout << std::endl;

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> V;
		Vtl::NumAlg::Lin::ublas::c_vector<double, 3> W;
		Vtl::NumAlg::Lin::ublas::c_vector<double, 3> RV1;

		Vtl::NumAlg::Lin::svd::decomp(out, W, V, RV1);

		double min = fabs(W(0));
		int k = 0;
		
		for (int i = 1; i < 3; i++)
		{
			if (fabs(W(i)) < min)
			{
				min = fabs(W(i));
				k = i;
			}
		}
		
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3 ,3> d(3, 3);

		d(0, 0) = W(0);
		d(0, 1) = 0;
		d(0, 2) = 0;
		d(1, 0) = 0;
		d(1, 1) = W(1);
		d(1, 2) = 0;
		d(2, 0) = 0;
		d(2, 1) = 0;
		d(2, 2) = W(2);
		
		d(k, k) = 0;
		
		V = boost::numeric::ublas::trans(V);
		out = boost::numeric::ublas::prod(out, d);
		out = boost::numeric::ublas::prod(out, V);
  }

	///////////////////////////////////////////////////////////////////////////////////////
	// Finds fundamental matrix
	// out is an array of resulting matrices (1 or 3)
	// &in_System is a matrix of system to solve obtained by CreateSystem
	void C_FundamentalMatrix::Find(std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out,
								   const Vtl::NumAlg::Lin::ublas::matrix<double> &in_System)
	{
		if (in_System.size2() == 7)
		{
		  SevenPointsAlg(out, in_System);
		}
		else
		{
		  Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> tmp;
			EightPointsAlg(tmp, in_System);
			out.push_back(tmp);
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_FundamentalMatrix::FromHomographies
			        (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_FM,
			         const std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &in_H)
	{
		std::cout << std::endl << "----------------------------------------"
			<< std::endl << "COMPUTE FUNDAMENTAL MATRIX FROM HOMOGRAPHIES" << std::endl;

		Vtl::NumAlg::Lin::ublas::matrix<double> system;
		C_FundamentalMatrix::CreateSystem(system, in_H);
		C_FundamentalMatrix::Find(out_FM, system);

	}

	///////////////////////////////////////////////////////////////////////////////////////
	double C_FundamentalMatrix::Dist
											(const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> &in_line,
											 const C_Point2D &in_point)
	{
		return 
				fabs(in_line(0, 0) * in_point.X() + in_line(1, 0) * in_point.Y() + in_line(2, 0)) /
				sqrt(in_line(0, 0) * in_line(0, 0) + in_line(1, 0) * in_line(1, 0)) /*/
				sqrt(in_p2.X() * in_p2.X() + in_p2.Y() * in_p2.Y())
						/ sqrt(in_p3->X() * in_p3->X() + in_p3->Y() * in_p3->Y())*/;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	double C_FundamentalMatrix::MistFunc(const Vtl::NumAlg::Lin::ublas::vector<double> 
		                                                                          &in_matrix,
			                                 const C_Point2D &in_p1,
			                                 const C_Point2D &in_p2)
	{
		return /*abs*/(in_p1.X() * in_p2.X() * in_matrix[0] + 
			         in_p2.X() * in_p1.Y() * in_matrix[1] + in_p2.X() * in_matrix[2] + 
							 in_p1.X() * in_p2.Y() * in_matrix[3] + 
							 in_p1.Y() * in_p2.Y() * in_matrix[4] + in_p2.Y() * in_matrix[5] + 
						   in_p1.X() * in_matrix[6] + in_p1.Y() * in_matrix[7] + in_matrix[8]);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_FundamentalMatrix::DerivativeOfMistFunc
		                           (Vtl::NumAlg::Lin::ublas::vector<double> &out,
			                          const Vtl::NumAlg::Lin::ublas::vector<double> &in_matrix,
			                          const C_Point2D &in_p1,
			                          const C_Point2D &in_p2)
	{
	  out.resize(9);
				
		out[0] = in_p1.X() * in_p2.X();
		out[1] = in_p2.X() * in_p1.Y();
		out[2] = in_p2.X();
		out[3] = in_p1.X() * in_p2.Y();
		out[4] = in_p1.Y() * in_p2.Y();
		out[5] = in_p2.Y();
		out[6] = in_p1.X();
		out[7] = in_p1.Y();
		out[8] = 1;
	}

  
	///////////////////////////////////////////////////////////////////////////////////////
	void C_FundamentalMatrix::NormMatrix
		                       (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_M,
													  Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_M,
														const double in_err)
	{
		out_M = boost::numeric::ublas::trans(in_M); 
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_FundamentalMatrix::ApplyRansac
			         (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_matrix,
		            std::vector<T_Points2DSet<C_Point2D>> &out_inliers,
								std::vector<T_Points2DSet<C_Point2D>> &in_pointsSet,
		            const T_ImagesSet<AImageMem8> &in_img,
		            const int in_i,
		            const double in_sigma)
	{
		std::cout << std::endl << "====================================================="
			<< std::endl << "RANSAC FOR FUNDAMENTAL MATRIX" << std::endl;

		T_Model<C_FundamentalMatrix> Model(in_pointsSet[in_i]);
		Ransac<C_FundamentalMatrix>(out_matrix[in_i], out_inliers[in_i], Model,
			                          in_pointsSet[in_i], in_pointsSet[in_i].Size(), in_sigma);

		std::cout << std::endl << std::endl;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> forPrint = out_matrix[in_i];
		T_MatrixOperations::Scale(forPrint, forPrint);
		T_MatrixOperations::Print(forPrint);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_FundamentalMatrix::RecomputeWithAllPoints
			         (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_matrix,
		            std::vector<T_Points2DSet<C_Point2D>> &out_inliers,
								std::vector<C_Points2DParamSet> &in_pointsParamSet,
								const T_ImagesSet<AImageMem8> &in_img,
		            const int in_i,
								const double in_sigma,
								const double in_t)
	{
		std::cout << std::endl << "======================================================="
		<< std::endl << "RECOMPUTE FUNDAMENTAL MATRIX WITH WHOLE SETS OF POINTS" << std::endl;

		T_Model<C_FundamentalMatrix> Model;
		RecompMatrix<C_FundamentalMatrix>(out_matrix[in_i], out_inliers[in_i], Model,
								   in_pointsParamSet[in_i], in_pointsParamSet[in_i + 1], in_sigma, in_t);

		PrintAndSave("FundM", out_matrix, in_i, in_i);
		DrawEpipolarLines(in_img, out_matrix[in_i], out_inliers, in_i, 20);
	}
}