#ifndef HOMOGRAPHY_MATRIX_H
#define HOMOGRAPHY_MATRIX_H

#include <vector>

// library includings
#include "nummet.h"
#include "num_linalg.h"
#include "aimage.h"

// own includings
#include "MatrixOperations.h"
#include "Point2D.h"
#include "Points2DSet.h"
#include "Points2DParamSet.h"
#include "SolveSystem.h"
#include "ImagesSet.h"
#include "Model.h"
#include "Ransac.h"
#include "RecompMatrix.h"

namespace Basics3D
{
	void MarkCorrespondences(const char *in_name,
		                       const int j,  // addition to name
		                       const T_ImagesSet<AImageMem8> &in_img,
		                       const std::vector<T_Points2DSet<C_Point2D>> &in_PointsSet,
													 const int i); // number of points to mark
													               // -1 means all points
	void MakeDifferenceImage
		      (const T_ImagesSet<AImageMem8> &in_img,
		       const std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &in_Matrix,
		       const int i,
					 const int j,
					 const int k);

	void PrintAndSave
		      (const char *in_name,
           const std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &in_Matrix,
					 const int i,
					 const int j);

	class C_HomographyMatrix// : Basics3D::I_SoughtMatrix
	{
	public:
		// Returns minimal number of points pairs requared for computation of homogr. matrix
	 static int SampleSize();
			
			// Returns the distance threshold
			// in_sigma is inaccuracy of feature points detector
		static double DistThresh(double in_sigma);

 		static void CreateSystem(Vtl::NumAlg::Lin::ublas::matrix<double> &out_M,
														 /*const std::vector<C_Point2D> &in_Points,*/
														 const T_Points2DSet<C_Point2D> &in_Points);

		static void Find(std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out,
										 const Vtl::NumAlg::Lin::ublas::matrix<double> &in_System);

		static double Dist(const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> &in_p1,
											 const C_Point2D &in_p2);

		static double MistFunc(const Vtl::NumAlg::Lin::ublas::vector<double> &in_matrix,
			                     const C_Point2D &in_p1,
			                     const C_Point2D &in_p2);

		static void DerivativeOfMistFunc(Vtl::NumAlg::Lin::ublas::vector<double> &out,
			                               const Vtl::NumAlg::Lin::ublas::vector<double> &in_matrix,
			                               const C_Point2D &in_p1,
			                               const C_Point2D &in_p2);

		static void NormMatrix(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_M,
			                     Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_M,
													const double in_err);
		static void ApplyRansac
			         (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_matrix,
		            std::vector<T_Points2DSet<C_Point2D>> &out_inliers,
								std::vector<T_Points2DSet<C_Point2D>> &in_pointsSet,
		            const T_ImagesSet<AImageMem8> &in_img,
		            const int in_i,
		            const int in_j,
		            const int in_size,
		            const double in_sigma);

		static void RecomputeWithAllPoints
			         (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_matrix,
		            std::vector<T_Points2DSet<C_Point2D>> &out_inliers,
								std::vector<C_Points2DParamSet> &in_pointsParamSet,
								const T_ImagesSet<AImageMem8> &in_img,
		            const int in_i,
								const int in_j,
								const double in_sigma,
								const double in_t);
	};
}
#endif // HOMOGRAPHY_MATRIX_H
