
#include "Newton-Raphson.h"

namespace Basics3D
{
	double rtnewt(const IFuncWithDer *g,
								double x1,
								double x2,
								double xacc)
	{
		double rtn = 0.5 * (x1 + x2);		//Initial guess.

		for (int j = 0; j < JMAX; j++)
		{
			double f = g->Func(rtn);
			double df = g->Deriv(rtn);

			double dx = f / df;
			rtn -= dx;

			if ((x1 - rtn) * (rtn - x2) < 0.0)
			{
				throw "Jumped out of brackets in rtnewt";
				return 0;
			}

			//Convergence
			if (fabs(dx) < xacc)
			{
				return rtn;
			}
		}

		throw "Maximum number of iterations exceeded in rtnewt";
		//Never get here.
		return 0.0; 
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	double rtsafe(const IFuncWithDer *g,
								double x1,
								double x2,
								double xacc)
	{
		double fl = g->Func(x1);
		double fh = g->Func(x2);

		if ((fl > 0.0 && fh > 0.0) || (fl < 0.0 && fh < 0.0))
		{
			throw "Root must be bracketed in rtsafe";
			return 0.0;
		}

		if (fl == 0.0)
		{
			return x1;
		}

		if (fh == 0.0)
		{
			return x2;
		}

		// Orient the search so that f(xl) < 0.
		double xh, xl;

		if (fl < 0.0)
		{
			xl = x1;
			xh = x2;
		} else 
		{
			xh = x1;
			xl = x2;
		}

		// Initialize the guess for root,
		double rts = 0.5 * (x1 + x2);
		double dxold = fabs(x2 - x1);	//the \stepsize before last,"
		double dx = dxold;			      //and the last step.
		double f = g->Func(rts);
		double df = g->Deriv(rts);
		
		// Loop over allowed iterations.
		for (int j = 0; j < MAXIT; j++)
		{ 
			std::cout << "rts = " << rts;
			if ((((rts - xh) * df - f) * ((rts - xl) * df - f) > 0.0) // Bisect if Newton out of range,
					|| (fabs(2.0 * f) > fabs(dxold * df)))								// or not decreasing fast enough.
			{
				dxold = dx;
				dx = 0.5 * (xh - xl);
				rts = xl + dx;
				// Change in root is negligible.
				if (xl == rts)
				{
					return rts;
				}
			} 
			//Newton step acceptable. Take it.
			else
			{
				dxold = dx;
				dx = f / df;
				double temp = rts;
				rts -= dx;
				if (temp == rts)
				{
					return rts;
				}
			}

			//  Convergence criterion.
			if (fabs(dx) < xacc)
			{
				return rts;
			}

			f = g->Func(rts);
			df = g->Deriv(rts);
			std::cout << "  f = " << f << "  df = " << df << std::endl;
			// The one new function evaluation per iteration.
			// Maintain the bracket on the root.
			if (f < 0.0)
			{
				xl = rts;
			}
			else
			{
				xh = rts;
			}
		}
		throw "Maximum number of iterations exceeded in rtsafe";
		// Never get here.
		return 0.0;
	}
}