
#include "ProjectiveMatrix.h"

namespace Basics3D
{
	void C_ProjectiveMatrix::GetW(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_W)
	{
		out_W(0, 0) = 0.0; out_W(0, 1) = -1.0; out_W(0, 2) = 0.0;
		out_W(1, 0) = 1.0; out_W(1, 1) =  0.0; out_W(1, 2) = 0.0;
		out_W(2, 0) = 0.0; out_W(2, 1) =  0.0; out_W(2, 2) = 1.0;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_ProjectiveMatrix::GetP
		                       (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &out_P,
			                      const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_UWV,
									          const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_U,
									          const int in_k,
									          const int in_sgn)
	{
		// first 3 columns
		out_P(0,0) = in_UWV(0,0); out_P(0,1) = in_UWV(0,1); out_P(0,2) = in_UWV(0,2);                                                
		out_P(1,0) = in_UWV(1,0); out_P(1,1) = in_UWV(1,1); out_P(1,2) = in_UWV(1,2); 
		out_P(2,0) = in_UWV(2,0); out_P(2,1) = in_UWV(2,1); out_P(2,2) = in_UWV(2,2); 
		// last column
		out_P(0,3) = in_sgn * in_U(0,in_k);
		out_P(1,3) = in_sgn * in_U(1,in_k);
		out_P(2,3) = in_sgn * in_U(2,in_k);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_ProjectiveMatrix::Right
		                (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4>> &out_P,
							       const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_E)
	{
		// decompose essential matrix into E = -c * U * diag(1, 1, 0) * V^T
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> U;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> V;
		Vtl::NumAlg::Lin::ublas::c_vector<double, 3> D;
		Vtl::NumAlg::Lin::ublas::c_vector<double, 3> RV1;
		U = in_E;
		Vtl::NumAlg::Lin::svd::decomp(U, D, V, RV1);

		std::cout << std::endl << std::endl << "U" << std::endl; T_MatrixOperations::Print(U);
		std::cout << std::endl << std::endl << "D" << std::endl; std::cout << D(0) << "  " << D(1) << "  " << D(2);
		std::cout << std::endl << std::endl << "V" << std::endl; T_MatrixOperations::Print(V);

		// sort singular values
		SortS<double> s[3];
		s[0] = SortS<double>(D(0),0);
		s[1] = SortS<double>(D(1),1);
		s[2] = SortS<double>(D(2),2);
		std::sort(s, s+3);
		
		std::cout << std::endl << "k0 = " << s[0].index << "  k1 = " << s[1].index
			<< "  k2 = " << s[2].index;

		T_MatrixOperations::SortColumns(U, U, s);
		T_MatrixOperations::SortColumns(V, V, s);

		// compose W, W^T, V^T
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> W;
		GetW(W);
		
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> WT;
		WT = boost::numeric::ublas::trans(W);
				
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> VT;
		VT = boost::numeric::ublas::trans(V);
		
		// compose right camera matrices
		out_P.resize(4);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> M;
		M = boost::numeric::ublas::prod(U, W);
		M = boost::numeric::ublas::prod(M, VT);
				
		GetP(out_P[0], M, U, s[0].index, 1);
		GetP(out_P[1], M, U, s[0].index, -1);
		
		M = boost::numeric::ublas::prod(U, WT);
		M = boost::numeric::ublas::prod(M, VT);
		
		GetP(out_P[2], M, U, s[0].index, 1);
		GetP(out_P[3], M, U, s[0].index, -1);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_ProjectiveMatrix::Left(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &out)
	{
		out(0, 0) = 1.0;  out(0, 1) = 0.0;  out(0, 2) = 0.0;  out(0, 3) = 0.0;
		out(1, 0) = 0.0;  out(1, 1) = 1.0;  out(1, 2) = 0.0;  out(1, 3) = 0.0;
		out(2, 0) = 0.0;  out(2, 1) = 0.0;  out(2, 2) = 1.0;  out(2, 3) = 0.0;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_ProjectiveMatrix::Right(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &out,
			                           const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_F)
	{
		// compute right epipole
		Vtl::NumAlg::Lin::ublas::c_vector<double, 3> e1;
		C_Epipole::Right(e1, in_F);

		// [e]x
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> E;
		C_TransformMatrices::SkewSymmetric(E, e1);

		E = boost::numeric::ublas::prod(E, in_F);
		double c =  E(2, 2);

		out(0,0) = E(0,0) / c; out(0,1) = E(0,1) / c; out(0,2) = E(0,2) / c; out(0,3) = e1(0);
		out(1,0) = E(1,0) / c; out(1,1) = E(1,1) / c; out(1,2) = E(1,2) / c; out(1,3) = e1(1);
		out(2,0) = E(2,0) / c; out(2,1) = E(2,1) / c; out(2,2) = E(2,2) / c; out(2,3) = e1(2);

		std::cout << std::endl << std::endl << "P" << std::endl;
		T_MatrixOperations::Print(out);

		// check Result 9.12 (Hartley, p.273)
		// P1^T * F * P is skew-symmetric
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> P;
		Left(P);
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 3> P1T;
		P1T = boost::numeric::ublas::trans(out);
		P = boost::numeric::ublas::prod(in_F, P);
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> S;
		S = boost::numeric::ublas::prod(P1T, P);

		std::cout << std::endl << std::endl << "Skew-symmetric check"
			<< std::endl << "S" << std::endl;
		T_MatrixOperations::Print(S);
		std::cout << std::endl << "S^T" << std::endl;
		S = boost::numeric::ublas::trans(S);
		std::cout << std::endl << std::endl;
		T_MatrixOperations::Print(S);

	}
}