
#include "RQDecomposition.h"

namespace Basics3D
{
	void RQDecomposition(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_R,
			                 Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_Q,
			                 const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_M)
	{
		double eps = 0.00001;

		// X rotation => A32 = 0
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> Qx;
		if (fabs(in_M(2, 1)) > eps)
		{
			double dx = sqrt(in_M(2, 1) * in_M(2, 1) + in_M(2, 2) * in_M(2, 2));
			C_TransformMatrices::RotationX(Qx, - in_M(2, 2) / dx, in_M(2, 1) / dx);
		}
		else
		{
			C_TransformMatrices::Identity(Qx);
		} 
		out_R = boost::numeric::ublas::prod(in_M, Qx);

		///*temp*/std::cout << std::endl << std::endl << "Qx" << std::endl; T_MatrixOperations::Print(Qx);
		///*temp*/std::cout << std::endl << std::endl << "M * Qx => (3, 2) = 0" << std::endl;
		///*temp*/T_MatrixOperations::Print(out_R);

		// Y rotation => A31 = 0
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> Qy;
		if (fabs(out_R(2, 0)) > eps)
		{
			double dy = sqrt(out_R(2, 0) * out_R(2, 0) + out_R(2, 2) * out_R(2, 2));
			C_TransformMatrices::RotationY(Qy, out_R(2, 2) / dy, out_R(2, 0) / dy);
		}
		else
		{
			C_TransformMatrices::Identity(Qy);
		}
		out_R = boost::numeric::ublas::prod(out_R, Qy);

		///*temp*/std::cout << std::endl << std::endl << "Qy" << std::endl; T_MatrixOperations::Print(Qy);
		///*temp*/std::cout << std::endl << std::endl << "M * Qx * Qy => (3, 1) = 0" << std::endl;
		///*temp*/T_MatrixOperations::Print(out_R);

		// Z rotation => A21 = 0
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> Qz;
		if (fabs(out_R(1, 0)) > eps)
		{
			double dz = sqrt(out_R(1, 0) * out_R(1, 0) + out_R(1, 1) * out_R(1, 1));
			C_TransformMatrices::RotationZ(Qz, - out_R(1, 1) / dz, out_R(1, 0) / dz);
		}
		else
		{
			C_TransformMatrices::Identity(Qz);
		}
		out_R = boost::numeric::ublas::prod(out_R, Qz);

		///*temp*/std::cout << std::endl << std::endl << "Qz" << std::endl; T_MatrixOperations::Print(Qz);
		///*temp*/std::cout << std::endl << std::endl << "M * Qx * Qy * Qz => (2, 1) = 0" << std::endl;
		///*temp*/T_MatrixOperations::Print(out_R);

		// Form Q
		Qx = boost::numeric::ublas::trans(Qx);
		Qy = boost::numeric::ublas::trans(Qy);
		Qz = boost::numeric::ublas::trans(Qz);

		out_Q = boost::numeric::ublas::prod(Qz, Qy);
		out_Q = boost::numeric::ublas::prod(out_Q, Qx);

		for (int j = 0; j < 2; j++)
		{
			if (out_R(j, j) < 0)
			{
				for (int i = 0; i < 3; i++)
				{
					out_R(j, i) *= -1.0;
					out_Q(i, j) *= -1.0;
				}
			}
		}

		///*temp*/std::cout << std::endl << std::endl << "R * Q" << std::endl;
		///*temp*/T_MatrixOperations::Print(boost::numeric::ublas::prod(out_R, out_Q));
		///*temp*/Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> QT;
		///*temp*/QT = boost::numeric::ublas::trans(out_Q);
		///*temp*/QT = boost::numeric::ublas::prod(out_Q, QT);
		///*temp*/std::cout << std::endl << std::endl << "Q * Q^T" << std::endl;
		///*temp*/T_MatrixOperations::Print(QT);
	}
}