#include "TransformMatrices.h"

namespace Basics3D
{
	///////////////////////////////////////////////////////////////////////////////////////
	// Builds matrix normalizing points MeanX -> 0, Mean -> 0, MeanDev -> sqrt(2)
	double C_TransformMatrices::Normalization
			                             (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
																		double in_MeanX,
																		double in_MeanY,
																		double in_MeanDev,
																		double in_pointErr)
	{
		double coef = sqrt(2.0) / in_MeanDev;

		out(0, 0) = coef;  out(0, 1) = 0;     out(0, 2) = -in_MeanX * coef;
		out(1, 0) = 0;     out(1, 1) = coef;  out(1, 2) = -in_MeanY * coef;
    out(2, 0) = 0;     out(2, 1) = 0;     out(2, 2) = 1;

		double coefErr = sqrt(2.0) * in_pointErr / 
									(in_MeanDev * in_MeanDev - in_pointErr * in_pointErr);

		double Err = sqrt(2 * coefErr * coefErr * (1 + in_pointErr * in_pointErr));
		
		return Err; 
	}

	///////////////////////////////////////////////////////////////////////////////////////
	double C_TransformMatrices::NormalizationInverse
			                             (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
													          double in_MeanX,
													          double in_MeanY,
													          double in_MeanDev,
											 		          double in_pointErr)
	{
	  double coef = 1.0 / (sqrt(2.0) / in_MeanDev);

		out(0, 0) = coef;  out(0, 1) = 0;     out(0, 2) = in_MeanX;
		out(1, 0) = 0;     out(1, 1) = coef;  out(1, 2) = in_MeanY;
		out(2, 0) = 0;     out(2, 1) = 0;     out(2, 2) = 1;

		double coefErr = sqrt(2.0) * in_pointErr / 
									(in_MeanDev * in_MeanDev - in_pointErr * in_pointErr);

		double Err = sqrt(2 * (coefErr * coefErr + in_pointErr * in_pointErr));
		
		return Err; 
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_TransformMatrices::Translational
		                               (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
													          const double in_X,
													          const double in_Y)
	{
	  out(0, 0) = 1;  out(0, 1) = 0;  out(0, 2) = -in_X;
		out(1, 0) = 0;  out(1, 1) = 1;  out(1, 2) = -in_Y;
		out(2, 0) = 0;  out(2, 1) = 0;  out(2, 2) = 1;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_TransformMatrices::TranslationalInverse
		                               (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
													          const double in_X,
													          const double in_Y)
	{
	  out(0, 0) = 1;  out(0, 1) = 0;  out(0, 2) = in_X;
		out(1, 0) = 0;  out(1, 1) = 1;  out(1, 2) = in_Y;
		out(2, 0) = 0;  out(2, 1) = 0;  out(2, 2) = 1;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_TransformMatrices::Rotation(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
				                             Vtl::NumAlg::Lin::ublas::c_vector<double, 3> &in_e)
	{
		out(0, 0) = in_e(0);   out(0, 1) = in_e(1);  out(0, 2) = 0;
		out(1, 0) = -in_e(1);  out(1, 1) = in_e(0);  out(1, 2) = 0;
		out(2, 0) = 0;         out(2, 1) = 0;        out(2, 2) = 1;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_TransformMatrices::RotationX(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
		                                  const double in_cos,
												              const double in_sin)
	{
		out(0, 0) = 1.0;	out(0, 1) = 0.0;    out(0, 2) = 0.0;
		out(1, 0) = 0.0;	out(1, 1) = in_cos;	out(1, 2) = -in_sin;
		out(2, 0) = 0.0;	out(2, 1) = in_sin;	out(2, 2) = in_cos;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_TransformMatrices::RotationY(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
		                                  const double in_cos,
												              const double in_sin)
	{
		out(0, 0) = in_cos; 	out(0, 1) = 0.0;	out(0, 2) = in_sin;
		out(1, 0) = 0.0;	    out(1, 1) = 1.0;	out(1, 2) = 0.0;
		out(2, 0) = -in_sin;	out(2, 1) = 0.0;	out(2, 2) = in_cos;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_TransformMatrices::RotationZ(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
		                                  const double in_cos,
												              const double in_sin)
	{
		out(0, 0) = in_cos;	out(0, 1) = -in_sin; out(0, 2) = 0.0;
		out(1, 0) = in_sin;	out(1, 1) = in_cos;	 out(1, 2) = 0.0;
		out(2, 0) = 0.0;    out(2, 1) = 0.0;	   out(2, 2) = 1.0;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_TransformMatrices::SkewSymmetric
		                           (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
			                          const Vtl::NumAlg::Lin::ublas::c_vector<double, 3> &in_v)
	{
		out(0, 0) = 0.0;      out(0, 1) = -in_v(2); out(0, 2) = in_v(1);
		out(1, 0) = in_v(2);  out(1, 1) = 0.0;      out(1, 2) = -in_v(0);
		out(2, 0) = -in_v(1); out(2, 1) = in_v(0);  out(2, 2) = 0.0;
	}
}