#ifndef NORMALIZATION_MATRIX_H_
#define NORMALIZATION_MATRIX_H_

#include <math.h>

// library includings
#include "num_linalg.h"

// own includings
#include "MatrixOperations.h"

namespace Basics3D
{
	// Creates matrix normalizing points MeanX -> 0, Mean -> 0, MeanDev -> sqrt(2)
	class C_TransformMatrices
	{
	public:
		static double Normalization(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
													      const double in_MeanX,
													      const double in_MeanY,
													      const double in_MeanDev,
											 		      const double in_pointErr);

		static double NormalizationInverse(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
													             const double in_MeanX,
													             const double in_MeanY,
													             const double in_MeanDev,
											 		             const double in_pointErr);

		static void Scaling(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
			                  const double in_coef);

    static void Translational(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
													    const double in_X,
													    const double in_Y);

		static void TranslationalInverse(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
													           const double in_X,
													           const double in_Y);

		static void Rotation(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
				                 Vtl::NumAlg::Lin::ublas::c_vector<double, 3> &in_e);

		static void RotationX(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
			                    const double in_cos,
													const double in_sin);

		static void RotationY(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
			                    const double in_cos,
													const double in_sin);

		static void RotationZ(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
			                    const double in_cos,
													const double in_sin);

		static void SkewSymmetric(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
			                        const Vtl::NumAlg::Lin::ublas::c_vector<double, 3> &in_v);

		template <class MatrixType>
		static void Identity(MatrixType &out)
		{
			for (size_t i = 0; i < out.size1(); i++)
			{
				for (size_t j = 0; j < out.size2(); j++)
				{
					out(i, j) = 0.0;
				}
				out(i, i) = 1.0;
			}
		}
	};
}
#endif // NORMALIZATION_MATRIX_H_
