#ifndef _SFM_DBG_H_
#define _SFM_DBG_H_
////////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <iomanip>
#include "loki_typeManip.h"
#include "loki_static_check.h"
#include "num_linalg.h"
////////////////////////////////////////////////////////////////////////////////////

///   
namespace Dbg
{
	namespace Format
	{
		//  printf("%+12.5f",value);
		struct F
		{
			const int width_;
			const int prec_;
			const bool sign_;
			F(int width,int prec,bool sign) : width_(width),prec_(prec),sign_(sign) {}
			void eachtime(std::ostream& os)
			{
				os<< std::setw(width_);
			}
		};
	}//namespace Format
	template<typename T,int format_type=0> struct Fmt
	{
		const T & value;
		explicit Fmt(const T& v) : value(v) {}
		std::ostream& write(std::ostream& os)
		{
			os << value;
			return os;
		}

	};



	template<typename T,int N,int M,int format>
	struct Fmt<Vtl::NumAlg::Lin::ublas::c_matrix<T,M,N>,format >
	{
		typedef Vtl::NumAlg::Lin::ublas::c_matrix<T,M,N> Matrix;
		const Matrix& value;
		enum Format
		{
			Default,
			RowsAndCols
		};
		//Format f;
		explicit Fmt(Matrix& m): value(m) {}
		std::ostream& write(std::ostream& os) const
		{
			write_(os,Loki::Int2Type<format>() );
			return os;
		}
	private:
		void write_(std::ostream& os,Loki::Int2Type<Default> ) const
		{
			os<< value;
		}
		void write_(std::ostream& os,Loki::Int2Type<RowsAndCols> ) const
		{
			//std::ios::iostate state=os.rdstate();
			Dbg::Format::F fm(5+9,5,true);
			os<< fm;
			for(Matrix::size_type i=0; i< value.size1(); ++i)
			{
				for(Matrix::size_type j=0; j< value.size2(); ++j)
				{
					fm.eachtime(os); 
					os << value(i,j);
				}
				os << std::endl;
			}
			//os.setstate(state,false);
		}

	};
}//namespace Dbg
////////////////////////////////////////////////////////////////////////////////////
template<typename T,int format>
std::ostream& operator<<(std::ostream& os,const Dbg::Fmt<T,format>& fmt)
{
	return fmt.write(os);
}
std::ostream& operator<<(std::ostream& os,const Dbg::Format::F& fmt)
{
	if(fmt.sign_)
		os << std::setiosflags(std::ios::showpoint|std::ios::fixed|std::ios::showpos);
	else
		os << std::setiosflags(std::ios::showpoint|std::ios::fixed);
	os << std::setprecision(fmt.prec_);
	std::setw(fmt.width_);
	return os;
}
////////////////////////////////////////////////////////////////////////////////////
#endif //_SFM_DBG_H_
