// file: #include "sfm_structure_from_2_views.h"
#ifndef _SFM_STRUCTURE_FROM_2_VIEWS_H 
#define _SFM_STRUCTURE_FROM_2_VIEWS_H 
////////////////////////////////////////////////////////////////////////////////////////
#include "num_linalg.h"

#include "sfm_calibration.h"
#include "sfm_reconsruction.h"
#include "sfm_export_3D_model.h" 

#include "Point2D.h"
#include "Points2DSet.h"
////////////////////////////////////////////////////////////////////////////////////////
namespace Basics3D
{
namespace Structure
{
	/// Algorithm:
	/// 1) Build calibration matrices K and Essential matrix
	/// 2) Build camera matrices P=K(R|t), few hypothesis 
	/// 3) Preliminary hypothesis check
	/// 4) For each reasonable hypothesis reconstruct via DLT and calc percent of points in from of cameras
	/// 5) Select best hypothesis
	/// 6) refine hypothesis
	/// 7) build mesh
	/// 8) Save VRML
	template<typename value_type,int N3_>
	void FromFM(const T_Points2DSet<C_Point2D> &in_points2D
		                ,const Vtl::NumAlg::Lin::ublas::c_matrix<value_type, N3_, N3_> &F
						,const value_type in_sigma)
	{
		using namespace Vtl::NumAlg::Lin;
		enum { N3=N3_};
		enum { N4=4};
		typedef typename Types<value_type>::Vector_2 Vector_2;
		typedef typename Types<value_type>::Vector_3 Vector_3;
		typedef typename Types<value_type>::Vector_4 Vector_4;
		typedef typename Types<value_type>::Matrix_3x3 Matrix_3x3;
		typedef typename Types<value_type>::Matrix_3x4 Matrix_3x4;
		typedef typename Types<value_type>::Matrix_4x4 Matrix_4x4;

		//typedef Vtl::NumAlg::Lin::ublas::c_matrix<value_type, N3, N3> Matrix_3x3;
		//typedef Vtl::NumAlg::Lin::ublas::c_vector<value_type, N3> Vector_3;
		//typedef ublas::c_matrix<value_type, N3, N4 > Matrix_3x4;
		//typedef ublas::c_vector<value_type, N4> Vector_4;
		//typedef Vtl::NumAlg::Lin::ublas::c_vector<value_type, 2> Vector_2;

		std::cout << std::endl;
		std::cout << "#######################################################" << std::endl;
		std::cout << "###### YURIN Structure From FM  ----- START ###########" <<std::endl;
		std::cout << "#######################################################" << std::endl;
		// 1) Build calibration matrices K and Essential matrix
		std::cout << std::endl << "====== Fm2Essential ================"<<std::endl;

		Fm2Essential::DirectMinimization<Matrix_3x3> f2e(F);
		f2e();


		Matrix_3x3 E=f2e.E;
		Matrix_3x3 K1=f2e.K1;
		Matrix_3x3 K2=f2e.K2;
		Fm2Essential::KruppaEq<Matrix_3x3> kruppa(F);


		// 2) Build camera matrices P=K(R|t), few hypothesis 
		std::cout << std::endl << "====== CameraPoseFromEssential ================"<<std::endl;
		Fm2Essential::CameraPoseFromEssential<value_type> cp(E);

		// 3) Preliminary hypothesis check

		// 4) For each reasonable hypothesis reconstruct via DLT and calc percent of points in from of cameras
		std::cout << std::endl << "====== FmTools<value_type>::Build_P ================"<<std::endl;
		Matrix_3x4 P1;
		Matrix_3x4 P2;
		std::cout << "K1="<< K1 << std::endl;
		std::cout << "R1="<< cp.R0 << std::endl;
		std::cout << "t1="<< cp.t0 << std::endl;
		std::cout << "K2="<< K2 << std::endl;
		std::cout << "R2="<< cp.R << std::endl;
		std::cout << "t2="<< cp.t << std::endl;
		//K1(0,0)=K1(1,1)=K2(0,0)=K2(1,1)=1.5;
		//FmTools<value_type>::Build_P(P2,K2,cp.R ,cp.t );
		//FmTools<value_type>::Build_P(P1,K1,cp.R0,cp.t0);
		FmTools<value_type>::Build_P(P2,cp.R0,cp.R ,cp.t );
		FmTools<value_type>::Build_P(P1,cp.R0,cp.R0,cp.t0);
		

		std::cout << "P1="<< P1 << std::endl;
		std::cout << "P2="<< P2 << std::endl;
		std::cout << " ----- P1 =" << std::endl;
		std::cout << Dbg::Fmt<Matrix_3x4 ,Dbg::Fmt<Matrix_3x4>::RowsAndCols>(P1) <<std::endl; 
		std::cout << " ----- P2 =" << std::endl;
		std::cout << Dbg::Fmt<Matrix_3x4 ,Dbg::Fmt<Matrix_3x4>::RowsAndCols>(P2) <<std::endl; 

		std::cout << std::endl << "====== Find H : Pm = P * H ================"<<std::endl;
//		ProjectiveToMetric<value_type> p2m(K1(0,0),P1,P2);
		value_type focus=4.2;
		ProjectiveToMetric<value_type> p2m(focus,P1,P2);
		Matrix_4x4 H;
		p2m(H);
		P1=prod(P1,H);
		P2=prod(P2,H);

		//      H,   P   
		// R2_cor=K^(-1) * (R2 *K  - t2 * p^T * K )
		//  t2_cor = K^(-1) * t
		Matrix_3x3 R2_cor;
		Vector_3 t2_cor;
		{
			R2_cor=ublas::subrange(P2, 0,3, 0,3);  
			t2_cor=ublas::column(P2,3);  
			Matrix_3x3 Kinv=ublas::identity_matrix<value_type>(3,3);
			Kinv(0,0)=Kinv(1,1)=1/focus;
			R2_cor=ublas::prod(Kinv,R2_cor);
			t2_cor=ublas::prod(Kinv,t2_cor);
		}
		std::vector<CameraPos<value_type> > cam;
		cam.resize(2);
		cam[0].t=cp.t0;
		cam[0].R=cp.R0;
		cam[0].focus=focus;
		cam[0].id=0;
//		cam[1].t=cp.t;
//		cam[1].R=cp.R;
		cam[1].t=t2_cor;
		cam[1].R=R2_cor;
		cam[1].focus=focus;
		cam[1].id=1;


		std::cout << std::endl << "====== Structure::Dlt ================"<<std::endl;
		Vector_4 X;
		Structure::Dlt<value_type> dlt(P1,P2);
		{//--call test
			Vector_2 v2;
			v2(0)=0; v2(1)=0;
			dlt(v2,v2,X);
			std::cout << "X = " << X <<std::endl; 
		}
		Comp3D<value_type> c3d(in_points2D,dlt);
		c3d();
		// 5) Select best hypothesis

		// 6) refine hypothesis

		// 7) build mesh

		// 8) Save VRML
		SaveVRML_AsPointsSet("yurin.wrl",c3d.points,cam);


		std::cout << "#######################################################" << std::endl;
		std::cout << "###### YURIN Structure From FM  ----- FINISHED ########" <<std::endl;
		std::cout << "#######################################################" << std::endl;
		std::cout << std::endl;
	}
}//namespace Structure
}//namespace Basics3D
////////////////////////////////////////////////////////////////////////////////////////
#endif //_SFM_STRUCTURE_FROM_2_VIEWS_H 
