#ifndef MASK_BLOB_RIDGE_H
#define MASK_BLOB_RIDGE_H

// std includes
#include <limits.h>

// own includes
#include "EdgeDetection/Hysteresis.h"

template <typename T>
void MaskBlobRidge(T* out_mask,
                   T* in_lambda1,
                   T* in_lambda2,
                   int w,
                   int h,
                   int nls,
                   double maxExcentric = 10.0,
                   T blobVal = T(255),
                   T ridgeVal = T(127))
{
  // calculate ratio |lambda1| / |lambda2|
  int size = w * h;

  std::vector<double> lambdaRatio(size);

  double* ratio = &(lambdaRatio[0]);
  
  for (int i = 0; i < size; ++i)
  {
    if (std::abs(double(in_lambda2[i])) > 0.00001)
    {
      ratio[i] = std::abs(double(in_lambda1[i])) / std::abs(double(in_lambda2[i]));
    }
    else
    {
      ratio[i] = 255;
    }
  }

  // extend areas with low lambda ratio via hysteresis
  Hysteresis<double> hysteresis(ratio, w, h);
  hysteresis.Apply(maxExcentric, 3.0, ratio, maxExcentric);
 
  // form mask of blobs and ridges
  for (int y = 0; y < h; ++y)
  {
    for (int x = 0; x < w; ++x)
    {
      // saddle
      if ((in_lambda1[x] * in_lambda2[x] < 0) && 
          (2 * maxExcentric * std::abs(in_lambda2[x]) > std::abs(in_lambda1[x])))
		  {
			  out_mask[x]= 0;
		  }
      // no feature
		  else if (std::abs(in_lambda1[x]) <= 0.001f )  // -
		  {
			  out_mask[x]= 0;
		  }
      else
      {
        // blob
				if(ratio[x] < maxExcentric)
        {
				  out_mask[x] = blobVal;
        }
        // ridge
			  else 
        {
				  out_mask[x] = ridgeVal;
        }
		  }
    }

    out_mask += nls;
    in_lambda1 += nls;
    in_lambda2 += nls;
    ratio += nls;
  }
}


#endif // MASK_BLOB_RIDGE_H