#ifndef MODEL_TYPES_H
#define MODEL_TYPES_H

// system includings
#include <vector>
#include <math.h>

// library includings
#include "aimage.h"
#include "alg_bresenham.h"

// own includings
#include "Point.h"
#include "PointsContainer.h"

namespace Vtl
{
namespace CurveDetection
{

	/*class Circle
	{
	private:
		double m_a;
		double m_b;
		double m_r;

	public:
		static const int sampleSize = 4;
		static const int distThresh = 3;

	public:
		Circle()
		{
			m_a = m_b = m_r = 1;
		}

		//---------------------------------------------------------------------------
		Circle(double a,
					 double b,
					 double r)
		{
			SetA(a);
			SetB(b);
			SetR(r);
		}

		//---------------------------------------------------------------------------
		double A()
		{
			return m_a;
		}

		double B()
		{
			return m_b;
		}

		double R()
		{
			return m_r;
		}

		//---------------------------------------------------------------------------
		void SetA(double a)
		{
			if (a >= 0)
			{
				m_a = a;
			}
			else
			{
				m_a = 0;
			}
		}

		void SetB(double b)
		{
			if (b >= 0)
			{
				m_b = b;
			}
			else
			{
				m_b = 0;
			}
		}

		void SetR(double r)
		{
			if (r >= 0)
			{
				m_r = r;
			}
			else
			{
				m_r = 0;
			}
		}

		//---------------------------------------------------------------------------
		Circle(std::vector<const Point *> &sample,
					 bool *ok)
		{
			if (sample.size() == sampleSize)
			{
				double x1 = sample[0]->x;
				double y1 = sample[0]->y;
				double x2 = sample[1]->x;
				double y2 = sample[1]->y;
				double x3 = sample[2]->x;
				double y3 = sample[2]->y;

				if (x1*y2 + x2*y3 + x3*y1 - x3*y2 - x2*y1 - x1*y3 < 0.01)
				{
					//std::cout << std::endl << "points are situated on the line";
					m_a = m_b = 0;
					m_r = 1;
					*ok = false;
				}
				else
				{
					double c = x2*x2 + y2*y2 - x1*x1 - y1*y1;
					double d = x3*x3 + y3*y3 - x1*x1 - y1*y1;
					double e = (y2 - y1);
					double f = (y3 - y1);
					double g = (x2 - x1);
					double h = (x3 - x1);
					double i = 2 * (g*f - h*e);
					
					m_a = (c*f - d*e) / i;
					m_b = (g*d - h*c) / i;
					m_r = sqrt((x1 - m_a)*(x1 - m_a) + (y1 - m_b)*(y1 - m_b));
					*ok = true;
				}
			}  
		}

		//---------------------------------------------------------------------------
		double Dist(const Point &p)
		{
			double x = p.x;
			double y = p.y;

			return (fabs(sqrt((x - m_a)*(x - m_a) + (y - m_b)*(y - m_b)) - m_r));
		}

		//---------------------------------------------------------------------------
		int DoesPointSatisfyModel(const Point &p,
															double prec)
		{ 
			return (Dist(p) <= prec);
		}

		//---------------------------------------------------------------------------
		Circle operator= (Circle c)
		{
			m_a = c.A();
			m_b = c.B();
			m_r = c.R();

			return *this;
		}

		//---------------------------------------------------------------------------
		char operator< (Circle c)
		{
			if (abs(m_r - c.R()) < 5)
			{
				if (abs(m_a - c.A()) < 3)
				{
					return (m_b < c.B());
				}
				return (m_a < c.A());
			}
			return (m_r < c.R());
		}

		//---------------------------------------------------------------------------
		char IsCloseToModel(Circle c)
		{
			if (sqrt((m_a - c.A()) * (m_a - c.A()) + 
				 (m_b - c.B()) * (m_b - c.B())) / m_r < 0.05)
				if (abs(m_r - c.R()) / m_r < 0.05)
			{
				return 1;
			}
			return 0;
		}
	  //-------------------------------------------------------------------------
		template <typename Img>
		void Draw(Img *output,
							int pointsNumb,
							int shiftX = 0,
							int shiftY = 0)
		{
			const double PI = 3.1415926;

			int w = output->Width();
			int h = output->Height();
			int bits = output->Bits();

			// pen is gray
			AImageMem8::Pixel color_pen = AImageMem8::Pixel(128); 

			typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor;
			DrawFunctor draw_functor(color_pen);

			typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel, DrawFunctor>
																																	 SetPixelFunctor;
			SetPixelFunctor set_pix((*output)[0], output->NextLineShift(), draw_functor);

			Vtl::ImAlg::Bres::Line<SetPixelFunctor> line(set_pix);

			std::vector<int> x;
			std::vector<int> y;
			//std::cout << std::endl << "x = " << int(x.size()) << "  y = " << int(y.size())
			//	<< "  points = " << pointsNumb;
			x.resize(pointsNumb + 1);
			y.resize(pointsNumb + 1);

			for(int i = 0; i <= pointsNumb ; i ++)
			{
				x[i] = int(m_a + m_r * cos(2 * PI / pointsNumb * i) + 0.5) + shiftX;
				y[i] = int(m_b + m_r * sin(2 * PI / pointsNumb *i ) + 0.5) + shiftY;
				if (x[i] < 0 || x[i] >= w || y[i] < 0 || y[i] >= h)
				{
					if (i == 0)
					{
						x[0] = y[0] = 0;
					}
					else
					{
						x[i] = x[i - 1];
						y[i] = y[i - 1];
					}
				}
			}

			//std::cout << std::endl << x[0] << "  " << y[0];
			set_pix.Set0(x[0], y[0]);

			for(int i = 1; i <= pointsNumb ; i ++)
			{
				//std::cout << std::endl << x[i] << "  " << y[i];
				line.Draw(x[i] - x[i - 1], y[i] - y[i - 1]);
			}
		}
	};*/

	///////////////////////////////////////////////////////////////////////////////
	template <typename T>
	class Line
	{
	public:
		T cosFi;
		T sinFi;
		T ro;
		
		static const int degOfFr = 2;
		static const int sampleSize = 2;
		static const int minDist = 5;
		static const int maxDist = 20;
		
	public:
		Line()
		{
		  DefaultInit();
		}

		//-------------------------------------------------------------------------
		Line(int in_cosFi,
			   int in_sinFi,
				 int in_ro)
		{
		  cosFi = in_cosFi;
			sinFi = in_sinFi;
			ro = in_ro;
		}

		//-------------------------------------------------------------------------
		template <typename IterType, typename PointType>
		Line(const std::vector<PointType *> sample,
			   const IterType gradX,
				 const IterType gradY,
				 int w,
				 int h,
			   bool *ok)
		{
			//std::cout << std::endl << "line";
			// check if the sample has proper size
		   if (sample.size() != 2)
			 {
			   *ok = false;
				 DefaultInit();
         return;
			 }

			 // check if points of sample have simular gradient direction
			 int x[2];
			 int y[2];
			 T dirGradX[2];
			 T dirGradY[2];
			 
			 for (int i = 0; i < 2; i++)
			 {
				 x[i] = sample[i]->x;
			   y[i] = sample[i]->y;
			   
				 dirGradX[i] = *(gradX + y[i] * w + x[i]);
			   dirGradY[i] = *(gradY + y[i] * w + x[i]);
			   T modGrad = sqrt(dirGradX[i] * dirGradX[i] + dirGradY[i] * dirGradY[i]);
			   dirGradX[i] /= modGrad;
			   dirGradY[i] /= modGrad;
		   }

			 if (fabs(dirGradX[0] - dirGradX[1]) / fabs(dirGradX[0]) > 0.1 ||
				   fabs(dirGradY[0] - dirGradY[1]) / fabs(dirGradY[0]) > 0.1)
			 {
			   *ok = false;
				 DefaultInit();
         return;
			 }

			 // calculate line parameters
			 T A = T(y[0] - y[1]);
			 T B = T(x[1] - x[0]);
			 T C = T(x[0] * y[1] - x[1] * y[0]);
			 
			 Transform(A, B, C);

       *ok = true;
			 return;
		}
		//-------------------------------------------------------------------------
		void DefaultInit()
		{
		  cosFi = ro = 0;
			sinFi = 1;
		}

		//-------------------------------------------------------------------------
		void Transform(T A,
			             T B,
									 T C)
		{
		  int sgnB = (B >= 0) ? 1 : -1;
			T den = sgnB * sqrt(A*A + B*B);
      ro = - C / den;
			cosFi = A / den;
			sinFi = B / den;

			//int sgnC = (C >= 0) ? 1 : -1;
			//den = sgnC * sqrt(A*A + B*B);
			//std::cout << std::endl << "cos = " << cosFi << "  sin = " << sinFi
			//	<< "  ro = " << ro << "  |  cos = " << A / den << "  sin = "
			//	<< B / den << "  ro = " << C / den;
		}
		
		//-------------------------------------------------------------------------
		Line &operator = (const Line &l)
		{
		  cosFi = l.cosFi;
			sinFi = l.sinFi;
			ro = l.ro;
			return *this;
		}

		//-------------------------------------------------------------------------
		template <typename PointType>
		bool DoesPointSatisfyModel(const PointType &p,
			                         T prec) const 
		{
		  return (Dist(p) <= prec);
		}
		//-------------------------------------------------------------------------
		template <typename PointType>
		T Dist(const PointType &p) const
		{
			return fabs(T(p.x) * cosFi + T(p.y) * sinFi - ro);
		}

		//-------------------------------------------------------------------------
		template <typename ImgType, typename PointType>
		void Draw(ImgType iter,
			        int w,
							int h,
							int nls,
							const PointType &p1,
			        const PointType &p2) const
		{
			int x1, y1, x2, y2;

			if (fabs(cosFi) > 0.99)
			{
			  y1 = p1.y;
				x1 = int(X(y1) + 0.5);
        y2 = p2.y;
				x2 = int(X(y2) + 0.5);

				if (x1 >= w || x1 < 0 || x2 >= w || x2 < 0)
				{
				  return;
				}
			}
			else
			{
			  x1 = p1.x;
			  y1 = int(Y(x1) + 0.5);
        x2 = p2.x;
			  y2 = int(Y(x2) + 0.5);

				if (y1 >= h  || y1 < 0 || y2 >= h || y2 < 0)
				{
				  return;
				}
			}

			Draw(iter, nls, x1, y1, x2, y2);
		}

		//-------------------------------------------------------------------------
		template <typename ImgType>
		void Draw(ImgType iter,
			        int w,
							int h,
							int nls) const 
		{                                  
			//std::cout << std::endl << cosFi << "  " << ro;
			int x1, y1, x2, y2;
			Ends(&x1, &y1, &x2, &y2, w, h);
			//std::cout << " | x1 = " << x1 << "  y1 = " << y1 << "  x2 = " << x2
			//	<< "  y2 = " << y2;
			Draw(iter, nls, x1, y1, x2, y2);
		}
		
		//-------------------------------------------------------------------------
		template <typename ImgType>
		void Draw(ImgType iter,
							int nls,
							int x1,
							int y1,
							int x2,
							int y2) const 
		{
			typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor;
			typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel, DrawFunctor> SetPixelFunctor;

			AImageMem8::Pixel color_pen = AImageMem8::Pixel(255);
			DrawFunctor draw_functor(color_pen);
			SetPixelFunctor set_pix(iter, nls, draw_functor);
			Vtl::ImAlg::Bres::Line<SetPixelFunctor> line(set_pix);

			set_pix.Set0(x1, y1);
			line.Draw(x2 - x1, y2 - y1);
		}

		//-------------------------------------------------------------------------
		void Ends(int *x1,
							int *y1,
							int *x2,
							int *y2,
							int w,
							int h) const
		{
			if (fabs(cosFi) > 0.99)
			{
			  (*y1) = 0;
        (*x1) = int(Line<T>::X(0) + 0.5);
				(*y2) = h - 1;
				(*x2) = int(Line<T>::X(h - 1) + 0.5);
			}
			else
			{
				(*x1) = 0;
				(*y1) = int(Line<T>::Y(*x1) + 0.5);
				(*x2) = w - 1;
				(*y2) = int(Line<T>::Y(*x2) + 0.5);
			}

			if ((*y1) > h - 1)
			{
				(*y1) = h - 1;
				(*x1) = int(Line<T>::X(h - 1) + 0.5);
			}

			if ((*y1) < 0)
			{
				(*y1) = 0;
				(*x1) = int(Line<T>::X(0) + 0.5);
			}

			if ((*y2) > h - 1)
			{
				(*y2) = h - 1;
				(*x2) = int(Line<T>::X(h - 1) + 0.5);
			}

			if ((*y2) < 0)
			{
				(*y2) = 0;
				(*x2) = int(Line<T>::X(0) + 0.5);
			}

			if ((*x1) > w - 1)
			{
				(*x1) = w - 1;
				(*y1) = int(Line<T>::Y(w - 1));
			}

			if ((*x1) < 0)
			{
				(*x1) = 0;
				(*y1) = int(Line<T>::Y(0));
			}

			if ((*x2) > w - 1)
			{
				(*x2) = w - 1;
				(*y2) = int(Line<T>::Y(w - 1));
			}

			if ((*x2) < 0)
			{
				(*x2) = 0;
				(*y2) = int(Line<T>::Y(0));
			}
		}

		//---------------------------------------------------------------------------
		T Y(int x) const
		{
			return (ro - x * cosFi) / sinFi;
		}

		//---------------------------------------------------------------------------
		T X(int y) const
		{
			return (ro - y * sinFi) / cosFi;
		}
		//-------------------------------------------------------------------------
		static T Ro(int x,
			          int y,
								T fi)
		{
		  return T(x*cos(double(fi)) + y*sin(double(fi)));
		}
	};

	/////////////////////////////////////////////////////////////////////////////
	template <typename ModelType>
	class ModelPointers
	{
	public:
		std::vector<const Point *> points;
		ModelType params;
	};

	/////////////////////////////////////////////////////////////////////////////
	template <typename ModelType>
	class ModelPoints
	{
	public:
		std::vector<Point> points;
		ModelType params;

	public:
    ModelPoints(const ModelType &in_params,
			          const std::vector<int> &pointsNumbs,
								const std::vector<Point> &in_points)
		{
		  params = in_params;

			int size = pointsNumbs.size();
			for (int i = 0; i < size; i++)
			{
			  int j = pointsNumbs[i];
				points.push_back(in_points[j]);
			}
		}

		//-------------------------------------------------------------------------
		template <typename ImgType>
		void Draw(ImgType iter,
			        int w,
							int h,
			        int nls)
		{
			std::sort(points.begin(), points.end());

			int n = points.size() - 1;
			
			for (int i = 0; i < n; i++)
			{
				double d = points[i].Dist(points[i+1]);
			  if ( d < 15)
				{
				  params.Draw<ImgType>(iter, w, h, nls, points[i], points[i+1]);
				}
			}
		}

		//-------------------------------------------------------------------------
		float AproximationError() const
		{
			int n = points.size();
			//std::cout << std::endl << "n = " << n << "  cos = " << params.cosFi
			//	<< "  sin = " << params.sinFi << "  ro = " << params.ro;
			float E = 0;

			for (int i = 0; i < n; i++)
			{
				E += params.Dist(points[i]);
			}
			return E;
		}

		//-----------------------------------------------------------------
		float ChiSqr(float E,
								 float sigma) const
		{
			float s2 = sigma * sigma;
			return (AproximationError() / s2);
		}

		//-----------------------------------------------------------------
		float Probability(float sigma = 1.0) const
		{
			int v = int(points.size()) - ModelType::degOfFr;
			if (v < 0)
			{
				return 0;
			}
			
			float Chi_2 = ChiSqr(AproximationError(), sigma);
			//std::cout << std::cout << "v = " << v << "  chi = " << Chi_2;
			float p = Vtl::NumAlg::Spf::Gamma::Q(float(v) / 2.0f, Chi_2 / 2.0f);
			
			//std::cout << std:: endl << "  E = " << AproximationError() 
			//	<< "  E / n = " << AproximationError() / points.size();
			//	<< "  ChiProb = " << p;

			return p;
		}

	};
}
}
#endif //MODEL_TYPES_H