#ifndef RANDOMIZED_FEATURE_DETECTION_H
#define RANDOMIZED_FEATURE_DETECTION_H

// library includings
#include "nummet.h"
#include "num_linalg.h"

// own includings
#include "ModelTypes.h"
#include "Basics3D/SolveSystem.h"

namespace Vtl
{
namespace CurveDetection
{
  void GetNumberOfSelections(int *N,
		                         int numbInl,
														 int numbPoints,
		                         double prob)
	{
		double a = double(numbInl) / double(numbPoints);
	  double eps = 1.0 - a; // probability that any selected data
																							 // point is outlier
		if (eps <= 0.999)
		{
			double newN = log10(1.0 - prob) / log10(1.0 - pow(1.0 - eps, 4.0));
			
			if ((*N) > newN)
			{
				(*N) = int(newN + 0.5);
				//std::cout << " newN = " << newN;
			}
		}
	}

	/////////////////////////////////////////////////////////////////////////////
	template <typename T, typename ModelType, typename IterType>		
	int FindFeature(std::vector<Point> &points,
		              std::vector< ModelPoints<ModelType> > &models,
									const IterType gradX,
									const IterType gradY,
									int w,
									int h,
									T prec)
	{
		int pointsSize = points.size();
		
		// number of samples we will check
    int N = pointsSize;
    
		std::vector<int> bestInliers(0);
		ModelType bestModel;

		int c = 0;

    for (int counter = 0; counter < N; counter++)
		{
			std::cout << std::endl << "i = " << counter;
      // get sample of n random points
		  std::vector<const Point *> sample;

			// if points are too close or too far 
		  if (!GetSample<ModelType>(points, sample))
		  {                              
				std::cout << std::cout << "Points are too close / too far";
			  continue;
		  }

      // obtain model parameters
			bool ok;
			ModelType model(sample, gradX, gradY, w, h, &ok);

			if (!ok)
			{
				std::cout << std::endl << "   points have diferent gradient direction";
				//counter--;
				continue;
			}
                       
			//std::cout << std::endl << sample[0]->x << " " << sample[0]->y
			//	<< "  " << sample[1]->x << " " << sample[1]->y << " | model: cos = "
			//	<< model.cosFi << " ro = " << model.ro;

			//numbers of points which satisfy to the found model

			c++;

			std::vector<int> inliers;

			// check if points satisfy model
			for (int i = 0; i < pointsSize; i++)
			{
				if (model.DoesPointSatisfyModel(points[i], prec))
				{
					inliers.push_back(i);
				}
			}
			std::cout << " => inlNumb = " << inliers.size();

			//if (inliers.size() == 165)
			//	std::cout << "";
			
			// if current model is better than the previous best model
			if (int(inliers.size()) > int(bestInliers.size()))
			{
				std::cout << std::endl << sample[0]->x << " " << sample[0]->y << "  " 
					<< sample[1]->x << " " << sample[1]->y << " | model: cos = "
					<< model.cosFi << " ro = " << model.ro << " => inlNumb = "
					<< inliers.size() << " NEW BEST";

				bestInliers = inliers;
				bestModel = model;
				//GetNumberOfSelections(&N, inliers.size(), pointsSize, prob);

				int newN = int( - pointsSize * log10(10 * double(bestInliers.size()) / double(pointsSize)) );
				std::cout << " N = " << N << "  newN = " << newN;

				if (newN < N)
				{
				  N = newN;
				}
			}
		}

		if (c > 0)
		{
			std::cout << std::endl << "N = " << N << "  c = " << c;
			models.push_back(ModelPoints<ModelType>(bestModel, bestInliers, points));

			//std::cout << std::endl << std::endl << "pointsNumb = " << points.size() << std::endl;
			int k = bestInliers.size();
			for (int i = k - 1; i >= 0; i--)
			{
				int j = bestInliers[i];
				//std::cout << j;
				points.erase(points.begin() + j);
				//std::cout << " ok  ";
			}
			//std::cout << " -> " << points.size() << "  " << k;
			return k;
		}
		return 0;
	}

	/////////////////////////////////////////////////////////////////////////////
	template <typename T, typename ModelType>
	void SelectTrueModels(std::vector< ModelPoints<ModelType> > &models,
		                    int absTh,
		                    T persTh,
												T probTh)
	{
	  // find model containing the biggest number of points
		int max = 0;

    for (int i = 0; i < int(models.size()); i++)
		{
		  int m = models[i].points.size();

			if (m > max)
			{
			  max = m;
				std::cout << std::endl << "new max: cos = " << models[i].params.cosFi
					<< "  ro = " << models[i].params.ro;
			}
		}

		// set threshold
		T thresh = std::max(persTh * max, T(absTh));
		std::cout << std::endl << "max = " << max << "  thresh = " << thresh;

		std::cout << std::endl << "mNumb = " << models.size();
		// select models containing more than thresh points
		for (int i = models.size() - 1; i >= 0; i--)
		{
		  if (models[i].points.size() < thresh || models[i].Probability() < probTh)
			{
			  models.erase(models.begin() + i);
			}
		}
		std::cout << std::endl << "mNumb = " << models.size();
	}

	/////////////////////////////////////////////////////////////////////////////
	template <typename IterType, typename ModelType >
	void DrawModels(const IterType iterEdges,
		              const IterType iterOrig,
		              std::vector< ModelPoints<ModelType> > &models,
		              int w,
						      int h,
									int nls)
	{
		// create destination images
	  std::auto_ptr<AImage> _dst1(::AImageCreate(0, AImage::IMGC_GRAYSCALE, w, h, 8));
		AImageMem8* dst1 = dynamic_cast<AImageMem8*>(_dst1.get());

		std::auto_ptr<AImage> _dst2(::AImageCreate(0, AImage::IMGC_GRAYSCALE, w, h, 8));
		AImageMem8* dst2 = dynamic_cast<AImageMem8*>(_dst2.get());

		std::auto_ptr<AImage> _dst3(::AImageCreate(0, AImage::IMGC_GRAYSCALE, w, h, 8));
		AImageMem8* dst3 = dynamic_cast<AImageMem8*>(_dst3.get());

		// copy input image as background
		IterType p_s1 = iterEdges;
		IterType p_s2 = iterOrig;
		AImageMem8::Pixel *p_d1 = (*dst1)[0];
		AImageMem8::Pixel *p_d2 = (*dst2)[0];
		AImageMem8::Pixel *p_d3 = (*dst3)[0];

		for (int y = 0; y < h; y++)
		{
			for (int x = 0; x < w; x++)
			{
				(*p_d1) = (*p_d2) = AImageMem8::Pixel(fabs(*p_s1));
				(*p_d3) = AImageMem8::Pixel(fabs(*p_s2));

				p_s1++;
				p_s2++;
				p_d1++;
				p_d2++;
				p_d3++;
			}
		}
		
		// draw models
    int count = 0;
		
		for (int i = 0; i < int(models.size()); i++)
		{
			std::cout << std::endl << std::endl << "pNumb = " << models[i].points.size()
				<< "  " << models[i].Probability();
			
			models[i].Draw((*dst1)[0], w, h, nls);//              std::cout << "    1";
			models[i].params.Draw((*dst2)[0], w, h, nls);// std::cout << "  2";
			models[i].params.Draw((*dst3)[0], w, h, nls);// std::cout << "  3";
			count++;
		}

    std::cout << std::endl << std::endl << "lines number:  " << count;
		::AImageSave(dst1, "Lines1.tif");	
		::AImageSave(dst2, "Lines2.tif");	
		::AImageSave(dst3, "Lines3.tif");					
	}

	/////////////////////////////////////////////////////////////////////////////
	template <typename T, typename ModelType>
	void Refinement(ModelPoints<ModelType> &model)
	{
		std::cout << std::endl << "pr = ";
		std::cout << model.Probability() << "  cos = "
			<< model.params.cosFi << "  ro = " << model.params.ro << "  |  ";

		int pointsNumb = model.points.size();
    Vtl::NumAlg::Lin::ublas::matrix<T> systM(pointsNumb, 3, false);
		
		for (int i = 0; i < pointsNumb; i++)
		{
		  systM(i, 0) = T(model.points[i].x);
      systM(i, 1) = T(model.points[i].y);
			systM(i, 2) = T(1);
		}

		Vtl::NumAlg::Lin::ublas::c_vector<T, 3> params;

		Basics3D::SolveSystem(params, systM);

		//std::cout << std::endl << params(0) << "  " << params(1) << "  " << params(2);
		model.params.Transform(params(0), params(1), params(2));

		std::cout << std::endl << "pr = " << model.Probability() << "  cos = "
			<< model.params.cosFi << "  ro = " << model.params.ro;
	}
	
	/////////////////////////////////////////////////////////////////////////////
  template <typename T, typename IterType, typename ImgType,
            typename ModelType, typename PointType>
	void RandPatternDetection(std::vector<PointType> &points,
		                        const IterType gradX,
														const IterType gradY,
														ImgType imgEdges,
														ImgType imgOrig,
		                        int width,
											      int height,
														int nls,
											      T prec,
														int absTh,
														T relTh,
														T probTh)
	{
		// container to save all found models
		std::vector< ModelPoints<ModelType> > models;
 
		// try to find one more feature
		int max = 0;
		int k = 10 * ModelType::sampleSize;
		
		while(int(points.size()) > k)
		{
			std::cout << std::endl << "points numb = " << points.size();

		  int m = FindFeature(points, models, gradX, gradY, width, height, prec);
			Refinement<T, ModelType>(models[models.size() - 1]);

			if (m > max)
			{
			  max = m;
			}
			else if (m < 0.5 * relTh * max || m < absTh)
			{
				std::cout << "";
				break;
			}
		}

		SelectTrueModels(models, absTh, relTh, probTh);
		DrawModels(imgEdges, imgOrig, models, width, height, nls);
	}
}
}

#endif // RANDOMIZED_FEATURE_DETECTION_H0