#ifndef SAVE_IMAGE_ARRAY_H
#define SAVE_IMAGE_ARRAY_H

// system includings
#include <iomanip>
#include <string>

// lifrary includings
#include "aimage.h"
#include "yfilesys.h"
#include "plf_functors.h"


namespace Tests
{
	inline void CreateFileName(const char *fname,
											const char *postfix,
											const char *ext,
											std::string &res)
	{
		std::ostringstream os;
		os <<	YFileSys::getNameWithoutExt(fname) << std::setprecision(2)
			<< std::setiosflags(std::ios::showpoint | std::ios::fixed) 
			<< postfix << ext;
		res = os.str();
	}

	/////////////////////////////////////////////////////////////////////////////

	inline Vtl::Plf::F11::Comb< uint8_t, Vtl::Plf::F11::GammaCorMult<float>,
											 Vtl::Plf::F11::Convert2ByteBounded > GammaCorFunctor()
	{
		//    ()
		const float g2_Mult = 1;
		Vtl::Plf::F11::GammaCorMult<float> lin_3mod(255, 2.2, 1.0 * g2_Mult);

		//     float    byte. , 
		//   float     0-255 -    
		//    
		Vtl::Plf::F11::Convert2ByteBounded c_to_Byte;

		//  ,   
		Vtl::Plf::F11::Comb< uint8_t, Vtl::Plf::F11::GammaCorMult<float>,
						Vtl::Plf::F11::Convert2ByteBounded > adj_3mod(lin_3mod, c_to_Byte);

		return adj_3mod;
	}

	/////////////////////////////////////////////////////////////////////////////

	template< typename T >
	void SaveImageArray(const T *data,
											int w,
											int h,
											const char *baseName,
											const char *postfix)
	{
		// Create destination image
		std::auto_ptr<AImage> _img(AImageCreate(0, AImage::IMGC_GRAYSCALE, w, h, 8));
		AImageMem8* img = dynamic_cast<AImageMem8*>(_img.get());

		//        
		//    ()
		Vtl::ImAlg::Tools::copy_1_1(data, w, h, w, (*img)[0], img->NextLineShift(),GammaCorFunctor());

		std::string outfname;
		CreateFileName(baseName, postfix, ".tif", outfname);

		::AImageSave(img, outfname.c_str());
	}

  /////////////////////////////////////////////////////////////////////////////

	template< typename T >
	void SaveColourImageArray(const T *dataR,
                            const T *dataG,
                            const T *dataB,
											      int w,
											      int h,
											      const char *baseName,
											      const char *postfix)
	{
		// Create destination image
		std::auto_ptr<AImage> _img(AImageCreate(0, AImage::IMGC_RGB, w, h, 8));
		AImageMemRgb8* img = dynamic_cast<AImageMemRgb8*>(_img.get());

    AImageMem8* r = img->GetPlane8(2);
		AImageMem8* g = img->GetPlane8(1);
		AImageMem8* b = img->GetPlane8(0);

    int nls = r->NextLineShift();

		//        
		//    ()
		Vtl::ImAlg::Tools::copy_1_1(dataR, w, h, w, (*r)[0], nls, GammaCorFunctor());
    Vtl::ImAlg::Tools::copy_1_1(dataG, w, h, w, (*g)[0], nls, GammaCorFunctor());
    Vtl::ImAlg::Tools::copy_1_1(dataB, w, h, w, (*b)[0], nls, GammaCorFunctor());

		std::string outfname;
		CreateFileName(baseName, postfix, ".bmp", outfname);

		::AImageSave(img, outfname.c_str());
	}

	/////////////////////////////////////////////////////////////////////////////

	template< typename T >
	void NormalizeImage(const T *src,
											T * dst,
											int w,
											int h,
                      T minDes = T(0),
                      T maxDes = T(255))
	{
		// define max and min values in the source array
		T max = 0;
		T min = 0;

		for (int y = 0; y < h; y++)
		{
			for (int x = 0; x < w; x++)
			{
				T val = *(src + y * w + x);
				if (val > max)
				{
					max = val;
				}
				else if (val < min)
				{
					min = val;
				}
			}
		}

		std::cout << std::endl << "max = " << max << ",  min = " << min;

		// shift and scale array values into range [0, 255]
		if (max != min)
		{
			float c = (maxDes - minDes) / (max - min);
	    
			for (int y = 0; y < h; y++)
			{
				for (int x = 0; x < w; x++)
				{
					T val = *(src + y * w + x);
					*(dst + y * w + x) = T((val - min) * c + minDes);
				}
			}
		}
	}

	/////////////////////////////////////////////////////////////////////////////

	template< typename T >
	void SaveNormImageArray(const T *data,
													int w,
													int h,
													const char *baseName,
													const char *postfix)
	{
		// shift and scale image into the range [0, 255]
		std::vector<T> temp(w * h);
		NormalizeImage(&(data[0]), &(temp[0]), w, h);

		// save image
		SaveImageArray(&(temp[0]), w, h, baseName, postfix);
	}

	/////////////////////////////////////////////////////////////////////////////

	template< typename T >
	void SaveAbsImageArray(const T *data,
												 int w,
												 int h,
												 const char *baseName,
												 const char *postfix)
	{
		//        
		std::vector<T> temp(w * h);
			
		// Calculate absolute value
		Vtl::Plf::F11::Abs<T> abs;
		Vtl::ImAlg::Tools::copy_1_1(data, w, h, w, &(temp[0]), w, abs);

		// save image
		SaveImageArray(&(temp[0]), w, h, baseName, postfix);
	}
	 
	/////////////////////////////////////////////////////////////////////////////
  // function is used while curvature estimation, NEEDS CORRECTION
 	template<typename T,class ConvFunctor>
	inline void SaveImage(AImageMem8* img,
								 const T *data,
								 int w,
								 int h,
								 int nls,
								 ConvFunctor func,
								 float sigma,
								 const char *base_name,
								 const char *postfix)
	{
		//          ()
		std::vector<T> _temp(w * h);
		T *temp = &(_temp[0]);
		
		Vtl::Plf::F11::Abs<T> abs;
		Vtl::ImAlg::Tools::copy_1_1(data,w,h,w,temp,w,abs);

		Vtl::ImAlg::Tools::copy_1_1(temp,w,h,w,(*img)[0],img->NextLineShift(),func);

		//    
		{
			//  :
			std::ostringstream os;
			os <<	YFileSys::getNameWithoutExt(base_name)
				<< std::setprecision(2) << std::setiosflags(std::ios::showpoint|std::ios::fixed)
				<< "_#_"<<postfix<<"_sigma=" ;
			if(sigma<10)
				os << "0";
			os << sigma 
				<< ".tif";//YFileSys::getExt(base_name);
			std::string outfname=os.str();
			// 
			::AImageSave(img,outfname.c_str());
		}
	}

	/////////////////////////////////////////////////////////////////////////////
  // function is used while curvature estimation, NEEDS CORRECTION
	inline void SaveImage(AImageMem8* img, float sigma, const char *base_name, const char *postfix)
	{
		//          ()
		
		//    
		{
			//  :
			std::ostringstream os;
			os <<	YFileSys::getNameWithoutExt(base_name)
				<< std::setprecision(2) << std::setiosflags(std::ios::showpoint|std::ios::fixed)
				<< "_#_"<<postfix<<"_sigma=" ;
			if(sigma<10)
				os << "0";
			os << sigma 
				<< YFileSys::getExt(base_name);
			std::string outfname=os.str();
			// 
			::AImageSave(img,outfname.c_str());
		}
	}
} // namespace Tests

#endif // SAVE_IMAGE_ARRAY_H