#ifndef EDGE_CURVATURE_H
#define EDGE_CURVATURE_H

// own includes
#include "CurvatureScale.h"

namespace Vtl
{
namespace Curvature
{
  template< typename T >
  class EdgeCurvature
  {
    //  Arrays for keeping auxiliary data (images)
    std::vector<T> Lx;
	  std::vector<T> Ly;
	  std::vector<T> Lxxx;
	  std::vector<T> Lxxy;
	  std::vector<T> Lxyy;
	  std::vector<T> Lyyy;
	  std::vector<T> Lxxxx;
	  std::vector<T> Lyyyy;
	  std::vector<T> Lxxyy;
	  std::vector<T> Lxxxy;
	  std::vector<T> Lxyyy;

    T* data;

    int w;
    int h;
    int size;

	  //-------------------------------------------------------------------------
  public:
	  EdgeCurvature(T* in_data,
                  int in_w,
		              int in_h)
	  {
      data = in_data;

      w = in_w;
      h = in_h;
      size = w * h;

		  Lx.resize(size);
		  Ly.resize(size);
      Lxxx.resize(size);
      Lxxy.resize(size);
      Lxyy.resize(size);
      Lyyy.resize(size);
      Lxxxx.resize(size);
      Lyyyy.resize(size);
      Lxxyy.resize(size);
      Lxxxy.resize(size);
      Lxyyy.resize(size);
	  }

    //-------------------------------------------------------------------------
    void CalculateDerivatives_direct(T sigma,
                                     T nSigmasEnv)
    {
      //    
		  memcpy(&(Lx[0]), data, size * sizeof(T));
		  memcpy(&(Ly[0]), data, size * sizeof(T));
  	
		  memcpy(&(Lxxx[0]), data, size * sizeof(T));
		  memcpy(&(Lxxy[0]), data, size * sizeof(T));
		  memcpy(&(Lxyy[0]), data, size * sizeof(T));
		  memcpy(&(Lyyy[0]), data, size * sizeof(T));

		  memcpy(&(Lxxxx[0]), data, size * sizeof(T));
		  memcpy(&(Lxxxy[0]), data, size * sizeof(T));
		  memcpy(&(Lxxyy[0]), data, size * sizeof(T));
		  memcpy(&(Lxyyy[0]), data, size * sizeof(T));
		  memcpy(&(Lyyyy[0]), data, size * sizeof(T));

		  //  ,    
      //     1,3,4 
      //  
		  typedef Vtl::ImAlg::WndFun::Gauss<float, 0, false> FF0; 
		  FF0 f0(sigma, nSigmasEnv);
		  typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K0;
		  K0 ker0(f0);
		  Vtl::ImAlg::Conv::ConvInplace<float, K0> g0(ker0);

		  // 1 
		  typedef Vtl::ImAlg::WndFun::Gauss<float, 1, false> FF1; 
		  FF1 f1(sigma, nSigmasEnv);
		  typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K1;
		  K1 ker1(f1);
		  Vtl::ImAlg::Conv::ConvInplace<float, K1> g1(ker1);

      //2 
			typedef Vtl::ImAlg::WndFun::Gauss<float, 2, false> FF2; 
			FF2 f2(sigma, nSigmasEnv);
			typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K2;
			K2 ker2(f2);
			Vtl::ImAlg::Conv::ConvInplace<float,K2> g2(ker2);

		  // 3 
		  typedef Vtl::ImAlg::WndFun::Gauss<float, 3, false> FF3; 
		  FF3 f3(sigma, nSigmasEnv);
		  typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K3;
		  K3 ker3(f3);
		  Vtl::ImAlg::Conv::ConvInplace<float, K3> g3(ker3);

		  // 4 
		  typedef Vtl::ImAlg::WndFun::Gauss<float, 4, false> FF4; 
		  FF4 f4(sigma, nSigmasEnv);
		  typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K4;
		  K4 ker4(f4);
		  Vtl::ImAlg::Conv::ConvInplace<float, K4> g4(ker4);

		  //  
		  // 1 
		  //  x:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lx[0]), w, h, w, g1, g0);
		  //  y:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Ly[0]), w, h, w, g0, g1);

		  // 3 
		  //  xxx:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxxx[0]), w, h, w, g3, g0);
		  //  xxy:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxxy[0]), w, h, w, g2, g1);
		  //  xyy:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxyy[0]), w, h, w, g1, g2);
		  //  yyy:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lyyy[0]), w, h, w, g0, g3);

		  // 4 
		  //  xxxx:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxxxx[0]), w, h, w, g4, g0);
		  //  xxxy:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxxxy[0]), w, h, w, g3, g1);
		  //  xxyy:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxxyy[0]), w, h, w, g2, g2);
		  //  xyyy:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxyyy[0]), w, h, w, g1, g3);
		  //  yyyy:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lyyyy[0]), w, h, w, g0, g4);
    }

    //-------------------------------------------------------------------------
    T operator() (int i)
	  {
		  T Lx2 = sqr(Lx[i]);
		  T Ly2 = sqr(Ly[i]);
		  T M = sqrt(Lx2 + Ly2);
		  //if (M > 0.0001)
		  //{
			  T LxLy = Lx[i] * Ly[i];
			  T LxLy2 = sqr(LxLy) ;
			  T nom = (Lxxxx[i] + Lyyyy[i]) * LxLy2
          + Lxxyy[i] * (sqr(Lx2) + sqr(Ly2) - 4 * LxLy2) 
          + 2 * LxLy * (Ly2 - Lx2) * (Lxxxy[i] - Lxyyy[i]);
			  T den = Lxxx[i] * Lx[i] * Lx2 + Lyyy[i] * Ly[i] * Ly2
          + 3 * LxLy * (Lxxy[i] * Lx[i] + Lxyy[i] * Ly[i]);
			  den *= (-M);
			  T R = CurvatureScale<T>::kappa_to_radius(nom, den);
			  return CurvatureScale<T>::radius_to_value(R);
		  //}
		  //return CurvatureScale<T>::radius_to_value(0);
	  }
  };

} // Curvature
} // Vtl

#endif //EDGE_CURVATURE_H