#ifndef EDT_EDGE_BLUR_H
#define EDT_EDGE_BLUR_H

// system includings
#define _USE_MATH_DEFINES
#include <cmath>

template <typename T>
void EdtEdgeBlur(T* edges,
                 T* Lx,
                 T* Ly,
                 T* edt,
                 int* edgeX,
                 int* edgeY,
                 int w,
                 int h,
                 T sigma,
                 T lev)
{
  int size = w * h;
  T sigma2 = sigma * sigma;

  for (int y = 0; y < h; y++)
  {
    for (int x = 0; x < w; x++)
    {
      int shift = y * w + x;

      T dx = T(x - *(edgeX + shift));
      T dy = T(y - *(edgeY + shift));
      T l = sqrt(dx * dx + dy * dy);

      dx /= l;
      dy /= l;

      T d2 = *(edt + shift);
      T mod = lev * std::exp(-d2 / 2 / sigma2 );
      *(edges + shift) = mod;
      *(Lx + shift) = mod * dx;
      *(Ly + shift) = mod * dy;
    }
  }
}

#endif EDT_EDGE_BLUR_H