#ifndef PREPROCESSING_H
#define PREPROCESSING_H

namespace Vtl
{
namespace Curvature
{
  template<typename T>
	void preproc_image_DoNothing(const T* data,
		                           const int w,
															 const int h,
															 T* result)
  {
		const int sz=w*h;
		memcpy(result,data,sz*sizeof(T));
	}

	/////////////////////////////////////////////////////////////////////////////

	template<typename T>
	void preproc_image_StatDiff(const T* data,
		                          const int w,
															const int h,
															T* result,
															float sigma,
															float nsigmas_env,
															float amplification = 12.0f,
															float sigma_desired = 85.0f,
															float mean_desired = 128.0f,
															float alpha = 0.8f)
  {
		const int sz = w * h;

		//     (). 
		//   1-    .
		std::vector<T> _data0(sz);
		std::vector<T> _data1(sz);
		std::vector<T> _data2(sz);

		//        .
		T *mean = &(_data0[0]);
		T *sq = &(_data1[0]);
		T *GradModul = &(_data2[0]);

		//    
		memcpy(mean, data, sz * sizeof(T));
		Vtl::ImAlg::Tools::copy_1_1(data, w, h, w, sq, w, Vtl::Plf::F11::sqr<T>);
		
		// ,       

		// 
		typedef Vtl::ImAlg::WndFun::Gauss<float,0, false> FF0; 
		FF0 f0(sigma,nsigmas_env);
		typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K0;
		K0 ker0(f0);
		Vtl::ImAlg::Conv::ConvInplace<float,K0> g0(ker0);

		//  
		Vtl::ImAlg::Conv::Conv2D::conv2DInplace(mean, w, h, w, g0, g0);
		Vtl::ImAlg::Conv::Conv2D::conv2DInplace(sq, w, h, w, g0, g0);

		//    .
		Vtl::Plf::F21::Disp<T> dsp;
		Vtl::ImAlg::Tools::copy_2_1(mean, w, h, w, sq, w, sq, w, dsp);
		
		//  
		Vtl::Plf::F31::StatDiff<T> stdif(amplification, sigma_desired,
			                                                    mean_desired, alpha); 
		Vtl::ImAlg::Tools::copy_3_1(data, w, h, w, mean, w, sq, w, result, w, stdif);
  }


	/////////////////////////////////////////////////////////////////////////////

	template<typename T>
	T step_edge_zero(T Lx,T Ly,T Lxx,T Lxy,T Lyy)
	{
		using namespace std;
		using namespace Vtl;//::Plf::F11;
		T Lx2 = sqr( Lx );
		T Ly2 = sqr( Ly );
		T M = sqrt( Lx2 + Ly2 );
		T nom = Lxx*Lx2 + 2*Lxy*Lx*Ly + Lyy*Ly2;
		T den=sqr(M);
		nom=abs(nom);
		T v = (nom<den*1000.0f) ? nom/den/0.26f : 0;
		return v ;
	}

  //---------------------------------------------------------------------------

	template<typename T>
	void preproc_image_ZeroCross(const T* data,const int w,const int h,T* result,float sigma,float nsigmas_env)
	{
		const int sz=w*h;
		//     (). 
		//   1-    ,   , ,  .
		std::vector<T> _data0(sz);
		std::vector<T> _data1(sz);
		std::vector<T> _data2(sz);
		std::vector<T> _data3(sz);
		std::vector<T> _data4(sz);

		//        .
		T *Lx=&(_data0[0]);
		T *Ly=&(_data1[0]);
		T *Lxx=&(_data2[0]);
		T *Lxy=&(_data3[0]);
		T *Lyy=&(_data4[0]);
		//    
		memcpy(Lx   ,data,sz*sizeof(T));
		memcpy(Ly   ,data,sz*sizeof(T));
		memcpy(Lxx  ,data,sz*sizeof(T));
		memcpy(Lxy  ,data,sz*sizeof(T));
		memcpy(Lyy  ,data,sz*sizeof(T));

		// ,         1,3,4 
		// 
		typedef Vtl::ImAlg::WndFun::Gauss<float,0, false> FF0; 
		FF0 f0(sigma,nsigmas_env);
		typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K0;
		K0 ker0(f0);
		Vtl::ImAlg::Conv::ConvInplace<float,K0> g0(ker0);
		//1 
		typedef Vtl::ImAlg::WndFun::Gauss<float,1,false> FF1; 
		FF1 f1(sigma,nsigmas_env);
		typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K1;
		K1 ker1(f1);
		Vtl::ImAlg::Conv::ConvInplace<float,K1> g1(ker1);
		//2 
		typedef Vtl::ImAlg::WndFun::Gauss<float,2,false> FF2; 
		FF2 f2(sigma,nsigmas_env);
		typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K2;
		K2 ker2(f2);
		Vtl::ImAlg::Conv::ConvInplace<float,K2> g2(ker2);
		//  
		// 1 
		//  Lx:
		Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lx,w,h,w,g1,g0);
		//  Ly:
		Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Ly,w,h,w,g0,g1);
		// 2 
		//  Lxx:
		Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lxx,w,h,w,g2,g0);
		//  Lxy:
		Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lxy,w,h,w,g1,g1);
		//  Lyy:
		Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lyy,w,h,w,g0,g2);

		// LiLijLj
		for(int i=0; i<sz; ++i)
			result[i] = sigma*sigma*Vtl::Curvature::step_edge_zero(Lx[i],Ly[i],Lxx[i],Lxy[i],Lyy[i]);
	}
} // namespace Curvature
} // namespace Vtl


#endif // PREPROCESSING_H