#include <memory>//std::auto_ptr
#include <vector>
#include <cmath>
#include "aimage.h"
#include "consolap.h"
#include "alg_bresenham.h"

#include "tests.h"

using  namespace Vtl;

int Tests::test_alg_bresenham(int argc,char* argv[])
{
	//   
	const int w=1024;
	const int h=1024;
	const int bits=8;
	std::auto_ptr<AImage> _im(::AImageCreate(0,AImage::IMGC_GRAYSCALE,w,h,bits));
	AImageMem8* img=dynamic_cast<AImageMem8*>(_im.get());
	//   -  
	if(img)
	{
		//      
		AImageMem8::Pixel color_back=0; 
		img->SetTo(color_back);
		//   
		AImageMem8::Pixel color_pen = AImageMem8::Pixel((1<<bits)-1); 

#if 1
		typedef Vtl::ImAlg::Bres::DrawDotted<AImageMem8::Pixel> DrawFunctor;
		DrawFunctor draw_functor(color_pen,color_pen/2,0xFFFF0000);
#else
		typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor;
		DrawFunctor draw_functor(color_pen);
#endif

		typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel,DrawFunctor > SetPixelFunctor;
		SetPixelFunctor set_pix((*img)[0],img->NextLineShift(),draw_functor);

		Vtl::ImAlg::Bres::Line<SetPixelFunctor> line(set_pix);

		//  
		const int c_x=w/2;
		const int c_y=h/2;
		int dx=w/6;
		int dy=h/3;

		//  
		set_pix.Set0(c_x,c_y);
		//   dx,dy  
		line.Draw(+dx,+dy);
		//   dx,dy  
		line.Draw(+dx,-dy);
		//   dx,dy  
		line.Draw(-dx,-dy);
		//   dx,dy  
		line.Draw(-dx,+dy);

		dy=h/11;
		//  
		set_pix.Set0(c_x,c_y);
		//   dx,dy  
		line.Draw(-dx,-dy);
		//   dx,dy  
		line.Draw(-dx,+dy);
		//   dx,dy  
		line.Draw(+dx,+dy);
		//   dx,dy  
		line.Draw(+dx,-dy);

		{
			const int circ_x=w/4;
			const int circ_y=h/4;
			const int radius=h*3/16;
			const int points=8;//5 degree steps in circle
			std::vector<int> x;
			std::vector<int> y;
			x.resize(points+1);
			y.resize(points+1);
			for(int i=0; i<= points ; ++i)
			{
				x[i]=int(circ_x+radius*cos(2*3.1415926/points*i)+0.5);
				y[i]=int(circ_y+radius*sin(2*3.1415926/points*i)+0.5);
			}
			set_pix.Set0(x[0],y[0]);
			for(int i=1; i<= points ; ++i)
			{
				line.Draw(x[i]-x[i-1],y[i]-y[i-1]);
			}
		}


		//  
		AImageSave(img,"alg_bresenham.pgm",indicator);
	}
	return 0;
}

