//#include <stdlib.h>
#include "tests.h"
#include "Basics3D/Includings.h"

// arguments:
// 1 2 Map_rect1.pgm Map_rect2.pgm Map_rect1.key Map_rect2.key Map_rect1-Map_rect2.dist.key
// 1 2 Han_rect0.pgm Han_rect1.pgm Han_rect0.key Han_rect1.key Han_rect0-Han_rect1.dist.key
// 1 2 Pl_rect0.pgm Pl_rect1.pgm Pl_rect0.key Pl_rect1.key Pl_rect0-Pl_rect1.dist.key

using namespace Basics3D;

///////////////////////////////////////////////////////////////////////////////////////

int Tests::test_basics3D(int argc, char* argv[])
{
	if (argc < 5)
	{
		std::cout << "Arguments:" << std::endl << "0/\"else number\" if the scene is/isn't plane," <<
			std::endl << "n - number of input images," << std::endl << "n input images *.pgm," <<
			std::endl << "n *.key files - output of sift," << 	std::endl <<
			"(n - 1) *.dist.key files - output of Corr" << std::endl;
		return -1;
	}

	const int    PLANE      = atoi(argv[1]);   // 0/"else number" if the scene is/isn't plane
	const int    N          = atoi(argv[2]);   // number of input images
	const double T          = 3.5;             // threshold for difference between the best 
											   // corresponding point and next point
	const double SIGMA      = 0.5;             // inaccuracy of feature points detector
	const int    PLANE_NUMB = 2;               // number of planes on images

	if (N < 2)
	{
		std::cout << "You have to use 2 or more images" << std::endl;
		return -1;
	}
	std::cout << "number of input images = " << N << std::endl;
	std::cout << "Is the scene plane?  " << PLANE << std::endl;

	/////////////////////////////////////////////////////////////////////////////////////
	// Loadings

	T_ImagesSet<AImageMem8> img(argv, N);
	
	std::vector<C_Points2DParamSet> PointsParamSet;
	std::vector<T_Points2DSet<C_Point2D>> PointsSet;
	LoadPoints(PointsParamSet, PointsSet, argv, N, T, img[0]->Height());

	///////////////////////////////////////////////////////////////////////////////////////
	MarkCorrespondences("Corr_", -1, img, PointsSet, 15);

	std::vector<matrix_3_3> Matrix(PLANE_NUMB);
// ???????   ?     (3,3),     33     0,1,2 !!!!
//        .   ,    .   , 
//    Debug-       .  Debug-  Boost          .
//		for (int i = 0; i < PLANE_NUMB; i++)
//		{
//			Matrix[i](3, 3);
//		}

	std::vector<T_Points2DSet<C_Point2D>> Inliers(N - 1);
	for (int i = 0; i < N - 1; i++)
	{
		Inliers[i].InitGroupSize(2);
	}

	/////////////////////////////////////////////////////////////////////////////////////
	if (PLANE)
	{
		for (int i = 0; i < N - 1; i++)
		{
			std::cout << std::endl << "=====================================================" 
				<< std::endl << "Images " << i << " and " << i + 1 << std::endl;

			int size = PointsSet[i].Size();

			// HOMOGRAPHIES
			for (int j = 0; j < PLANE_NUMB; j++)
			{
				C_HomographyMatrix::ApplyRansac(Matrix, Inliers, PointsSet, img,
												  i, j, size, SIGMA);
	  C_HomographyMatrix::RecomputeWithAllPoints(Matrix, Inliers, PointsParamSet,
															 img, i, j, SIGMA, T);
			}
			
			if (PLANE_NUMB >= 2)
			{
				// FUNDAMENTAL MATRIX
				T_MatrixOperations::TransformIntoNormCoord(Matrix, img[0]->Width(),
															 img[0]->Height(), SIGMA);
				std::vector<matrix_3_3> FundM;
				C_FundamentalMatrix::FromHomographies(FundM, Matrix);
				
				T_MatrixOperations::TransformIntoPixelCoord(FundM, img[0]->Width(),
															 img[0]->Height(), SIGMA);

				DrawEpipolarLines(img, FundM[0], PointsSet, 0, 20);
				T_MatrixOperations::Scale(FundM[0], FundM[0]);
				T_MatrixOperations::Save(FundM[0], "FundMfromH.txt");

				T_Points2DSet<C_Point2D> inlFund(2);
				T_Model<C_FundamentalMatrix> model;
				PointsParamSet[i].EnableAll();
				PointsParamSet[i + 1].EnableAll();
				model.CountInliers(inlFund, PointsParamSet[i], PointsParamSet[i + 1],
									 FundM[0], SIGMA, T);

				std::cout << std::endl << std::endl << "number of inliers for FM:  "
					<< inlFund.Size() << std::endl;
				MarkCorrespondences("Corr_FM", img, inlFund, N, -1); 

				///*temp*/std::cout << std::endl << "Inliers for FM: " << Inliers[0].Size();

				///*temp*/T_Points2DSet<C_Point2D> inl(2);
				///*temp*/T_Model<C_FundamentalMatrix> model0(Inliers[0]);
				///*temp*/Inliers[0].EnableAll();
				///*temp*/int numbInl = 0;
				///*temp*/for (int m = 0; m < Inliers[0].Size(); m++)
				///*temp*/{
				///*temp*/	if (model0.IsInlier(FundM[0], m, SIGMA))
				///*temp*/	{
				///*temp*/		numbInl ++;
				///*temp*/	}
				///*temp*/}
				///*temp*/std::cout << std::endl << std::endl << "number of inliers for FM:  "
				///*temp*/	<< numbInl;
				
				// STRUCTURE COMPUTATION
				C_StructureComputation structure(Inliers[0], FundM[0], img[0]->Width(),
																		 img[0]->Height(),SIGMA);
				structure.Reconstruction(1.0 / 1.15);
				//structure.OptimalSolution();
				//structure.BuildModelFaces("ModelFaces.wrl", "triang.txt");
				//structure.BuildModelPoints("Model.wrl");
			}
		}
	}
	// if the the scene isn't plane
//*************************************************************************
// NOW THIS PART WORKS
//*************************************************************************
	else
	{
		// FUNDAMENTAL MATRIX
		for (int i = 0; i < N - 1; i++)
		{
			C_FundamentalMatrix::ApplyRansac(Matrix, Inliers, PointsSet, img, i, SIGMA);
			C_FundamentalMatrix::RecomputeWithAllPoints(Matrix, Inliers, PointsParamSet,
														  img, i, SIGMA, T);
			MarkCorrespondences("Corr_FM", img, Inliers[i], N, 15);
		}
		std::cout << std::endl << std::endl << "number of inliers for FM:  "
					<< Inliers[0].Size() << std::endl;
		// STRUCTURE COMPUTATION
		C_StructureComputation structure(Inliers[0], Matrix[0], img[0]->Width(),
																		img[0]->Height(), SIGMA);
		structure.Reconstruction(1.0 / 1.15);
		//structure.OptimalSolution();
	}

	/////////////////////////////////////////////////////////////////////////////////////
	// Find common feature points
	/////////////////////////////////////////////////////////////////////////////////////
	
	//T_Points2DSet<C_Point2D> commonPoints;
	//FindCommonFeaturePoints(commonPoints, Inliers, N);
	//MarkCorrespondences("CommonCorr_", img, commonPoints, N, 100);

	return 0;
}



namespace Basics3D
{

	///////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////




    void LoadPoints(std::vector<C_Points2DParamSet> &out_PointsParamSet,
                    std::vector<T_Points2DSet<C_Point2D>> &out_PointsSet,
                    char *in_argv[],
                    const int in_N,
                    const double in_T,
										const int in_imHeight)
	{
		out_PointsParamSet.resize(in_N);
		out_PointsSet.resize(in_N - 1);

		for( int i = 0; i < in_N; i++)
		{
			out_PointsParamSet[i].Load(in_argv[3 + in_N + i], in_imHeight);
		}
		
		for( int i = 0; i < in_N - 1; i++)
		{
			out_PointsSet[i].InitGroupSize(2);
			out_PointsSet[i].Load(in_argv[3 + 2 * in_N + i], in_T, in_imHeight);
		}
		
		std::cout << "Points set is loaded - ";
		for( int i = 0; i < in_N - 1; i++)
		{
			std::cout << "PointsSet" << i << ": " << out_PointsSet[i].Size() << 
				" pairs of points." << std::endl << "                       ";
		}
		for( int i = 0; i < in_N; i++)
		{
			std::cout << "PointsParamSet" << i << ":  " << out_PointsParamSet[i].Size() << 
				" points" << std::endl << "                       ";
		}
		std::cout << std::endl << "========================================" << std::endl;
	}

	///////////////////////////////////////////////////////////////////////////////////////
    void MarkCorrespondences(const char *in_name,
                             const int j, // addition to name
                             const T_ImagesSet<AImageMem8> &in_img,
                             const std::vector<T_Points2DSet<C_Point2D>> &in_PointsSet,
                             const int i) // number of points to mark
                                          // -1 means all points
	{
	  try
		{
			C_Combination Comb(in_img);
			Comb.Make();
			Comb.MarkCorrespondences(in_PointsSet, i);

			std::ostringstream os1;
			os1 << in_name << j << ".pgm";
			std::string filename = os1.str();

			Comb.Save(filename.c_str());

			std::cout << "Combination with correspondences is made" << std::endl
								<< "========================================" << std::endl;
		}
		catch(const char* report)
		{
			std::cout << report;
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////

    void MarkCorrespondences(const char *in_name,
                             const T_ImagesSet<AImageMem8> &in_img,
                             const T_Points2DSet<C_Point2D> &in_PointsSet,
                             const int N,
                             const int i) // number of points to mark
                                          // -1 means all points
	{
	  try
		{
			C_Combination Comb(in_img);
			Comb.Make();
			Comb.MarkCorrespondences(in_PointsSet, N, i);

			std::ostringstream os1;
			os1 << in_name << ".pgm";
			std::string filename = os1.str();

			Comb.Save(filename.c_str());

			std::cout << "Combination with correspondences is made" << std::endl
								<< "========================================" << std::endl;
		}
		catch (const char* report)
		{
			std::cout << report;
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////
    void MakeDifferenceImage(const T_ImagesSet<AImageMem8> &in_img,
                             const std::vector<matrix_3_3> &in_Matrix,
                             const int i, // function works with images i and i + 1
                             const int j, // number of matrix and addition to name
                             const int k) // addition to name
	{
	  try
		{
			// Transform
			C_Transform Transform2(in_img[i + 1]);
			AImageMem8* transf2 = Transform2.Apply(in_Matrix[j]);
			
			// Difference
			C_AbsDifference Difference2(in_img[i], transf2);
			Difference2.Apply();

			std::ostringstream os1;
			os1 << "AbsDiff_" << i << "_" << i + 1 << "_" << j << "_" << k << ".pgm";
			std::string filename = os1.str();
			
			Difference2.Save(filename.c_str());
		}
		catch (const char* report)
		{
			std::cout << report;
		}
	}

  ///////////////////////////////////////////////////////////////////////////////////////
    void PrintAndSave(const char *in_name,
                      const std::vector<matrix_3_3> &in_Matrix,
                      const int i,
                      const int j)
	{
		matrix_3_3 forPrint;
    T_MatrixOperations::Scale(forPrint, in_Matrix[j]);
    T_MatrixOperations::Print(forPrint);

	  std::ostringstream os1;
		os1 << in_name << i << "_" << j << ".txt";
		std::string filename = os1.str();
								
		T_MatrixOperations::Save(forPrint, filename.c_str());
	}

	///////////////////////////////////////////////////////////////////////////////////////
    void DrawEpipolarLines(const T_ImagesSet<AImageMem8> &in_img,
                           const matrix_3_3 &in_Matrix,
                           const std::vector<T_Points2DSet<C_Point2D>> &in_PointsSet,
                           const int i,
                           const int n)
	{
		C_DrawEpipolarLines Lines(in_img[i], in_img[i + 1]);
		Lines.Apply(in_Matrix, in_PointsSet[i], n);
		Lines.Save("EpLines", i);
	}

	///////////////////////////////////////////////////////////////////////////////////////
    void FindCommonFeaturePoints(T_Points2DSet<C_Point2D> &out_commonPoints,
                                 const std::vector<T_Points2DSet<C_Point2D>> &in_Inliers,
                                 const int N)
	{
		std::cout << std::endl;
	  for (int i = 0; i < N - 1; i++)
		{
			std::cout << in_Inliers[i].Size() << "  ";
		}
		std::cout << std::endl;

		out_commonPoints = in_Inliers[0];

		for (int i = 0; i < N - 2; i++)
		{
			T_Points2DSet<C_Point2D>::Combine
				                  (out_commonPoints, out_commonPoints, in_Inliers[i + 1], i + 2);
		}

		std::cout << std::endl << "Common points " << out_commonPoints.Size() << std::endl;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//_____________________________________________________________________________

					//std::cout << std::endl << "----------------------------------------"
					//	<< std::endl << "DFP MINIMIZATION" << std::endl;

					//const int M = 9; // number of elements to minimaze
					//Vtl::NumAlg::Lin::ublas::vector<double> p(M);

					//for (int k = 0; k < 3; k++)
					//{
					//	for (int l = 0; l < 3; l++)
					//	{
					//		p[k * 3 + l] = Matrix[j](k, l);
					//	}
					//}

					//const double GTOL = 0.01;
					//int iter;
					//double fret;

					//std::cout << std::endl;
					//std::cout << "4   " << Inliers[i].Size() << "  ";
					//std::cout << Inliers[i].GroupSize() << "  ";
					//std::cout << M;

					//T_MistakeFunc<C_HomographyMatrix> Mist(Inliers[i], M);

					//_dfpmin<double>(&p, M, GTOL, &iter, &fret, &Mist);

					//Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> V;
					//std::cout << "\n";
		   //  
					//for (int k = 0; k < 3; k++)
					//{
					//	for (int l = 0; l < 3; l++)
					//	{
					//		std::cout << p[k * 3 + l] << "  ";
					//		V(k, l) = p[k * 3 + l];
					//	}
					//}
					//std::cout << "\n" << fret << "\n" << std::endl;

					//T_MatrixOperations::Print(V);
					//std::cout << std::endl << std::endl;

					//T_MatrixOperations::Scale(V, V);
					//T_MatrixOperations::Print(V);
					//std::cout << std::endl << std::endl;
		
		///////////////////////////////////////////////////////////////////////////////////////
/*		// Count inliers
		int numbInl = 0;
		
		double er = 0.0005;
	  
		double dist = 0;
		double distInl = 0;
		
		int size = PointsSet.Size() / 2;

		for (int i=0; i < size; i++)
		{
			C_Matrix Point(3, 1);
			Point = PointsSet[2 * i].Matrix();
			Point = M.Mult(Point);
			Point.SetElem(0, 0, Point.GetElem(0, 0) / Point.GetElem(2, 0));
			Point.SetElem(1, 0, Point.GetElem(1, 0) / Point.GetElem(2, 0));
			Point.SetElem(2, 0, 1);
								
			double d = 
				abs(Point.GetElem(0, 0) * PointsSet[2 * i + 1].X() + 
						Point.GetElem(1, 0) * PointsSet[2 * i + 1].Y() +
						Point.GetElem(2, 0)) /
				sqrt(Point.GetElem(0, 0) * Point.GetElem(0, 0) + 
						 Point.GetElem(1, 0) * Point.GetElem(1, 0));
			
			dist += d;

			//std::cout << "er = " << er << "   tres = " <<  tres << "\n";

			if (d < er)
			{
				numbInl++;
				distInl += d;
			}
			//std::cout << PointsSet[2 * i].X() << "  " << PointsSet[2 * i].Y() << " | " 
			//			  << PointsSet[2 * i + 1].X() << "  " << PointsSet[2 * i + 1].Y() << " | " 
			//					<< Point.GetElem(0,0) << "  " << Point.GetElem(1,0) << " | " << d << "\n";
		}

			dist /= size;
			distInl /= numbInl;
			std::cout << "\n_____________________________________________________\n"
										 << "DIST = " << dist << "   DIST INL = " << distInl << "\n"
										 << "Inl = " << numbInl << "\n";
		
			M.Print(); std::cout << "\n\n";*/
				
