
// std includes
#include <string>
#include <sstream>
#include <memory> 
#include <list>
#include <iostream>
#include <fstream>
#include <iomanip>

#include <time.h>
#include <assert.h>

#define _USE_MATH_DEFINES
#include <cmath>

// library includes
#include "yfilesys.h"
#include "aimage.h"
#include "color_spaces.h"
#include "convolve.h"
#include "imalg_tools.h"
#include "imalg_resampling.h"
#include "img_registration.h"

#include "plf_functors.h"
#include "plf_kernels_1d.h"

#include "tests.h"
#include "consolap.h"

//#include "Hanina/img_utils.h"
//#include "Hanina/ss_ridge_det.h"

// own includes
#include "Semeikina/SaveImageArray.h"
#include "Semeikina/DrawEllipse.h"
#include "Semeikina/Blobs&Ridges/DrawBlob.h"
#include "Semeikina/Blobs&Ridges/MaskBlobRidge.h"


int getTestFlag();
void setTestFlag(int new_flag);

static int my_test_flag_global_field=0;
int getTestFlag()
{
	return 0;//my_test_flag_global_field;
}
void setTestFlag(int new_flag)
{
	my_test_flag_global_field=new_flag;
}

using namespace Vtl::NumAlg::Lin;

namespace Vtl
{
namespace Plf
{
namespace Fx
{
		template<typename T>
		class EigenVal
		{
			T mult;
		public:
			//typedef boost::numeric::ublas::c_matrix<T,2,2> RotateMatrix;
			EigenVal(T sigma) 
				: mult(sigma*sigma/16.0f) //because result is proportional to second derivative i.e. d^2/dx^2
				// constant is arbitrary taken
			{
			}
			void operator()(T v11, T v12, T v22, T& type, T& lambda_1, T& lambda_2, T& ev1_x, T& ev1_y) const 
      {
				const int max_excentric=8;
				T a=(v11-v22)/2;
				T b=(v11+v22)/2;
				T D = sqrt(sqr(a) + sqr(v12));
				if(b >= 0)
				{
					lambda_1= (b+D);
					lambda_2= (b-D);
				}
				else
				{
					lambda_1= (b-D);
					lambda_2= (b+D);
				}
				//T ratio=0;
				ev1_x = v12;
				ev1_y = lambda_1-v11;
				lambda_1 *= mult;
				lambda_2 *= mult;
//				direct_cos=0;
			//	if( ( lambda_1*lambda_2 < 0 ) 
			//	 && ( 2*max_excentric*std::abs(lambda_2) > std::abs(lambda_1)  )
			//	  )  // 
			//	{
			//		type=0;
			//	}
			//	else if( std::abs(lambda_1) <=0.001f )  // -
			//	{
			//		type=0;
			//	}
			//	else
			//	{
			//		//ratio=lambda_2/lambda_1;

			//		////if(ratio<0)
			//		////	std::cout << "RAtio is negative! " << lambda_2 << " / " << lambda_1 <<std::endl;
			//		//T d=sqrt(sqr(ev1_x)+sqr(ev1_y));
			//		//ev1_x /=d; 
			//		//ev1_y /=d; 
			//		if(std::abs(lambda_1) <= max_excentric*std::abs(lambda_2))
			//			type=255;//blob
			//		else 
			//			type=127;//ridge
			//	}
			}
		};
}//namespace Fx
namespace F33
{
	template<typename T>
	class NormEuclidean
	{
		T thr2;
	public:
		NormEuclidean(T zero_threshold) : thr2(zero_threshold*zero_threshold)
		{
		}
		void operator()(T v1,T v2,T v3,T& o1,T& o2,T& o3)
		{
			T q=Vtl::sqr(v1)+Vtl::sqr(v2)+Vtl::sqr(v3);
			if(q > thr2)
			{
				q=sqrt(q);
				o1=v1/q;
				o2=v2/q;
				o3=v3/q;
			}
			else
			{
				o1=o2=o3=T(0);
			}
		}

	};
}//namespace F33
namespace F61
{
	template<typename T>
	struct ScalarProduct
	{
		T operator()(T a1,T a2,T a3,T b1,T b2,T b3)
		{
				return a1*b1+a2*b2+a3*b3;
		}

	};
}
}//namespace Plf
namespace ColorBlob
{
	template<typename T>
	class Derivative2
	{
		typedef Vtl::ImAlg::WndFun::Gauss<T,0, false> FF0; 
		typedef Vtl::Plf::Kernel::FKern<T,Vtl::Plf::Kernel::KT_Conv,true> K0;
		//1 
		typedef Vtl::ImAlg::WndFun::Gauss<T,1,false> FF1; 
		typedef Vtl::Plf::Kernel::FKern<T,Vtl::Plf::Kernel::KT_Conv,true> K1;
		//2 
		typedef Vtl::ImAlg::WndFun::Gauss<T,2,false> FF2; 
		typedef Vtl::Plf::Kernel::FKern<T,Vtl::Plf::Kernel::KT_Conv,true> K2;

		float sigma;
		FF0 f0;
		K0 ker0;
		Vtl::ImAlg::Conv::ConvInplace<T,K0> g0;
		//1 
		FF1 f1;
		K1 ker1;
		Vtl::ImAlg::Conv::ConvInplace<T,K1> g1;
		//2 
		FF2 f2;
		K2 ker2;
		Vtl::ImAlg::Conv::ConvInplace<T,K2> g2;

	public:
		Derivative2(float sigma_) 
			: sigma(sigma_) 
			,f0(sigma,4),ker0(f0),g0(ker0)
			,f1(sigma,4),ker1(f1),g1(ker1)	//1 
			,f2(sigma,4),ker2(f2),g2(ker2)	//2 
		{
		}
		template<typename Tin>
		void operator()(int w, int h,int nls , const Tin* img,T* L)
		{
			Vtl::Plf::F11::Convert<T> c_to_float;
			Vtl::ImAlg::Tools::copy_1_1(img,w,h,nls,L,w,c_to_float);	
		}
		void operator()(int w, int h,T* Lxx,T* Lxy,T* Lyy)
		{
			std::copy(Lxx,Lxx+w*h,Lxy);
			std::copy(Lxx,Lxx+w*h,Lyy);
			// 2 
			// Gxx :
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lxx,w,h,w,g2,g0);
			// Gyy :
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lyy,w,h,w,g0,g2);
			// Gxy :
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lxy,w,h,w,g1,g1);
		}
		
	};
	template<typename T> 
	class MaxIn3x3x3
	{
			int nls;
			T thr;
		public:
			int Nls() const {return nls;}
			typedef const T* Planes_t;
			MaxIn3x3x3(int nls_set,T threshold=0.1) //   !!!!
				:nls(nls_set),thr(threshold)
			{
			}

			bool operator()(const Planes_t* v_image, size_t ofs_to_cur)
			{
				T center=v_image[1][ofs_to_cur];
//				T thresh=1.0e-10f;
//				center -= thresh;
				if(center<thr)
					return false;
				if(getTestFlag()) std::cout << v_image[0][ofs_to_cur-1-nls] << "    ";
				if( center < v_image[0][ofs_to_cur-1-nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[0][ofs_to_cur  -nls] << "    ";
				if( center < v_image[0][ofs_to_cur  -nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[0][ofs_to_cur+1-nls] << "    ";
				if(getTestFlag()) std::cout << std::endl;
				if( center < v_image[0][ofs_to_cur+1-nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[0][ofs_to_cur-1    ] << "    ";
				if( center < v_image[0][ofs_to_cur-1    ] )
					return false;
				if(getTestFlag()) std::cout << v_image[0][ofs_to_cur      ] << "    ";
				if( center < v_image[0][ofs_to_cur      ] )
					return false;
				if(getTestFlag()) std::cout << v_image[0][ofs_to_cur+1    ] << "    ";
				if(getTestFlag()) std::cout << std::endl;
				if( center < v_image[0][ofs_to_cur+1    ] )
					return false;
				if(getTestFlag()) std::cout << v_image[0][ofs_to_cur-1+nls] << "    ";
				if( center < v_image[0][ofs_to_cur-1+nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[0][ofs_to_cur  +nls] << "    ";
				if( center < v_image[0][ofs_to_cur  +nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[0][ofs_to_cur+1+nls] << "    ";
				if(getTestFlag()) std::cout << std::endl;
				if( center < v_image[0][ofs_to_cur+1+nls] )
					return false;
				//----------------------------------
				if(getTestFlag()) std::cout << "-------"<< std::endl;
				if(getTestFlag()) std::cout << v_image[1][ofs_to_cur-1-nls] << "    ";
				if( center < v_image[1][ofs_to_cur-1-nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[1][ofs_to_cur  -nls] << "    ";
				if( center < v_image[1][ofs_to_cur  -nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[1][ofs_to_cur+1-nls] << "    ";
				if(getTestFlag()) std::cout << std::endl;
				if( center < v_image[1][ofs_to_cur+1-nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[1][ofs_to_cur-1    ] << "    ";
				if( center < v_image[1][ofs_to_cur-1    ] )
					return false;
				if(getTestFlag()) std::cout << v_image[1][ofs_to_cur      ] << "    ";
				//if( center < v_image[1][ofs_to_cur      ] )
				//	return false;
				if(getTestFlag()) std::cout << v_image[1][ofs_to_cur+1    ] << "    ";
				if(getTestFlag()) std::cout << std::endl;
				if( center < v_image[1][ofs_to_cur+1    ] )
					return false;
				if(getTestFlag()) std::cout << v_image[1][ofs_to_cur-1+nls] << "    ";
				if( center < v_image[1][ofs_to_cur-1+nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[1][ofs_to_cur  +nls] << "    ";
				if( center < v_image[1][ofs_to_cur  +nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[1][ofs_to_cur+1+nls] << "    ";
				if(getTestFlag()) std::cout << std::endl;
				if( center < v_image[1][ofs_to_cur+1+nls] )
					return false;
				//----------------------------------
				if(getTestFlag()) std::cout <<"------------"<< std::endl;
				if(getTestFlag()) std::cout << v_image[2][ofs_to_cur-1-nls] << "    ";
				if( center < v_image[2][ofs_to_cur-1-nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[2][ofs_to_cur  -nls] << "    ";
				if( center < v_image[2][ofs_to_cur  -nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[2][ofs_to_cur+1-nls] << "    ";
				if(getTestFlag()) std::cout << std::endl;
				if( center < v_image[2][ofs_to_cur+1-nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[2][ofs_to_cur-1    ] << "    ";
				if( center < v_image[2][ofs_to_cur-1    ] )
					return false;
				if(getTestFlag()) std::cout << v_image[2][ofs_to_cur      ] << "    ";
				if( center < v_image[2][ofs_to_cur      ] )
					return false;
				if(getTestFlag()) std::cout << v_image[2][ofs_to_cur+1    ] << "    ";
				if(getTestFlag()) std::cout << std::endl;
				if( center < v_image[2][ofs_to_cur+1    ] )
					return false;
				if(getTestFlag()) std::cout << v_image[2][ofs_to_cur-1+nls] << "    ";
				if( center < v_image[2][ofs_to_cur-1+nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[2][ofs_to_cur  +nls] << "    ";
				if( center < v_image[2][ofs_to_cur  +nls] )
					return false;
				if(getTestFlag()) std::cout << v_image[2][ofs_to_cur+1+nls] << "    ";
				if( center < v_image[2][ofs_to_cur+1+nls] )
					return false;
				if(getTestFlag()) std::cout << std::endl;
				if(getTestFlag()) std::cout << "==============="<<std::endl;

				//=========================================
				int count_of_eq=0;
				//=========================================
				if( center == v_image[0][ofs_to_cur-1-nls] ) ++count_of_eq;
				if( center == v_image[0][ofs_to_cur  -nls] ) ++count_of_eq;
				if( center == v_image[0][ofs_to_cur+1-nls] ) ++count_of_eq;
				if( center == v_image[0][ofs_to_cur-1    ] ) ++count_of_eq;
				if( center == v_image[0][ofs_to_cur      ] ) ++count_of_eq;
				if( center == v_image[0][ofs_to_cur+1    ] ) ++count_of_eq;
				if( center == v_image[0][ofs_to_cur-1+nls] ) ++count_of_eq;
				if( center == v_image[0][ofs_to_cur  +nls] ) ++count_of_eq;
				if( center == v_image[0][ofs_to_cur+1+nls] ) ++count_of_eq;
				//----------------------------------
				if( center == v_image[1][ofs_to_cur-1-nls] ) ++count_of_eq;
				if( center == v_image[1][ofs_to_cur  -nls] ) ++count_of_eq;
				if( center == v_image[1][ofs_to_cur+1-nls] ) ++count_of_eq;
				if( center == v_image[1][ofs_to_cur-1    ] ) ++count_of_eq;
				//if( center == v_image[1][ofs_to_cur      ] ) ++count_of_eq;
				if( center == v_image[1][ofs_to_cur+1    ] ) ++count_of_eq;
				if( center == v_image[1][ofs_to_cur-1+nls] ) ++count_of_eq;
				if( center == v_image[1][ofs_to_cur  +nls] ) ++count_of_eq;
				if( center == v_image[1][ofs_to_cur+1+nls] ) ++count_of_eq;
				//----------------------------------
				if( center == v_image[2][ofs_to_cur-1-nls] ) ++count_of_eq;
				if( center == v_image[2][ofs_to_cur  -nls] ) ++count_of_eq;
				if( center == v_image[2][ofs_to_cur+1-nls] ) ++count_of_eq;
				if( center == v_image[2][ofs_to_cur-1    ] ) ++count_of_eq;
				if( center == v_image[2][ofs_to_cur      ] ) ++count_of_eq;
				if( center == v_image[2][ofs_to_cur+1    ] ) ++count_of_eq;
				if( center == v_image[2][ofs_to_cur-1+nls] ) ++count_of_eq;
				if( center == v_image[2][ofs_to_cur  +nls] ) ++count_of_eq;
				if( center == v_image[2][ofs_to_cur+1+nls] ) ++count_of_eq;
				//=========================================
//				if(count_of_eq>0 && count_of_eq< 3)
//					std::cout << "in class MaxIn3x3x3: count_of_eq =" << count_of_eq << "   center = " << center << count_of_eq <<std::endl;
				if(count_of_eq > 1) 
					return false;
				//=========================================
				return true;
			}
	};
	template<typename T> 
	class FillFPointInfo
	{
		const int nls;
		T sigma;
		const T* Mask;
		const T* Ev1;
		const T* Ev2;
		const T* Ev_x;
		const T* Ev_y;
		T* strength;
		T* ellipticity;
		T* direction_cos;
		T* scale;

		public:
			int Nls() const {return nls;}
			FillFPointInfo(int nls_set,T sigma_
			  ,const T* Mask_,const T* Ev1_,const T* Ev2_,const T* Ev_x_,const T* Ev_y_,T* blob_strength
			  ,T* blob_ellipticity,T* blob_cos,T* blob_scale) 
				: nls(nls_set),sigma(sigma_)
				 ,Mask(Mask_),Ev1(Ev1_),Ev2(Ev2_),Ev_x(Ev_x_),Ev_y(Ev_y_)
				 ,strength(blob_strength),ellipticity(blob_ellipticity),direction_cos(blob_cos),scale(blob_scale)
			{			
			}
			void operator()(size_t ofs_to_cur,int x,int y)
			{
				if(Mask[ofs_to_cur]>254)
				{
					strength[ofs_to_cur]      = Ev2[ofs_to_cur];
					ellipticity[ofs_to_cur]   = Ev1[ofs_to_cur]/Ev2[ofs_to_cur];
					T d=sqrt(sqr(Ev_x[ofs_to_cur])+sqr(Ev_y[ofs_to_cur]));
					direction_cos[ofs_to_cur] = ((Ev_y[ofs_to_cur]<0) ? -Ev_x[ofs_to_cur] : Ev_x[ofs_to_cur]) /d;
					scale[ofs_to_cur] = sigma;
					std::cout << "Point!  (" <<x<<","<<y<<")   "<<  strength[ofs_to_cur] << "  " << ellipticity[ofs_to_cur] << "  " << direction_cos[ofs_to_cur] << "  " << sigma <<std::endl;
				}
			}
	};
};//namespace ColorBlob
namespace ImAlg
{
namespace Tools
{	
	template <typename PtrToArray, class F1, class F2> 
	void envTransf_26(PtrToArray planes, int width, int height, int nls,  F1 functor_cond,F2 functor_todo)
	{
		int offs = 0;
		for(int y=1; y<(height-1); y++)
		{
			offs = y*nls;
		    for(int x=1; x<(width-1); x++)
			{
				//if( (x==132)&& (y==47) )
				//{
				//	std::cout << x << " " << y << std::endl;
				//}
				int i = offs+x;

				if(x==100 && y==100)
					setTestFlag(1);

				if(functor_cond(planes, i) )
					functor_todo(i,x,y);

				if(x==100 && y==100)
					setTestFlag(0);
			}
		}
	}
}//namespace ImAlg

}//namespace Tools
}//namespace Vtl

namespace Tests
{
	int test_color_blob(int argc,char *argv[])
	{
		if(argc < 2) 
      {
				std::cout << "Command string:\n test.exe input_img output_img" <<std::endl;
            return 1;
      }
		
    // ,  auto_ptr        {}
		std::auto_ptr<AImage> _img(::AImageLoad(0,true,argv[1],indicator));

		//     ,    8 ,   
		//     -         1.
		AImageMemRgb8* img=dynamic_cast<AImageMemRgb8*>(_img.get());
		if(!img)
			return 1;
		AImageMem8* r_=img->GetPlane8(2);
		AImageMem8* g_=img->GetPlane8(1);
		AImageMem8* b_=img->GetPlane8(0);

		//   
		const int w=img->Width();
		const int h=img->Height();
		const int sz=w*h;
		const int nls=r_->NextLineShift();

    typedef float Real_t;
		using namespace Vtl::ColorBlob;

		std::vector<Real_t> data(sz*(3*3+3+3+10+2+4+4+1));
		Real_t * Rxx = &(data[0]);
		Real_t * Ryy = Rxx+sz* 1;
		Real_t * Rxy = Rxx+sz* 2;
		Real_t * Gxx = Rxx+sz* 3;
		Real_t * Gyy = Rxx+sz* 4;
		Real_t * Gxy = Rxx+sz* 5;
		Real_t * Bxx = Rxx+sz* 6;
		Real_t * Byy = Rxx+sz* 7;
		Real_t * Bxy = Rxx+sz* 8;

		Real_t * Cr  = Rxx+sz* 9;
		Real_t * Cg  = Rxx+sz*10;
		Real_t * Cb  = Rxx+sz*11;

		Real_t * Lxx = Rxx+sz*12;
		Real_t * Lyy = Rxx+sz*13;
		Real_t * Lxy = Rxx+sz*14;

		Real_t * Ev1_0  = Rxx+sz*15;
		Real_t * Ev1_1  = Rxx+sz*16;
		Real_t * Ev1_2  = Rxx+sz*17;
		Real_t * Ev2_0  = Rxx+sz*18;
		Real_t * Ev2_1  = Rxx+sz*19;
		Real_t * Ev2_2  = Rxx+sz*20;
		Real_t * Ev_x_0  = Rxx+sz*21;
		Real_t * Ev_x_1  = Rxx+sz*22;
		Real_t * Ev_y_0  = Rxx+sz*23;
		Real_t * Ev_y_1  = Rxx+sz*24;

		Real_t * Mask_last  = Rxx+sz*25;
		Real_t * Mask_prev  = Rxx+sz*26;

		Real_t * blob_strength     = Rxx+sz*27;
		Real_t * blob_ellipticity  = Rxx+sz*28;
		Real_t * blob_cos          = Rxx+sz*29;
		Real_t * blob_scale        = Rxx+sz*30;

		//to generate pseudo gray image, no need for calculations, only demo!
		Real_t * R  = Rxx+sz*31;
		Real_t * G  = Rxx+sz*32;
		Real_t * B  = Rxx+sz*33;
		Real_t * L  = Rxx+sz*34;

		Real_t * Ridge  = Rxx+sz*35;

		Derivative2<Real_t> D2(1.0f);
		D2(w,h,nls,(*r_)[0],R); //  
		D2(w,h,nls,(*g_)[0],G); //  
		D2(w,h,nls,(*b_)[0],B); //  


		Real_t * Ev1    = Ev1_0;  
		Real_t * Ev1_p1 = Ev1_1;  
		Real_t * Ev1_p2 = Ev1_2;
		Real_t * Ev2    = Ev2_0;  
		Real_t * Ev2_p1 = Ev2_1;  
		Real_t * Ev2_p2 = Ev2_2;
		Real_t * Mask  = Mask_last;
		Real_t * Mask_p1  = Mask_prev;
		Real_t * Ev_x  = Ev_x_0;
		Real_t * Ev_x_p1  = Ev_x_1;
		Real_t * Ev_y  = Ev_y_0;
		Real_t * Ev_y_p1  = Ev_y_1;


		std::fill( blob_strength, blob_strength+sz,0.0f);
		std::fill( blob_ellipticity, blob_ellipticity+sz,0.0f);
		std::fill( blob_cos, blob_cos+sz,0.0f);
		std::fill( blob_scale, blob_scale+sz,0.0f);


		float sigma_prev =1;
		for( float sigma=2.0f ; sigma < 30.0f ; sigma*=sqrt(sqrt(2.0f)) )
		{
			std::cout << " ==== sigma= " << sigma << std::endl;
			Derivative2<Real_t> D2(sigma);
			D2(w,h,nls,(*r_)[0],Rxx);
			D2(w,h,Rxx,Rxy,Ryy);
			D2(w,h,nls,(*g_)[0],Gxx);
			D2(w,h,Gxx,Gxy,Gyy);
			D2(w,h,nls,(*b_)[0],Bxx);
			D2(w,h,Bxx,Bxy,Byy);

			//build Laplasians first iteration (round laplasian shape)
			Vtl::Plf::F21::Sum<Real_t> sum;
			Vtl::ImAlg::Tools::copy_2_1(Rxx,w,h,w,Ryy,w,Cr,w,sum);
			Vtl::ImAlg::Tools::copy_2_1(Gxx,w,h,w,Gyy,w,Cg,w,sum);
			Vtl::ImAlg::Tools::copy_2_1(Bxx,w,h,w,Byy,w,Cb,w,sum);


	//		for(int iter=0 ; iter<1 ; iter++)
			{
				//norming vector C 
				Vtl::Plf::F33::NormEuclidean<Real_t> ne(0.001f);
				Vtl::ImAlg::Tools::copy_3_3(Cr,w,h,w,Cg,w,Cb,w,Cr,w,Cg,w,Cb,w,ne);
				//Calc derivatives of aqdaptive grayscale image 
				Vtl::Plf::F61::ScalarProduct<Real_t> sprod;
				Vtl::ImAlg::Tools::copy_6_1(Rxx,w,h,w,Gxx,w,Bxx,w,Cr,w,Cg,w,Cb,w,Lxx,w,sprod);
				Vtl::ImAlg::Tools::copy_6_1(Rxy,w,h,w,Gxy,w,Bxy,w,Cr,w,Cg,w,Cb,w,Lxy,w,sprod);
				Vtl::ImAlg::Tools::copy_6_1(Ryy,w,h,w,Gyy,w,Byy,w,Cr,w,Cg,w,Cb,w,Lyy,w,sprod);
				
				Vtl::ImAlg::Tools::copy_6_1(R,w,h,w,G,w,B,w,Cr,w,Cg,w,Cb,w,L,w,sprod);

				Vtl::ImAlg::Tools::xchg(Ev1_p2,Ev1_p1);// Ev_p1 -> Ev_p2
				Vtl::ImAlg::Tools::xchg(Ev1_p1,Ev1);//Ev->Ev_p1
				Vtl::ImAlg::Tools::xchg(Ev2_p2,Ev2_p1);// Ev_p1 -> Ev_p2
				Vtl::ImAlg::Tools::xchg(Ev2_p1,Ev2);//Ev->Ev_p1
				Vtl::ImAlg::Tools::xchg(Mask,Mask_p1);
				Vtl::ImAlg::Tools::xchg(Ev_x,Ev_x_p1);
				Vtl::ImAlg::Tools::xchg(Ev_y,Ev_y_p1);
				// now build new Ev
				Vtl::Plf::Fx::EigenVal<Real_t> ev(sigma);
				Vtl::ImAlg::Tools::copy_3_5(Lxx,w,h,w,Lxy,w,Lyy,w,Mask,w,Ev1,w,Ev2,w,Ev_x,w,Ev_y,w,ev);

        MaskBlobRidge(Mask, Ev1, Ev2, w, h, w);
			}

			if(sigma> 3.0f )
			{
				MaxIn3x3x3<Real_t> func(w);
				Real_t* planes[]={Ev2_p2,Ev2_p1,Ev2}; 
				FillFPointInfo<Real_t> todo(w,sigma_prev,Mask_p1,Ev1_p1,Ev2_p1,Ev_x_p1,Ev_y_p1
					,blob_strength,blob_ellipticity,blob_cos,blob_scale);

		//					SaveImage(img,planes[1],w,h,w,sigma_prev,"aab.tif","eig3Dss");
				Vtl::ImAlg::Tools::envTransf_26(planes,w,h,w,func,todo);

				//Ridge
				Vtl::Plf::Fpie::CannyNMS<float> nms(0);
				Vtl::Plf::Extent ext=nms.ext();
				Vtl::ImAlg::Tools::envTransf_3e1_1(ext,Ev1_p1,w,h,w,Ev_x_p1,w,Ev_y_p1,w,Ridge,w,nms);
			}
			if(std::abs(sigma-10.0f)< 1.0f ) //    : sigma =10.0f
			{
				Vtl::Plf::F11::Lin<Real_t> lin(16.0f,128.0f);
				//     float    byte. ,   float  
				//     0-255 -       
				// Hess_Eig_diff  HessianDetSqrt  -         .
				//   Hess_Eig_diff
				Vtl::Plf::F11::Convert<uint8_t> c_to_Byte;
				//  ,      
				Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::Lin<Real_t>,Vtl::Plf::F11::Convert<uint8_t> > adj(lin,c_to_Byte);
					//          ()
				Vtl::ImAlg::Tools::copy_1_1(Ev2_p1,w,h,w,(*r_)[0],r_->NextLineShift(),adj);
				::AImageSave(r_,"col_detect_lambda_2.tif",indicator);
				Vtl::ImAlg::Tools::copy_1_1(Mask_p1,w,h,w,(*r_)[0],r_->NextLineShift(),c_to_Byte);
				::AImageSave(r_,"col_detect_mask.tif",indicator);
				Vtl::Plf::F11::Lin<Real_t> lin2(0.40f,128.0f);
				Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::Lin<Real_t>,Vtl::Plf::F11::Convert<uint8_t> > adj2(lin2,c_to_Byte);
				Vtl::ImAlg::Tools::copy_1_1(L,w,h,w,(*r_)[0],r_->NextLineShift(),adj2);
				::AImageSave(r_,"col_detect_pseudogray.tif",indicator);
				Vtl::ImAlg::Tools::copy_1_1(Ridge,w,h,w,(*r_)[0],r_->NextLineShift(),adj);
				::AImageSave(r_,"col_detect_ridge.tif",indicator);
			}
			sigma_prev=sigma;
		}

		Vtl::Plf::F11::Lin<Real_t> lin(32.0f,0.0f);
		Vtl::Plf::F11::Convert<uint8_t> c_to_Byte;
		//  ,      
		Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::Lin<Real_t>,Vtl::Plf::F11::Convert<uint8_t> > adj(lin,c_to_Byte);
			//          ()
		Vtl::ImAlg::Tools::copy_1_1(blob_strength,w,h,w,(*r_)[0],r_->NextLineShift(),adj);
		::AImageSave(r_,"col_detect_blob_strength.tif",indicator);
			
//		,blob_strength,blob_ellipticity,blob_cos,blob_scale);

		return 0;
	}
}

	
