// system includings
//#include <memory> 
#include <vector>
#include <iostream>

// library includings
#include "aimage.h"
#include "consolap.h"
#include "imalg_tools.h"
#include "plf_functors.h"
#include "plf_kernels_1d.h"

// own includings
#include "SaveImageArray.h"


namespace Tests {
int TestDerivatives(int argc, char *argv[])
{
	// Load input image
  if(argc < 2) 
  {
	  std::cout << "Argument - input image" << std::endl;
    return 0;
  }
  
  const char *fname = argv[1];

  std::auto_ptr<AImage> _img(::AImageLoad(0, true, fname, indicator));
	AImageMem8* img = dynamic_cast<AImageMem8*>(_img.get());
	if(!img)
	{
		return 0;
	}

	int w = img->Width();
	int h = img->Height();
  int nls = img->NextLineShift();
	int size = w * h;

  std::vector<float> _data(size);
	float *data  = &(_data[0]);

	Vtl::Plf::F11::Convert<float> c_to_float;
	Vtl::ImAlg::Tools::copy_1_1((*img)[0], w, h, nls, data, w, c_to_float);

	const float sigma = 7.0f;
	const float nSigmasEnv = 3.0;

  std::vector<float> Lx(size);
	std::vector<float> Ly(size);
  std::vector<float> Lxx(size);
  std::vector<float> Lxy(size);
  std::vector<float> Lyy(size);

  memcpy(&(Lx[0]), data, size * sizeof(float));
	memcpy(&(Ly[0]), data, size * sizeof(float));
  memcpy(&(Lxx[0]), data, size * sizeof(float));
  memcpy(&(Lxy[0]), data, size * sizeof(float));
  memcpy(&(Lyy[0]), data, size * sizeof(float));

  //  ,    
  //     1,3,4 
  //  
  typedef Vtl::ImAlg::WndFun::Gauss<float, 0, false> FF0; 
  FF0 f0(sigma, nSigmasEnv);
  typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K0;
  K0 ker0(f0);
  Vtl::ImAlg::Conv::ConvInplace<float, K0> g0(ker0);

  // 1 
  typedef Vtl::ImAlg::WndFun::Gauss<float, 1, false> FF1; 
  FF1 f1(sigma, nSigmasEnv);
  typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K1;
  K1 ker1(f1);
  Vtl::ImAlg::Conv::ConvInplace<float, K1> g1(ker1);

  //2 
	typedef Vtl::ImAlg::WndFun::Gauss<float, 2, false> FF2; 
	FF2 f2(sigma, nSigmasEnv);
	typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K2;
	K2 ker2(f2);
	Vtl::ImAlg::Conv::ConvInplace<float,K2> g2(ker2);

  //  
  // 1 
  //  x:
  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lx[0]), w, h, w, g1, g0);
  //  y:
  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Ly[0]), w, h, w, g0, g1);

  // 2 
  //  xx:
  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxx[0]), w, h, w, g2, g0);
  //  xy:
  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxy[0]), w, h, w, g1, g1);
  //  yy:
  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lyy[0]), w, h, w, g0, g2);

  SaveNormImageArray(&(Lx[0]), w, h, fname, "_Lx");
  SaveNormImageArray(&(Ly[0]), w, h, fname, "_Ly");
  SaveNormImageArray(&(Lxy[0]), w, h, fname, "_Lxy");
  SaveNormImageArray(&(Lxx[0]), w, h, fname, "_Lxx");
  SaveNormImageArray(&(Lyy[0]), w, h, fname, "_Lyy");
}
}