#include <iostream>
#include <boost/numeric/ublas/matrix.hpp>
#include "loki_NullType.h" 
#include "loki_typeManip.h"

#include "vtl_meta_base.h"
#include "tests.h"


namespace Tests
{
int test_meta_determinant(int argc,char* argv[])
{
	using namespace boost::numeric::ublas;
	using namespace Vtl::Meta;
		{
			const int N=1;
			typedef c_matrix<float,N,N> Matrix;
			Matrix a;
			a(0,0)=33.0f;  
			typedef det<Matrix> Functor;
			Functor d(a);
			ForEachPermutation<N,Functor>(d);
			std::cout <<" det(perm)="<< d.result << ((d.result==33)? "   result Ok":"   result BAD") <<std::endl;
		}
		{
			typedef IL::MakeInRange<0,7>::Result MyLst;
			 enum{ x = IL::Elem<5, MyLst >::value };
			//next line generates compile time error intentionally
			// enum{ y = IL::Elem<10, MyLst >::value };
		}
		{
			const int N=1;
			typedef c_matrix<float,N,N> Matrix;
			Matrix a;
			a(0,0)=33.0f;  
			typedef Vtl::Meta::det<Matrix> Functor;
			Functor d(a);
			ForEachPermutation<N,Functor>(d);
			std::cout <<" det(perm)="<< d.result << ((d.result==33)? "   result Ok":"   result BAD") <<std::endl;
		}
		{
			enum { N=2 };
			typedef c_matrix<float,N,N> Matrix;
			Matrix a;
			a(0,0)=5.0f;  a(0,1)=13.0f;
			a(1,0)=11.0f; a(1,1)=7.0f;
			typedef det<Matrix> Functor;
			Functor d(a);
			ForEachPermutation<N,Functor>(d);
			std::cout <<" det(perm)="<< d.result << ((d.result==-108)? "   result Ok":"   result BAD") <<std::endl;
		}
		{
			const int N=3;
			typedef c_matrix<float,N,N> Matrix;
			Matrix a;
			a(0,0)=1.0f;  a(0,1)=2.0f; a(0,2)=3.0f;
			a(1,0)=4.0f;  a(1,1)=5.0f; a(1,2)=6.0f;
			a(2,0)=7.0f;  a(2,1)=8.0f; a(2,2)=0.0f;
			typedef det<Matrix> Functor;
			Functor d(a);
			ForEachPermutation<N,Functor>(d);
			std::cout <<" det(perm)="<< d.result << ((d.result==27)? "   result Ok":"   result BAD") <<std::endl;
		}
		{
			typedef Loki::NullType il1;
			typedef IL::MakeInRange<1,4>::Result il2;
			typedef IL::MakeInRange<1,5>::Result il4;
			{
				typedef IL::PushBack<33,il1>::Result tmp;
				std::cout << "PushBack :" << IL::type2ostream<tmp>() <<std::endl;
			}
			printPermutations dsj;

			ForEachPermutation<4,printPermutations>(dsj);
		}
		{
			const int N=0xFF00F001;
			std::cout <<" Meta::bitsIn<" << N << ">()="<< BitsIn<N>::value <<  std::endl;
		}
		{
			const int N=12;
			std::cout <<" Meta::bitsIn<" << N << ">()="<< BitsIn<N>::value <<  std::endl;
		}
		{
			const int N=0x1201;
			std::cout <<" Meta::bitsIn<" << N << ">()="<< BitsIn<N>::value <<  std::endl;
		}
		{
			const int N=2;
			typedef c_matrix<float,N,N> Matrix;
			Matrix a;
			a(0,0)=1.0f; a(0,1)=2.0f;
			a(1,0)=3.0f; a(1,1)=4.0f;
			Matrix b;
			b(0,0)=6.0f; b(0,1)=5.0f;
			b(1,0)=7.0f; b(1,1)=8.0f;
			typedef det_poly_coef<Matrix,N> Functor;
			Functor d(a,b);
			ForEachPermutation<N,Functor>(d);
			print_array<N+1>(d.coef,"Poly",false);
			std::cout <<" Desired:  {  -2, 3, 13 }             i.e. f := -2+3*x+13*x^2 " <<  std::endl;
		}
		{
			const int N=3;
			typedef c_matrix<float,N,N> Matrix;
			Matrix a;
			a(0,0)=9.0f; a(0,1)=8.0f; a(0,2)=4.0f;
			a(1,0)=7.0f; a(1,1)=1.0f; a(1,2)=3.0f;
			a(2,0)=5.0f; a(2,1)=3.0f; a(2,2)=6.0f;
			Matrix b;
			b(0,0)=1.0f; b(0,1)=2.0f; b(0,2)=3.0f;
			b(1,0)=4.0f; b(1,1)=5.0f; b(1,2)=6.0f;
			b(2,0)=7.0f; b(2,1)=8.0f; b(2,2)=0.0f;
			typedef det_poly_coef<Matrix,N> Functor;
			Functor d(a,b);
			ForEachPermutation<N,Functor>(d);
			print_array<N+1>(d.coef,"Poly",false);
			std::cout <<" Desired:  {  -179, 243, 42, 27 }     i.e. f := -179+243*x+42*x^2+27*x^3" <<  std::endl;
		}
		std::cout << std::endl;
		std::cout << "-------------" << std::endl;
		std::cout << "Another way to calculate determinant - via formula with row expansion:" << std::endl;
		std::cout << std::endl;
		{
			const int N=1;
			typedef c_matrix<float,N,N> Matrix;
			Matrix a;
			a(0,0)=33.0f;
			float res=determinant<N>(a);
			std::cout <<" det=" << res << ((res==33)? "   result Ok":"   result BAD") <<std::endl;
		}
		{
			const int N=2;
			typedef c_matrix<float,N,N> Matrix;
			Matrix a;
			a(0,0)=5.0f;  a(0,1)=13.0f;
			a(1,0)=11.0f; a(1,1)=7.0f;
			float res=determinant<N>(a);
			std::cout <<" det="<< res << ((res==-108)? "   result Ok":"   result BAD") <<std::endl;
		}
		{
			const int N=3;
			typedef c_matrix<float,N,N> Matrix;
			Matrix a;
			a(0,0)=1.0f;  a(0,1)=2.0f; a(0,2)=3.0f;
			a(1,0)=4.0f;  a(1,1)=5.0f; a(1,2)=6.0f;
			a(2,0)=7.0f;  a(2,1)=8.0f; a(2,2)=9.0f;
			float res=determinant<N>(a);
			std::cout <<" det="<< res << ((res==0)? "   result Ok":"   result BAD") <<std::endl;
			a(2,2)=0.0f;
			res=determinant<N>(a);
			std::cout <<" det="<< res << ((res==27)? "   result Ok":"   result BAD") <<std::endl;
		}

	return 0;
}
}//namespace Tests

