#include <string>
#include <sstream>
#include <memory> //auto_ptr()
#include <assert.h>

#include "aimage.h"
#include "yfilesys.h"
#include "consolap.h"

#include "plf_filters.h"
//#include "plf_filters_complex.h"
#include "plf_sys.h"
//#include <time.h>
//#include "Sidorkina/plt_filter_image.h"
#include "color_spaces.h" 

#include "tests.h"
#include <fstream>
#include <cstdlib>
#include "math.h"

namespace Students
{
namespace Rudyak
{
	inline double RatioLog(double chrom,double green)
	{
		return log(chrom/green);
	}
	// ,  value   -1..+1
	inline double Scale(double value,double scale)
	{
//		return scale* ( (value>0) ? log(1+1000.*value): -log(1-1000.*value) )/log(1000.);
		return scale*2/M_PI*atan(20*value);
	}
}//namespace Rudyak
}//namespace Students
namespace Tests
{
int test_shadow_detection_invariant(int argc,char *argv[])
{		
		if(argc<2)
		{
			std::cout << "Command string: this.exe input_rgb_file " <<std::endl;
			return 1;
		}

		double lnx,lny;	
		double fi=-0.4999; //    
		double cosfi=cos(fi);
		double sinfi=sin(fi);
		double t=2.0002;//  
		double cost=cos(t);
		double sint=sin(t);

			std::string fname_in=argv[1];		
			std::string fname_out=YFileSys::getNameWithoutExt(fname_in);

			std::auto_ptr<AImage> _in(::AImageLoad(0,true,fname_in.c_str(),indicator));
			AImageMemRgb8 *in=dynamic_cast<AImageMemRgb8 *>(_in.get());
			if(in==0)
			{
				std::ostringstream os;
				os << "Can't load RGB 8 bit input image "<< argv[1] ;
				throw os.str();
			}
			AImageMem8 * r=in->GetPlane8(0);
			AImageMem8 * g=in->GetPlane8(1);
			AImageMem8 * b=in->GetPlane8(2);

			int nls=r->NextLineShift();
			int w=r->Width();
			int h=r->Height();

			uint8_t *pr=(*r)[0];
			uint8_t *pg=(*g)[0];
			uint8_t *pb=(*b)[0];
		
			const size_t bits=r->Bits();
			const uint8_t Max_color=uint8_t((1<<bits)-1);
			const uint8_t bias=uint8_t(1<<(bits-1));
			const uint8_t scale=Max_color-bias;
			//      ( ) 
			const double max_out=Students::Rudyak::RatioLog(Max_color,1.0);
			const double pre_max_out=Students::Rudyak::RatioLog(Max_color-1,1.0);
			const double min_out=Students::Rudyak::RatioLog(1.0,Max_color);
			const double pre_min_out=Students::Rudyak::RatioLog(2.0,Max_color);;
			//    (  )   
			// : (    
			//           ,   ,     ,   
			//              
			//          log(r/g)  -  log((r+1)/(g+1)),    
			//           -  ,   )
			//           ,
			//    ,        -   ,    
			//  ,   -   ,    .
			const double ext_max_out=max_out+(max_out-pre_max_out);
			const double ext_min_out=min_out+(min_out-pre_min_out);
			const double range=std::max(ext_max_out,abs(ext_min_out));
			const double coef=scale/range;
			for(int y=0; y<h; y++)
			{
				for(int x=0; x<w; x++)
				{
					if(pg[x]>0)
					{
						if(pr[x]>0) //  ""   ,       
							lnx=Students::Rudyak::RatioLog(pr[x],pg[x]);//   ,  
						else
							lnx=ext_min_out;
						if(pb[x]>0)					 //     r/g  b/g
							lny=Students::Rudyak::RatioLog(pb[x],pg[x]);
						else
							lny=ext_min_out;
					} 
					else
					{
						lnx=ext_max_out;
						lny=ext_max_out;
					}
					double inv=(sinfi*lnx-cosfi*lny);
					double non_inv=(sint*lnx-cost*lny);
					//pb[x]=inv*coef+bias;
					//pg[x]=non_inv*coef+bias;
					//     ,   .   
					//  ,      ,    
					//   ,      (.   ).
					pb[x]=uint8_t(Students::Rudyak::Scale(  inv  / range ,scale) + bias);
					pg[x]=uint8_t(Students::Rudyak::Scale(non_inv/ range ,scale) + bias);
				}
				pr+=nls;
				pg+=nls;
				pb+=nls;
			}

		std::string fn=fname_out+"_inv.pgm";
		::AImageSave(b,fn.c_str(),indicator);
		std::string fn2=fname_out+"_non_inv.pgm";
		::AImageSave(g,fn2.c_str(),indicator);

		return 0;
}
}