#include <iostream>
#include <iomanip>
#include <string>
#include <sstream>
#include <memory> //auto_ptr()
#include <assert.h>

#include "aimage.h"
#include "yfilesys.h"
#include "consolap.h"

#include <typeinfo.h>

#include "num_linalg.h"
#include "vtl_meta_base.h"
#include "vtl_meta_SizeList.h"
#include "vtl_image.h"

#include "tests.h"



namespace Tests
{
	void test_meta_tools()
	{
		std::cout << ">>>===== Vtl::Meta::Tools  tests =====<<<" << std::endl;
		{
			using namespace Vtl;

			const int Ndim=3;//Compile time constant !
			const int Nbands=2;//Compile time constant !
			int a[Ndim];
			int b[]={1,2,3,4,5};
			Meta::set<Ndim>(a,33);
			Meta::print_array<Ndim>(a,"aa - 33");
			Meta::copy<Ndim>(a,b+1);
			Meta::print_array<Ndim>(a,"aa - 33");
		}
		{//EnumTraits tests
			std::cout << "EnumTraits::max " << Vtl::Meta::EnumTraits::max << std::endl;
			std::cout << "EnumTraits::min " << Vtl::Meta::EnumTraits::min << std::endl;
			std::cout << "EnumTraits::isSigned " << Vtl::Meta::EnumTraits::isSigned << std::endl;
			std::cout << "EnumTraits::bits " << Vtl::Meta::EnumTraits::bits << std::endl;
		}
	}
	void test_meta_IndexList()
	{
		std::cout << ">>>===== Vtl::Meta::IndexList  tests =====<<<" << std::endl;
		using namespace Vtl;
		using namespace Vtl::Meta;
		//typedef IndexList<3,IndexList<7,IndexList<4> > > il1;
		typedef IL::Make<3,7,4>::Result il1;
		std::cout << "Length = "<< IL::Length<il1>::value << std::endl;
		std::cout << "IndexOf 3 = "<< IL::IndexOf<3,il1>::value << std::endl;
		std::cout << "IndexOf 7 = "<< IL::IndexOf<7,il1>::value << std::endl;
		std::cout << "IndexOf 4 = "<< IL::IndexOf<4,il1>::value << std::endl;
		std::cout << "IndexOf 2 = "<< IL::IndexOf<2,il1>::value << std::endl;
		std::cout << "IndexOf -1 = "<< IL::IndexOf<-1,il1>::value << std::endl;
		std::cout << "Elem 1 = "<< IL::Elem<1,il1>::value << std::endl;
		std::cout << "Elem 2 = "<< IL::Elem<2,il1>::value << std::endl;
		std::cout << "Elem 0 = "<< IL::Elem<0,il1>::value << std::endl;
//		std::cout << "Elem 3 = "<< Elem<3,il1>::value << std::endl; //Generate error, it is DESIRED!
//		std::cout << "Elem -1 = "<< Elem<-1,il1>::value << std::endl;//Generate error, it is DESIRED!
		typedef IL::MakeInRange<2,6>::Result il2;
		std::cout << "Length = "<< IL::Length<il2>::value << std::endl;
		std::cout << "Elem 0 = "<< IL::Elem<0,il2>::value << std::endl;
		std::cout << "Elem 0 = "<< IL::Elem<1,il2>::value << std::endl;
		std::cout << "Elem 0 = "<< IL::Elem<2,il2>::value << std::endl;
		std::cout << "Elem 0 = "<< IL::Elem<3,il2>::value << std::endl;
		std::cout << "Erase 0 = "<< IL::Elem<1,IL::Erase<3,il2>::Result>::value << std::endl;
//		std::cout << "Erase 0 = "<< Elem<1,Erase<6,il2>::Result>::value << std::endl;//Generate error, it is DESIRED!

		std::cout << "il2 :" << IL::type2ostream<il2>() <<std::endl;
		typedef IL::EraseAt<2,il2>::Result il3;
		std::cout << "il3 :" << IL::type2ostream<il3>() <<std::endl;
		typedef IL::EraseAt<0,il3>::Result il4;
		std::cout << "il4 :" << IL::type2ostream<il4>() <<std::endl;

	}
	void test_meta_SizeList()
	{
		using namespace Vtl;
		std::cout << ">>>===== Vtl::Meta::SizeList =====<<<" << std::endl;
		{
			typedef Vtl::Meta::IL::Make<128,0,256,0,10>::Result SList;
			enum { sz=Vtl::Meta::IL::Length<SList>::value };
			std::cout << sz << std::endl;

			std::cout << "Elem["<<0<<"]  is " << Vtl::Meta::IL::Elem<0,SList>::value << std::endl;
			std::cout << "Elem["<<1<<"]  is " << Vtl::Meta::IL::Elem<1,SList>::value << std::endl;
			std::cout << "Elem["<<2<<"]  is " << Vtl::Meta::IL::Elem<2,SList>::value << std::endl;
			std::cout << "Elem["<<3<<"]  is " << Vtl::Meta::IL::Elem<3,SList>::value << std::endl;
			std::cout << "Elem["<<4<<"]  is " << Vtl::Meta::IL::Elem<4,SList>::value << std::endl;
			//Next line generate compile error intentionally
			//std::cout << "SizeList["<<5<<"]  is " << Vtl::Meta::IL::Elem<5,SList>::value << std::endl;
			std::cout << "sizeof(SList)= " << sizeof(SList) <<std::endl;
		}
		{
			typedef Vtl::Meta::IL::Make<1,0,1,0,1>::Result SizeList_2;
			std::cout << "Vtl::Meta::IL::Count<0,SizeList_2>::value= " << Vtl::Meta::IL::Count<0,SizeList_2>::value <<std::endl;
			std::cout << "Vtl::Meta::IL::Count<0,SizeList_2>::value= " << Vtl::Meta::IL::Count<1,SizeList_2>::value <<std::endl;
			std::cout << "Vtl::Meta::IL::ICount<0,1,SizeList_2>::value= " << Vtl::Meta::IL::ICount<0,1,SizeList_2>::value <<std::endl;
			std::cout << "Vtl::Meta::IL::ICount<0,3,SizeList_2>::value= " << Vtl::Meta::IL::ICount<0,3,SizeList_2>::value <<std::endl;
		}
		typedef Vtl::Meta::SizeList<128,0,256,0,10> SZLst;
		SZLst sl(128,222,256,444,10);
		std::cout << "sl.isEnum<0>() = " << sl.isEnum<0>() <<std::endl;
		std::cout << "sl.isEnum<1>() = " << sl.isEnum<1>() <<std::endl;
		std::cout << "sl.isEnum<2>() = " << sl.isEnum<2>() <<std::endl;
		std::cout << "sl.isEnum<3>() = " << sl.isEnum<3>() <<std::endl;
		std::cout << "sl.isEnum<4>() = " << sl.isEnum<4>() <<std::endl;
//		std::cout << "sl.isEnum<5>() = " << sl.isEnum<5>() <<std::endl;

		std::cout << "sizeof(sl) = " << sizeof(sl) <<std::endl;
		//std::cout << "sizeof(Vtl::Meta::SizeList_base<SizeList>) = " << sizeof(Vtl::Meta::SizeList_base<SizeList>) <<std::endl;
		//std::cout << "sizeof(Vtl::Meta::SizeList_base<SizeList>) = " << sizeof(Vtl::Meta::SizeList_base<SizeList>) <<std::endl;

		std::cout << "sl.x<0>() = " << sl.x<0>() <<std::endl;
		std::cout << "sl.x<1>() = " << sl.x<1>() <<std::endl;
		std::cout << "sl.x<2>() = " << sl.x<2>() <<std::endl;
		std::cout << "sl.x<3>() = " << sl.x<3>() <<std::endl;
		std::cout << "sl.x<4>() = " << sl.x<4>() <<std::endl;
		sl.set<0>(123);
		sl.set<1>(124);
		sl.set<2>(125);
		sl.set<3>(126);
		sl.set<4>(127);
		std::cout << "sl.x<0>() = " << sl.x<0>() <<std::endl;
		std::cout << "sl.x<1>() = " << sl.x<1>() <<std::endl;
		std::cout << "sl.x<2>() = " << sl.x<2>() <<std::endl;
		std::cout << "sl.x<3>() = " << sl.x<3>() <<std::endl;
		std::cout << "sl.x<4>() = " << sl.x<4>() <<std::endl;
		std::cout << "sl.hasInts = " << sl.hasInts <<std::endl;
		std::cout << "sl.size = " << sl.size <<std::endl;

	}
	void test_meta_Align()
	{
		using namespace Vtl;
		std::cout << ">>>===== Vtl::Meta::Align =====<<<" << std::endl;
		enum {align_to=4};
		typedef short TT;
		std::cout << "Meta::Align<char,sz,align_to>::value" <<std::endl;
		std::cout << Meta::Align<TT,1,align_to>::value <<std::endl;
		std::cout << Meta::Align<TT,2,align_to>::value <<std::endl;
		std::cout << Meta::Align<TT,3,align_to>::value <<std::endl;
		std::cout << Meta::Align<TT,4,align_to>::value <<std::endl;
		std::cout << Meta::Align<TT,5,align_to>::value <<std::endl;
		std::cout << Meta::Align<TT,6,align_to>::value <<std::endl;
		std::cout << Meta::Align<TT,7,align_to>::value <<std::endl;
		std::cout << Meta::Align<TT,8,align_to>::value <<std::endl;
	}

	template<int I0,int I1,int I2,int I3,int I4>
	struct Test_meta_Layout_5D
	{
		enum {NBands=3};
//		enum {Interleave=ndim};
		enum {Interleave=1};
//		enum {align_to=1};
		enum {align_to=8};
	public:
		template<int ii>
		const char * ParamPRN() const { return ii?":enum":":var "; }
		Test_meta_Layout_5D(int I0_var,int I1_var,int I2_var,int I3_var,int I4_var)
		{
			using namespace Vtl;
			std::cout << ">>>===== Vtl::Meta:: Image Layout =====<<<" << std::endl;

			std::cout << "  with I0:"<<ParamPRN<I0>() 
				           << "  I1:"<<ParamPRN<I1>()
				           << "  I2:"<<ParamPRN<I2>()
				           << "  I3:"<<ParamPRN<I3>()
				           << "  I4:"<<ParamPRN<I4>()
						   << std::endl << std::endl;

			typedef Vtl::Meta::SizeList<I0,I1,I2,I3,I4> SZLst;
			SZLst sl(I0_var,I1_var,I2_var,I3_var,I4_var);

			enum {ndim=SZLst::size}; 
			std::cout << "sl.isEnum<0>() = " << sl.isEnum<0>() <<std::endl;
			std::cout << "sl.isEnum<1>() = " << sl.isEnum<1>() <<std::endl;
			std::cout << "sl.isEnum<2>() = " << sl.isEnum<2>() <<std::endl;
			std::cout << "sl.isEnum<3>() = " << sl.isEnum<3>() <<std::endl;
			std::cout << "sl.isEnum<4>() = " << sl.isEnum<4>() <<std::endl;
			std::cout << "sizeof(sl) = " << sizeof(sl) <<std::endl;
			std::cout << "sl.x<0>() = " << sl.x<0>() <<std::endl;
			std::cout << "sl.x<1>() = " << sl.x<1>() <<std::endl;
			std::cout << "sl.x<2>() = " << sl.x<2>() <<std::endl;
			std::cout << "sl.x<3>() = " << sl.x<3>() <<std::endl;
			std::cout << "sl.x<4>() = " << sl.x<4>() <<std::endl;
			std::cout << "sl.hasInts = " << sl.hasInts <<std::endl;

	//		typedef Meta::BuildDefaultLayout<float,SZLst::SList,NBands,Interleave,align_to > LayList;
	//		std::cout << "LayList :" << Meta::IL::type2ostream<LayList::Result>() <<std::endl;

			typedef Vtl::Meta::LayoutListDefault<float,SZLst,3,Interleave,5,align_to> LayoutTypeDef;
			LayoutTypeDef ll(sl,3);
			LayoutTypeDef::LayoutType lay(ll);
			std::cout << "lay.stride<0>() =\t" << lay.stride<0>() <<std::endl;
			std::cout << "lay.stride<1>() =\t" << lay.stride<1>() <<std::endl;
			std::cout << "lay.stride<4>() =\t" << lay.stride<4>() <<std::endl;
			std::cout << "lay.stride_band() =\t" << lay.stride_band() <<std::endl;
			std::cout << "lay.Bands() =\t" << lay.Bands() <<std::endl;
			std::cout << "lay.Ndim() =\t" << lay.Ndim <<std::endl;
			typedef Vtl::Img::ImageNDref<SZLst,LayoutTypeDef::LayoutType> ImgRef_Type;
			ImgRef_Type iref(0,sl,lay);
			ImgRef_Type::Iter1D<1>::Type it;
			std::cout << typeid(ImgRef_Type::Iter1D<0>::Type).name() <<std::endl;
			std::cout << typeid(ImgRef_Type::Iter1D<1>::Type).name() <<std::endl;
			std::cout << typeid(ImgRef_Type::Iter1D<2>::Type).name() <<std::endl;
			std::cout << typeid(ImgRef_Type::Iter1D<3>::Type).name() <<std::endl;
			std::cout << typeid(ImgRef_Type::Iter1D<4>::Type).name() <<std::endl;
			std::cout << typeid(ImgRef_Type::Iter1D<5>::Type).name() <<std::endl;
			ImgRef_Type::IterND ni = iref.NIter();
			ni.print();
			++ni.x<0>();
			ni.print();
			++ni.x<1>();
			ni.print();
			++ni.x<2>();
			ni.print();
			++ni.x<3>();
			ni.print();
		}
	};
	template<int NBands,int Interleave,int Granulation,int align_to,int I0,int I1,int I2,int I3>
	struct Test_meta_BuildDefaultLayout_4D
	{
	public:
		Test_meta_BuildDefaultLayout_4D(int I0_var,int I1_var,int I2_var,int I3_var)
		{
			typedef float T;
			typedef Vtl::Meta::SizeList<I0,I1,I2,I3> SZLst;
			SZLst sl(I0_var,I1_var,I2_var,I3_var);
			enum {ndim=SZLst::size}; 
			typedef Vtl::Meta::BuildDefaultLayout<T,SZLst::IList,NBands,Interleave,Granulation,align_to>::Result temp_IList;
			std::cout << "il2 :" << Vtl::Meta::IL::type2ostream<temp_IList>() <<std::endl;
		}
	};


int test_vtl_ND_image(int argc,char* argv[])
{
	using namespace Vtl;
	test_meta_tools();
			std::cout << std::endl << std::endl;
	test_meta_IndexList();
			std::cout << std::endl << std::endl;
	test_meta_SizeList();
			std::cout << std::endl << std::endl;
	test_meta_Align();
			std::cout << std::endl << std::endl;
	{
		Test_meta_BuildDefaultLayout_4D<3,0,4,1,5,7,11,13> dl0(5,7,11,13);
		Test_meta_BuildDefaultLayout_4D<3,1,4,1,5,7,11,13> dl1(5,7,11,13);
		Test_meta_BuildDefaultLayout_4D<3,2,4,1,5,7,11,13> dl2(5,7,11,13);
		Test_meta_BuildDefaultLayout_4D<3,3,4,1,5,7,11,13> dl3(5,7,11,13);
		Test_meta_BuildDefaultLayout_4D<3,4,4,1,5,7,11,13> dl4(5,7,11,13);
		//Test_meta_BuildDefaultLayout_4D<3,5,1,5,7,11,13> dl5(5,7,11,13);
		std::cout << "    Desired   : Length=6,  Elems = { 3, 15, 105, 1155, 1, 3 }" <<std::endl; 
		std::cout << "    Desired   : Length=6,  Elems = { 1, 15, 105, 1155, 5, 3 }" <<std::endl; 
		std::cout << "    Desired   : Length=6,  Elems = { 1, 5, 105, 1155, 35, 3 }" <<std::endl; 
		std::cout << "    Desired   : Length=6,  Elems = { 1, 5, 35, 1155, 385, 3 }" <<std::endl; 
		std::cout << "    Desired   : Length=6,  Elems = { 1, 5, 35, 385, 5005, 3 }" <<std::endl; 
			std::cout << std::endl;
	}
	{
		Test_meta_BuildDefaultLayout_4D<3,0,4,8,5,7,11,13> dl0(5,7,11,13);
		Test_meta_BuildDefaultLayout_4D<3,1,4,8,5,7,11,13> dl1(5,7,11,13);
		Test_meta_BuildDefaultLayout_4D<3,2,4,8,5,7,11,13> dl2(5,7,11,13);
		Test_meta_BuildDefaultLayout_4D<3,3,4,8,5,7,11,13> dl3(5,7,11,13);
		Test_meta_BuildDefaultLayout_4D<3,4,4,8,5,7,11,13> dl4(5,7,11,13);
		//Test_meta_BuildDefaultLayout_4D<3,5,4,8,5,7,11,13> dl5(5,7,11,13);
		std::cout << "    Desired   : Length=6,  Elems = { 3, 16, 112, 1232, 1, 3 }" <<std::endl; 
		std::cout << "    Desired   : Length=6,  Elems = { 1, 18, 126, 1386, 6, 3 }" <<std::endl; 
		std::cout << "    Desired   : Length=6,  Elems = { 1, 6, 126, 1386, 42, 3 }" <<std::endl; 
		std::cout << "    Desired   : Length=6,  Elems = { 1, 6, 42, 1386, 462, 3 }" <<std::endl; 
		std::cout << "    Desired   : Length=6,  Elems = { 1, 6, 42, 462, 6006, 3 }" <<std::endl; 
			std::cout << std::endl;
	}
	{
		Test_meta_Layout_5D<3,5,7,11,13> tml_1(3,5,7,11,13);
			std::cout << std::endl << std::endl;
		Test_meta_Layout_5D<3,5,0, 0,13> tml_2(3,5,7,11,13);
			std::cout << std::endl << std::endl;
		Test_meta_Layout_5D<0,0,0, 0, 0> tml_3(3,5,7,11,13);
	}
			std::cout << std::endl << std::endl;
	return 0;
}
}//namespace Tests
