//#pragma warning( disable : 4786 )
#include <iostream>
#include <memory>//auto_ptr
#include <windows.h>//CharToOem()
#include <stdlib.h>//setenv()

#include "ywinini.h"
#include "xml_interface.h"

#include "tests.h"

namespace Tests
{

	std::ostream& operator<<(std::ostream& os,const char* str) 
	{ 
		std::string s(str);
		CharToOem(str,(char *)s.c_str());//trust, that string size are equal and no reallocation take place.
		os<<s;

		return os;
	}

	std::ostream& operator<<(std::ostream& os,const DataString& str)
	{
		return os<< (char*)str.c_str(); 
	}

	#define USE_XML

	#ifdef USE_XML
		typedef YXML::DNode TreeNode;
	#else
		typedef WiNode TreeNode;
	#endif


	static void CommonIniXmlInterfaceDemo(TreeNode n);
		
		
		
		
		
	int test_YWinIni(int argc, char* argv[])
	{
		std::cout << "Demo tree-like data" << std::endl;


	#ifdef USE_XML
		//YXML     ,    ,    !
		//YXML::XMLdoc doc("e:/LIB/SHARED_S/psp.xml","Passport",true);
		YXML::XMLdoc doc(".\\psp.xml","Passport",true);
	#else
		WinIniFile doc(".\\psp.ini");
	#endif

		TreeNode n=doc.Root();

		//  YXML::DNode  WiNode ,      .
		//     ,  ,
		//    ,    .
		CommonIniXmlInterfaceDemo(n);

		//Specific XML demo
	#ifdef USE_XML
		n=doc.Root()("ImageSet");
		n["transf"]="HRPT";
		n["device"]="AVHRR";

		//Append childs without checking it already in document
		for(int i=0; i<5 ; i++)
		{
			YXML::DNode n2=n.AppendChild(YXML::DNode::ELEMENT,"image");
			n2["chan"]=i+1;
			n2.AppendChild(YXML::DNode::ELEMENT,"clb")="calibration data";
		}

		{
			std::cout << "Find Element nodes with attr in some bound" << std::endl;
			n=doc.Root()("ImageSet");
			//  
			YXML::DNode::IsCond cond(YXML::IsNameEq("image") && YXML::IsAttr<int,std::less<int> >("chan",3));
			// ,  
			YXML::DNode n2=n.FirstChild(cond);
			while(n2.IsNode())
			{
				std::cout << "<" << n2.getName() << " chan="<<n2.GetAttr("chan") <<">" << std::endl;
				n2=n2.Next(cond);
			}
		}
		{
			std::cout << "Find Element nodes with attr in some bound -another condition" << std::endl;
			n=doc.Root()("ImageSet");
			//  
			YXML::DNode::IsCond cond(YXML::IsNameEq("image") && YXML::IsAttrIn<int>("chan",3,4));
			// ,  
			YXML::DNode n2=n.FirstChild(cond);
			while(n2.IsNode())
			{
				std::cout << "<" << n2.getName() << " chan="<<n2.GetAttr("chan") <<">" << std::endl;
				n2=n2.Next(cond);
			}
		}
		{
			std::cout << "Find Element nodes with attr in some bound -one more complex condition" << std::endl;
			n=doc.Root()("ImageSet");
			//  
			YXML::DNode::IsCond cond=YXML::IsNameEq("image") && ( YXML::IsAttrIn<int>("chan",3,4) || YXML::IsAttr<int,std::equal_to<int> >("chan",1) );
			// ,  
			YXML::DNode n2=n.FirstChild(cond);
			while(n2.IsNode())
			{
				std::cout << "<" << n2.getName() << " chan="<<n2.GetAttr("chan") <<">" << std::endl;
				n2=n2.Next(cond);
			}
		}

		
		//Append childs WITH checking it already in document, and modify if exist.
		//Two equal name childs differ only by attribute
		n=doc.Root();
		YXML::DNode n2=n.FirstChild("Station-XML");
		if(n2.IsNode())
			n=n2; //already exist
		else
		{//Not exist, insert before "Station"
			n2=n.FirstChild("Station");
			if(n2.IsNode())
			{
				n=n2.InsertBefore(YXML::DNode::ELEMENT,"Station-XML");
				n.InsertBefore(YXML::DNode::COMMENT,"Inserted comment for Station-XML");
			}
			else
				n=n.AppendChild(YXML::DNode::ELEMENT,"Station-XML");
		}
		//now n=doc.Root()("Station-XML");
		n2=n.FirstChild("name");
		bool rusn=false;
		bool engn=false;
		while(n2.IsNode())
		{
			if(n2.GetAttr("lang")=="ru")
			{
				DataString s=n2;
				n2=s+"_1";
				rusn=true;
			}
			else if(n2.GetAttr("lang")=="en")
			{
				DataString s=n2;
				n2=s+"_2";
				engn=true;
			}
			n2=n2.Next("name");
		}
		if(!rusn)
		{
			n2=n.AppendChild(YXML::DNode::ELEMENT,"name");
			n2["lang"]="ru";
			n2=" ";
		}
		if(!engn)
		{
			n2=n.AppendChild(YXML::DNode::ELEMENT,"name");
			n2["lang"]="en";
			n2="Roskartographia";
		}

	//Alternative variant

		//  
		YXML::DNode::IsCond cond(YXML::IsNameEq("name") && YXML::IsAttr<std::string,std::equal_to<std::string> >("lang","ru"));
		// ,  
		n2=n.FirstChild(cond);
		if(n2.IsNode())
			std::cout << "Found russian version:" << n2 << std::endl; 
		//      ,  
		n2=n.FirstChild(YXML::IsNameEq("name") && YXML::IsAttr<std::string,std::equal_to<std::string> >("lang","en"));
		if(n2.IsNode())
			std::cout << "Found english version:" << n2 << std::endl; 

		
		n2=n.FirstChild("name","lang","ru");
		if(n2.IsNode())
			std::cout << "Found russian version:" << n2 << std::endl; 
		n2=n.FirstChild("name","lang","en");
		if(n2.IsNode())
			std::cout << "Found english version:" << n2 << std::endl; 

	#endif
		return 0;
	}



	void CommonIniXmlInterfaceDemo(TreeNode root)
	{
		TreeNode n=root;
		n("Program")["AlcorSat"]=1;
		n("Program")["version"]="1.3.7";
		n("Program")="  ";
		

		std::string seance("Seance");
		std::string image("image");
		DataString ds="2003-03-12T19:21:3.65432Z+03:30";
		YDataTime dt1=ds;
		YDataTime dt2=dt1;
		YTimeInterval ti;
		ti.Set(0,1,0,0.001);
		dt2+=ti;

		{
			TreeNode n1=n.GetNode(seance).GetNode("Zap");
			n1("time1")=dt1;
			n1("time2")=dt2;
			n1["Trace"]=43331;
			n1["corr"]=123.456;
			n1(image)["width"]=2048;
			n1(image)["height"]=5123;
			n1(image)["flip_vert"]=1;
			n1(image)["flip_hor"]=0;
		}
		{
			TreeNode n1=n.GetNode(seance).GetNode("Vos");
			YTimeInterval ti;
			ti.Set(0,1,40,10.001);
			dt1-=ti;
			dt2+=ti;
			n1("time1")=dt1;
			n1("time2")=dt2;
		}
		{
			TreeNode n1=n.GetNode("Station");
			YDataTime dt;
			n1["id"]=17;
			dt.Set(0,0,0,3,30,0.0);
			n1("time_zone")=dt;
			n1("name")("ru")=" ";
			n1("name")("en")="Roskartographia";
		}

		n=n.GetNode(std::string("s1-0")).GetNode(std::string("s1-1"));
		n=n.GetNode("s1-2");
		n[std::string("i1")]=std::string("data in the item 1");
		n["i2"]="data in the item 2";
	//-----------
		n=root;
		n=n("s1-0")("s1-1");
		n=n("s1-2");
		n["i3"]="data in the item 3";
		n["i4"]="data in the item 4";
	//-----------
		n=root("s1-0");
		n=n("s2-1");
		n=n("s2-2");
		n["i3"]="data in the item 3";
		n["i4"]="data in the item 4";
	//-----------
	  n=root("s1-0");
		n=n("s1-1");
		n=n("s1-2");
	  n["i5"] = "s1- 0/1/2/ i5";
	  std::string a1 = DataString(n["i5"]);
		n("s1-3")["i6"] = "s1- 0/1/2/3 i5";
	//-----------
		n=root("s1-0")("s1-1")("s1-2");
		n["i1"]="data in the item 1";
		n ="asdasdasdqwe_empty";
	//	n << "asdasdasdqwe_empty";
		root("s1-0")("s1-1")("s1-3")="1234123";
	//-----------different output syntax --------------
		n=root.GetNode("s1-0");
		std::cout << "read from ini:" << n.GetNode("s1-1").GetNode("s1-2")["i1"] << std::endl;
		std::cout << "read from ini:" << n("s1-1")("s1-2")["i2"] << std::endl;
		std::cout << "read from ini:" << n("s1-1")("s1-2")["i3"] << std::endl;
		std::cout << "read from ini:" << n("s1-1")("s1-2")["i4"] << std::endl;
		std::cout << "read from ini:" << n.GetNode("s1-1")("s1-2") << std::endl; 

		std::cout << "read from ini:" << root("Program") << std::endl; 
		DataString s=root("Program");
		std::cout << "read from ini:" << s << std::endl; 
		const char *ss=s.c_str();
		std::cout << "read from ini:" << ss << std::endl; 
	}


}//namespace Tests
