#include <stdio.h>
#include <string>
#include <windows.h>
#include "tests.h"
#include "xml_interface.h"

using namespace YXML;

namespace Tests
{
	static void offset(int level)
	{
		while(level--)
			printf("  ");
	}
	void Print(DNode n,int level=0)
	{
		char buf[256];
		offset(level); 
		::CharToOem(n.getName().c_str(),buf);
		printf("%d:%s",int(n.getType()),buf);
		if(DNode::ELEMENT==n.getType())
		{
			std::string val=n.GetAttr("i1");
			if(!val.empty())
			printf("{attribute i1=%s, size=%d}",val.c_str(),val.size());
		}
		else if(DNode::TEXT==n.getType() || DNode::COMMENT==n.getType())
		{
			::CharToOem(n.getValue().c_str(),buf);
			printf(":%s",buf);
		}

		printf("\n");
		n=n.FirstChild();
		while(n.IsNode())
		{
			Print(n,level+1);
			n=n.Next();
		}
	}

	void Print2(DNode n,int level=0)
	{
		char buf[256];
		offset(level);
		
		::CharToOem(n.getName().c_str(),buf);
		printf("%d:%s",int(n.getType()),buf);
		if(DNode::ELEMENT==n.getType())
		{
			std::string val=n.GetAttr("i1");
			if(!val.empty())
			printf("{attribute i1=%s, size=%d}",val.c_str(),val.size());
		}
		else if(DNode::TEXT==n.getType() || DNode::COMMENT==n.getType())
		{
			::CharToOem(n.getValue().c_str(),buf);
			printf(":%s",buf);
		}

		printf("\n");

	/*	NodeList list=n.Childs();

		for(NodeList::iterator it=list.begin(); it<list.end(); it.next() )
		{
			Print(it.get(),level+1);
		}
	*/
	}


	int test_XML(int argc, char* argv[])
	{
		XMLdoc doc;
		//doc.Create("doc_new.xml","MyDoc");

		doc.Load("qq2.xml");
		DNode n=doc.Root();
		n=n.FirstChild();
		while(n.IsNode())
		{
			if(DNode::ELEMENT==n.getType())
				if(n.getName()==std::string("Program"))
				{
					n.SetAttr("i1","qwerty");
					break;
				}

			n=n.Next();
		}
		Print(doc.Root());
	//	Print2(doc.Root());


		doc.Save("doc_yxml.xml",true);
	//-------------------------------
		doc.Create("doc_new.xml","Passport");
		//doc.Load("doc_new.xml");
		{
			DNode n=doc.Root();
			DNode n2=n.AppendChild(DNode::ELEMENT,"Sat");
			n2.AppendChild(DNode::ELEMENT,"Name").AppendChild(DNode::TEXT,"NOAA 16");
			DNode n3=n2.AppendChild(DNode::ELEMENT,"Description");
			n3.AppendChild(DNode::TEXT,"NOAA");
			n3.SetAttr("number","16");
			n2=n.AppendChild(DNode::ELEMENT,"Data");
			n2.AppendChild(DNode::COMMENT,"HRPT data for AVHRR");
			n2=n2.AppendChild(DNode::ELEMENT,"ImageSet");
			n2.SetAttr("type","received");
			n2.SetAttr("transf","HRPT");
			n2.SetAttr("dev","AVHRR");
			for(int i=1; i<=5; i++)
			{
				//char buf[256];
				n3=n2.AppendChild(DNode::ELEMENT,"Image");
				{
				//::sprintf(buf,"%d",i);
					std::ostringstream os;
					os << i;
				//n3.SetAttr("chan",buf);
					n3.SetAttr("chan",os.str().c_str());
				}
				std::ostringstream os;
				os << "d43331_" << i << ".oz";
				//::sprintf(buf,"d43331_%d.oz",i);
				DNode n4=n3.AppendChild(DNode::ELEMENT,"file").AppendChild(DNode::TEXT,os.str());
				//DNode n4=n3.AppendChild(DNode::ELEMENT,"file").AppendChild(DNode::TEXT,buf);
				n4=n3.AppendChild(DNode::ELEMENT,"sw").AppendChild(DNode::TEXT,"5456 | 1 5456");
				n4=n3.AppendChild(DNode::ELEMENT,"CalibCof");
				n4.SetAttr("Type",(i<3)?"VISIBLE":"IR");
				n4.AppendChild(DNode::TEXT,"here is data");
			}
		}
		doc.Save("doc_new.xml",true);
		doc.Load("doc_new.xml");
		{
			DNode n=doc.Root();
			//find node 'Data'
			n=n.FirstChild();
			while(n.IsNode())
			{
				if(DNode::ELEMENT==n.getType())
				{
					std::string name=n.getName();
					if(name==std::string("Data"))
						break;
				}
				n=n.Next();
			}
			//find node 'ImageSet'
			n=n.FirstChild();
			while(n.IsNode())
			{
				if(DNode::ELEMENT==n.getType())
					if(n.getName()=="ImageSet")
						break;
				n=n.Next();
			}
			if(n.IsNode())
			{//found ok
				if(n.GetAttr("dev")=="AVHRR" && n.GetAttr("type")=="received")
				{//AVHRR device original data, process
					//find 3 channel data
					DNode n3=n.FirstChild();
					while(n3.IsNode())
					{
						if(DNode::ELEMENT==n3.getType())
						{
							std::string name=n3.getName();
							if(name=="Image")
							{
								if(n3.GetAttr("chan")=="3")
								break;
							}
						}
						n3=n3.Next();
					}
					if(n3.IsNode())
					{//found 3 ch node OK
						DNode file=n3.FirstChild();
						while(file.IsNode())
						{
							if(DNode::ELEMENT==file.getType())
							{
								std::string name=file.getName();
								if(name=="file")
								{
									file.FirstChild().setValue("new_file.tif");
									break;
								}
							}
							file=file.Next();
						}

					}
				}
			}
		}
		doc.Save("doc_new2.xml",true);
		doc.Load("doc_new2.xml");
		{
			::printf(">---FirstChild(name),Next(name) demo:\n");
			DNode n=doc.Root();
			printf("Name=%s\n",n.getName().c_str());
			n=n.FirstChild("Data");
			printf("Name=%s\n",n.getName().c_str());
			n=n.FirstChild("ImageSet");
			printf("Name=%s\n",n.getName().c_str());
			n=n.FirstChild("Image");
			printf("List of childs named \"%s\"\n","Image");
			while(n.IsNode())
			{
				char *at="chan";
				::printf("<%s %s=%s>\n",n.getName().c_str(),at,n.GetAttr(at).c_str());
				n=n.Next("Image");
			};

		}
		doc.Save("doc_new3.xml");

		return 0;
	}
}//namespace Tests
