/*! \page dox_gen_file_mapping_to_mem How to ...         

 ,   ,   ,  ,   -   , 
                  .
       (,  , . \n
@b :  32-      4 ,    Intel-Pentium - 2 .\n
   ,        PGM, PPM ( P5  P6 ),
 BMP,  IMG/HDR (raw),  ENVI       ,   
    .\n
     Windows,  .          FileAsArray,
     ,  UNIX   ,     ,    
       (        ).

@code

#include <windows.h>
    //    __pin ,   
    //    .   ,    
class FileAsArray
{
    HANDLE hFile;
    HANDLE hFileMapping;
    uint8_t *ptr;
    size_t file_size;
public:
    FileAsArray(const char *fname) 
        :hFile(INVALID_HANDLE_VALUE),hFileMapping(0)
        ,ptr(0), file_size(0)
    {
    //  ,    ,   
    //   /    - 
        hFile = CreateFile( fname, GENERIC_READ, 0, NULL,             OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
        if(hFile!=INVALID_HANDLE_VALUE)
        {
            //    , .   
            file_size = GetFileSize(hFile, NULL);
            //    
            hFileMapping = CreateFileMapping(hFile, NULL,PAGE_READONLY, 0, 0, NULL);
            if(hFileMapping !=0)
                ptr = (uint8_t *) MapViewOfFile(hFileMapping,FILE_MAP_READ, 0, 0, 0);
            //   -  ptr   
        }
    }
    virtual ~FileAsArray() {close();}
    //   close(),     
    //     .       //      .
    void close()
    {
        if(ptr!=0)
        UnmapViewOfFile(ptr);
        if(hFileMapping!=0)
        {
            CloseHandle(hFileMapping);
            hFileMapping=0;
        }
        if(hFile!=INVALID_HANDLE_VALUE)
        {
            CloseHandle(hFile);
            hFile=INVALID_HANDLE_VALUE;
        }
    }
    //  ,      
    // ,         .
    size_t size() const { return file_size;}
    const void* GetPtr() const {return ptr}
    //    read()   
    //  
    size_t read(size_t offset,void* data,size_t len) 
    {
        memcpy(data,ptr+offset,len);
        return len;
    }
};

//:
void main()
{
    FileAsArray fa("c:/mydir/my_file.img");
    //          2-
    //    , : 
    const uint16_t * img=( const uint16_t *)fa.GetPtr() ;
    //      ,  
    //    . 
    unsigned sum;
    int len=fa.size()/sizeof(img[0]); 
    for(int i=0 ; i<100 ; i++)
        sum+=img[i];
    printf("sum=%d"(int)sum);
    //       , 
    // ..      -
    //          
    //       . .  
    //  .
    img[len/2]=12123;
}

@endcode

 */
