#ifndef _APIPE_H
#define _APIPE_H
///////////////////////////////////////////////////////////////////////
#include <string>
#include <vector>
#include <map>
#include <windows.h>


class ReadPipe
{
	friend class childPipe;
    HANDLE handle_;              
private:
	ReadPipe() : handle_(INVALID_HANDLE_VALUE) {}
	ReadPipe(HANDLE handle) : handle_(handle) {}
	ReadPipe(const ReadPipe&);//no body, protect from copy
	ReadPipe& operator=(const ReadPipe&);//no body, protect from copy
	void open(HANDLE handle) 
	{
		if(handle_!=INVALID_HANDLE_VALUE)
			close();
		handle_=handle;
	}
public:
	~ReadPipe() { close(); }
	bool isEof() const {return handle_==INVALID_HANDLE_VALUE;}
	void close();
	unsigned readNoWait(char *data,unsigned len);
	unsigned readWait(char *data,unsigned len);
};

class WritePipe
{
	friend class childPipe;
    HANDLE handle_;              
private:
	WritePipe() : handle_(INVALID_HANDLE_VALUE) {}
	WritePipe(HANDLE handle) : handle_(handle) {}
	WritePipe(const WritePipe&);//no body, protect from copy
	WritePipe& operator=(const WritePipe&);//no body, protect from copy
	void open(HANDLE handle) 
	{
		if(handle_!=INVALID_HANDLE_VALUE)
			close();
		handle_=handle;
	}
public:
	~WritePipe() { close(); }
	bool isEof() const {return handle_==INVALID_HANDLE_VALUE;}
	void close();
	unsigned writeNoWait(const char *data,unsigned len);
	unsigned writeWait(const char *data,unsigned len);
};

class childPipe 
{
public://types and constants
	enum Flags 
	{
		F_StdIn =0x0001,
		F_StdOut=0x0002,//>      NUL
		F_StdErr=0x0004,//>      NUL
		F_Split =0x0008,//>     F_StdOut  F_StdErr.   F_Split  ,     F_StdOut
		F_StdInOut=F_StdIn|F_StdOut,
		F_StdInErr=F_StdIn|F_StdErr,
		F_StdOutErr=F_StdOut|F_StdErr,
		F_StdOutErrSplit=F_StdOutErr|F_Split,
		F_StdInOutErr=F_StdIn|F_StdOut|F_StdErr,
		F_StdInOutErrSplit=F_StdInOutErr|F_Split
	};
	typedef std::map<std::string,std::string> Env;
	static size_t envCalcBinBlockSize(const childPipe::Env& env);
	static void* BuildEnvBlock(std::vector<char>& e,const childPipe::Env* env);
	static void envBuildBinBlock(const childPipe::Env& env,char *e,char delim=0,char delim_line=0);
	static void envBuildFromBinBlock(childPipe::Env& env,const char *e);//>ONLY if build with delim=0,delim_line=0
	static char * envMakeBinBlock(childPipe::Env* env);
	static void envFreeBinBlock(char* env_blk);
private:
    PROCESS_INFORMATION pi;
    mutable DWORD exitStatus;
	WritePipe stdin_;
//    HANDLE hStdin;              // write, child's STDIN
    ReadPipe stdout_;             // read, child's STDOUT
    ReadPipe stderr_;             // read, child's STDOUT
	bool the_same_stderr_stdout;
private:
	void nullify();
	static bool Is(Flags flags_to_check,Flags flags) {return (flags_to_check&flags)==flags; }
public:
	childPipe() {nullify();}
	childPipe(std::string command,const Env *env=0,Flags flags=F_StdInOutErr) 
		{nullify();Init(command ,env,flags );}
	~childPipe() {Close();}
	void Init(std::string command,const Env *env=0,Flags flags=F_StdInOutErr);
	void Close();
	DWORD ExitStatus() {return exitStatus;}
    bool isRunning();
    void terminate();//  ,  ,     -    ,      DLL.
    void wait();
	WritePipe& Stdin() {return stdin_;}
	ReadPipe& Stdout() {return stdout_; }
	ReadPipe& Stderr() {return the_same_stderr_stdout ? stdout_:stderr_; }
};

class ReadString
{
	ReadPipe& rdpipe_;
	const unsigned max_len;
	std::vector<char> buffer;
	unsigned first_pos2;
	unsigned last_pos2;
	unsigned pos;
	bool null_detected;
	bool eof_detected;
	char last_char;
private:
	void copy();
public:
	ReadString(ReadPipe& rdpipe,unsigned max_string_len);
	char *gets(char *buf);
	bool IsAvail();
	bool IsEOF() const {return eof_detected; }
};

///////////////////////////////////////////////////////////////////////
#endif //_APIPE_H
