#include <math.h> //sqrt
#include <stdlib.h> //div_t
#include "ytypes.h"
#include "ygeometr.h"

inline void swap( int* pa, int* pb )
{
	int t= *pa;
	*pa = *pb;
	*pb = t;
}

// LineFun functions

bool LineFun::areIntersecting(const YPointInt& begin1, const YPointInt& end1, const YPointInt& begin2, const YPointInt& end2, bool strict)
{
	int dx1 = end1.x - begin1.x;
	int dy1 = end1.y - begin1.y;
	int dx2 = end2.x - begin2.x;
	int dy2 = end2.y - begin2.y;

	if( dx1 == 0 && dy1 == 0 )
		return ( !strict && LineFun::isInLine( begin2, end2, begin1) );
	if( dx2 == 0 && dy2 == 0 )
		return ( !strict && LineFun::isInLine( begin1, end1, begin2) );

	if( strict )
		if( ((dx1*(begin2.y-begin1.y)-dy1*(begin2.x-begin1.x))*(dx1*(end2.y-begin1.y)-dy1*(end2.x-begin1.x))) < 0
		 && ((dx2*(begin1.y-begin2.y)-dy2*(begin1.x-begin2.x))*(dx2*(end1.y-begin2.y)-dy2*(end1.x-begin2.x))) < 0 )
			return true;
	if( !strict )
		if( ((dx1*(begin2.y-begin1.y)-dy1*(begin2.x-begin1.x))*(dx1*(end2.y-begin1.y)-dy1*(end2.x-begin1.x))) <= 0
		 && ((dx2*(begin1.y-begin2.y)-dy2*(begin1.x-begin2.x))*(dx2*(end1.y-begin2.y)-dy2*(end1.x-begin2.x))) <= 0 )
			return true;

	return false;
}

bool LineFun::isInLine( const YPointInt& begin, const YPointInt& end, const YPointInt& what, bool strict)
{
	int minx;
	int maxx;
	int miny;
	int maxy;
	int x = what.x;
	int y = what.y;

	if (begin.x<end.x)
	{
		minx = begin.x;
		maxx = end.x;
	} else
	{
		minx = end.x;
		maxx = begin.x;
	}

	if (begin.y<end.y)
	{
		miny = begin.y;
		maxy = end.y;
	} else
	{
		miny = end.y;
		maxy = begin.y;
	}

	if( x > maxx || x < minx || y > maxy || y < miny ) return false;
	if(strict)
		if( x >= maxx || x <= minx || y >= maxy || y <= miny ) return false;

	return LineFun::isInLineBres4( begin, end, what, strict);
}

bool LineFun::isInLineBres4( const YPointInt& begin, const YPointInt& end, const YPointInt& p, bool strict)
{
	if( strict && (begin == p || end == p) ) return false;
	int x1 = begin.x;
	int y1 = begin.y;
	int x2 = end.x;
	int y2 = end.y;                  /* endpoints */
	int d,dx,dy;
	int Aincr,Bincr,yincr;
	int x,y;
	bool flag = false;
	bool xincf;
	bool yincf;

	dx = abs( x2-x1 );                       /* initialize constants */
	dy = abs( y2-y1 );

	if( dx >= dy )
	{
		if (x1 > x2)                        /* force x1 < x2 */
		{
			swap( &x1, &x2 );
			swap( &y1, &y2 );
		}

		d = 2 * dy - dx;

		if (y2 > y1)                        /* determine increment for y */
		yincr = 1;
		else
			yincr = -1;

		Aincr = 2 * (dy - dx);
		Bincr = 2 * dy;

		x = x1;                             /* initial x and y */
		y = y1;

		if( p.x == x && p.y == y ) return true;

		for (x=x1+1; x<=x2; x++)            /* do from x1+1 to x2 */
		{
			if (d >= 0)
			{
				y += yincr;                     /* set pixel A */
				d += Aincr;
				yincf = true;
			}
			else                              /* set pixel B */
			{
				d += Bincr;
				yincf = false;
			}

			if( (p.x == x && p.y == y) || (yincf && p.x == x && p.y == y - yincr) )
			{
				flag = true;
				break;
			}
		}
	} else
	{
		if (y1 > y2)                        /* force x1 < x2 */
		{
			swap( &x1, &x2 );
			swap( &y1, &y2 );
		}

		d = 2 * dx - dy;

		if (x2 > x1)                        /* determine increment for y */
		yincr = 1;
		else
			yincr = -1;

		Aincr = 2 * (dx - dy);
		Bincr = 2 * dx;

		x = x1;                             /* initial x and y */
		y = y1;

		if( p.x == x && p.y == y ) return true;

		for (y=y1+1; y<=y2; y++)            /* do from x1+1 to x2 */
		{
			if (d >= 0)
			{
				x += yincr;                     /* set pixel A */
				d += Aincr;
				xincf = true;
			}
			else                              /* set pixel B */
			{
				d += Bincr;
				xincf = false;
			}

			if( (p.x == x && p.y == y) || (xincf && p.x == x - yincr && p.y == y) )
			{
				flag = true;
				break;
			}
		}
	}
	
	return flag;
}

YPointInt LineFun::getIntersectionPoint( const YPointInt& begin1, const YPointInt& end1, const YPointInt& begin2, const YPointInt& end2)
{
	int x11 = begin1.x;
	int y11 = begin1.y;
	int x21 = end1.x;
	int y21 = end1.y;

	int x12 = begin2.x;
	int y12 = begin2.y;
	int x22 = end2.x;
	int y22 = end2.y;
	
	int x10;
	int y10;
	int x20;
	int y20;

	div_t resx1;
	div_t resy1;
	div_t resx2;
	div_t resy2;

	bool match = false;

	while( abs(x11-x21) >= 100 || abs(y11-y21) >= 100 )
	{
		if( abs(x21-x11) >= abs(y21-y11) )
		{
			resx1 = div( x21-x11, 2);
			x10 = x11 + resx1.quot;
			y10 = LineFun::getIntersectionPointBres4Y( YPointInt(x11,y11), YPointInt(x21,y21), x10);
		}
		else
		{
			resy1 = div( y21-y11, 2);
			y10 = y11 + resy1.quot;
			x10 = LineFun::getIntersectionPointBres4X( YPointInt(x11,y11), YPointInt(x21,y21), y10);
		}

		if( LineFun::isInLineBres4( begin2, end2, YPointInt(x10,y10)) )
		{
			match = true;
			break;
		}

		if(LineFun::areIntersecting( begin2, end2, YPointInt(x11, y11), YPointInt(x10,y10)))
		{
			x21 = x10;
			y21 = y10;
		} else
		{
			x11 = x10;
			y11 = y10;
		}

	}

	if(match) return YPointInt(x10,y10);

	while( abs(x12-x22) >= 100 || abs(y12-y22) >= 100 )
	{
		if( abs(x22-x12) >= abs(y22-y12) )
		{
			resx2 = div( x22-x12, 2);
			x20 = x12 + resx2.quot;
			y20 = LineFun::getIntersectionPointBres4Y( YPointInt(x12,y12), YPointInt(x22,y22), x20);
		}
		else
		{
			resy2 = div( y22-y12, 2);
			y20 = y12 + resy2.quot;
			x20 = LineFun::getIntersectionPointBres4X( YPointInt(x12,y12), YPointInt(x22,y22), y20);
		}

		if( LineFun::isInLineBres4( YPointInt(x11,y11), YPointInt(x21,y21), YPointInt(x20,y20)) )
		{
			match = true;
			break;
		}

		if(LineFun::areIntersecting( YPointInt(x11,y11), YPointInt(x21,y21), YPointInt(x12, y12), YPointInt(x20,y20)))
		{
			x22 = x20;
			y22 = y20;
		} else
		{
			x12 = x20;
			y12 = y20;
		}
	}

	if(match) return YPointInt(x20,y20);

	return LineFun::getIntersectionPointBres4( YPointInt(x11,y11), YPointInt(x21,y21), YPointInt(x12,y12), YPointInt(x22,y22));
}

YPointInt LineFun::getIntersectionPointBres4( const YPointInt& begin1, const YPointInt& end1, const YPointInt& begin2, const YPointInt& end2)
{
	int x1 = begin1.x;
	int y1 = begin1.y;
	int x2 = end1.x;
	int y2 = end1.y;                  /* endpoints */
	int d,dx,dy;
	int Aincr,Bincr,yincr;
	int x,y;

	dx = abs( x2-x1 );                       /* initialize constants */
	dy = abs( y2-y1 );

	if( dx >= dy )
	{

		if (x1 > x2)                        /* force x1 < x2 */
		{
			swap( &x1, &x2 );
			swap( &y1, &y2 );
		}

		d = 2 * dy - dx;

		if (y2 > y1)                        /* determine increment for y */
		yincr = 1;
		else
			yincr = -1;

		Aincr = 2 * (dy - dx);
		Bincr = 2 * dy;

		x = x1;                             /* initial x and y */
		y = y1;

		if( LineFun::isInLineBres4( begin2, end2, YPointInt(x,y)) )
			return YPointInt(x,y);

		for (x=x1+1; x<=x2; x++)            /* do from x1+1 to x2 */
		{
			if (d >= 0)
			{
				y += yincr;                     /* set pixel A */
				d += Aincr;
			}
			else                              /* set pixel B */
				d += Bincr;

			if( LineFun::isInLineBres4( begin2, end2, YPointInt(x,y)) )
				break;
		}
	} else
	{
		if (y1 > y2)                        /* force x1 < x2 */
		{
			swap( &x1, &x2 );
			swap( &y1, &y2 );
		}

		d = 2 * dx - dy;

		if (x2 > x1)                        /* determine increment for y */
		yincr = 1;
		else
			yincr = -1;

		Aincr = 2 * (dx - dy);
		Bincr = 2 * dx;

		x = x1;                             /* initial x and y */
		y = y1;

		if( LineFun::isInLineBres4( begin2, end2, YPointInt(x,y)) )
			return YPointInt(x,y);

		for (y=y1+1; y<=y2; y++)            /* do from x1+1 to x2 */
		{
			if (d >= 0)
			{
				x += yincr;                     /* set pixel A */
				d += Aincr;
			}
			else                              /* set pixel B */
				d += Bincr;

			if( LineFun::isInLineBres4( begin2, end2, YPointInt(x,y)) )
				break;
		}
	}
	return YPointInt(x,y);
}

bool LineFun::onSameSide( const YPointInt& from, const YPointInt& to, const YPointInt& p1, const YPointInt& p2)
{
	double a1, b1, c1;
	if( (from.y-to.y) == 0 )
	{
		b1 = 1;
		a1 = 0;
		c1 = -1*from.y;
	} else
	{
		a1 = 1;
		b1 = (double(to.x)-double(from.x))/(double(from.y)-double(to.y));
		c1 = -double(from.x) - b1*double(from.y);
	}
	if( (a1*double(p1.x)+b1*double(p1.y)+c1)*(a1*double(p2.x)+b1*double(p2.y)+c1) > 0)
		return true;
	else
		return false;
}

int LineFun::getIntersectionPointBres4X( const YPointInt& begin, const YPointInt& end, int Y)
{
	int x1 = begin.x;
	int y1 = begin.y;
	int x2 = end.x;
	int y2 = end.y;                  /* endpoints */
	int d,dx,dy;
	int Aincr,Bincr,yincr;
	int x,y;

	dx = abs( x2-x1 );                       /* initialize constants */
	dy = abs( y2-y1 );

	if( dy == 0 )
		if (y1 == Y )
			return x1;
		else
			return x2;

	if( dx >= dy )
	{
		if (x1 > x2)                        /* force x1 < x2 */
		{
			swap( &x1, &x2 );
			swap( &y1, &y2 );
		}

		d = 2 * dy - dx;

		if (y2 > y1)                        /* determine increment for y */
		yincr = 1;
		else
			yincr = -1;

		Aincr = 2 * (dy - dx);
		Bincr = 2 * dy;

		x = x1;                             /* initial x and y */
		y = y1;

		if( Y == y ) return x;

		for (x=x1+1; y!=Y; x++)            /* do from x1+1 to x2 */
		{
			if( x > x2 )
				return x2+1;
			if (d >= 0)
			{
				y += yincr;                     /* set pixel A */
				d += Aincr;
			}
			else                              /* set pixel B */
			{
				d += Bincr;
			}
		}
	} else
	{
		if (y1 > y2)                        /* force x1 < x2 */
		{
			swap( &x1, &x2 );
			swap( &y1, &y2 );
		}

		d = 2 * dx - dy;

		if (x2 > x1)                        /* determine increment for y */
		yincr = 1;
		else
			yincr = -1;

		Aincr = 2 * (dx - dy);
		Bincr = 2 * dx;

		x = x1;                             /* initial x and y */
		y = y1;

		if( Y == y ) return x;

		for (y=y1+1; y!=Y; y++)            /* do from x1+1 to x2 */
		{
			if( y > y2 )
				return x2+1;
			if (d >= 0)
			{
				x += yincr;                     /* set pixel A */
				d += Aincr;
			}
			else                              /* set pixel B */
			{
				d += Bincr;
			}
		}
	}
	return x;
}

int LineFun::getIntersectionPointBres4Y( const YPointInt& begin, const YPointInt& end, int X)
{
	int x1 = begin.x;
	int y1 = begin.y;
	int x2 = end.x;
	int y2 = end.y;                  /* endpoints */
	int d,dx,dy;
	int Aincr,Bincr,yincr;
	int x,y;

	dx = abs( x2-x1 );                       /* initialize constants */
	dy = abs( y2-y1 );

	if( dx == 0 )
		if (x1 == X )
			return y1;
		else
			return y2;

	if( dx >= dy )
	{
		if (x1 > x2)                        /* force x1 < x2 */
		{
			swap( &x1, &x2 );
			swap( &y1, &y2 );
		}

		d = 2 * dy - dx;

		if (y2 > y1)                        /* determine increment for y */
		yincr = 1;
		else
			yincr = -1;

		Aincr = 2 * (dy - dx);
		Bincr = 2 * dy;

		x = x1;                             /* initial x and y */
		y = y1;

		if( X == x ) return y;

		for (x=x1+1; x!=X; x++)            /* do from x1+1 to x2 */
		{
			if( x > x2 )
				return y2+1;
			if (d >= 0)
			{
				y += yincr;                     /* set pixel A */
				d += Aincr;
			}
			else                              /* set pixel B */
			{
				d += Bincr;
			}
		}
	} else
	{
		if (y1 > y2)                        /* force x1 < x2 */
		{
			swap( &x1, &x2 );
			swap( &y1, &y2 );
		}

		d = 2 * dx - dy;

		if (x2 > x1)                        /* determine increment for y */
		yincr = 1;
		else
			yincr = -1;

		Aincr = 2 * (dx - dy);
		Bincr = 2 * dx;

		x = x1;                             /* initial x and y */
		y = y1;

		if( X == x ) return y;

		for (y=y1+1; x!=X; y++)            /* do from x1+1 to x2 */
		{
			if( y > y2 )
				return y2+1;
			if (d >= 0)
			{
				x += yincr;                     /* set pixel A */
				d += Aincr;
			}
			else                              /* set pixel B */
			{
				d += Bincr;
			}

			if( X == x )
				break;
		}
	}
	return y;
}

static const double e = 0.00000000000001;

bool equalDouble( double n1, double n2)
{
	return ABS(n1-n2) < e;
}

bool PolyFun::isPointInPolygon(
	const YPoint<double>& point,
	const YPoint<double>* polygon, int n)
{
	int count = 0;
	for( int i = 0; i < n; i++)
	{
		if( equalDouble(polygon[i].x,point.x) && equalDouble(polygon[i].y,point.y) )
			return true;
		int j = (i + 1) % n;
		if( equalDouble(polygon[i].y,polygon[j].y) )
		{
			if( equalDouble(point.y,polygon[i].y)
			&& point.x < (MAX(polygon[i].x,polygon[j].x) - e) 
			&& point.x > (MIN(polygon[i].x,polygon[j].x) + e) )
				return true;
			continue;
		}
		if( (polygon[i].y - e) > point.y && (polygon[j].y - e) > point.y )
			continue;
		if( (polygon[i].y + e) < point.y  && (polygon[j].y + e) < point.y )
			continue;
		if( equalDouble(MAX(polygon[i].y,polygon[j].y),point.y) )
		{
			if( equalDouble(polygon[i].y,point.y) && (polygon[i].x - e) < point.x )
				count++;
			if( equalDouble(polygon[j].y,point.y) && (polygon[j].x - e) < point.x )
				count++;
			continue;
		}
		else
			if( equalDouble(MIN(polygon[i].y,polygon[j].y),point.y) )
				continue;
			else
			{
				double t = ( point.y - polygon[i].y ) / ( polygon[j].y - polygon[i].y );
				double xdbl = polygon[i].x + t*(polygon[j].x - polygon[i].x);
				if( (t + e) > 0 && (t - e) < 1 && (xdbl + e) > point.x )
				{
					if( equalDouble(xdbl,point.x) )
						return true;
					count++;
				}
			}
	}
	return count&1;
}

double distance( const double x1, const double y1, const double x2, const double y2)
{
	const double dx = x1 - x2;
	const double dy = y1 - y2;
	return ::sqrt(dx*dx + dy*dy);
}

double PointOnAxisProjectionLength( const double axis_begin_x, const double axis_begin_y,
										const double axis_end_x, const double axis_end_y,
										const double point_x, const double point_y)
{
	const double a = distance( axis_begin_x, axis_begin_y, point_x, point_y);
	const double b = distance( axis_end_x, axis_end_y, point_x, point_y);
	const double c = distance( axis_end_x, axis_end_y, axis_begin_x, axis_begin_y);
	if(a == 0) return 0;
	const double cos = (a*a + c*c - b*b)/(2*a*c);
	return a*cos;
}

double TriangleArea( 
	const YPoint<double>& a,
	const YPoint<double>& b,
	const YPoint<double>& c)
{
	return ABS(((b.x-a.x)*(c.y-a.y)-(c.x-a.x)*(b.y-a.y))/2);
}

double QuadrangleArea(
	const YPoint<double>& a,
	const YPoint<double>& b,
	const YPoint<double>& c,
	const YPoint<double>& d)
{
	return ::TriangleArea(a,b,c) + ::TriangleArea(a,c,d);
}

//static 
bool LineFun::areIntersecting( 
	const YPoint<double>& begin1,
	const YPoint<double>& end1,
	const YPoint<double>& begin2,
	const YPoint<double>& end2)
{
	double x11 = begin1.x;
	double y11 = begin1.y;
	double x12 = end1.x;
	double y12 = end1.y;

	double x21 = begin2.x;
	double y21 = begin2.y;
	double x22 = end2.x;
	double y22 = end2.y;
	
	double maxx1 = MAX(x11, x12), maxy1 = MAX(y11, y12);
	double minx1 = MIN(x11, x12), miny1 = MIN(y11, y12);
	double maxx2 = MAX(x21, x22), maxy2 = MAX(y21, y22);
	double minx2 = MIN(x21, x22), miny2 = MIN(y21, y22);

	if (minx1 > maxx2 || maxx1 < minx2 || miny1 > maxy2 || maxy1 < miny2)
	return false;  // ,      ...
	 

	double dx1 = x12-x11, dy1 = y12-y11; //       x  y
	double dx2 = x22-x21, dy2 = y22-y21; //       x  y
	double dxx = x11-x21, dyy = y11-y21;
	double div, mul;


	if ((div = (dy2*dx1-dx2*dy1)) == 0) 
		return false; //  ...
	if (div > 0)
	{
		if ((mul = (dx1*dyy-dy1*dxx)) < 0 || mul > div)
			return false; //      ...
		if ((mul = (dx2*dyy-dy2*dxx)) < 0 || mul > div)
			return false; //      ...
	}

	if ((mul = -(dx1*dyy-dy1*dxx)) < 0 || mul > -div)
	return false; //      ...
	if ((mul = -(dx2*dyy-dy2*dxx)) < 0 || mul > -div)
	return false; //      ...

	return true;
}



