#ifndef _YTIME_H
#define _YTIME_H
///////////////////////////////////////////////////////////////////////
#include "ytypes.h"
/** @defgroup TimeGroup    . 
 *  @ingroup GeneralGroup
 *  @{
 */
///////////////////////////////////////////////////////////////////////
/**
  * month = 1-12 , day_in_month =1-31 ,year - 0 is forbidden
  */
class YDtTm
{
public:
	static int MonthInYear();
	static bool IsLeapYear(int _year);
	static int  DaysInYear(int _year);
	static int32_t SecondsInYear(int _year);
	static int64_t MicroSecondsInYear(int _year);
	static int  DaysInMonth(int _year,int _month);
	static int  YearDay(int _year,int _month,int _day);
	static int  YearDayMonth(int _year_day,int _year);
	static int  YearDayDay(int _year_day,int _year);
	static void BCDdate2date(int _bcd,int& _year,int& _month,int& _day);// bcd like 19991231
	static int  BCDdate2date(int _bcd,int& _year_day);// returns year. bcd like 19991231
	static int  date2BCDdate(int _year,int _month,int _day);// returns bcd date like 19991231
	static int  date2BCDdate(int _year,int _year_day);// returns bcd date like 19991231
};
///////////////////////////////////////////////////////////////////////
class YTimeInterval
{
	friend class YDataTime;
	int64_t microsec;
public:
	YTimeInterval() { microsec=0; }
	YTimeInterval(int32_t mks) { microsec=mks; }
	YTimeInterval(int64_t mks) { microsec=mks; }
	YTimeInterval(int days,int hour,int minutes,double seconds) {Set(days,hour,minutes,seconds);}
	YTimeInterval& Set(int32_t mks) { microsec=mks; return *this;}
	YTimeInterval& Set(int64_t mks) { microsec=mks; return *this;}
	YTimeInterval& Set(int days,int hour,int minutes,double seconds);
	const YTimeInterval& Get(int& days,int& hour,int& minutes,double& seconds) const;
	void Get(double& seconds) const;
//	Next functions may use int as argument, because type cast to YTimeInterval occured due to constructor
	YTimeInterval  operator-() const { return YTimeInterval(-microsec); }
	YTimeInterval& operator=(const YTimeInterval& val) { microsec=val.microsec; return *this;}
	YTimeInterval& operator+=(const YTimeInterval& val) { microsec+=val.microsec; return *this;}
	YTimeInterval& operator-=(const YTimeInterval& val) { microsec-=val.microsec; return *this;}
	YTimeInterval& operator*=(const YTimeInterval& val) { microsec*=val.microsec; return *this;}
	YTimeInterval& operator/=(const YTimeInterval& val) { microsec/=val.microsec; return *this;}
	YTimeInterval  operator+(const YTimeInterval& val) const { YTimeInterval ti=*this; ti+=val; return ti; }
	YTimeInterval  operator-(const YTimeInterval& val) const { YTimeInterval ti=*this; ti-=val; return ti; }
	YTimeInterval  operator*(const YTimeInterval& val) const { YTimeInterval ti=*this; ti*=val; return ti; }
	YTimeInterval  operator/(const YTimeInterval& val) const { YTimeInterval ti=*this; ti/=val; return ti; }
	bool operator<(const YTimeInterval& ti)  const { return (microsec<ti.microsec); }
	bool operator>(const YTimeInterval& ti)  const { return (microsec>ti.microsec); }
	bool operator<=(const YTimeInterval& ti) const { return (microsec<=ti.microsec); }
	bool operator>=(const YTimeInterval& ti) const { return (microsec>=ti.microsec); }
	bool operator==(const YTimeInterval& ti) const { return (microsec==ti.microsec); }
	bool operator!=(const YTimeInterval& ti) const { return (microsec!=ti.microsec); }
};
///////////////////////////////////////////////////////////////////////
class YDataTime
{
	int16_t year;     //sign value, negative is before Crist, 0-year is forbidden,nessesary for leap year monitoring.
	int64_t microsec; // time in microseconds from NewYear
public:
	YDataTime() { year=0; microsec=0; }
	YDataTime(int _year,int _month,int _day,int _hour,int _minutes,double _seconds) { Set(_year,_month,_day,_hour,_minutes,_seconds); }
	void Set(int _year,int _month,int _day,int _hour,int _minutes,double _seconds);
	void Get(int& _year,int& _month,int& _day,int& _hour,int& _minutes,double& _seconds) const;
	bool IsOk() const { return year!=0; }
	YTimeInterval  operator-(const YDataTime& dt) const;
	YDataTime&  operator+=(const YTimeInterval& ti);
	YDataTime&  operator-=(const YTimeInterval& ti) {(*this)+=(-ti); return (*this);}
	YDataTime   operator+(const YTimeInterval& ti) const {YDataTime d=*this; d+=ti; return d; }
	bool operator<(const YDataTime& dt2)  const;
	bool operator<=(const YDataTime& dt2) const;
	bool operator>(const YDataTime& dt2)  const {	return dt2 <(*this);}
	bool operator>=(const YDataTime& dt2) const { return dt2<=(*this);}
	bool operator==(const YDataTime& dt2) const;
	bool operator!=(const YDataTime& dt2) const {return !(dt2==(*this));}
	void Between(const YDataTime& dt1,const YDataTime& dt2,int cnt1,int cnt2,int cnt);
	bool IsLeapYear() const;
	static int MonthInYear();
	int  DaysInMonth() const;
	int  DaysInYear() const;
	int32_t SecondsInYear() const;
	int64_t MicroSecondsInYear() const;
	int  Year() const;
	int  Month() const;
	int  Day() const;
	int  Hour() const;
	int  Minutes() const;
	double  Seconds() const;
	int  YearDay() const;
	double  YearDaySec() const; //returns yearday with fractional accuracy up to microseconds
	int  date2BCDdate() const;// returns bcd date like 19991231
};
class YDataTimeZone
{
	YDataTime dt;
	YTimeInterval zone;
public:
	YDataTimeZone(){} 
	YDataTimeZone(const YDataTime& dt_,const YTimeInterval& zone_):dt(dt_),zone(zone_) {} 
	YDataTimeZone(const YDataTimeZone& val) { Set(val.DataTime(),val.Zone()); }
	void Set(const YDataTimeZone& val) { Set(val.DataTime(),val.Zone()); }
	void Set(const YDataTime& dt_,const YTimeInterval& zone_)
	{
		dt=dt_;
		zone=zone_;
	} 
	YDataTime DataTime() const { return dt; }
	YTimeInterval Zone() const {return zone;}
	YDataTime Grinvich() const {YDataTime t=dt; t-=zone; return t;}
};

/** @} */ // end of TimeGroup  
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
#endif //_YTIME_H
