#ifndef _YWININI_H
#define _YWININI_H
////////////////////////////////////////////////////////////////////////////////////////
// file: 
// #include "ywinini.h"
////////////////////////////////////////////////////////////////////////////////////////
#include <string>
#include <iostream>
#include "datastring.h"
////////////////////////////////////////////////////////////////////////////////////////
/*! @addtogroup XMLandINI_DoxGroup   XML  INI    
 *  @{
 */
//------------------Interfaces to use-------------------------------------------
class WiNode;
class WinIniFile;
//------------------Interfaces to use-------------------------------------------
class WinIniFile
{
	std::string fname;
public:
	WinIniFile() {}
	WinIniFile(std::string filename);
	void Load(std::string filename);//connect to new (another) ini file
	std::string GetFName() const  {return fname;}
	WiNode Root() const;
};
//------------------Interfaces to use-------------------------------------------
class WiNode 
{
	std::string fname;
	std::string name;
private:
	class AttrRef; 
	friend class WinIniFile;
	WiNode(const std::string filename,const std::string node) : fname(filename),name(node) {}
public:
	WiNode() {}
	WiNode(const WiNode& n) : fname(n.fname), name(n.name) {}
	WiNode& operator=(const WiNode& n) { fname=n.fname; name=n.name; return *this;}
/*! @addtogroup Common_XML_INI_DoxGroup  
 *  @ingroup XMLandINI_DoxGroup
 *   /       -  ,
 *   -    . \n
 *  INI -       /,\n
 *  XML -    GetNode()        -  ,\n
 *   -    ,        -  
 *     ,     "INI and XML common interface" !!! 
 *     XML,      
 *    XML.
 *  @{
 */
    /** @defgroup BaseFunctionsINI_DoxGroup   INI
      *      
      *   @see INIAlternative_XML_INI_DoxGroup,    
      *   
      * @{ 
      */
	bool IsNode() const { return !name.empty();} ///<     NULL.
	WiNode GetNode(const std::string node_name);
	const WiNode GetNode(const std::string node_name) const;
	void SetValue(const DataString data);
	DataString GetValue() const;
	void SetAttr(const std::string attr_name,const DataString attr_value);
	DataString GetAttr(const std::string attr_name) const;
    /**@} */ 
    /** @defgroup INIAlternative_XML_INI_DoxGroup INI: Alternative INI and XML common interface
      *  INI and XML common interface,    .
      * @{ 
      */
	WiNode  operator()(const std::string node_name) {return GetNode(node_name);}
	const WiNode  operator()(const std::string node_name) const {return GetNode(node_name);}
	AttrRef operator[](const std::string& attr);
	const AttrRef operator[](const std::string& attr) const;
	WiNode& operator=(const DataString data) { SetValue(data); return *this; }
	operator DataString() const { return GetValue();}
    /**@} */ 
/** @} */ // end of Common_XML_INI
	friend std::ostream& operator<<(std::ostream& s,WiNode& in) { return s<<in.operator DataString();}
};
//------------------Implementation, used indirectly via WiNode operator[] ------
class WiNode::AttrRef 
{
	friend class WiNode; 
	WiNode n;
	std::string attrib;
private:
	AttrRef() {}
	AttrRef(const AttrRef& ar) 
		:n(ar.n),attrib(ar.attrib) {} //Normally should be forbidden< but required by VC7.0
	AttrRef operator=(const AttrRef&);
	AttrRef(WiNode node,const std::string& attr) 
		:n(node),attrib(attr) {}
public:
//	operator std::string() const { return n.GetAttr(attrib);}
	operator DataString() const { return n.GetAttr(attrib);}
	void operator=(const DataString& s) { n.SetAttr(attrib,s);} 
	friend std::ostream& operator<<(std::ostream& s,WiNode::AttrRef& ir) { return s<<ir.operator DataString();}
};
//------------------Interfaces to use-------------------------------------------
inline const WiNode::AttrRef WiNode::operator[](const std::string& attr) const { return WiNode::AttrRef(*this,attr); }
//inline const DataString WiNode::operator[](const std::string& attr) const { return DataString(WiNode::AttrRef(*this,attr));}
inline WiNode::AttrRef WiNode::operator[](const std::string& attr)  { return WiNode::AttrRef(*this,attr); }
/** @} */ // end of XMLandINI_DoxGroup
////////////////////////////////////////////////////////////////////////////////////////
#endif //_YWININI_H
