#ifndef _VTL_META_PERMUTATIONS_H 
#define _VTL_META_PERMUTATIONS_H 
///////////////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include "vtl_meta_Tools.h" 
#include "vtl_meta_IndexList.h" 
///////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
namespace Meta
{
	///      n .   \ref Lit_AlgTricks, . 76.
	template<unsigned n> 
	struct BitsIn
	{
		enum { temp=(n/2)+BitsIn<(n/2)>::temp };
		enum { value=n-unsigned(temp) };
	};
	template<> 
	struct BitsIn<0>
	{
		enum { temp=0 };
		enum { value=0 };
	};



	namespace Impl
	{
		template<int N,class Il1,class Il2,int j>
		struct determinant
		{
			typedef typename IL::EraseAt<0,Il1>::Result IndexInCol;
			typedef typename IL::EraseAt<j,Il2>::Result IndexInRow;
			template<class Matrix>
			static typename Matrix::value_type inRow(const Matrix& mat)
			{
				return MultMinusOnePowerTo<j>(
					  mat(IL::Elem<0,Il1>::value,IL::Elem<j,Il2>::value)
					* determinant<N-1,IndexInCol,IndexInRow,0>::inRow(mat) ) 
					+ determinant<N,Il1,Il2,j+1>::inRow(mat) ;
			}
		};
		template<int N,class Il1,class Il2>
		struct determinant<N,Il1,Il2,N >
		{
			template<class Matrix>
			static typename Matrix::value_type  inRow(const Matrix& mat)
			{
				return 0;
			}
		};
		template<class Il1,class Il2>
		struct determinant<1,Il1,Il2,0>
		{
			template<class Matrix>
			static typename Matrix::value_type inRow(const Matrix& mat)
			{
				return mat(IL::Elem<0,Il1>::value,IL::Elem<0,Il2>::value);
			}
		};

	}//namespace Impl
	///   ,     .    
	///      ()   .  RUNTIME    
	///   , ..       i,j -  ,
	///   i,j  ( 3,1 , 0,2  ..)      .\n
	///  ,    22,       ,      
	/// \code
	/// result=a[0]*a[3]-a[1]*a[2];
	/// \endcode
	///         ! \n 
	///       ,     NxN    N! /
	///            ,    ,
	///     .        -    .
	/// .   ForEachPermutation     det.
	template<int N,class Matrix>
	typename Matrix::value_type determinant(const Matrix & matrix)
	{
		typedef IL::MakeInRange<0,N>::Result Il;
		return Impl::determinant<N,Il,Il,0>::inRow(matrix);

	};
	///
	namespace Impl
	{
		///      ( !).  ,   
		///           f. 
		///    -       .\n
		///  - .  ForEachPermutation.         . 
		///      ,        ForEachPermutation,
		///         .
		template<class IList1,class IList2,int isOdd,int k,int i> 
		struct permut
		{
			enum { value=IL::Elem<i,IList2>::value };
			typedef typename IL::EraseAt<i,IList2>::Result Tail;
			typedef typename IL::PushBack<value,IList1>::Result Head;
			enum { is_Odd = (1&(isOdd+i)) };

			template<class Func>
			static void forEach(Func& f)
			{
				permut<IList1,IList2,isOdd,IL::Length<IList2>::value,i+1>::forEach(f);
				permut<Head,Tail,is_Odd,IL::Length<Tail>::value,0>::forEach(f);
			}
		};
		template<class IList1,class IList2,int isOdd,int k> 
		struct permut<IList1,IList2,isOdd,k,k > 
		{
			template<class Func>
			static void forEach(Func& f) {}
		};
		template<class IList1,int isOdd> 
		struct permut<IList1,Loki::NullType,isOdd,0,0> 
		{
			template<class Func>
			static void forEach(Func& f)
			{ 
				f.f<isOdd,IList1>(); 
			}
		};
	}//namespace Impl

	///         IndexList ,   Func.
	///    :
	/// \code
	/// class MyFunctorName
	/// {
	/// public:
	///   template<int isOdd,class IList>
	///   void f()
	///   {
	///     //  ,        
	///     std::cout << (isOdd?"- ":"+ ") << IL::type2ostream<IList>() << std::endl;
	///   }
	/// };
	/// \endcode
	///  f()           
	///   IndexList.\n
	///   ,       
	///    . . printPermutations, det , det_poly_coef \n
	///        0,1,...,N-1     
	///   - .    ForEachPermutation
	template<class IList,class Func>
	void ForEachPermutationOfIndexList(Func& f)
	{
		Impl::permut<Loki::NullType,IList,0,IL::Length<IList>::value,0>::forEach(f);
	}
	///      0,1,...,N-1 ,   Func.
	///    :
	/// \code
	/// class MyFunctorName
	/// {
	/// public:
	///   template<int isOdd,class IList>
	///   void f()
	///   {
	///     //  ,        
	///     std::cout << (isOdd?"- ":"+ ") << IL::type2ostream<IList>() << std::endl;
	///   }
	/// };
	/// \endcode
	///   ,       
	///    . . printPermutations, det , det_poly_coef
	/// 
	template<int N,class Func>
	void ForEachPermutation(Func& f)
	{
		typedef IL::MakeInRange<0,N>::Result IList;
		//ForEachPermutationOfIndexList<IList>(f); //Next line is another way to say the same except at this line compiler calculates N as Length of IList
		Impl::permut<Loki::NullType,IList,0,N,0>::forEach(f);
	}
	///       
	///   ForEachPermutationOfIndexList, ForEachPermutationOfIndexList, Impl::permut.
	struct printPermutations
	{
		template<int isOdd,class IList>
		void f()
		{
			std::cout << (isOdd?"- ":"+ ") << IL::type2ostream<IList>() << std::endl;
		}
	};
	///   ForEachPermutation .  ,     .    
	///             ()  . 
	///  RUNTIME      , ..       i,j -  ,
	///   i,j  ( 3,1 , 0,2  ..)      .\n
	///  ,    22,       ,      
	/// \code
	/// result=a[0]*a[3]-a[1]*a[2];
	/// \endcode
	///         ! \n 
	///       ,     NxN    N! /
	///            ,    ,
	///     .        -    .
	/// .   determinant .
	template<class Matrix>
	class det
	{
		typedef typename Matrix::value_type value_type;
		const Matrix& A;

		template<class Il,int i>
		value_type a()
		{
			return A(i,IL::Elem<i,Il>::value);
		}
		template<class Il,int i> 
		value_type mul(Loki::Int2Type<i>  )
		{
			return a<Il,i>()*mul<Il,i-1>(Loki::Int2Type<i-1>());
		}
		template<class Il,int i> 
		value_type mul(Loki::Int2Type<0> )
		{ 
			return a<Il,i>(); 
		}
	public:
		value_type result;
		det(const Matrix& mat) : A(mat) , result(0) {}
		template<int isOdd,class IList>
		void f()
		{
			enum {last=IL::Length<IList>::value-1};
			AddOrSub<isOdd>(result,mul<IList,last>(Loki::Int2Type<last>()) );
		}
	};
	///    A  B  NxN.    Meta::ForEachPermutation   
	/// det(Ax+B),  x -  .     coef[N+1].  :\n
	/// det(Ax+B) = coef[0] + coef[1] x + coef[2] x^2 + coef[3] x^3 + ... + coef[N] x^N
	template<class Matrix,int N>
	class det_poly_coef
	{
		typedef typename Matrix::value_type T;
		const Matrix& A;
		const Matrix& B;
		enum { mask = (1<<N)-1 };

		template<class Il,int i>
		T ab(Loki::Int2Type<0> )
		{ 
			return A(i,IL::Elem<i,Il>::value);
		}
		template<class Il,int i>
		T ab(Loki::Int2Type<1> )
		{ 
			return B(i,IL::Elem<i,Il>::value); 
		}
		template<class Il,int i,int is_b>
		T ab()
		{ 
			return ab<Il,i>(Loki::Int2Type<is_b>()); 
		}

		template<class Il,int i,int m> 
		T mul(Loki::Int2Type<i> )
		{
			enum { m2 = m/2 };
			return ab<Il,i,m&1>() * mul<Il,(i-1),m2>( Loki::Int2Type<(i-1)>() );
		}
		template<class Il,int i,int m> 
		T mul(Loki::Int2Type<0> )
		{ 
			return ab<Il,i,m&1>(); 
		}
		template<int m,int isOdd,class IList>
		void one_comp()
		{
			enum {last=IL::Length<IList>::value-1};
			AddOrSub<isOdd>(coef[BitsIn<m>::value] ,mul<IList,last,m>(Loki::Int2Type<last>()) );
		}
		template<int m,int isOdd,class IList>
		void comp(Loki::Int2Type<1> )
		{
			one_comp<m,isOdd,IList>();
			comp<m-1,isOdd,IList>(Loki::Int2Type<((m-1)?1:0)>());
		}
		template<int m,int isOdd,class IList>
		void comp(Loki::Int2Type<0> )
		{
			one_comp<m,isOdd,IList>();
		}
	public:
		T coef[N+1];//!<    NN,   N-  N+1 . 
		det_poly_coef(const Matrix& A_,const Matrix& B_): A(A_), B(B_) 
		{ 
			set<N+1>(coef,0); 
		}
		template<int isOdd,class IList>
		void f()
		{
			comp<mask,isOdd,IList>(Loki::Int2Type<1>());
		}
	};
}//namespace Meta
}//namespace Vtl
///////////////////////////////////////////////////////////////////////////////////////////
#endif //_VTL_META_PERMUTATIONS_H 
