#ifndef _VTL_META_TOOLS_H 
#define _VTL_META_TOOLS_H
///////////////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include "loki_EmptyType.h" 
#include "loki_NullType.h" 
#include "loki_typeManip.h"
#include "loki_Typelist.h" //currently not used but included for potential user usage
#include "loki_static_check.h"
///////////////////////////////////////////////////////////////////////////////////////////
namespace Loki
{
	/// Analogous to Loki::Int2Type but for <b>boolean</b> type. Used to simplify code: instead using Loki::Int2Type<((a>b)?1:0)>() now use Loki::Bool2Type<(a>b)>()
	template <bool v>
	struct Bool2Type
	{
		enum { value = v };
	};
};
namespace Vtl
{
namespace Meta
{
	///   enum {} :   ,   ,   .
	class EnumTraits
	{
		template<bool next=true,unsigned i=1>
		struct MaxEnum;
		template<unsigned i>
		struct MaxEnum<false,i>
		{
		public:
			enum { value = i };
		};
		template<unsigned i>
		struct MaxEnum<true,i>
		{
		private:
			enum { cur=i}; //!< nesesary because i is unsigned but comparison in next line should be preformed on enum, not for unsigned!
			enum { next = ((cur<<1)|cur) };
		public:
			enum { value = MaxEnum<(next>cur),((next>cur)?next:cur)>::value};
		};
		struct IsEnum
		{
			enum {a=-1};
			enum {b=0};
			enum {Signed=(a<b)? 1:0};
		};
		template<unsigned mask=1>
		struct Bits
		{
			enum { value = 1+Bits<(mask<<1)>::value};
		};
		template<>
		struct Bits<0>
		{
			enum { value = 0 };
		};
	public:
		enum { max=MaxEnum<>::value};
		enum { isSigned=IsEnum::Signed};
		enum { min=isSigned? (((max<<1)|1)^max) : 0};
		enum { bits=Bits<>::value};
	};
	///
	///  sz-    T.
	/// \code
	/// T array[sz];
	/// \endcode
	///     size2 ,  size2 >= sz      align_to .
	///
	template<typename T>
	static int align(int sz,int align_to)
	{
		while((sz*sizeof(T))%align_to)
			++sz;
		return sz;
	}
	///
	///   ,    int align(int sz,int align_to),   sz  align_to    
	///      ,    runtime.
	///  -   value. 
	/// \code
	/// enum { size=11};
	/// enum { new_size = Align<float,size,8>::value };
	/// \endcode
	///
	template<typename T,int sz,int align_to>
	class Align
	{
		template<int next,int size>
		struct Align_
		{
			enum { sz2=size+1};
			enum { is_next = ((sz2*sizeof(T))%align_to)  };
			enum { value = Align_<is_next,sz2>::value  };
		};
		template<int size>
		struct Align_<0,size>
		{
			enum { value = size };
		};
		enum { is_next = ((sz*sizeof(T))%align_to)  };
	public:
		enum { value = Align_<is_next,sz>::value  };
	};
	///  result+=value;  isSubtract==0,  result-=value;
	template<int isSubtract,typename T>
	void AddOrSub(T& result,T value)
	{
		AddOrSub(result,value,Loki::Int2Type<(isSubtract?1:0)>() );
	}
	template<typename T>
	void AddOrSub(T& result,T value,Loki::Int2Type<1> )
	{ 
		result-=value; 
	}
	template<typename T>
	void AddOrSub(T& result,T value,Loki::Int2Type<0> )
	{ 
		result+=value; 
	}
	///
	///  (-1):N * value,   value  N   -value  N  ;
	template<int N,typename T>
	T MultMinusOnePowerTo(T value)
	{
		return MultMinusOnePowerTo(value,Loki::Int2Type<(N&1)>());
	}
	template<typename T>
	T MultMinusOnePowerTo(T value,Loki::Int2Type<0> )
	{
		return value;
	}
	template<typename T>
	T MultMinusOnePowerTo(T value,Loki::Int2Type<1> )
	{
		return -value;
	}
	///        ( ,   ).
	template<int N,typename T>
	void copy(T* dst,const T* src)
	{
		copy(dst,src,Loki::Int2Type<N>());
	}
	template<int N,typename T>
	void copy(T* dst,const T* src,Loki::Int2Type<N>)
	{
		*dst=*src;
		copy(dst+1,src+1,Loki::Int2Type<N-1>());
	}
	template<typename T>
	void copy(T* dst,const T* src,Loki::Int2Type<1>)
	{
		*dst=*src;
	}
	template<typename T>
	void copy(T* dst,const T* src,Loki::Int2Type<0>)
	{
	}
	///           value ( ,   ).
	template<int N,typename T,typename TT>
	void set(T* dst,TT value)
	{ 
		set(dst,T(value),Loki::Int2Type<N>()); 
	}
	template<typename T,int N>
	void set(T* dst,T value,Loki::Int2Type<N>)
	{
		*dst=value;
		set(dst+1,value,Loki::Int2Type<N-1>());
	}
	///     ,     -  1  
	template<typename T>
	void set(T* dst,T value,Loki::Int2Type<1>)
	{
		*dst=value;
	}
	///    ,   N=0   -   , 
	///      .      0 -  
	///            ,    
	///  ,      .
	template<typename T>
	void set(T* dst,T value,Loki::Int2Type<0>) 
	{
	}
	///         
	///    v,v+s,v+s,v+s,... .      
	///   .
	template<int N,typename T,typename TT>
	void set_seq(T* dst,TT value_first=0,TT value_step=1)
	{
		set_seq(dst,T(value_first),T(value_step),Loki::Int2Type<N>());
	}
	template<typename T,int N>
	void set_seq(T* dst,T value,T step,Loki::Int2Type<N>)
	{
		*dst=value;
		set_seq(dst+1,value+step,step,Loki::Int2Type<N-1>());
	}
	template<typename T>
	void set_seq(T* dst,T value,T step,Loki::Int2Type<1>)
	{
		*dst=value;
	}
	template<typename T>
	void set_seq(T* dst,T value,T step,Loki::Int2Type<0>)
	{
	}
	///           value ( ,   ).
	template<int N,typename TT,typename T>
	TT sum(T* dst)
	{ 
		return sum(dst,Loki::Int2Type<N>()); 
	}
	template<typename TT,int N,typename T>
	TT sum(T* dst,Loki::Int2Type<N>)
	{
		return *dst+sum<TT>(dst+1,Loki::Int2Type<N-1>());
	}
	///     ,     -  1  
	template<typename TT,typename T>
	TT sum(T* dst,Loki::Int2Type<1>)
	{
		return *dst;
	}
	///    ,   N=0   -   , 
	///      .      0 -  
	///            ,    
	///  ,      .
	template<typename TT,typename T>
	TT sum(T* dst,Loki::Int2Type<0>) 
	{
		return TT(0);
	}
	///  .   RUNTIME        
	///    . 
	/// \param src -    , 
	/// 
	/// \param name - ,     .
	/// 
	/// \param inColumn -  true - g   ,  -   . 
	/// 
	/// N -  ,    , :
	/// \code
	/// int x[3]={12,33,73}; 
	/// print_array<3>(x,"My array 'x' content:");
	/// \endcode
	///  ( )  :
	/// \code
	/// const int N=3; //,  ,    
	/// //enum {N=3};
	///
	/// int x[N];
	/// //...   ...
	/// print_array<N>(x,"My array 'x' content:");
	/// \endcode

	template<int N,typename T>
	void print_array(const T* src,const char *name,bool inColumn=true)
	{
		std::cout << name <<" ["<<N<<"] : {" ;
		if(inColumn)
		{
			std::cout << std::endl;
			for(int i=0; i<N; ++i)
				std::cout << i << ":  " << src[i] << std::endl;
			std::cout << "}" <<std::endl;
		}
		else
		{
			for(int i=0; i<N; ++i)
				std::cout << "  " << src[i] ;
			std::cout << " }" <<std::endl;
		}
	}
}//namespace Meta
}//namespace Vtl
#endif //_VTL_META_TOOLS_H
