////////////////////////////////////////////////////////////////////////////////////////
// MSliEqlz.cpp - SlidingEqualization filter class source file                        // 
// by Mintchenkov M.V.                                                                //
////////////////////////////////////////////////////////////////////////////////////////
#include <stdlib.h>
#include <memory.h>
//#include <stdio.h> 
#include <sstream>
#include "aimage.h"
#include "mslieqlz.h"
//#include "statmpl.cpp"

///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
// templates used by class MSlidingEqualization ///////////////////
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////

template<class PIX> void tmpl_RightLeft(int Wd,int Or_frame_sz_x, int Frame_sz_y,int Down_sz_y, int Half_framW,
					  long* hist,  PIX* p, PIX* dst_pt,int src_shift,
					  int* max_cval,int* min_cval, const unsigned int N)
{//  

	int max=*max_cval;
	int min=*min_cval;
	int NumPoints=0;
	int cnt=0;
	int col=0;
	unsigned long s=0;
	unsigned int val=0;
	long* hist_col=0;
	int Frame_sz_x=2+Half_framW;// left_half_frame+center
	int corner_shift=src_shift*Down_sz_y-((Or_frame_sz_x-1)>>1);
	int sh=src_shift*Frame_sz_y-corner_shift;
	int to_move=Or_frame_sz_x>>1;
	while(to_move--)
	{
		NumPoints=Frame_sz_y*Frame_sz_x++;//1) num of points 2) incr size
		p+=corner_shift;//move pointer to button left
		cnt=Frame_sz_y;
		while(cnt--)
		{	// adding points to hist - moving pointer p to top left
			col=*p;
			p-=src_shift;
			if(col>max)
				max=col;
			if(col<min)
				min=col;
			hist[col]++;
		};	
		//calculate rang
		p+=sh;		
		cnt=*p--; //1) get color of current center point 
				  //2) go to next frame position on src image
		cnt-=min;//get count of color less than central one
		hist_col=hist+min;//      
		s=0;
		while(cnt--)
			s+=*hist_col++;
		s+=((*hist_col)>>1);
		val=(s*N+(NumPoints>>1))/NumPoints;
		if (val>=N)
			val=N-1;					
		*dst_pt--=val; // go to next frame position on dest image
	}
	to_move=Wd-Or_frame_sz_x;		
	Frame_sz_x--;
	
	sh=Frame_sz_x+corner_shift;
	while(to_move--)
	{
		p+=corner_shift;//move pointer to button left
		int cnt=Frame_sz_y;
		int col;
		while(cnt--)
		{// adding points to hist - moving pointer p to top left
			col=*p;
			p-=src_shift;
			if(col>max)
				max=col;
			if(col<min)
				min=col;
			hist[col]++;
		};
		cnt=Frame_sz_y;
		p+=Frame_sz_x;//move pointer to button right+1 (line adjacent to frame)
		while(cnt--)
		{// removing points from hist (line adjacent to frame)	
			p+=src_shift;//moving pointer to button right adjacent line
			col=*p;				
			hist[col]--;
		};
		p-=sh;
		//correct min/max values
		while(!hist[min]) min++;
		while(!hist[max]) max--;
		//calculate rang
		cnt=*p--; //go to next frame position on src image
		cnt-=min;
		hist_col=hist+min;//      
		s=0;
		while(cnt--)
			s+=*hist_col++;
		s+=(*hist_col)>>1;
		val=(s*N+(NumPoints>>1))/NumPoints;
		if (val>=N) 
			val=N-1;
		*dst_pt--=val; // go to next frame position on dest image
	}
	to_move=Half_framW;
	corner_shift=src_shift*Down_sz_y+(Or_frame_sz_x>>1)+1;
	sh=src_shift*Frame_sz_y-corner_shift;
	while(to_move--)
	{		
		Frame_sz_x--;
		NumPoints=Frame_sz_y*Frame_sz_x;
		cnt=Frame_sz_y;
		p+=corner_shift;//move pointer to button right corner (adj. line)
		while(cnt--)
		{// removing points from hist (line adjacent to frame)			
			col=*p;
			p-=src_shift;
			hist[col]--;
		};	
		//correct min/max values
		while(!hist[min]) min++;
		while(!hist[max]) max--;
		//calculate rang
		p+=sh;
		cnt=*p--; //go to next frame position on src image
		cnt-=min;
		hist_col=hist+min;//      
		s=0;
		while(cnt--)
			s+=*hist_col++;
		s+=(*hist_col)>>1;
		val=(s*N+(NumPoints>>1))/NumPoints;
		if (val>=N)	val=N-1;					
		*dst_pt--=val; // go to next frame position on dest image
	}
	*max_cval=max;
	*min_cval=min;
}

template<class PIX> void tmpl_LeftRight(int Wd,int Or_frame_sz_x, int Frame_sz_y,int Down_sz_y, int Half_framW,
					  long* hist,  PIX* p, PIX* dst_pt,int src_shift,
					  int* max_cval,int* min_cval, const unsigned int N)
{
	int max=*max_cval;
	int min=*min_cval;
	int NumPoints=0;
	int cnt=0;
	int col=0;
	unsigned long s=0;
	unsigned int val=0;
	long* hist_col=0;
	int to_move=Half_framW;
			
	int Frame_sz_x=(Or_frame_sz_x>>1)+2;//initial size = visible part of frame+1
	int corner_shift=src_shift*Down_sz_y+(Or_frame_sz_x>>1);
	int sh=Frame_sz_y*src_shift-corner_shift;
	
	while(to_move--)
	{		
		NumPoints=Frame_sz_y*Frame_sz_x++;//1) num of points 2) incr size
		p+=corner_shift;// move pointer to button right
		cnt=Frame_sz_y;
		while(cnt--)
		{	// adding points to hist - moving pointer p to top right
			col=*p;
			p-=src_shift;
			if(col>max)	max=col;
			if(col<min) min=col;
			hist[col]++;
		};	
		//calculate rang
		p+=sh;
		cnt=*p++; 
		cnt-=min;//get count of color less than central one
		hist_col=hist+min;//      
		s=0;
		while(cnt--)
			s+=*hist_col++;
		s+=(*hist_col)>>1;
		val=(s*N+(NumPoints>>1))/NumPoints;
		if (val>=N) val=N-1;					
		*dst_pt++=val; // go to next frame position on dest image
	}


	to_move=Wd-Or_frame_sz_x;		
	Frame_sz_x--;
	sh=Frame_sz_x-corner_shift;
	while(to_move--)
	{
		p+=corner_shift;//move pointer to button right
		int cnt=Frame_sz_y;
		int col;
		while(cnt--)
		{// adding points to hist - moving pointer p to top right
			col=*p;
			p-=src_shift;
			if(col>max)	max=col;
			if(col<min) min=col;
			hist[col]++;
		};
		cnt=Frame_sz_y;
		p-=Frame_sz_x;//move pointer to button left-1 (line adjacent to frame)
		while(cnt--)
		{// removing points from hist (line adjacent to frame)			
			p+=src_shift;//moving pointer to button left adjacent line
			col=*p;			
			hist[col]--;
		};
		//correct min/max values
		while(!hist[min]) min++;
		while(!hist[max]) max--;
		//calculate rang
		p+=sh;// move pointer to center of frame
		cnt=*p++; //go to next frame position on src image
		cnt-=min;
		hist_col=hist+min;//      
		s=0;
		while(cnt--)
			s+=*hist_col++;
		s+=(*hist_col)>>1;
		val=(s*N+(NumPoints>>1))/NumPoints;
		if (val>=N)	val=N-1;		
		*dst_pt++=val;		
	}

	to_move=Or_frame_sz_x>>1;
	corner_shift=src_shift*Down_sz_y-((Or_frame_sz_x-1)>>1)-1;//to button left corner (adj. line)
	sh=src_shift*Frame_sz_y-corner_shift;
	while(to_move--)
	{		
		Frame_sz_x--;
		NumPoints=Frame_sz_y*Frame_sz_x;
		cnt=Frame_sz_y;
		p+=corner_shift;//move pointer to button left corner (adj. line)
		while(cnt--)
		{// removing points from hist (line adjacent to frame)			
			col=*p;
			hist[col]--;
			p-=src_shift;
		};
		//correct min/max values
		while(!hist[min]) min++;
		while(!hist[max]) max--;
		//calculate rang		
		p+=sh;
		cnt=*p++; //go to next frame position on src image
		cnt-=min;
		hist_col=hist+min;//      
		s=0;
		while(cnt--)
			s+=*hist_col++;
		s+=(*hist_col)>>1;
		val=(s*N+(NumPoints>>1))/NumPoints;
		if (val>=N)	val=N-1;					
		*dst_pt++=val; // go to next frame position on dest image
	}
	*max_cval=max;
	*min_cval=min;	
}


template<class PIX> void tmpl_VertLeftMiddle( int frame_sz_x, int frame_sz_y, int y_down,long* hist,
				PIX* p, PIX* dst_pt, int src_shift, int* max_val,
				int* min_val, const unsigned int N)
				 //frame_sz_x - size of frame in x-direction
				 //frame_sz_y - size of frame in x-direction (depends on Y!!!!!)
				 //src_pt - pointer to initial position frame center			
				 //dst_pt - pointer to initial position frame center								
{
	int NumPoints=frame_sz_x*frame_sz_y;	
	p+=(y_down+1)*src_shift;//move pointer to button position (left==center_of_frame)
	int cnt=frame_sz_x;
	int col;
	while(cnt--)
	{// adding points to hist - moving pointer p to button right
		col=*p++;
		if(col>(*max_val))
			(*max_val)=col;
		if(col<(*min_val))
			(*min_val)=col;
		hist[col]++;
	};
	cnt=frame_sz_x;
    p-=(frame_sz_y)*src_shift;//move pointer to top right+1 (line adjacent to frame)
	while(cnt--)
	{// removing points from hist (line adjacent to frame)
		p--;
		col=*p;
		hist[col]--;
	};
	//correct min/max values
	while(!hist[(*min_val)]) (*min_val)++;
	while(!hist[(*max_val)]) (*max_val)--;
	//calculate rang
	p+=(frame_sz_y-y_down)*src_shift;//move to new center;
	cnt=*p;
	cnt-=(*min_val);
	long* hist_col=hist+(*min_val);//      
	long s=0;
	while(cnt--)
		s+=*hist_col++;
	s+=(*hist_col)>>1;
	unsigned int val=(s*N+(NumPoints>>1))/NumPoints;
	if (val>=N)
		val=N-1;					
	*dst_pt=val; 		
}

template<class PIX>  void tmpl_VertLeftUp( int frame_sz_x, int frame_sz_y, int y_down,long* hist,
				PIX* p, PIX* dst_pt, int src_shift, int* max_val,
				int* min_val, const unsigned int N)
				 //frame_sz_x - size of frame in x-direction
				 //frame_sz_y - size of frame in x-direction (depends on Y!!!!!)
				 //src_pt - pointer to initial position frame center			
				 //dst_pt - pointer to initial position frame center								
{
	int NumPoints=frame_sz_x*frame_sz_y;	
	p+=(y_down+1)*src_shift;//move pointer to button position (left==center_of_frame)
	int cnt=frame_sz_x;
	int col;
	while(cnt--)
	{// adding points to hist - moving pointer p to button right
		col=*p++;
		if(col>(*max_val))
			(*max_val)=col;
		if(col<(*min_val))
			(*min_val)=col;
		hist[col]++;
	};	
	p-=frame_sz_x;
	p-=y_down*src_shift;//move to center;
	//correct min/max values
	while(!hist[(*min_val)]) (*min_val)++;
	while(!hist[(*max_val)]) (*max_val)--;
	//calculate rang
	cnt=*p;
	cnt-=(*min_val);
	long* hist_col=hist+(*min_val);//      
	long s=0;
	while(cnt--)
		s+=*hist_col++;
	s+=(*hist_col)>>1;
	unsigned int val=(s*N+(NumPoints>>1))/NumPoints;
	if (val>=N) val=N-1;					
		*dst_pt=val; 		
}

template<class PIX>  void tmpl_VertLeftDown( int frame_sz_x, int frame_sz_y, int y_down,long* hist,
				PIX* p, PIX* dst_pt, int src_shift, int* max_val,
				int* min_val, const unsigned int N)
				 //frame_sz_x - size of frame in x-direction
				 //frame_sz_y - size of frame in x-direction (depends on Y!!!!!)
				 //src_pt - pointer to initial position frame center			
				 //dst_pt - pointer to initial position frame center								
				 //src_shift- next line shift for src image
{
	int NumPoints=frame_sz_x*frame_sz_y;	
	p-=(frame_sz_y-y_down)*src_shift;//move pointer to top position (left==center_of_frame)
	int cnt=frame_sz_x;
	int col;
    while(cnt--)
	{// removing points from hist (line adjacent to frame)
		col=*p++;
		hist[col]--;
	};
	p-=frame_sz_x;
	p+=(frame_sz_y-y_down+1)*src_shift;//move pointer to top position (left==center_of_frame)
	//correct min/max values
	while(!hist[(*min_val)]) (*min_val)++;
	while(!hist[(*max_val)]) (*max_val)--;
	//calculate rang
	cnt=*p;
	cnt-=(*min_val);
	long* hist_col=hist+(*min_val);//      
	long s=0;
	while(cnt--)
		s+=*hist_col++;
	s+=(*hist_col)>>1;
	unsigned int val=(s*N+(NumPoints>>1))/NumPoints;
	if (val>=N) val=N-1;					
		*dst_pt=val; 		
}


template<class PIX>  void tmpl_VertRightMiddle( int frame_sz_x, int frame_sz_y, int y_down,long* hist,
				PIX* p, PIX* dst_pt, int src_shift, int* max_val,
				int* min_val, const unsigned int N)
				 //frame_sz_x - size of frame in x-direction
				 //frame_sz_y - size of frame in x-direction (depends on Y!!!!!)
				 //src_pt - pointer to initial position frame center			
				 //dst_pt - pointer to initial position frame center								
{
	int NumPoints=frame_sz_x*frame_sz_y;	
	p+=(y_down+1)*src_shift;//move pointer to button position
	int cnt=frame_sz_x;
	int col;
	while(cnt--)
	{// adding points to hist - moving pointer p to top right
		col=*p--;
		if(col>(*max_val))
			(*max_val)=col;
		if(col<(*min_val))
			(*min_val)=col;
		hist[col]++;
	};
	cnt=frame_sz_x;

    p-=(frame_sz_y)*src_shift;//move pointer to top left-1 (line adjacent to frame)
	while(cnt--)
	{// removing points from hist (line adjacent to frame)
		col=*(++p);
		hist[col]--;
	};
	p+=(frame_sz_y-y_down)*src_shift;//move to center;

	//correct min/max values
	while(!hist[(*min_val)]) (*min_val)++;
	while(!hist[(*max_val)]) (*max_val)--;
	
	//calculate rang
	cnt=*p;
	cnt-=(*min_val);
	long* hist_col=hist+(*min_val);//      
	long s=0;
	while(cnt--)
		s+=*hist_col++;
	s+=(*hist_col)>>1;
	unsigned int val=(s*N+(NumPoints>>1))/NumPoints;
	if (val>=N) val=N-1;					
		*dst_pt=val; 		
}

template<class PIX>  void tmpl_VertRightDown( int frame_sz_x, int frame_sz_y, int y_down,long* hist,
				PIX* p, PIX* dst_pt, int src_shift, int* max_val,
				int* min_val, const unsigned int N)
				 //frame_sz_x - size of frame in x-direction
				 //frame_sz_y - size of frame in x-direction (depends on Y!!!!!)
				 //src_pt - pointer to initial position frame center			
				 //dst_pt - pointer to initial position frame center								
{
	int NumPoints=frame_sz_x*frame_sz_y;	
	p-=(frame_sz_y-y_down)*src_shift;//move pointer to top position
	int cnt=frame_sz_x;
	int col;
	while(cnt--)
	{// removing points from hist (line adjacent to frame)
		col=*p--;
		hist[col]--;
	};
	p+=frame_sz_x;
	p+=(frame_sz_y-y_down+1)*src_shift;//move to center;

	//correct min/max values
	while(!hist[(*min_val)]) (*min_val)++;
	while(!hist[(*max_val)]) (*max_val)--;
	
	//calculate rang
	cnt=*p;
	cnt-=(*min_val);
	long* hist_col=hist+(*min_val);//      
	long s=0;
	while(cnt--)
		s+=*hist_col++;
	s+=(*hist_col)>>1;
	unsigned int val=(s*N+(NumPoints>>1))/NumPoints;
	if (val>=N) val=N-1;					
		*dst_pt=val; 		
}

template<class PIX>  void tmpl_VertRightUp( int frame_sz_x, int frame_sz_y, int y_down,long* hist,
				PIX* p, PIX* dst_pt, int src_shift, int* max_val,
				int* min_val, const unsigned int N)
				 //frame_sz_x - size of frame in x-direction
				 //frame_sz_y - size of frame in x-direction (depends on Y!!!!!)
				 //src_pt - pointer to initial position frame center			
				 //dst_pt - pointer to initial position frame center								
{
	int NumPoints=frame_sz_x*frame_sz_y;	
	p+=(y_down+1)*src_shift;//move pointer to button position
	int cnt=frame_sz_x;
	int col;
	while(cnt--)
	{// adding points to hist - moving pointer p to button left
		col=*p--;		
		if(col>(*max_val))
			(*max_val)=col;
		else if(col<(*min_val))
			(*min_val)=col;
		hist[col]++;
	};
	p+=frame_sz_x;//return it to button right
	p-=y_down*src_shift;//return it to center

	//correct min/max values
	while(!hist[(*min_val)]) (*min_val)++;
	while(!hist[(*max_val)]) (*max_val)--;
	
	//calculate rang
	cnt=*p;
	cnt-=(*min_val);
	long* hist_col=hist+(*min_val);//      
	long s=0;
	while(cnt--)
		s+=*hist_col++;
	s+=(*hist_col)>>1;
	unsigned int val=(s*N+(NumPoints>>1))/NumPoints;
	if (val>=N) val=N-1;					
		*dst_pt=val; 		
}


template<class IMAGE, class PIXEL> bool tmpl_Run(const IMAGE* Src, IMAGE* Dst,
							   int Original_XFrmSz, int Original_YFrmSz,
							   	long *hist, ProgressCounter* count)
{
	const int H=Src->Height();
	const int W=Src->Width();
	const int Ncolor=1<<Src->Bits();
	const int src_shift=Src->NextLineShift();
	const int dst_shift=Dst->NextLineShift();
	::memset(hist,0, sizeof(hist[0])*Ncolor);

	if (count)
	{
		//char name[100];
		//::sprintf(name,"Image Equaliziation (%d bits per pixel image).", Src->Bits());
		std::ostringstream os;
		os << "Image Equaliziation ("<< Src->Bits() << " bits per pixel image)." << std::endl;
		count->start(os.str().c_str(),0,H-1);
	}

	int min_val=Ncolor-1;
	int max_val=0;

	PIXEL* src_center=(*Src)[0];
	PIXEL* dst_center=(*Dst)[0];

	const int half_framH=(Original_YFrmSz-1)>>1;//distance from center up top (not including center)
	const int half_framW=(Original_XFrmSz-1)>>1;
	int NumPoints=0;
	int Frame_sz_y=Original_YFrmSz-half_framH; //initital height of frame !!!
	int down_sz_y=Frame_sz_y-1;// distance between center and button lines
					//   3*3 - 


	///LINE N 0
	int Frame_sz_x=((Original_XFrmSz)>>1)+1;//initial size = visible part of frame+center 
	PIXEL* p=src_center;
	int col=0;
	int j;
	for (j=0; j<Frame_sz_y; j++)//first frame position
	{
		for (int x=0; x<Frame_sz_x; x++)
		{
			col=*(p+x);
			if(col>max_val)
				max_val=col;
			if(col<min_val)
				min_val=col;
			hist[col]++;
			NumPoints++;
		}
		p+=src_shift;
	}
	j=src_center[0]-min_val;
	long* hst_min_row = hist+min_val;
	long s=0;
	while(j--)
		s+=*hst_min_row++;
	s+=(*hst_min_row)>>1;
	PIXEL val=PIXEL((s*Ncolor+(NumPoints>>1))/NumPoints);
	if (val>=Ncolor) val=Ncolor-1;					
		*dst_center=val; 

	int to_move=half_framW;
	int center_shft=1;	
	Frame_sz_x++;//initial size = visible part of frame 
	
	tmpl_LeftRight<PIXEL>(W,Original_XFrmSz,Frame_sz_y,down_sz_y,half_framW,hist,src_center+1,
		dst_center+1,src_shift,&max_val,&min_val,Ncolor);

	Frame_sz_x=half_framW+1;
	tmpl_VertRightUp<PIXEL>(Frame_sz_x,Frame_sz_y+1,down_sz_y,hist,src_center+W-1,
		dst_center+W-1+dst_shift,src_shift,&max_val,&min_val,Ncolor);	
	
	int y;
	for(y=1; y<half_framH; y++)
	{
		if (count) count->current(y);
		
		src_center=(*Src)[y];
		dst_center=(*Dst)[y];
		Frame_sz_y=y+Original_YFrmSz-half_framH; //height of frame !!!
		if (y&1)
		{// -    			
			src_center+=W-2;
			dst_center+=W-2;

			tmpl_RightLeft<PIXEL>(W,Original_XFrmSz,Frame_sz_y,down_sz_y, half_framW,hist, 
				src_center,dst_center,src_shift,&max_val,&min_val,Ncolor);

			src_center=(*Src)[y];
			dst_center=(*Dst)[y];
			Frame_sz_x=(Original_XFrmSz>>1)+1;// right_half_frame + center
			tmpl_VertLeftUp<PIXEL>(Frame_sz_x,Frame_sz_y+1,down_sz_y,hist,src_center,
					dst_center+dst_shift,src_shift,&max_val,&min_val,Ncolor);
					
		}
    	else 
		{//-   

			tmpl_LeftRight<PIXEL>(W,Original_XFrmSz,Frame_sz_y,down_sz_y,half_framW,hist,src_center+1,
				dst_center+1,src_shift,&max_val,&min_val,Ncolor);

			Frame_sz_x=half_framW+1;
			tmpl_VertRightUp<PIXEL>(Frame_sz_x,Frame_sz_y+1,down_sz_y,hist,src_center+W-1,
				dst_center+W-1+dst_shift,src_shift,&max_val,&min_val,Ncolor);	
		}
	}

	int down_limit=((Original_YFrmSz)>>1);	
	Frame_sz_y=Original_YFrmSz; //height of frame !!!	

	for(y=half_framH; y<H-down_limit-1; y++)
	{
		src_center=(*Src)[y];
		dst_center=(*Dst)[y];
		if (count) count->current(y);

		if (y&1)
		{// -    			
			src_center+=W-2;
			dst_center+=W-2;

			tmpl_RightLeft<PIXEL>(W,Original_XFrmSz,Frame_sz_y,down_sz_y, half_framW,hist, 
				src_center,dst_center,src_shift,&max_val,&min_val,Ncolor);

			src_center=(*Src)[y];
			dst_center=(*Dst)[y];
			Frame_sz_x=(Original_XFrmSz>>1)+1;// right_half_frame + center
			tmpl_VertLeftMiddle<PIXEL>(Frame_sz_x,Frame_sz_y,down_sz_y,hist,src_center,
					dst_center+dst_shift,src_shift,&max_val,&min_val,Ncolor);
					
		}
    	else 
		{//-   
			tmpl_LeftRight<PIXEL>(W,Original_XFrmSz,Frame_sz_y,down_sz_y,half_framW,hist,src_center+1,
				dst_center+1,src_shift,&max_val,&min_val,Ncolor);

			Frame_sz_x=half_framW+1;
			tmpl_VertRightMiddle<PIXEL>(Frame_sz_x,Frame_sz_y,down_sz_y,hist,src_center+W-1,
				dst_center+W-1+dst_shift,src_shift,&max_val,&min_val,Ncolor);	
		}
	}

	Frame_sz_y++;
    down_sz_y++;
	for(y=H-down_limit-1; y<H; y++)
	{
		src_center=(*Src)[y];
		dst_center=(*Dst)[y];
		if (count) count->current(y);

		down_sz_y--;
		Frame_sz_y--; //height of frame !!!
		
		if (y&1)
		{// -    			
			src_center+=W-2;
			dst_center+=W-2;

			tmpl_RightLeft<PIXEL>(W,Original_XFrmSz,Frame_sz_y,down_sz_y, half_framW,hist, 
				src_center,dst_center,src_shift,&max_val,&min_val,Ncolor);
			if (y!=H-1)
			{
				src_center=(*Src)[y];
				dst_center=(*Dst)[y];
				Frame_sz_x=(Original_XFrmSz>>1)+1;// right_half_frame + center
				tmpl_VertLeftDown<PIXEL>(Frame_sz_x,Frame_sz_y-1,down_sz_y,hist,src_center,
						dst_center+dst_shift,src_shift,&max_val,&min_val,Ncolor);				
			}			
		}
    	else 
		{//-   

			tmpl_LeftRight<PIXEL>(W,Original_XFrmSz,Frame_sz_y,down_sz_y,half_framW,hist,src_center+1,
				dst_center+1,src_shift,&max_val,&min_val,Ncolor);

			if (y!=H-1)
			{
				Frame_sz_x=half_framW+1;
				tmpl_VertRightDown<PIXEL>(Frame_sz_x,Frame_sz_y-1,down_sz_y,hist,src_center+W-1,
					dst_center+W-1+dst_shift,src_shift,&max_val,&min_val,Ncolor);	
			}
		}
	}
	if (count) count->finish();
	return true;	
}
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
// class MSlidingEqualization /////////////////////////////////////
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
MSlidingEqualization::MSlidingEqualization()
{
	src_image = 0; hist = 0; N_colors = 0;
	Hi = 0; Wd = 0; Bt = 0;
	ClassType = AImage::IMGC_GRAYSCALE;  
}

//virtual 
void MSlidingEqualization::Init(const AImage *src)
{
	if(!src) throw "MSlidingEqualization::Init() : null input image";
	if(!src->IsOk()) throw "MSlidingEqualization::Init() : input image is not OK";
	Close();
	try
	{
		src_image=src;
		int _bits = src_image->Bits();
		int k=1<<_bits;
		if(_bits<=0) throw "MSlidingEqualization::Init() : invalid parameter bits<=0";
		hist=new long[k];
		if(!hist) throw 1;
	}
	catch(...)
	{
		throw "MSlidingEqualization::Init(): Insufficient memory";
		Close();
	};
	Hi = src->Height();
	Wd = src->Width();
	Bt = src->Bits();
	ClassType = src->ImageClass();
}


void MSlidingEqualization::CloseThis()
{
	if(hist)
	{
		delete[] hist;
		hist = 0;	
		N_colors = 0;
	}
	N_colors = 0;
	Hi = 0;
	Wd = 0;
	Bt = 0;
	ClassType = AImage::IMGC_GRAYSCALE;

}

//virtual 
bool MSlidingEqualization::Run(AImage* dst, int Original_XFrmSz, int Original_YFrmSz,
							   ProgressCounter* Ind)
{
	if(!dst || !src_image) throw "MSlidingEqualization::Run() : null input images";
	if(!dst->IsOk() || !src_image->IsOk()) throw "MSlidingEqualization::Run() : Not OK input images";
	if((src_image->ImageClass()!=dst->ImageClass())) throw "MSlidingEqualization::Run() : Images Are Not the same class;";
	if((src_image->Height()!=dst->Height())||(src_image->Width()!=src_image->Width())) throw "MSlidingEqualization::Run() : Different Sizes Images";
	if(src_image->Bits()!=dst->Bits()) throw "MSlidingEqualization::Run() : Different Bits Per Pixel";
	if(src_image->ImageClass()!=AImage::IMGC_GRAYSCALE) throw "MSlidingEqualization::Run() : Images Are Not Grayscale;";

	const AImageMem8 *s8=dynamic_cast<const AImageMem8 *>(src_image);
	if(s8)
	{
		AImageMem8 *d8=dynamic_cast<AImageMem8 *>(dst);
		if(d8)
			return tmpl_Run<AImageMem8,Pixel8>(s8,d8,Original_XFrmSz,Original_YFrmSz, hist,Ind);
		else
			throw "MSlidingEqualization::Run(): src is AImageMem8 but dst is not";
	}
	else
	{
		const AImageMem16 *s16=dynamic_cast<const AImageMem16 *>(src_image);
		if(s16)
		{
			AImageMem16 *d16=dynamic_cast<AImageMem16 *>(dst);
			if(d16)
				return tmpl_Run<AImageMem16,Pixel16>(s16,d16,Original_XFrmSz,Original_YFrmSz, hist,Ind);
			else
				throw "MSlidingEqualization::Run(): src is AImageMem16 but dst is not";
		}
		else
		{
				throw "MSlidingEqualization::Run(): src is not AImageMem8 or AImageMem16";
		}
	}
	return false;
}

