#ifndef _NUM_BOOST_ADDS_H
#define _NUM_BOOST_ADDS_H
////////////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
namespace NumAlg
{
namespace Lin
{ 
	/*!
	 *         , 
	 *     .
	 *
	 *****************************************************************************************/
	template<class Iter>
	Iter max_abs_element(Iter begin,Iter end)
	{
		Iter m=begin;
		typedef Iter::value_type T;
		T a=std::abs(*begin++);
		for( ; begin!=end ; ++begin )
		{
			T b=std::abs(*begin);
			if(b > a)
			{
				m=begin;
				a=b;
			}
		}
		return m;
	}
	/*!
	 *    A   ,   diag_matrix, 
	 *    .
	 *
	 *  -   A ( A      ).
	 *
	 *       Boost.
	 ****************************************************************************************/
	template<class M,class V>
	void prod_diag(ublas::matrix_container<M> & A,const ublas::vector_container<V>&  diag_matrix)
	{
		typedef M* M_PTR;
		typedef const V* V_PTR;
		const size_t n=(*static_cast<M_PTR>(&A)).size2();
		for(size_t i=0; i< n ; ++i)
		{
			ublas::matrix_column<M>(*static_cast<M_PTR>(&A),i)*=(*static_cast<V_PTR>(&diag_matrix))[i];
		}
	}
	/*!
	 *    ,   diag_matrix, 
	 *       A.
	 *
	 *  -   A ( A      ).
	 *
	 *       Boost.
	 *****************************************************************************************/
	template<class V,class M>
	void prod_diag(const ublas::vector_container<V>&  diag_matrix,ublas::matrix_container<M> & A)
	{
		typedef M* M_PTR;
		typedef const V* V_PTR;
		const size_t n=(*static_cast<M_PTR>(&A)).size1();
		for(size_t i=0; i< n; ++i)
		{
			ublas::matrix_row<M>(*static_cast<M_PTR>(&A),i)*=(*static_cast<V_PTR>(&diag_matrix))[i];
		}
	}
	/*!
	 *        , ..
	 *
	 * return (value>eps)? 1/value : 0;
	 *
	 * ,  value >=0
	 *****************************************************************************************/
	template<typename T>
	class PseudoInverseElemPositive
	{
	   T eps;   // The value to multiply by
	public:
	   PseudoInverseElemPositive( const T& _eps ) : eps( _eps ) {  }
	   void operator ( ) ( T& elem ) const
	   {
			elem =(elem >= eps ) ? 1/elem : 0;
	   }
	};
	/*!
	 *        ,    , ..
	 *
	 * return (abs(value)>eps)? 1/value : 0;
	 *****************************************************************************************/
	template<typename T>
	class PseudoInverseElem
	{
	   T eps;   // The value to multiply by
	public:
	   PseudoInverseElem( const T& _eps ) : eps( _eps ) {  }
	   void operator ( ) ( T& elem ) const
	   {
		   elem =(std::abs(elem) >= eps ) ? 1/elem : 0;
	   }
	};
	/*!
	 *           , ..
	 * return (value>eps)? 1/value : 0;
	 *
	 * ,       >=0
	 *
	 *        .
	 *****************************************************************************************/
	template<class VectorW>
	static void inverse_diag_matrix_positive(VectorW& W,double relerr)
	{
		typedef VectorW::value_type T;
		T tau=*std::max_element(W.begin(),W.end());
		tau*=T(relerr);
		std::for_each(W.begin(),W.end(), PseudoInverseElemPositive<T>(tau) );
	}

	/*!
	 *           ,
	 *    , ..
	 *
	 * return (abs(value)>eps)? 1/value : 0;
	 *
	 *          ..  ..
	 *****************************************************************************************/
	template<class VectorW>
	static void inverse_diag_matrix(VectorW& W,double relerr)
	{
		typedef VectorW::value_type T;
		T tau=std::abs(*max_abs_element(W.begin(),W.end()));
		tau*=T(relerr);
		std::for_each(W.begin(),W.end(), PseudoInverseElem<T>(tau) );
	}
	/*!
	 *         . 
	 * max_{i,j}|Aij| -norm
	 *******************************************************************************/
	template<class M>
	typename M::value_type matrix_max_abs_element(M & A)
	{
		const int m = A.size1();
		const int n = A.size2();
		typename M::value_type max_el=std::abs(A(0,0));
		for(int i=0; i<m; ++i)
		for(int j=0; j<n; ++j)
			max_el=std::max(max_el,std::abs(A(i,j)));
		return max_el;
	}



////////////////////////////////////////////////////////////////////////////////////////////////
}//namespace Lin
}//namespace NumAlg
}//namespace Vtl
////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_NUM_BOOST_ADDS_H
