#ifndef _NUM_IMPLICIT_H
#define _NUM_IMPLICIT_H
////////////////////////////////////////////////////////////////////////////////////////////////
#include "_num_boost_adds.h"
////////////////////////////////////////////////////////////////////////////////////////////////

namespace Vtl
{
namespace NumAlg
{
///\brief       .   .   
///	 ,    ,      . 
namespace Implicit
{
	namespace Pred
	{
		///  GreaterThan: 
		template<class T>
		struct gt;
		///  LessThan: 
		template<class T>
		struct lt;
		///  GreaterThan: Specialization for boost::ublas::c_vector
		template<class T_,int N_>
		struct gt<Vtl::NumAlg::Lin::ublas::c_vector<T_,N_> >
		{
			enum {N=N_};
			typedef T_ value_type;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T_,N_> Vector;
			const Vector& v;

			gt(const Vector & vec) : v(vec) {}
			bool operator()(int elem1,int elem2) const 
			{
				return v[elem1] > v[elem2];
			}
		};
		///  LessThan: Specialization for boost::ublas::c_vector
		template<class T_,int N_>
		struct lt<Vtl::NumAlg::Lin::ublas::c_vector<T_,N_> >
		{
			enum {N=N_};
			typedef T_ value_type;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T_,N_> Vector;

			const Vector& v;
			lt(const Vector & vec) : v(vec) {}
			bool operator()(int elem1,int elem2) const 
			{
				return v[elem1] < v[elem2];
			}
		};
	}//namespace Pred
	///       Implicit::Sort
	///  N_=0 - (  )       std::vector,
	///  N_>0 -         int index[N_];      .
	template<int N_>
	class Sort_base
	{
	public:
		enum {N=N_};
	private:
		int index[N];
	public:
		template<class Pred2_>
		Sort_base(const Pred2_& p,int )
		{
			Vtl::Meta::set_seq<N>(index,0);
			std::sort(index,index+N,p);
		}
		int operator[](int i) const {return index[i];}
		operator const int*() const {return index;}
	};
	///       Implicit::Sort (    
	///       ).
	template<>
	class Sort_base<0>
	{
	public:
		const int N;
	private:
		std::vector<int> index;
	public:
		template<class Pred2_>
		Sort_base(const Pred2_& p,int n)
			: N(n)
		{
			index.resize(N);
			for(int i=0; i<N; ++i) 
				index[i]=i;
			std::sort(index.begin(),index.end(),p);
		}
		int operator[](int i) const {return index[i];}
		operator const int*() const {return &(index[0]);}
	};


	/// \brief   .    
	/// ,   
	///   .    .    operator[]  
	///             
	///  const int*      (      
	///  , . .  Implicit::sort).  - .   
	///  T == boost::ublas::c_vector<TT,N>
	template<class T_,class Pred_=Pred::lt<T_> >
	struct Sort;

	/// 
	/// \brief Specialization for boost::ublas::c_vector.   .    
	/// ,   
	///   .    .    operator[]  
	///             
	///  const int*      (      
	///  , . .  Implicit::sort). \n
	/// : \n
	/// \code
	/// using namespace Vtl::NumAlg::Implicit;
	/// using namespace boost::numeric::ublas;
	/// enum {N=5};
	/// typedef vector_c<double,N> MyVector;
	/// MyVector v;
	/// for(int i=0< i<N ; ++i) 
	///    v(i) = i; 
	/// Sort<MyVector, Pred::gt<MyVector> > s(v);
	/// //  Pred::gt<MyVector> p(v); //  
	/// //  Sort<MyVector> s(v,p);   //  .
	/// std::cout << v(0) << " " << v(s[0]) << v(N-1) << " " << v(s[N-1]) <<std::endl;
	/// \endcode
	///  : \n
	/// 0 4 4 0 \n
	/// ..     0,  4,   ""  -   4,  0.
	/// 
	///    -  ,     . , 
	///        [].
	///   - ,   LessThan,..    "",       . 
	///     :\n
	/// \li  -     ,  .      _Pred(object).
	/// \li  - ,       -  _Pred2    . ,    
	///   - (            ).    - 
	///         .\n
	///
	///     :\n
	/// \code
	/// template<typename T1> struct X 
	/// { 
	///		template<typename T2> X(T1 t) {} 
	/// };
	/// X<float> x<int>(1.0f); //Compile error.
	/// \endcode
	///      C++      :
	/// \code
	/// X<float> x=X<float><int>(1.0f); //Compile error.
	/// \endcode
	///
	template<class T_,int N_,class Pred_>
	class Sort<Vtl::NumAlg::Lin::ublas::c_vector<T_,N_>,Pred_>
		: public Sort_base<((N_>30)?0:N_)> //    -        ,    !
	{
	public:
		typedef T_ value_type;
		typedef Vtl::NumAlg::Lin::ublas::c_vector<T_,N_> Vector;
	public:
		Sort(const Vector& v)
			: Sort_base(Pred_(v),N_)
		{
		}
		template<class Pred2_>
		Sort(const Vector& v,const Pred2_& p)
			: Sort_base(p,N_)
		{
		}
	};
	/// Reorder boost::numeric::ublas::c_matrix columns in order as indexes in array s.
	/// Array 's' usually obtained with Vtl::NumAlg::Implicit::Sort specialization derived 
	/// from Vtl::NumAlg::Implicit::Sort_base 
	template<typename T,size_t M,size_t N> //,class T_Vector,class T_Predictor> 
	inline void sort_columns(Vtl::NumAlg::Lin::ublas::c_matrix<T,M,N>& U_sorted
				,Vtl::NumAlg::Lin::ublas::c_matrix<T,M,N>& U_in
				,const int* s )
	{
		using namespace boost::numeric::ublas;
		typedef typename c_matrix<T, N, N> Matrix;
		for(int i=0; i<N; ++i)
		{
			matrix_column<Matrix> c( U_sorted, i );
			c=column<Matrix>( U_in, s[i] );
		}
	}
	/// Reorder boost::numeric::ublas::matrix columns in order as indexes in array s.
	/// Array 's' usually obtained with Vtl::NumAlg::Implicit::Sort specialization derived 
	/// from Vtl::NumAlg::Implicit::Sort_base 
	template<typename T> //,class T_Vector,class T_Predictor> 
	inline void sort_columns(Vtl::NumAlg::Lin::ublas::matrix<T>& U_sorted
				,const Vtl::NumAlg::Lin::ublas::matrix<T>& U_in
				,const int* s )
	{
		using namespace boost::numeric::ublas;
		typedef typename matrix<T> Matrix;
		if( ( U_sorted.size1() != U_in.size1() ) 
		 || ( U_sorted.size2() != U_in.size2() ) )
		 throw "Vtl::NumAlg::Implicit::sort() : Error: matrices are of different size.";
		for(int i=0; i < U_in.size2() ; ++i)
		{
			matrix_column<Matrix> c( U_sorted, i );
			c=column<Matrix>( U_in, s[i] );
		}
	}
	/// Reorder boost::numeric::ublas::c_vector components in order as indexes in array s
	/// Array 's' usually obtained with Vtl::NumAlg::Implicit::Sort specialization derived 
	/// from Vtl::NumAlg::Implicit::Sort_base 
	template<typename T,size_t N> //,class T_Vector,class T_Predictor> 
	inline void sort(Vtl::NumAlg::Lin::ublas::c_vector<T,N>& V_sorted
				,const Vtl::NumAlg::Lin::ublas::c_vector<T,N>& V_in
				,const int* s )
	{
		using namespace boost::numeric::ublas;
		for(int i=0; i<N; ++i)
			V_sorted(i)= V_in(s[i]);
	}
	/// Reorder boost::numeric::ublas::vector components in order as indexes in array s
	/// Array 's' usually obtained with Vtl::NumAlg::Implicit::Sort specialization derived 
	/// from Vtl::NumAlg::Implicit::Sort_base 
	template<typename T> //,class T_Vector,class T_Predictor> 
	inline void sort(Vtl::NumAlg::Lin::ublas::vector<T>& V_sorted
				,const Vtl::NumAlg::Lin::ublas::vector<T>& V_in
				,const int* s )
	{
		using namespace boost::numeric::ublas;
		if ( U_sorted.size() != U_in.size() ) 
		 throw "Vtl::NumAlg::Implicit::sort() : Error: vectors are of different size.";
		for(int i=0; i<U_in.size() ; ++i)
			V_sorted(i)= V_in(s[i]);
	}
}//namespace Implicit
}//namespace NumAlg
}//namespace Vtl

////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_NUM_IMPLICIT_H
