#include <math.h>
#include "nummet.h"
/////////////////////////////////////////////////////////
// Functions:
//   int nmPow(int a,int b); 
//   double nmPoly(double x,const double *coef,int n)
//--------------------------------------------------------
//   double QUANC8(double (*FUN)(double,void *),const double A,const double B,
//		                void *param,const double ABSERR,const double RELERR,
//		                double& ERREST,int& NOFUN,double& FLAG);
//   double FMIN(double (*FUN)(double,void *),double AX,double BX,void *param,double TOL);
//   double ZEROIN(double (*FUN)(double,void *),double AX,double BX,void *param,double TOL);
//   double SEVAL(int n,double u,const double *x,const double *y,const double *b,const double *c,const double *d);
//   void   SPLINE_FREE(int n,const double *x,const double *y,double *b,double *c,double *d);
//   void   SPLINE(int n,const double *x,const double *y,double *b,double *c,double *d);
//   int nmSVD(int NM,int M,int N,double *A,double *W,double *U,double *V,double *RV1);
//
// :
// ., ., ..    .
// . . 1980.  280 . .  . ...
//
// :
// Prentice-hall, inc.
// Englrwood clifs, N.J. 07632
// 1977
//
// Computer methods for mathematical computations.
// George E. Forsythe (Department of Computer Science University of Waterloo)
// Michael A. Malcolm (Department of Computer Science University of Waterloo)
// Cleve B/ Moler (Department of Mathematics and Statistics University of New Mexico)
//
//-----------------------------------------------------------------------------------
//   double nmGAMMLN(double xx);
//
// :
// Numerical resipes. The Art of Programming.
//
//-----------------------------------------------------------------------------------
//int nmVDCPS(int N,int *IPF,int *IEXP,int *IPWR);   N   
//void nmVTRAN(double *A,int N,int M); //      
//
// :
// IMSL
//
//////////////////////////////////////////////////////////

int nmPow(int a,int b)
{
	if(b==0)
		return 1;
	else if(b>0)
	{
		int res=1;
		while(b--)
			res*=a;
		return res;
	}
	return 0;
}
//=======================================================
//=======================================================
//=======================================================

double nmPoly(double x,const double *coef,int n)
{
// result=coef[0]+coef[1]*x+coef[2]*x*x+...+coef[n]*pow(x,n);
// !!! n - ,    =(n+1)   !!!!!
//
	double sum=0;
	if(n>=0)
	{
		sum+=coef[n];
		while(n>0)
		{
			sum*=x;
			sum+=coef[--n];
		}
	}
	return sum;
}

//=======================================================
//=======================================================
//=======================================================
double nmQUANC8(double (*FUN)(double,void *),const double A,const double B,
		void *param,const double ABSERR,const double RELERR,
		double& ERREST,int& NOFUN,double& FLAG)
{
//          
//          
//          ,   
//      - 8- 
//
//       
//
//      NOFUN  -    . 
//      ERREST -   
//      FLAG   -   :  FLAG=000.000,
//              RESULT   ;
//              FLAG=AAA.BBB,  AAA-  ,
//                 , 0.BBB- 
//               ,  NOFUN
//
//       :
//						RESULT -  
//
      double W0, W1, W2, W3, W4, AREA, X0, F0, STONE, STEP, COR11, TEMP;
      double QPREV, QNOW, QDIFF, QLEFT, ESTERR, TOLERR;
      double  QRIGHT[31], F[16], X[16], FSAVE[8][30], XSAVE[8][30];
      int  LEVMIN, LEVMAX, LEVOUT, NOMAX, NOFIN, LEV, NIM;
//***    ,   ***
        LEVMIN = 1;
        LEVMAX = 30;
        LEVOUT = 6;
        NOMAX  = 5000;
        NOFIN  = NOMAX-8*(LEVMAX-LEVOUT+nmPow(2,LEVOUT+1));
//        NOFUN>NOFIN,  
      W0 = 3956./14175.;
      W1 = 23552./14175.;
      W2 = -3712./14175.;
      W3 = 41984./14175.;
      W4 = -18160./14175.;
        FLAG   = 0.;
      double RESULT = 0.;
        COR11  = 0.;
        ERREST = 0.;
        AREA   = 0.;
        NOFUN  = 0;
      if(A == B) return RESULT;
//
      LEV   = 0;
      NIM   = 1;
      X0    = A;
      X[15] = B;
      QPREV = 0.;
      F0    = FUN(X0,param);
      STONE=(B-A)/16.;
      X[7] =(X0+X[15])/2.;
      X[3] =(X0+X[7])/2.;
      X[11]=(X[7]+X[15])/2.;
      X[1] =(X0+X[3])/2.;
      X[5] =(X[3]+X[7])/2.;
      X[9] =(X[7]+X[11])/2.;
      X[13]=(X[11]+X[15])/2.;
			{
				for(int j=1; j<16; j+=2)
				{
					F[j]=FUN(X[j],param);
				}
			}
      NOFUN=9;
//
//
//...  
//...  QPREV, X0, X2, ....X16,F0,F2,....F16,
//...  X1,X3,... X15, F1,F3,....F15,QLEFT,
//... QRIGHT,QNOW,QDIFF,AREA.
//
l30:  X[0]=(X0+X[1])/2.;
      F[0]=FUN(X[0],param);
			{
				for(int j=2; j<15; j+=2)
				{
					X[j]=(X[j-1]+X[j+1])/2.;
					F[j]=FUN(X[j],param);
				}
			}
      NOFUN=NOFUN+8;
      STEP=(X[15]-X0)/16.;
      QLEFT=(W0*(F0+F[7])+W1*(F[0]+F[6])+W2*(F[1]+F[5])+W3*(F[2]+F[4])+W4*F[3])*STEP;
      QRIGHT[LEV]=(W0*(F[7]+F[15])+W1*(F[8]+F[14])+W2*(F[9]+F[13])+W3*(F[10]+F[12])+W4*F[11])*STEP;
      QNOW=QLEFT+QRIGHT[LEV];
      QDIFF=QNOW-QPREV;
      AREA=AREA+QDIFF;
//
//
//...   
//
      ESTERR=ABS(QDIFF)/1023.;
      TOLERR=MAX(ABSERR,RELERR*ABS(AREA))*(STEP/STONE);
      if(LEV < LEVMIN) goto l50;
      if(LEV >= LEVMAX) goto  l62;
      if(NOFUN > NOFIN) goto l60;
      if(ESTERR <= TOLERR) goto l70;
//
//...  
//...  
//
l50:  NIM=NIM*2;
      LEV=LEV+1;
//
			{
				for(int j=0; j<8; j++)
				{
					FSAVE[j][LEV-1]=F[j+8];
					XSAVE[j][LEV-1]=X[j+8];
				}
			}
//
//
      QPREV=QLEFT;
			{
				for(int k=-1; k>=-8; k--)
				{
					F[2*k+17]=F[k+8];
					X[2*k+17]=X[k+8];
				}
			}
      goto l30;
//
l60:  NOFIN=2*NOFIN;
      LEVMAX=LEVOUT;
      FLAG=FLAG+(B-X0)/(B-A);
      goto l70;
l62:  FLAG=FLAG+1.;
//
l70:  RESULT=RESULT+QNOW;
      ERREST=ERREST+ESTERR;
      COR11=COR11+QDIFF/1023.;
//
l72:  if(NIM == 2*(NIM/2)) goto l75;
      NIM=NIM/2;
      LEV=LEV-1;
      goto l72;
l75:  NIM=NIM+1;
      if(LEV <= 0) goto l80;
//
      QPREV=QRIGHT[LEV-1];
      X0=X[15];
      F0=F[15];
			{
				for(int k=0; k<8; k++)
				{
					F[2*k+1]=FSAVE[k][LEV-1];
					X[2*k+1]=XSAVE[k][LEV-1];
				}
			}
      goto l30;
//
l80:  RESULT=RESULT+COR11;
//
      if(ERREST == 0.) return RESULT;
l82:  TEMP=ABS(RESULT)+ERREST;
      if(TEMP != ABS(RESULT)) return RESULT;
      ERREST=2.*ERREST;
      goto l82;
//      return RESULT;
}

//=======================================================
//=======================================================
//=======================================================
double nmFMIN(double (*FUN)(double,void *),double AX,double BX,void *param,double TOL)
{
//
//         X  ,  FUN  
//       [AX,BX].
//
//      :
//     """""""""""""""""""""
//     AX   -     
//     BX   -     
//     FUN(X) -  -,   FUN(X)  
//             X   [AX,BX].
//     TOL  -       -
//              / TOL >= 0. /.
//
//      :
//     """"""""""""""""""""""
//     FMIN -  ,  ,   FUN  .
//
//               -
//         .    
//     ,    .  FUN    -
//     ,    (    AX, 
//      BX),        
//     1.324...
//         FUN      ,   
//         EPS*ABS(X)+TOL/3 ,  EPS  
//          .  FUN - -
//          FUN  
//         ,  FMIN -
//         FUN   (AX,BX)  ,
//      3*EPS*ABS(X)+TOL.  FUN   ,   FMIN
//            , -
//     ,    .
//        -    
//      60 -  LOCALMIN,   Richard Brent , 
//     Algorithms for minimization without derivatives, Prentice-Hall,inc.
//     (1973).
//
//
//
//      C     ,    
//
	const double C=0.5*(3.-::sqrt(5.));
//
//     EPS       -
//      .
//
	double EPS=1.0;
	{
	  double TOL1;
		do
		{
			EPS=EPS/2.0;
			TOL1=1.0+EPS;
		} while(TOL1 > 1.0);
	}
  EPS=::sqrt(EPS);
//---------------<    >------
  double A=AX;
  double B=BX;
  double V=A+C*(B-A);
  double W=V;
  double X=V;
  double E=0.;
  double FX=FUN(X,param);
  double FV=FX;
  double FW=FX;

  double D,P,Q,R,U;
//***------------<     >-------
	while(1)
	{
		const double XM=0.5*(A+B);
		const double TOL1=EPS*ABS(X)+TOL/3.;
		const double TOL2=2.*TOL1;
		//===============<    >=========
		if( ABS(X-XM) <= ( TOL2-0.5*(B-A) ) )
		{//-------------**<    >**-----
			return X;
		};
		//
		//        
		//
    bool logic= ( ABS(E) <= TOL1 );
    if( ! logic )
		{
//---------------<   >--------
       R=(X-W)*(FX-FV);
       Q=(X-V)*(FX-FW);
       P=(X-V)*Q-(X-W)*R;
       Q=2.*(Q-R);
       if( Q > 0. ) P=-P;
       Q=ABS(Q);
       R=E;
       E=D;
//---------------<    >------------------
       logic=logic || ( ABS(P) >= ABS(0.5*Q*R) )
       || ( P <= Q*(A-X) ) || ( P >= Q*(B-X) );
    };
    if( logic ) 
		{
//-----------<    >------------
       if( X >= XM ) E=A-X;
       if( X < XM ) E=B-X;
       D=C*E;
		}
    else
		{
//----------------<    >---
       D=P/Q;
       U=X+D;
//-----------< FUN       AX  BX >--
       if( (U-A) < TOL2 ) D=SIGN(TOL1,XM-X);
       if( (B-U) < TOL2 ) D=SIGN(TOL1,XM-X);
    };
//-----------< FUN       X >-------
		if( ABS(D) >= TOL1 )
			U=X+D;
		else
			U=X+SIGN(TOL1,D);
		const double FU=FUN(U,param);
//===========<     A,B,V,W,X >======
    if( FU > FX ) 
		{
			if( U < X ) A=U;
			else        B=U;
      if( FU <= FW || W == X ) 
			{
				V=W;
				FV=FW;
				W=U;
				FW=FU;
			}
      else if( FU <= FV || V == X || V == W ) 
			{
				V=U;
				FV=FU;
      };
		}
    else
		{
      if( U >= X ) A=X;
      else         B=X;
      V=W;
      FV=FW;
      W=X;
      FW=FX;
      X=U;
      FX=FU;
    };
	};//end of while(1);
}

//=======================================================
//=======================================================
//=======================================================
double nmZEROIN(double (*FUN)(double,void *),double AX,double BX,void *param,double TOL)
{
//
//       ()   
//      AX , BX
//
//                
//
//      AX       
//      BX       
//      FUN   -,  
//            FUN(X)    X   AX,BX
//
//     TOL      
//             
//
//                
//
//      ZEROIN  ,    FUN
//                AX, BX
//
//           ,  FUN(AX)n i FUN(BX)
//       . ZEROIN   X 
//       AX, B     
//      4*MASHEPS*ABS(X)+TOL,  MASHEPS-
//      .
//                    :
//     1:       
//     2:          
//              
//                       B-A     2
//               ( LOG (------) )
//                    2  TOL1
//             TOL1=0.5*TOL+2.0*EPS*ABS(B)
//     3:      R,  , ,  FUN
//                 ,
//               (R-2*TOL1, R+2*TOL1).
//     4:           
//            10  
//
//
//           EPS
//
  double TOL1;
	double EPS=1.0;
	do
	{
	  EPS=EPS/2.0;
		TOL1=1.0+EPS;
	} while(TOL1 > 1.0);
//
//        
//
	double A=AX;
	double B=BX;
	double FA=FUN(A,param);
	double FB=FUN(B,param);
//
//      
//
  double C=A;
  double FC=FA;
  double D=B-A;
  double E=D;
	while(1)
	{
		if(ABS(FC) < ABS(FB)) 
		{
			A=B;
			B=C;
			C=A;
			FA=FB;
			FB=FC;
			FC=FA;
		}
		//  
    TOL1=2.0*EPS*ABS(B)+0.5*TOL;
    double XM=0.5*(C-B);
    if(ABS(XM) <= TOL1) break;
    if(FB == 0.0)       break;
		//
	  double P,Q;
		bool use_bisection;
		//    ?
    if( ABS(E) < TOL1 || ABS(FA) <= ABS(FB) ) 
			use_bisection=true;
		else
		{
			use_bisection=false;
			//    ?
			double S;
      if(A == C) 
			{//                        
				S=FB/FA;
				P=2.*XM*S;
				Q=1.0-S;
			}
			else
			{//       
				Q=FA/FC;
				double R=FB/FC;
				S=FB/FA;
				P=S*(2.0*XM*Q*(Q-R)-(B-A)*(R-1.0));
				Q=(Q-1.0)*(R-1.0)*(S-1.0);
			}
			//                         
      if(P>0.0) Q=-Q;
      P=ABS(P);
			//   
      if( 2.0*P >= (3.0*XM*Q-ABS(TOL1*Q)) ||  P >= ABS(0.5*E*Q) ) 
			{
				use_bisection=true;
			}
		}		
		//                    
		if(use_bisection)
		{//                     
	    D=XM;
      E=D;
		}
		else
		{//  
			E=D;
      D=P/Q;
		}
		//  
		A=B;
		FA=FB;
		if(ABS(D)>TOL1) B=B+D;
		if(ABS(D) <= TOL1) B=B+SIGN(TOL1,XM);
		FB=FUN(B,param);
		if( FB*(FC/ABS(FC)) > 0.0)
		{
			C=A;
			FC=FA;
			D=B-A;
			E=D;
		}
	};
//             
  return B;
}

//=======================================================
//=======================================================
//=======================================================
void   nmSPLINE(int n,const double *x,const double *y,double *b,double *c,double *d)
{
	if(n<2) return;
	if(n<3)
	{
		b[0]=(y[1]-y[0])/(x[1]-x[0]);
		c[0]=0.;
		d[0]=0.;
		b[1]=b[0];
		c[1]=0.;
		d[1]=0.;
		return;
	}
//
// build 3-diagonal sceem
// b- diagonal
//
	d[0]=x[1]-x[0];
	c[1]=(y[1]-y[0])/d[0];
	const int max_i=n-1;
	const int nm1=n-1;
	{
		for(int i=1; i<nm1; i++)
		{
			d[i]=x[i+1]-x[i];
			b[i]=2.*(d[i-1]+d[i]);
			c[i+1]=(y[i+1]-y[i])/d[i];
			c[i]=c[i+1]-c[i];
		}
	}
// granichnye usloviya
	b[0]=-d[0];
	b[max_i]=-d[max_i-1];
	c[0]=0.;
	c[max_i]=0.;
	if(n!=3)
	{
		c[0]=c[2]/(x[3]-x[1])-c[1]/(x[2]-x[0]);
		c[max_i]=c[max_i-1]/(x[max_i]-x[max_i-2])-c[max_i-2]/(x[max_i-1]-x[max_i-3]);
		c[0]=-c[0]*d[0]*d[0]/(x[3]-x[0]);
		c[max_i]=-c[max_i]*d[max_i-1]*d[max_i-1]/(x[max_i]-x[max_i-3]);
	}
//pryamoy hod
	{
		for(int i=1;i<n;i++)
		{
			double t=d[i-1]/b[i-1];
			b[i]=b[i]-t*d[i-1];
			c[i]=c[i]-t*c[i-1];
		}
	}
// obratnaya podstanovka
	c[max_i]=c[max_i]/b[max_i];
	{
		for(int i=n-2; i>=0; i--)
		{
			c[i]=(c[i]-d[i]*c[i+1])/b[i];
		}
	}
//now c[i]=SIGMA(i)
	const int nm2=nm1-1;
	b[max_i]=(y[max_i]-y[nm2])/d[nm2]+d[nm2]*(c[nm2]+2.*c[max_i]);
	{
		for(int i=0; i<nm1; i++)
		{
			b[i]=(y[i+1]-y[i])/d[i]-d[i]*(c[i+1]+2.*c[i]);
			d[i]=(c[i+1]-c[i])/d[i];
			c[i]=3.*c[i];
		}
	}
	c[max_i]=3*c[max_i];
	d[max_i]=d[max_i-1];
	return;
}

//=======================================================
//=======================================================
//=======================================================
void   nmSPLINE_FREE(int n,const double *x,const double *y,double *b,double *c,double *d)
{
	if(n<2) return;
	if(n<3)
	{
		b[0]=(y[1]-y[0])/(x[1]-x[0]);
		c[0]=0.;
		d[0]=0.;
		b[1]=b[0];
		c[1]=0.;
		d[1]=0.;
		return;
	}
//
// build 3-diagonal sceem
// b- diagonal
//
	d[0]=x[1]-x[0];
	c[1]=(y[1]-y[0])/d[0];
	const int max_i=n-1;
	const int nm1=n-1;
	for(int i=1; i<nm1; i++)
	{
		d[i]=x[i+1]-x[i];
		b[i]=2.*(d[i-1]+d[i]);
		c[i+1]=(y[i+1]-y[i])/d[i];
		c[i]=c[i+1]-c[i];
	}
// granichnye usloviya
	b[0]=-d[0];
	b[max_i]=-d[max_i-1];
	c[0]=0.;
	c[max_i]=0.;
//pryamoy hod
	{
		for(int i=1;i<n;i++)
		{
			double t=d[i-1]/b[i-1];
			b[i]=b[i]-t*d[i-1];
			c[i]=c[i]-t*c[i-1];
		}
	}
// obratnaya podstanovka
	c[max_i]=c[max_i]/b[max_i];
	{
		for(int i=n-2; i>=0; i--)
		{
			c[i]=(c[i]-d[i]*c[i+1])/b[i];
		}
	}
//now c[i]=SIGMA(i)
	const int nm2=nm1-1;
	b[max_i]=(y[max_i]-y[nm2])/d[nm2]+d[nm2]*(c[nm2]+2.*c[max_i]);
	{
		for(int i=0; i<nm1; i++)
		{
			b[i]=(y[i+1]-y[i])/d[i]-d[i]*(c[i+1]+2.*c[i]);
			d[i]=(c[i+1]-c[i])/d[i];
			c[i]=3.*c[i];
		}
	}
	c[max_i]=3*c[max_i];
	d[max_i]=d[max_i-1];
	return;
}

//=======================================================
//=======================================================
//=======================================================
double nmSEVAL(int n,double u,const double *x,const double *y,const double *b,const double *c,const double *d)
{
	static int i=0;
	if( i >= (n-1)) i=0;
	if( i < 0 )     i=0;
	if( u <  x[i])   goto lab10;
	if( u <= x[i+1]) goto lab30;
lab10:
	i=0;
	{
	 int j=n;
	 do
	 {
		 int k=(i+j)/2;
		 if(u <  x[k]) j=k;
		 if(u >= x[k]) i=k;
	 } while(j > i+1);
	}
lab30:
	double dx=u-x[i];
	double s = y[i] + dx*(b[i]+dx*(c[i]+dx*d[i]));
return s;
}

//=======================================================
//=======================================================
//=======================================================
double nmGAMMLN(double xx)
{
	static double cof[6]={76.18009173,-86.50532033,24.01409822,
		-1.231739516,0.120858003e-2,-0.536382e-5};

	double x=xx-1.0;
	double tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	double ser=1.0;
	for (int j=0;j<6;j++) 
	{
		x += 1.0;
		ser += cof[j]/x;
	}
	return -tmp+log(2.50662827465*ser);
}
//=======================================================
//=======================================================
//=======================================================
//int SVD(int NM,int M,int N,double **A,double *W,bool MATU,double **U,bool MATV,double **V,double *RV1)
int nmSVD(int NM,int M,int N,double *A,double *W,double *U,double *V,double *RV1)
{
//      double A(NM,N),W(N),U(NM,N),V(NM,N),RV1(N)
//MY      double A(M,N),W(N),U(NM,N),V(N,N),RV1(N)
//
//           
//                              T
//                         A=USV
//        A   M*N.
//          -
//         QR-.
//
//    :
// """""""""""""""""
//     M -    A  U.
//
//     N -    A  U   V.
//
//     NM =< max{ M,N }   -     -
//       ,      
//      .
//
//     A -    ,   -
//        .
//
//     MATU  =.TRUE.      U  .
//     MATU  =.FALSE.    .
//
//     MATV  =.TRUE.      V  .
//     MATV  =.FALSE.    .
//     !!!Yurin change:   MATU  =.FALSE.  
//                        U=NULL
//     !!!Yurin change:   MATV  =.FALSE.  
//                        V=NULL
//
//    :
//   """"""""""""""""
//     A   (      U  V).
//
//     W  N ()   A (-
//         S).   .    
//         ,     IERR+1 -:- IERR + N 
//           .
//
//     U   U (  )  -
//       ,    MATU    .TRUE.  
//         U    . U -
//          A.     ,  -
//        U ,        , 
//          .
//
//     V   U (  )  ,
//           MATV    .TRUE.
//           V   . V  
//          A  U  .  -
//          ,   V ,  
//          ,    .
//
//     IERR=0 -     .
//     IERR=K - K-       30
//           .
//     IERR - !!!!!
//
//     RV1-    .
//
//=====================================================================
    int I,J,K,L,II,I1,KK,LL,L1,MN,ITS;
    double C,F,G,H,S,X,Y,Z,SCALE,ANORM;
//
//#define AA(i,j) A[i-1][j-1]
//#define UU(i,j) U[i-1][j-1]
//#define VV(i,j) V[i-1][j-1]

#define AA(i,j) A[(i-1)*N+j-1]
#define UU(i,j) U[(i-1)*N+j-1]
#define VV(i,j) V[(i-1)*N+j-1]

	if(A!=U)
	{
    for( I=1; I<=M; I++)
		{
      for( J=1; J<=N; J++)
			{
         UU(I,J)=AA(I,J);
			}
		}
	}
//------<      >----
    G=0.;
    SCALE=0.;
    ANORM=0.;
//
		for(I=1; I<=N; I++)
		{
      L=I+1;
      RV1[I-1]=SCALE*G;
      G=0.;
      S=0.;
      SCALE=0.;
      if( I <= M )
			{
				for(K=I; K<=M; K++)
				{
					SCALE=SCALE+ABS(UU(K,I));
				}
				if( SCALE != 0. )
				{
					for(K=I; K<=M; K++)
					{
						UU(K,I)=UU(K,I)/SCALE;
						double t=UU(K,I);
						S=S+t*t;
					}
					F=UU(I,I);
					G=-SIGN(::sqrt(S),F);
					H=F*G-S;
					UU(I,I)=F-G;
					if( I != N )
					{
						for(J=L; J<=N; J++)
						{
							S=0.;
							for(K=I; K<=M; K++)
							{
								S=S+UU(K,I)*UU(K,J);
							}
							F=S/H;
							for(K=I; K<=M; K++)
							{
								UU(K,J)=UU(K,J)+F*UU(K,I);
							}
						}
					}
					for(K=I; K<=M; K++)
					{
						UU(K,I)=SCALE*UU(K,I);
					}
				}
			}
			W[I-1]=SCALE*G;
      G=0.;
      S=0.;
      SCALE=0.;
			if( !(I > M || I == N) ) 
			{
				for(K=L; K<=N; K++)
				{
					SCALE=SCALE+ABS(UU(I,K));
				}
				if( SCALE != 0. )
				{
					for( K=L; K<=N; K++)
					{
						UU(I,K)=UU(I,K)/SCALE;
						double t=UU(I,K);
						S=S+t*t;
					}
					F=UU(I,L);
					G=-SIGN(::sqrt(S),F);
					H=F*G-S;
					UU(I,L)=F-G;
					for( K=L; K<=N; K++)
					{
						RV1[K-1]=UU(I,K)/H;
					}
					if( I != M )
					{
						for( J=L; J<=M; J++)
						{
							S=0.;
							for( K=L; K<=N; K++)
							{
								S=S+UU(J,K)*UU(I,K);
							}
							for( K=L; K<=N; K++)
							{
								UU(J,K)=UU(J,K)+S*RV1[K-1];
							}
						}
					}
					for( K=L; K<=N; K++)
					{
						UU(I,K)=SCALE*UU(I,K);
					}
				}
			}
			ANORM=MAX(ANORM,ABS(W[I-1])+ABS(RV1[I-1]));
		}
//---------<    >--
    if( V ) 
		{//  I=N   -1  1  
      for( II=1; II<=N; II++)
			{
        I=N+1-II;
        if( I != N ) 
				{
					if( G != 0.)
					{
						for( J=L; J<=N; J++)
						{//       
							VV(J,I)=(UU(I,J)/UU(I,L))/G;
						}
						for( J=L; J<=N; J++)
						{
							S=0.;
							for( K=L; K<=N; K++)
							{
								S=S+UU(I,K)*VV(K,J);
							}
							for( K=L; K<=N; K++)
							{
								VV(K,J)=VV(K,J)+S*VV(K,I);
							}
						}
					}
					for( J=L; J<=N; J++)
					{
						VV(I,J)=0.;
						VV(J,I)=0.;
					}
				}
				VV(I,I)=1.;
        G=RV1[I-1];
        L=I;
			}
		}
//--------<    >--
    if( U ) 
		{ // I=min{ N,M }   -1  1  
      MN=N;
      if( M < N ) MN=M;
//
			for( II=1; II<=MN; II++)
			{
				I=MN+1-II;
				L=I+1;
				G=W[I-1];
				if( I != N )
				{
					for( J=L; J<=N; J++)
					{
						 UU(I,J)=0.;
					}
				}
				if( G == 0. ) 
				{
					for( J=I; J<=M; J++)
					{
						UU(J,I)=0.;
					}
				}
				else
				{
					if( I != MN )
					{
						for( J=L; J<=N; J++)
						{
							S=0.;
							for( K=L; K<=M; K++)
							{
								S=S+UU(K,I)*UU(K,J);
							}
				//--------------<       >---
							F=(S/UU(I,I))/G;
							for( K=I; K<=M; K++)
							{
								UU(K,J)=UU(K,J)+F*UU(K,I);
							}
						}
					}
					for( J=I; J<=M; J++)
					{
						UU(J,I)=UU(J,I)/G;
					}
				}
				UU(I,I)=UU(I,I)+1.;
			}
		}
//------<    >---
//--------<  K=N   -1  1  - >---
		for(KK=1; KK<=N ; KK++)
		{
			const int K1=N-KK;
			K=K1+1;
			ITS=0;
//--------<   .  L=K
//            -1  1  >----
L520:

			for(LL=1; LL<=K; LL++)
			{
				L1=K-LL;
				L=L1+1;
				if( ABS(RV1[L-1])+ANORM == ANORM ) goto L565;
				// RV1(0)   ,       . 
				if( ABS(W[L1-1])+ANORM == ANORM ) break;
			}
//  L ,  1,  RV1(L-1)  0.
			C=0.;
			S=1.;
      for(I=L; I<=K; I++)
			{
        F=S*RV1[I-1];
        RV1[I-1]=C*RV1[I-1];
        if( ABS(F)+ANORM == ANORM ) goto L565;
        G=W[I-1];
        H=::sqrt(F*F+G*G);
        W[I-1]=H;
        C=G/H;
        S=-F/H;
        if(  U ) 
				{
					for(J=1; J<=M; J++)
					{
						 Y=UU(J,L1);
						 Z=UU(J,I);
						 UU(J,L1)=Y*C+Z*S;
						 UU(J,I)=-Y*S+Z*C;
					}
				}
			}
//-------<     >------
L565:      
			Z=W[K-1];
      if( L != K )
			{ //        2.
				if( ITS == 30 ) 
				{
					//     -  30  
					//     . 
					return K;//(IERR)
				}
				ITS=ITS+1;
				X=W[L-1];
				Y=W[K1-1];
				G=RV1[K1-1];
				H=RV1[K-1];
				F=((Y-Z)*(Y+Z)+(G-H)*(G+H))/(2.*H*Y);
				G=::sqrt(F*F+1.);
				F=((X-Z)*(X+Z)+H*(Y/(F+SIGN(G,F))-H))/X;
	//-------<  QR- >---
				C=1.;
				S=1.;
				for(I1=L; I1<=K1; I1++)
				{
					I=I1+1;
					G=RV1[I-1];
					Y=W[I-1];
					H=S*G;
					G=C*G;
					Z=::sqrt(F*F+H*H);
					RV1[I1-1]=Z;
					C=F/Z;
					S=H/Z;
					F=X*C+G*S;
					G=-X*S+G*C;
					H=Y*S;
					Y=Y*C;
					if( V ) 
					{
						for( J=1; J<=N; J++)
						{
							X=VV(J,I1);
							Z=VV(J,I);
							VV(J,I1)=X*C+Z*S;
							VV(J,I)=-X*S+Z*C;
						}
					}
					Z=::sqrt(F*F+H*H);
					W[I1-1]=Z;
					//---<    ,  Z   >-- 
					if( Z != 0. ) 
					{
						C=F/Z;
						S=H/Z;
					}
					F=C*G+S*Y;
					X=-S*G+C*Y;
					if( U ) 
					{
						for( J=1; J<=M; J++)
						{
							Y=UU(J,I1);
							Z=UU(J,I);
							UU(J,I1)=Y*C+Z*S;
							UU(J,I) =-Y*S+Z*C;
						}
					}
				}
	//
				RV1[L-1]=0.;
				RV1[K-1]=F;
				W[K-1]=X;
				goto L520;
			}      
//-------<  >----
			if( Z < 0. )
			{// W[K-1]   
				W[K-1]=-Z;
				if( V ) 
				{
					for( J=1; J<=N; J++)
					{
						VV(J,K)=-VV(J,K);
					}
				}
			}
		}
	return 0;//(IERR=0)
}
//=======================================================
//=======================================================
//=======================================================
//
//C   IMSL ROUTINE NAME   - VDCPS
//C   LATEST REVISION     - JANUARY 1, 1978
//C   PURPOSE             - DECOMPOSE AN INTEGER INTO ITS PRIME FACTORS
//C   USAGE               - CALL VDCPS (N,NPF,IPF,IEXP,IPWR)
//C   ARGUMENTS    N      - INTEGER NUMBER TO BE FACTORED. (INPUT)
//C                NPF    - INTEGER NUMBER OF DIFFERENT PRIME FACTORS OF
//C                           THE ABSOLUTE VALUE OF N. (OUTPUT)
//C                           IF N IS -1, 0, OR 1, NPF IS SET TO 0.
//C                IPF    - INTEGER VECTOR OF LENGTH 13. (OUTPUT) IPF(I)
//C                           CONTAINS THE PRIME FACTORS OF THE ABSOLUTE
//C                           VALUE OF N, FOR I=1,2,...,NPF. THE
//C                           REMAINING 13-NPF LOCATIONS OF IPF ARE NOT
//C                           USED.
//C                IEXP   - INTEGER VECTOR OF LENGTH 13. (OUTPUT) IEXP(I)
//C                           CONTAINS THE EXPONENT OF IPF(I), FOR
//C                           I=1,2,...,NPF. THE REMAINING 13-NPF
//C                           LOCATIONS OF IEXP ARE NOT USED.
//C                IPWR   - INTEGER VECTOR OF LENGTH 13. (OUTPUT) IPWR(I)
//C                           CONTAINS THE QUANTITY IPF(I)**IEXP(I), FOR
//C                           I=1,2,...,NPF. THE REMAINING 13-NPF
//C                           LOCATIONS OF IPWR ARE NOT USED.
//C   PRECISION/HARDWARE  - SINGLE/ALL
//C   REQD. IMSL ROUTINES - NONE REQUIRED
//C   NOTATION            - INFORMATION ON SPECIAL NOTATION AND
//C                           CONVENTIONS IS AVAILABLE IN THE MANUAL
//C                           INTRODUCTION OR THROUGH IMSL ROUTINE UHELP
//C   REMARKS      THE OUTPUT FROM VDCPS SHOULD BE INTERPRETED IN THE
//C                FOLLOWING MANNER:
//C                ABS(N) = IPF(1)**IEXP(1) * IPF(2)**IEXP(2) * ...
//C                * IPF(NPF)**IEXP(NPF)
//C   DIMENSION IPF(13),IEXP(13),IPWR(13)
//
int nmVDCPS(int N,int *IPF,int *IEXP,int *IPWR)
{//NPF -return value
	int I = 0;
	int NF = ABS(N);
//If N is -1, 0, or 1, set NPF to 0.
	if(NF <= 1)
		return 0;
	int IFC = 0;
	int ID = 2;
	int IQ;
	while(1)
	{
		IQ = NF/ID;
		if(NF != ID*IQ)
		{	// ID is not a factor of NF. If NF is
			// not prime, continue to factor.
			if(IQ <= ID)
			{// NF is prime or 1
				if(NF > 1)
				{// NF is prime. Update the prime
				 // factor vector, the power vector,
				 // and the exponent vector.
	  		 //I = I+1;//moved bellow
					IPF[I] = NF;
					IPWR[I] = NF;
					IEXP[I] = 1;
					I = I+1;// moved to here because "C" array start at 0;
				}
  			return I;
			}
			// update the divisor
			ID=(ID > 2) ? ID+2 : 3;
		}
		else
		{ //  ID is a factor of NF
			NF = IQ;
			if(ID <= IFC)
			{ // ID factors N more than once. Update
				// the power vector and the exponent vector.
				IPWR[I-1] = ID*IPWR[I-1];
	      IEXP[I-1] = IEXP[I-1]+1;
			}
		  else
		  {	// ID is a new divisor
	      //I = I+1;//moved bellow
				// Update the prime factor vector,
				// the power vector, and set the exponent to 1.
				IPF[I] = ID;
				IPWR[I] = ID;
				IEXP[I] = 1;
				I = I+1;// moved to here because "C" array start at 0;
				IFC = ID;
			}
		}
	}
}

//=======================================================
//=======================================================
//=======================================================
//
//C   IMSL ROUTINE NAME   - VTRAN
//C   LATEST REVISION     - JANUARY 1, 1978
//C   PURPOSE             - TRANSPOSE A RECTANGULAR MATRIX
//C   USAGE               - CALL VTRAN (A,N,M)
//C   ARGUMENTS    A      - REAL MATRIX OF DIMENSION N BY M. ON INPUT A
//C                           CONTAINS THE MATRIX TO BE TRANSPOSED.
//C                           ON OUTPUT A CONTAINS THE TRANSPOSED MATRIX.
//C                N      - INTEGER NUMBER OF ROWS OF A. (INPUT)
//C                M      - INTEGER NUMBER OF COLUMNS OF A. (INPUT)
//C   PRECISION/HARDWARE  - SINGLE AND DOUBLE/H32
//C                       - SINGLE/H36,H48,H60
//C   REQD. IMSL ROUTINES - VDCPS
//C   NOTATION            - INFORMATION ON SPECIAL NOTATION AND
//C                           CONVENTIONS IS AVAILABLE IN THE MANUAL
//C                           INTRODUCTION OR THROUGH IMSL ROUTINE UHELP
//      SUBROUTINE VTRAN  (A,N,M)
//      DIMENSION          A(1),IPF(13),IEXP(13),IPWR(13),ITEMP(13)
//      REAL               A,TEMP,TEMP1
//
void nmVTRAN(double *A,int N,int M)
{
// If a is one column or one row return
//  if(N <= 1 || M <= 1) goto l9005;
  if(N <= 1 || M <= 1) return;
// check to see if a is square
  const int NM = N*M-1;
  if(N == M)
	{// Square matrices are handled separately.
    int K = 2;
    int NP1 = N+1;
    int NM1 = N-1;
		for(int I=N; I<=NM; I+=N)
		{
        int L = K+NM1;
				for(int J=K; J<=I; J++)
				{
          double TEMP = A[J-1];
          A[J-1] = A[L-1];
          A[L-1] = TEMP;
          L = L+N;
				}
        K = K+NP1;
		}
  	goto l9005;
	}
	else
	{
		// A is not square. Transposition of an N by M matrix amounts to
		// replacing the element at vector position I with the element at
		// position N*I (MOD N*M-1). Each subcycle of this permutation is
		// completed in order.
		// Decompose NM into its prime factors.
 		int IPF[13],IEXP[13],IPWR[13],ITEMP[13];
		int NPF=::nmVDCPS(NM,IPF,IEXP,IPWR);
		int i;
		for(i=0; i<NPF; i++)
			ITEMP[i] = 0;
		// Generate divisors of NM less than NM/2.
		int ID = 1;
l25:
		if(ID >= NM/2) goto l9005;
		int J = NM/ID;
		for(i=0; i<NPF; i++)
		{
		  if( ITEMP[i] != IEXP[i] )
				J = (J/IPF[i])*(IPF[i]-1);
		}
		// The starting point of a subcycle is divisible
		// only by ID and must not appear in any other subcycle.
		int ISTART = ID;
l35:
		int NMIS = NM-ISTART;
		if(ISTART != ID)
		{
		  int ISD = ISTART/ID;
			for(i=0; i<NPF; i++)
			{
		    if(ITEMP[i] == IEXP[i]) break;
		    if((ISD%IPF[i]) == 0) goto l70;
			}
		  int IT = ISTART;
			do
			{
		   	IT = (N*IT)%NM;
		    if(IT < ISTART || IT > NMIS) goto l70;
		  } while(IT > ISTART && IT < NMIS);
		}
		{
			double TEMP = A[ISTART];
			double TEMP1 = A[NMIS];
			int K = ISTART;
			while(1)
			{
				int KK = (N*K)%NM;
				int L = NM-K;
				int LL = NM-KK;
				J = J-2;
				// Move two elements. The second from
				// the negative subcycle. Check to
				// see if the subcycle is complete.
				if(KK == ISTART)
				{// Subcycle complete. Recompute ID.
		   		A[K] = TEMP;
					A[L] = TEMP1;
					goto l70;
				}
				else if(LL == ISTART)
				{// Subcycle complete. Recompute ID.
					A[K] = TEMP1;
					A[L] = TEMP;
					goto l70;
				}
				else
				{// Subcycle not complete. Move elements and update pointers.
					A[K] = A[KK];
					A[L] = A[LL];
					K = KK;
				}
			}
		}
// Subcycle complete. Recompute ID.
l70:
		ISTART = ISTART+ID;
		if(J > 0) goto l35;
		for(i=0; i<NPF; i++)
		{
			if(ITEMP[i] != IEXP[i])
			{
		    ITEMP[i] = ITEMP[i]+1;
		    ID = ID*IPF[i];
		    goto l25;
			}
		  ITEMP[i] = 0;
		  ID = ID/IPWR[i];
		}
	}
l9005:
	;
}

