#include <math.h>
#include "nummet.h"
//#define DEBUG_NUMERIC

double nmNormQRev(double p)
{
//"   "   .,..,
// ., 1979, 832 .
// .729,  26.2.23.
//    x,  Q(x)=p, 0<p<=0.5
//  +/-4.5e-4;
	static double c[3]={2.515517,0.802853,0.010328};
	static double d[4]={1.0,1.432788,0.189269,0.001308};

	if(p<=0 || p>0.5)
		throw "NormQRev: bad argument p, 0<p<=0.5";

	double t=::sqrt(-2*::log(p));
//	double t=::sqrt(::log(1./p/p));
	double x=t-::nmPoly(t,c,2)/::nmPoly(t,d,3);
	return x;
}

double nmTStudentPercPtAssimpt(double p,int n)
{
//"   "   .,..,
// ., 1979, 832 .
// .742,  26.7.5.
// t- :     .
//    n>20,    p.
	static double gc1[2]={  1./4.,    1./4.};
	static double gc2[3]={  3./96.,  16./96.,  5./96.};
	static double gc3[4]={-15./384., 17./384.,19./384.,   3./384.};
	static double gc4[5]={-945./92160., -1920./92160., 1482./92160., 776./92160.,79./92160.};

	double xp=::nmNormQRev((1-p)/2);
	double x2=xp*xp;
	double cc[5];
	cc[0]=1.;
	cc[1]=::nmPoly(x2,gc1,1);
	cc[2]=::nmPoly(x2,gc2,2);
	cc[3]=::nmPoly(x2,gc3,3);
	cc[4]=::nmPoly(x2,gc4,4);
	double tt=1./double(n);
	double t=::nmPoly(tt,cc,4);
	//double t=1+(::nmPoly(x2,gc1,1)+(::nmPoly(x2,gc2,2)+(::nmPoly(x2,gc3,3)+::nmPoly(x2,gc4,4)/n)/n)/n)/n;

	return t*xp;
}

double nmBetaFnForStudent(double n)
{ // B(1/2,n/2)==(1/2)(n/2)/((n+1)/2)
//    :
//"   "   .,..,
// ., 1979, 832 .
// .81,  6.2.2, 6.1.15., 6.1.8.  6.1.9. ( B(,)  ()-. (z+1)=z! )
//
	double b=1;
	while((n-2)>0)
	{
		n-=2;
		b*=n;
		b/=n+1;
	}
	if(n==1)
	{
		//b*=M_PI;
		double pi=2*acos(0.);
		b*=pi;
	}
	else //n==2
	{
		b*=2;
	}
	return b;
}
//////////////////////////////////////////////////////
// Student's percent points via analitical integration
// and numerical solving equation
//
///////////////////////////////////////////////////////
static double IntOnePlusXSquarePowerMinusN(double x,int n)
{
//    
//
//                1      n  
//           (-------- )
//            1 + X*X
//
// .., .., ... 
//   .  .
// ,,  1981. 800 .
//  .41.  1.2.10.12.  1.2.10.14.
//
	if(n<1)
		return 0; //Invalid argument
	else if(n==1)
		return atan(x);
	else
	{
		int nn=n-1;
		return (x/::pow(1+x*x,nn)+(2*n-3)*IntOnePlusXSquarePowerMinusN(x,n-1))/(2.*nn);
	}
}

static double IntOnePlusXSquarePowerMinusNAndAHalf(double x,int n)
{
//    
//
//                1      n+1/2  
//           (-------- )
//            1 + X*X
//
// .., .., ... 
//   .  .
// ,,  1981. 800 .
//  .92.  1.2.43.9. , 1.2.43.13.  1.2.43.17.
//
	if(n<0)
		return 0; //Invalid argument
	else if(n==0)
	{
		double v=x+::sqrt(1+x*x);
		return ::log(ABS(v));
	}
	else if(n==1)
		return x/::sqrt(1+x*x);
	else
	{
		int nn=2*n-1;
		return (x/::pow(1+x*x,n-0.5)+(nn-1)*IntOnePlusXSquarePowerMinusNAndAHalf(x,n-1))/nn;
	}
}


struct _T_Stud_Param
{
	int n;
	double p;
	double right;
};
static double pdfStud(double t,void *param)
{
//  :
//    :
//"   "   .,..,
// ., 1979, 832 .
// .742,  26.7.2. (    . .      .778.)
//
	int n=((_T_Stud_Param*)param)->n;
	double p=((_T_Stud_Param*)param)->p;

	double nsq=::sqrt(double(n));
	double x=t/nsq;
	double y;
	if(n%2)
	{//n-
		y=IntOnePlusXSquarePowerMinusN(x,(n+1)/2);
	}
	else
	{//n-
		y=IntOnePlusXSquarePowerMinusNAndAHalf(x,(n+1)/2);
	}
	//double B=exp(-::gammln((n+1.)/2.)+::gammln(n/2.)+::gammln(1./2.));
	//double B=BetaFnForStudent(n);
	return ((_T_Stud_Param*)param)->right - y;
}

double nmTStudentPercPtNlRoot(double p,int n)
{
// t- :  .
//    n<20,    p.
//    :
//"   "   .,..,
// ., 1979, 832 .
// .742,  26.7.2. (    .
// .      .778.).   ( )
//      ZEROIN. .    numeric0.cpp
	_T_Stud_Param param;
	{
		param.n=n;
		param.p=p;
		double B=nmBetaFnForStudent(n);
		param.right=B*p/2;
	}
	double A=0.;
	double B=1.;
	for(int i=0; i<100; i++)
	{
		double v=pdfStud(B,&param);
		if(v<0)
			break;
		A=B;
		B*=8.;
	}
	double tol=1.0e-10;
	double z=::nmZEROIN(pdfStud,A,B,&param,tol);
	return z;
}

double nmTStudentPercPt(double p,int n)
{//n>=1; 0<p<1
	if(n>30)
		return ::nmTStudentPercPtAssimpt(p,n);
	else
		return ::nmTStudentPercPtNlRoot(p,n);
}

/////////////////////////////////////////////////////////////////
// Stdent's percent points via direct  integration and  
// solving nonlinear equation
//
/////////////////////////////////////////////////////////////////
struct _T_Stud_Param2
{
	double n;
	double p;
	double right;
};

static double pdfStudDistr(double x,void *n_double)
{
	double n=*((double*)n_double);
	return ::pow(1+x*x/n,-(n+1.)/2.);
}

static double pdfStud2(double t,void *param)
{
	double n=((_T_Stud_Param2*)param)->n;
	double p=((_T_Stud_Param2*)param)->p;
	double right=((_T_Stud_Param2*)param)->right;

	double ERREST;
	int NOFUN;
	double FLAG;
	double y=nmQUANC8(pdfStudDistr,0,t,&n,1.0e-10,1.0e-20,ERREST,NOFUN,FLAG);

#ifdef DEBUG_NUMERIC
	if(FLAG!=0)
		printf("n=%lf p=%lf   NOFUN=%d   ERREST=%lf  FLAG=%lf\n",n,p,NOFUN,ERREST,FLAG);
#endif

	return right-y;
}


double nmTStudentPercPtViaIntegral(double p,double n)
{
	_T_Stud_Param2 param2;
	{
		param2.n=n;
		param2.p=p;
		//double B=BetaFnForStudent(n);
		double B=exp(-::nmGAMMLN((n+1.)/2.)+::nmGAMMLN(n/2.)+::nmGAMMLN(1./2.));
		param2.right=B*p*::sqrt(n)/2;
	}
	double A=0.;
	double B=1.;
	for(int i=0; i<100; i++)
	{
		double v=pdfStud2(B,&param2);
		if(v<0)
			break;
		A=B;
		B*=8.;
	}
	double tol=1.0e-10;
	double z=::nmZEROIN(pdfStud2,A,B,&param2,tol);
	return z;
}

/////////////////////////////////////////////////////////////////
double nmTStudentPercPtTauThreshold(double p,int n)
{
	double t=nmTStudentPercPt(p,n-2);	
	double tau=t*::sqrt((n-1.)/(n-2+t*t));
	return tau;
}

