		template<typename T,unsigned D> class Gauss_;
		///  ( )
		template<typename T> class Gauss_<T,0>
		{
			const T s2;
		public:
			Gauss_(T sigma) : s2(2*sigma*sigma) {}
			template<class TT> T operator()(TT x) const
			{
				return exp(-(x*x)/s2);
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,1>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma){}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T(-x*exp(-t/2.));
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,2>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma){}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T((t-1)*exp(-t/2.));
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,3>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma) {}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T(x*(3-t)*exp(-t/2.));
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,4>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma) {}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T((3+(t-6)*t)*exp(-t/2.));
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,5>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma) {}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T(-x*((t-10)*t+15)*exp(-t/2.));
				
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,6>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma) {}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T((((t-15)*t+45)*t-15)*exp(-t/2.));
			}
		};
		/*!
		 * \brief   (Hann).   
		 * \f$ w(x)= { 1 \over 2 } [ 1 - cos( { 2 \pi x \over size - 1 ) }] \f$
		 *  0 <= x <= size-1
		 *
		 *   \f$ c = { 2 \pi  \over size - 1 }  \f$ ,    
		 *   : \f[ w(x)= { 1 \over 2 }  [1-cos( c x )] \f]
		 *
		 *   : \n
		 * \f$ w'(x) = { d w(x) \over d x } = + { 1 \over 2 }  c  sin( c x ) \f$ \n
		 * \f$ w''(x)     = + { 1 \over 2 }  c^2 cos( c x )  \f$ \n
		 * \f$ w^{(3)}(x) = - { 1 \over 2 }  c^3 sin( c x )  \f$ \n
		 * \f$ w^{(4)}(x) = - { 1 \over 2 }  c^4 cos( c x )  \f$ \n
		 * \f$ w^{(5)}(x) = + { 1 \over 2 }  c^5 sin( c x )  \f$ \n
		 * :\n
		 *	\f[ w^{(n)}(x) = { 1 \over 2 } c^n  (-1)^{ [ n/2 ] } \left\lbrace   \begin{array}{cc}
		 *		sin( c x ) , \qquad if \quad n \quad is \quad odd \\ 
		 *		cos( c x ) , \qquad if \quad n \quad is \quad even
		 *	\end{array} \right. \f] 
		 *
		 *   \f[ dsin(x,n)= { d^n sin \over d x^n}(x) = (-1)^{ [ (n+1)/2 ] } \left\lbrace   \begin{array}{cc}
		 *		sin( x ) , \qquad if \quad n \quad is \quad even \\ 
		 *		cos( x ) , \qquad if \quad n \quad is \quad odd
		 *	\end{array} \right. \f] 
		 *  \f$ { [ ... ] } \f$         . 
		 * \see Vtl::Alg::dsin
		 * 
		 *   :
		 *	\f[ w^{(n)}(x) = { 1 \over 2 } c^n  dsin<n-1>( c x )  \f] 
		 *
		 *
		 *****************************************************************************/
		template<typename T,unsigned D> class Hann_
			: public Traits<T,1,D,D,0>
		{
			const T c;
		public:
			static const char * Name() { return "Hann"; }
			template<class TT> 
			Hann_(TT size_) : Traits(T(size_)), c(T(2*M_PI/(size_-1))) {}
			template<class TT> 
			Hann_(TT size_, const Hann_& ) : Traits(T(size_)), c(T(2*M_PI/(size_-1))) {}
			template<class TT> T operator()(TT x) const
			{
				return result( T(x), Loki::Int2Type<(D==0)>() );
			}
		private:
			T result(T x,Loki::Int2Type<true>) const
			{
				return T(0.5*(1-cos(c*x))); 
			}
			T result(T x,Loki::Int2Type<false>) const
			{
				return T(0.5*Vtl::Alg::x_to_N<D>::value(c)*Vtl::Alg::dsin<D-1>(c*x));;
			}
		};
		/*!
		 * \brief  e (Hamming).   
		 * \f$ w(x)=   0.54 - 0.46 cos( { 2 \pi x \over size - 1 ) } \f$ 
		 *  0 <= x <= size-1
		 *
		 *   \f$ c = { 2 \pi  \over size - 1 }  \f$ ,    
		 *   : \f[ w(x)= 0.54 - 0.46  cos( c x ) \f]
		 *
		 *   : \n
		 * \f$ w'(x) = { d w(x) \over d x } = 0.46  c sin( c x  )  \f$ \n
		 * \f$ w''(x)     = + 0.46  c^2 cos( c x  )  \f$ \n
		 * \f$ w^{(3)}(x) = - 0.46  c^3 sin( c x  )  \f$ \n
		 * \f$ w^{(4)}(x) = - 0.46  c^4 cos( c x  )  \f$ \n
		 * \f$ w^{(5)}(x) = + 0.46  c^5 sin( c x  )  \f$ \n
		 *
		 *   \f[ dsin(x,n)= { d^n sin \over d x^n}(x) = (-1)^{ [ (n+1)/2 ] } \left\lbrace   \begin{array}{cc}
		 *		sin( x ) , \qquad if \quad n \quad is \quad even \\ 
		 *		cos( x ) , \qquad if \quad n \quad is \quad odd
		 *	\end{array} \right. \f] 
		 *  \f$ { [ ... ] } \f$         , 
		 * \see Vtl::Alg::dsin
		 *  n >=1   :
		 *	\f[ w^{(n)}(x) = 0.46 c^n dsin<n-1>( c x )  \f] 
		 *
		 *
		 *****************************************************************************/
		template<typename T,unsigned D> class Hamming_
			: public Traits<T,1,D,D,0>
		{
			const T c;
		public:
			static const char * Name() { return "Hamming"; }
			template<class TT> 
			Hamming_(TT size_) : Traits(T(size_)), c(T(2*M_PI/(size_-1))) {}
			template<class TT> 
			Hamming_(TT size_, const Hamming_& ) : Traits(T(size_)), c(T(2*M_PI/(size_-1))) {}
			template<class TT> T operator()(TT x) const
			{
				return result( T(x), Loki::Int2Type<(D==0)>() );
			}
		private:
			T result(T x,Loki::Int2Type<true>) const
			{
				return T(0.54-0.46*cos(c*x));
			}
			T result(T x,Loki::Int2Type<false>) const
			{
				return T(0.46 * Vtl::Alg::x_to_N<D>::value(c) * Vtl::Alg::dsin<D-1>(c*x));
			}
		};
