#ifndef _PLF_PORT_H
#define _PLF_PORT_H
/////////////////////////////////////////////////////////////////////////////
#include <list>
#include "plf_strip_buffer.h"
#include "plf_functors.h"

namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
/** @addtogroup PortsAndBuffersDoxGroup  Ports and buffers 
 *  @ingroup PipeLineFiltersDoxGroup  
 *  @{
 */
/** @page PortAndBufDoxPage
	
	   :

	1)  (InPort)

	2)  (OutPortNoPostProc)

	OutPortNoPostProc -   (StripBuffer). StripBuffer -   
	 (  ,        , 
	      -    , 
	   ).      OutPortNoPostProc, 
	     , ..        
	         .
	InPort -    ,    OutPortNoPostProc      
	. 

	  :
    -     ; 
    -              ; 
    -   , : 
         - OutPortNoPostProc     StripBuffer (  ) 
         - InPort -     OutPortNoPostProc,   . 

	 :

	OutPortNoPostProc:

    -      (     InPort); 
    -    (       StripBuffer,   Y); 
    -       InPort (    \n
         -    ); 
    -   (,    )       ; 
    -     ( ). 
	
	InPort:

    -     (   OutPortNoPostProc);
    -     ,..    Y     \n
       Y-Extent.before  Y+Extent.forward  ( ); 
    -     Y       ;
    -   ( ,     OutPortNoPostProc \n
        )       ; 
    -     ( ). 

*/
	class FilterBase;
	class FilterElementary;
	class FilterComplex;

	class Port;
	class InPortBase;
	class OutPortBase;
	template<typename T> class InPort;
	template<typename T> class OutPortNoPostProc;
/////////////////////////////////////////////////////////////////////////////////
/// class Port
/// @brief     "\b ".  ,
///    \b  .
/////////////////////////////////////////////////////////////////////////////////	
	class Port
	{
	private:
		FilterElementary * filter; //> ,   
		int port; //>      (    )	
	private:
		Port(const Port&);//!>forbidden
		Port& operator=(const Port&);//!>forbidden
	protected:
		Port()
			:filter(0),port(0)
		{
		}
		void _init(FilterElementary * _filter,int _port)
		{
			filter=_filter;
			port=_port;
		}

	public:
		virtual ~Port() {}
		virtual FilterElementary * getFilter() {return filter;} //>   ,     
		virtual int getPort() {return port;} //>    ,     ,      .
		virtual void print()
		{
			printf("class Port: this=0x%08X  Filter=0x%08X port=%d \n",unsigned(this),unsigned(filter),port);
		}
	    ///  ,       , :
	    /// OutPortNoPostProc     StripBuffer (  )
		/// InPort -     OutPortNoPostProc,   .
		virtual ImgSize  getImageSize() const=0; 
 	};
/////////////////////////////////////////////////////////////////////////////////
/// class InPortBase
/// @brief     "\b  ".  ,
///    \b  .
/////////////////////////////////////////////////////////////////////////////////
	class InPortBase
		: public Port
	{
	public:
		friend 	class OutPortBase;
		int total_delay;
	private:
		InPortBase(const InPortBase&);//!>forbidden
		InPortBase& operator=(const InPortBase&);//!>forbidden
	private:
		Extent ext_required;
		///       y_first   y_first+lines,
		///    y_first+ext_required.before
		Strip strip; 
		bool finished;
	protected:
		InPortBase()
			:Port()
		{
			total_delay=0;
		}
		void _init(FilterElementary * _filter,int _port,Extent extent_required)
		{
			Port::_init(_filter,_port);
			ext_required=extent_required;
			//lines=ext_required.before+ext_required.forward;
			Rewind();
		}
	public:
		virtual void Init(FilterElementary * _filter,int _port,Extent extent_required)=0;
		virtual ~InPortBase() {}
		Extent ext() const { return ext_required;}
//		virtual ImgSize  getImageSize() const=0; 
		void Rewind()
		{
			int y=0;
			//     
			//forward>=before,     
			//  ,    ,
			//   ,    
			//       
			//    . .   
			// StripBuffer::init_2(const Extent& ext);
			int forw=std::max(ext_required.before,ext_required.forward);
			strip=Strip(y-ext_required.before,y+forw+1);
			finished=false;
		}
		void ToNextLine(int y)
		{
			if(y>=Y())
			{
				int delta=y-Y();
				strip+=delta;
			}
			if(Y() >= getImageSize().height)
				Finish();
		}
		bool isFinished() const {return finished; }
		void Finish()
		{
			finished=true;
		}
		virtual bool isDataReady()=0;
		int Y() const {return strip.last()+ext_required.before; }
		Strip getRequiredStrip() { return strip; }

/////////////
		virtual OutPortBase * BaseConnectedTo()=0; 
	};
/////////////////////////////////////////////////////////////////////////////////
/// class OutPortBase
/// @brief     "\b  ".  ,
///    \b  .
/////////////////////////////////////////////////////////////////////////////////
	class OutPortBase
		: public Port
	{
	private:
		OutPortBase(const OutPortBase&);//!>forbidden
		OutPortBase& operator=(const OutPortBase&);//!>forbidden
	public:
		typedef std::list<InPortBase*> List;
		struct Connection
		{
			FilterBase* filter; 
			int i_in;
			Connection(FilterBase* filter_,int i_in_) : filter(filter_),i_in(i_in_) {}
		};
	protected:
		OutPortBase()
			:Port()
		{
		}
		void ConnectTo_(InPortBase *in_port)
		{
			connected_to.push_back(in_port);
		}
	private:
		List connected_to; //!> ,       
		bool finished;
	public:
		virtual void Init(FilterElementary * _filter,int _port)
		{
			Port::_init(_filter,_port);
		}
		virtual ~OutPortBase() 
		{
		}

		virtual void Rewind()
		{
			finished=false;
		}
		virtual int this_delay()=0;
		Extent calcRequiredExtent()
		{
			Extent e(0,0,0,0);
			for(List::iterator it=connected_to.begin();it!=connected_to.end();it++)
			{
				InPortBase* in=(*it);
				e.max(in->ext());
			}
			return e;
		}
		///   4&5  6,       .
		virtual void set_total_delay(int before_delay)
		{
			before_delay+=this_delay();
			for(List::const_iterator it=connected_to.begin(); it!=connected_to.end(); ++it)
			{
				InPortBase *in=*it;
				in->total_delay=before_delay;
			}
		}
		virtual int NumConnectedTo() const {return connected_to.size();}

		void putConnectionsInQueue(std::list<Vtl::Plf::FilterElementary *>& Q)
		{
			for(List::const_iterator it=connected_to.begin(); it!=connected_to.end(); ++it)
			{
				InPortBase *in=*it;
				Q.push_back(in->getFilter());
			}
		}
		void putConnectionsInQueue(Fifo<Vtl::Plf::FilterElementary*>& Q)
		{
			for(List::const_iterator it=connected_to.begin(); it!=connected_to.end(); ++it)
			{
				InPortBase *in=*it;
				Q.put(in->getFilter());
			}
		}
		virtual Strip getRequestedStrips() 
		{
			Strip s;
			int cnt=0;
			for(List::const_iterator it=connected_to.begin(); it!=connected_to.end(); ++it)
			{
				InPortBase *in=*it;
				if( !in->isFinished() )
				{
					if(cnt)
						s|=in->getRequiredStrip();
					else
						s=in->getRequiredStrip();
					cnt++;
				}
			}
			if(cnt==0)
				Finish();
			return s;
		}
		///    ?  ,         , 
		///      .     - ,   ,  
		///    .
		bool isFinished() const {return finished; }
		///     ,   ,
		///     .   - ,   ,
		///          \ getImageSize(),
		///       FilterBase::processLine()   
		/// 0  1      .  -1,  0    
		///     ,     ,
		/// ,          , 
		///   FilterBase::processLine()      1 , 
		///     -    .
		void Finish()
		{
			finished=true;
		}
		///   .     true     
		/// -  FilterBase::processLine()   .
		bool isReadyForUpdate()
		{// y_first  y_first+lines      
			if(isFinished())
				return false;
			Strip s=getRequestedStrips();
			if(isFinished())
				return false;
			bool can=getStrip().canMove(s);
			return can;
		}
		virtual bool inPopTailStage()=0;
		virtual void PopTail()=0;

		virtual Strip getStrip() const =0;

		virtual void set_bufferImgSize(const ImgSize& ims)=0;
		virtual void init_bufferExtents()=0;
		virtual void init_buffer()=0;
		virtual void init_buffer(Extent ext)=0;

		virtual void increase_delay_by(int delta_delay)=0;
	};
/////////////////////////////////////////////////////////////////////////////////
/// class InPort
/// @brief     "\  "  , 
///   FilterBase   PipeLineFiltering.
///
///         
///	    -      -
/// . ,       
///	           
///	       .
/////////////////////////////////////////////////////////////////////////////////
	template<typename T> 
	class InPort
		: public InPortBase
	{
	public:
		friend class OutPortNoPostProc<T>;
		typedef StripBuffer<T> BufType; 
	private:
		OutPortNoPostProc<T> * input;//    
	private:
		InPort(const InPort&);//!>forbidden
		InPort& operator=(const InPort&);//!>forbidden
	public:
		InPort()
			: InPortBase()
		{
		}
		virtual void Init(FilterElementary * _filter,int _port,Extent extent_required)
		{
			InPortBase::_init(_filter,_port,extent_required);
		}
		virtual ~InPort()
		{
		}
		OutPortNoPostProc<T> * ConnectedTo() 
		{ 
			return input;
		}
		virtual OutPortBase * BaseConnectedTo() 
		{
			return ConnectedTo(); 
		}

		ImageRef<T> getImg() 
		{
			assert(input!=0);
			return input->getImg(ext(),getRequiredStrip());
		}
		virtual ImgSize  getImageSize() const 
		{
			assert(input!=0);
			return input->getImageSize(); 
		}
		virtual bool isDataReady()
		{// y_first  y_first+lines      
			Vtl::Plf::Strip s=BaseConnectedTo()->getStrip();
			Vtl::Plf::Strip s_req=getRequiredStrip();
			bool result=s.isContains(s_req);
			return result; 
//			return BaseConnectedTo()->getStrip().isContains(getRequiredStrip());
		}
	};
/////////////////////////////////////////////////////////////////////////////////
/// class InPort
/// @brief 
///	
/////////////////////////////////////////////////////////////////////////////////
	template<typename T> 
	class InPortComb
//		: public InPortBase
	{
	public:
		friend class OutPortNoPostProc<T>;
		typedef StripBuffer<T> BufType; 
		typedef std::list<InPort<T>*> List; 
	private:
//		OutPortNoPostProc<T> * input;//    
	public:
		List inps; 
	private:
		InPortComb(const InPortComb&);//!>forbidden
		InPortComb& operator=(const InPortComb&);//!>forbidden
	public:
		InPortComb()
//			: InPortBase()
		{
		}
//		virtual void Init(FilterElementary * _filter,int _port,Extent extent_required)
//		{
//			InPortBase::_init(_filter,_port,extent_required);
//		}
		void Clear()
		{
			inps.clear();
		}
		void Add(InPort<T>* inp)
		{
			assert(inp!=0);
			inps.push_back(inp);
		}
		void Add(InPortComb<T>* inp)
		{
			assert(inp!=0);
			for(List::const_iterator it=inp->inps.begin(); it!=inp->inps.end(); ++it)
				inps.push_back(*it);
		}
		virtual ~InPortComb()
		{
		}
/*		OutPortNoPostProc<T> * ConnectedTo() 
		{ 
			return input;
		}
		virtual OutPortBase * BaseConnectedTo() 
		{
			return ConnectedTo(); 
		}

		ImageRef<T> getImg() 
		{
			assert(input!=0);
			return input->getImg(ext(),getRequiredStrip());
		}
		virtual ImgSize  getImageSize() const 
		{
			assert(input!=0);
			return input->getImageSize(); 
		}
		virtual bool isDataReady()
		{// y_first  y_first+lines      
			Vtl::Plf::Strip s=BaseConnectedTo()->getStrip();
			bool result=true;
			for()
			InPort<T> *in;
			Vtl::Plf::Strip s_req=in->getRequiredStrip();
			result=result && s.isContains(s_req);
			return result; 
//			return BaseConnectedTo()->getStrip().isContains(getRequiredStrip());
		}
*/
	};
/////////////////////////////////////////////////////////////////////////////////
/// class OutPortNoPostProc
/// @brief      \ " "  , 
///   FilterBase   PipeLineFiltering.
///
///         
///	    -      -
/// . ,       
///	           
///	       .
/////////////////////////////////////////////////////////////////////////////////
	template<typename T> 
	class OutPortNoPostProc
		: public OutPortBase
	{
		friend class InPort<T>;
	public:
		typedef StripBuffer<T> BufType; 
		typedef typename BufType::Iter_Line Iter;
	private:
		BufType buffer; //!> ,     
//======    =======
	private:
		OutPortNoPostProc(const OutPortNoPostProc&);//!>forbidden
		OutPortNoPostProc& operator=(const OutPortNoPostProc&);//!>forbidden
	protected:
		OutPortNoPostProc()
			:OutPortBase()
		{
		}
	public:
		virtual ~OutPortNoPostProc() 
		{
		}
		///   1  6
		virtual void Init(FilterElementary * _filter,int _port)
		{
			OutPortBase::Init(_filter,_port);
		}
//======  =======
		///   2  6
		virtual void set_bufferImgSize(const ImgSize& ims)
		{
			buffer.init_1(ims);
		}
		///   3  6
		virtual void init_bufferExtents()
		{
			Extent ext=this->calcRequiredExtent();
			buffer.init_2(ext);
			buffer.init_3(0);
		}
		///   4&5  6,       .
		virtual void increase_delay_by(int delta_delay)
		{
			buffer.init_3(delta_delay);
		}
		///   6  6
		virtual void init_buffer()
		{
			buffer.init_4();
			Rewind();
		}
		// (!!!)  4&5&6
		virtual void init_buffer(Extent ext)
		{
			buffer.init_2(ext);
			//buffer.init_3(0);
			//buffer.init_4();
			//Rewind();
		}
		virtual int this_delay()
		{
			return this->buffer.delay();
		}
//======       =======
		void ConnectTo(InPort<T> *in_port)
		{
			assert(in_port!=0);
			in_port->input=this;
			ConnectTo_(in_port);
		}
		void ConnectTo(InPortComb<T> *in_ports)
		{
			assert(in_ports!=0);
			//std::list<InPort<T>*> inps; 
			for(std::list<InPort<T>*>::iterator it=in_ports->inps.begin(); it!=in_ports->inps.end(); ++it )
			{
				InPort<T> *in_port=*it;
				in_port->input=this;
				ConnectTo_(in_port);
			}
		}
//======    =======
		virtual void Rewind()
		{
			OutPortBase::Rewind();
			buffer.Rewind();
		}
		/// @brief  ,     
		///  OutPortNoPostProc    .   >=0. 
		///  Y_wait()>=getImageSize().height ,     .
		int Y_wait() const {return buffer.Y_wait(); }
		///     . 
		int Y() const {return buffer.Y(); }
		///  ,       .
		virtual Strip getStrip() const 
		{
			//return Strip(buffer.Y_last(),buffer.Y_wait());
			return buffer.getCurStrip();
		}
//======    =======
		virtual ImgSize  getImageSize() const {return buffer.ImageSize(); }
		ImageRef<T> getImg(const Extent& ex1,const Strip& strip) const {return buffer.getImg(ex1,strip); }

//======     =======
		/// @brief       Y_wait(),    .
		///     beginUpdate()  endUpdate()
		typename Iter begin() { return buffer.begin_line_insert(); }
		/// @brief          Y_wait(),    .
		///     beginUpdate()  endUpdate()
		typename Iter end() { return buffer.end_line_insert(); }

		virtual void beginUpdate()
		{
			assert(begin()==0);
			assert(buffer.Y_wait()<buffer.ImageSize().height);
			buffer.InsertNextLineStart();
		}
		virtual void endUpdate()
		{
			assert(begin()!=0);
			buffer.InsertNextLineFinished();
			if(buffer.Y_wait()>=buffer.ImageSize().height)
			{
				Finish();
			}
		}
		virtual bool inPopTailStage()
		{
			Strip s=getRequestedStrips();
			return ((buffer.Y_wait() >= buffer.ImageSize().height )
				&& (buffer.Y() < (buffer.ImageSize().height-1)) 
				&& getStrip().canMove(s));
		}
		virtual void PopTail()
		{
				buffer.PopTail();
		}

	};
/////////////////////////////////////////////////////////////////////////////////
/// class OutPort
/// @brief     \ " "  , 
///   FilterBase   PipeLineFiltering.   
///    OutPortNoPostProc.
///
///         
///	    -      -
/// . ,       
///	           
///	       .
/////////////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> > 
	class OutPort
		: public OutPortNoPostProc<T>
	{
		P functor_post_proc;
//======    =======
	private:
		OutPort(const OutPort&);//!>forbidden
		OutPort& operator=(const OutPort&);//!>forbidden
	public:
		OutPort()
			:OutPortNoPostProc<T>()
		{
		}
		OutPort(P func_post_proc)
			:OutPortNoPostProc<T>(),functor_post_proc(func_post_proc)
		{
		}
		void setPostProc(P func_post_proc)
		{
			functor_post_proc=func_post_proc;
		}
		virtual ~OutPort() 
		{
		}
		virtual void endUpdate()
		{
			assert(begin()!=0);
			proc(functor_post_proc);
			OutPortNoPostProc<T>::endUpdate();
		}
	private:
		template<class P> void proc(P functor)
		{
			for(Iter it=begin(); it!=end(); ++it)
				(*it)=functor(*it);
		}
		void proc(Vtl::Plf::F11::DoNothing<T> )
		{
		}
	};
/** @} */ // end of PipeLineFiltersDoxGroup      . 
} //namespace Plf
} //namespace Vtl
#endif //_PLF_PORT_H
