#ifndef _PLF_STRIP_BUFFER_SUPPORT_H
#define _PLF_STRIP_BUFFER_SUPPORT_H
//////////////////////////////////////////////////////////////////////////////////
#include <sstream>
#include <algorithm>
#include <list>
namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
/** @addtogroup PipeLineFiltersDoxGroup  Pipe Line Filters. 
 *  @{
 */
/** @} */ // end of PipeLineFiltersDoxGroup  
/** @addtogroup PortsAndBuffersDoxGroup  Ports and buffers 
 *  @ingroup PipeLineFiltersDoxGroup  
 *  @{
 */
	///////////////////////////////////////////////////////////////////////////
	/// struct ImgSize
	/// @brief       -   
	///
	///
	///////////////////////////////////////////////////////////////////////////
	struct ImgSize
	{
		int width;//>  
		int height;//>  
		ImgSize() : width(0),height(0) {}
		ImgSize(int w,int h): width(w),height(h) {}
		///   ,       ( ).
		bool operator!=(const ImgSize& is) 
		{
			return (width!=is.width) || (height!=is.height);
		}
		///   ,       ( ).
		bool operator==(const ImgSize& is) 
		{
			return !((*this)!=is);
		}
		///   : (W=640,H=480) 
		friend std::ostream& operator<<(std::ostream& out,const ImgSize& is)
		{
			return out<< "(W=" << is.width << ",H=" <<is.height<<")";
		}
	};
	///////////////////////////////////////////////////////////////////////////
	/// struct Extent
	/// @brief           (  )
	///
	///         
	///    ,      
	///          ,   
	///       ,     
	///      (      
	///    -  ).
	///////////////////////////////////////////////////////////////////////////
	struct Extent
	{
		int left; //>            
		int right;//>            
		int before;//>          
		int forward;//>          
		/// _left -            ;\n
		/// _right -             ;\n
		/// _before -           ;\n
		/// _forward -          ;\n
		Extent(int _left,int _right,int _before,int _forward) : left(_left),right(_right),before(_before),forward(_forward) {}
		/// 4  left,right,before,forward     radius.
		explicit Extent(int radius=0) : left(radius),right(radius),before(radius),forward(radius) {}
		Extent& max(const Extent& e)
		{
			left=std::max(left,e.left);
			right=std::max(right,e.right);
			before=std::max(before,e.before);
			forward=std::max(forward,e.forward);
			return *this;
		}
		void checkCompatible(const ImgSize& is) const
		{
			check(left,is.width,"left","width");
			check(right,is.width,"right","width");
			check(before,is.height,"before","height");
			check(forward,is.height,"forward","height");
		}
	private:
		static void check(int ex_param,int im_param,const char* ex_name,const char* im_name)
		{
			std::ostringstream os;
			if(ex_param<0)
			{
				os<< "Negative image Extent: " << ex_name << "=" << ex_param;
				throw os.str();
			}
			if(im_param<=0)
			{
				os<< "Negative or zero image dimension: " << im_name << "=" << im_param;
				throw os.str();
			}
			if(im_param<=ex_param)
			{
				os<< "Requested Extent '" << ex_name << "'=" << ex_param 
					<< " is greater or equal than image param '" 
					<< im_name << "'=" << im_param
					<< ", no sufficient data to reflect! ";
				throw os.str();
			}
		}
	};
	///////////////////////////////////////////////////////////////////////////
	/// class ImageRef
	/// @brief    '  '   , 
	///      StripBuffer,     
	///   OutPortNoPostProc.
	/// 
	///////////////////////////////////////////////////////////////////////////
	template<typename T>
	class ImageRef
	{
		T*     _data;
		int    _width;
		int    _nls;
		Extent _ex;
	public:
		typedef T Pixel;
	private:
		static void check(int a,int b,const char* name)
		{
			std::ostringstream os;
			if(a<0)
			{
				os<< "Negative image Extent: " << name;
				throw os.str();
			}
			if(b<0)
			{
				os<< "Negative new image Extent: " << name;
				throw os.str();
			}
			if(a<b)
			{
				os<< "Requested new image Extent: '" << name << "' is larger than original";
				throw os.str();
			}
		}
	public:
		ImageRef()
			:_data(0),_width(0),_nls(0),_ex()
		{}
		ImageRef(T* data_ptr,int width,int next_line_shift,Extent ex)
			:_data(data_ptr),_width(width),_nls(next_line_shift),_ex(ex)
		{
		}
		ImageRef(const ImageRef& ir,const Extent& new_ext,int y_offset)
		{
			check(ir._ex.left,new_ext.left,"left");
			check(ir._ex.right,new_ext.right,"right");
			check(ir._ex.before,new_ext.before,"before");
			check(ir._ex.forward,new_ext.forward,"forward");
			_data=ir._data+y_offset*ir._nls;
			_width=ir._width;
			_nls=ir._nls;
			_ex=new_ext;
		}
		int nls() const {return _nls;}
		int w() const {return _width;}
		int h() const {return _ex.before,+1+_ex.forward;}
		const Extent& ext() const {return _ex;}
		const T* operator[](int y)  const {return _data+_nls*y;}
	};
	///////////////////////////////////////////////////////////////////////////
	/// class Strip
	/// 
	/// @brief     ,  
	///  <tt>class StripBuffer</tt>  ,    
	///  <tt>class InPort</tt>.
	///
	/// ,  y_last <= y_wait    ,  
	///    T.  left>right,   ,
	///    ,     NULL.
	///
	///////////////////////////////////////////////////////////////////////////
	class Strip
	{
		int y_last;
		int y_wait;
	public:
		Strip(): y_wait(-1),y_last(0) {}
		Strip(int y_start,int y_after_end): y_wait(y_after_end),y_last(y_start) {}
		Strip(int y_current,Extent ext): y_wait(y_current-ext.before),y_last(y_current+ext.forward) {}
		Strip(const Strip& s): y_wait(s.y_wait),y_last(s.y_last) {}
		Strip& operator=(const Strip& s) { y_wait=s.y_wait; y_last=s.y_last; return *this;}
		///
		bool isEmpty() const {return y_last>y_wait; }
		int last() const {return y_last;}
//
		Strip& operator++() { ++y_wait;++y_last; return *this;}
		Strip& operator+=(int delta) { y_wait+=delta;y_last+=delta; return *this;}
		Strip& operator|=(const Strip& s) { y_last=std::min(y_last,s.y_last); y_wait=std::max(y_wait,s.y_wait); return *this;}
		friend static Strip operator|(const Strip& s1,const Strip& s2) { Strip s(s1); s|=s2; return s;}
		///  true,  s    
		bool isContains(const Strip& s) const {return ( y_last <= s.y_last ) && ( s.y_wait <= y_wait ); }
		///  true,  y    
		bool isContains(int y) const {return ( y_last <= y ) && ( y <= y_wait ); }
		///  true,      s
		bool isIn(const Strip& s) const {return s.isContains(*this); }
		///  true,    s ,   ,
		/// ..  <tt> y_last < s.y_last </tt>
		bool canMove(const Strip& s) const 
		{ 
			//Strip next(*this); 
			//++next; 
			//return next.isContains(s);
			return y_last < s.y_last;
		}
	};
	///////////////////////////////////////////////////////////////////////////
	/// class Fifo
	/// 
	/// @brief   
	///
	///////////////////////////////////////////////////////////////////////////
	template<class T> 
	class Fifo
	{
		std::list<T> lst;
	public:
		Fifo() {}
		~Fifo() {}
		bool empty() const { return lst.size()==0;}
		void put(T x)
		{
			lst.push_back(x);
		}
		T get()
		{
			assert(!lst.empty());
			T x=*(lst.begin());
			lst.pop_front();
			return x;
		}
	};
/** @} */ // end of PortsAndBuffersDoxGroup    . 
} //namespace Plf
} //namespace Vtl
//////////////////////////////////////////////////////////////////////////////////
#endif //_PLF_STRIP_BUFFER_SUPPORT_H
